/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of DIVA base library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA base library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

#ifndef DI_SLEEP_H
#define DI_SLEEP_H

#include <unistd.h>
#include <di_types.h>
#include <sys/time.h>
#include <time.h>


#define DI_BILLION 1000000000
#define DI_MILLION 1000000

class DI_Mutex;
class DI_TimeStamp;

enum DI_Timing {
  DI_TIMING_GOOD,
  DI_TIMING_LATE,
  DI_TIMING_EARLY
};


inline void DI_addTimeNs(struct timespec *tspec, long ns)
{
  tspec->tv_nsec += ns;
  if(tspec->tv_nsec >= DI_BILLION) {
    long tmp = tspec->tv_nsec / DI_BILLION;
    tspec->tv_sec += tmp;
    tspec->tv_nsec = tspec->tv_nsec - tmp * DI_BILLION;
  }
}

inline void DI_addTimeUs(struct timeval *tspec, long us)
{
  tspec->tv_usec += us;
  if(tspec->tv_usec >= DI_MILLION) {
    long tmp = tspec->tv_usec / DI_MILLION;
    tspec->tv_sec += tmp;
    tspec->tv_usec = tspec->tv_usec - tmp * DI_MILLION;
  }
}

inline void DI_addTime(struct timeval *tspec, const struct timeval *tspecAdd)
{
  tspec->tv_sec  += tspecAdd->tv_sec;
  tspec->tv_usec += tspecAdd->tv_usec;
  if(tspec->tv_usec >= DI_MILLION) {
    long tmp = tspec->tv_usec / DI_MILLION;
    tspec->tv_sec += tmp;
    tspec->tv_usec = tspec->tv_usec - tmp * DI_MILLION;
  }
}

inline long DI_timeDiffNs(const struct timespec *tspecOld,
			  const struct timespec *tspecNew)
{
  return (tspecNew->tv_sec - tspecOld->tv_sec) * DI_BILLION +
    tspecNew->tv_nsec - tspecOld->tv_nsec;
}

inline long DI_timeDiffUs(const struct timeval *tspecOld,
			  const struct timeval *tspecNew)
{
  return (tspecNew->tv_sec - tspecOld->tv_sec) * DI_MILLION +
    tspecNew->tv_usec - tspecOld->tv_usec;
}

/** Sleeping services. This class contains only static member
    functions. The constructor and destructor are included to prevent
    compiler warnings.*/
class DI_Sleep
{
public:

  DI_Sleep() { }

  ~DI_Sleep() {}

  /// Sleep for n seconds.
  static bool sleepS(uint secs) { return ! sleep(secs); }

  /** Sleep for n milliseconds. You cannot sleep more than one second
      with this function. */
  static bool sleepMs(uint msecs) 
  { 
    struct timespec tspec;
    tspec.tv_sec = 0;
    tspec.tv_nsec = (long) msecs * DI_MILLION;
    return nanosleep(&tspec, 0) != -1;
  }

  /** Sleep for n microseconds. You cannot sleep more than one second
      with this function. */
  static bool sleepUs(uint usecs)
  { 
    struct timespec tspec;
    tspec.tv_sec = 0;
    tspec.tv_nsec = (long) usecs * 1000;
    return nanosleep(&tspec, 0) != -1;
  }
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/// Synchronized sleeping.
class DI_SleepSync
{
public:
  /// The constructor resets the timing.
  DI_SleepSync() { resetTiming(); }

  /// Resets the reference time to current time.
  void resetTiming() 
  { gettimeofday(& m_timing, 0); m_initial = m_timing; }
  
  long sleepSynchroUs(long us);

  long sleepTo(const DI_TimeStamp *stamp, DI_Mutex *mutex = 0);

protected:  
  struct timeval m_initial, m_timing;
};

#endif
