/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */


#ifndef DI_AUDIO_BASE_H
#define DI_AUDIO_BASE_H

#include <di_types.h>
#include "di_config.h"
#include "di_io_defs.h"


/// Unspecified audio library sample rate.
const long DI_AUDIO_UNSPECIFIED_RATE = 0;

/** Audio Interface Base Class.

    DI_AudioBase is declared in "di_audio_base.h" and defined libdiio.

    {\large {\bf Introduction}}

    DI_AudioBase is a base class for all audio I/O classes.
    
*/

class DI_AudioBase
{
public:

  /**@name Constructors */

  //@{

  ///

  DI_AudioBase();

  ///

  virtual ~DI_AudioBase(void);

  //@}

  /**@name Access */

  //@{

  /// Get the name of an input interface.

  virtual const char *inName(void) const = 0;

  /// Get the name of an output interface.

  virtual const char *outName(void) const = 0;

  /// Get the label of an input interface.

  virtual const char *inLabel(void) const = 0;

  /// Get the label of an output interface.

  virtual const char *outLabel(void) const = 0;

  /// Return the IO mode.

  virtual DI_IoMode mode(void) const { return m_mode; }

  /// Return the input sample format.

  virtual DI_AudioSampleFormat inSampleFormat(void) const
  { return m_inFormat; }
  
  /// Return the output sample format.

  virtual DI_AudioSampleFormat outSampleFormat(void) const
  { return m_outFormat; }
  
  /// Return the input audio buffer size.

  virtual uint inBufSize(void) const { return m_inBufSize; }

  /// Return the output audio buffer size.

  virtual uint outBufSize(void) const { return m_outBufSize; }

  /// Return input sample rate.

  virtual uint inSampleRate(void) const { return m_inRate; }

  /// Return output sample rate.

  virtual uint outSampleRate(void) const { return m_outRate; }

  /// Return the number of input channels.

  virtual uint inChannels(void) const { return m_inChannels; }

  /// Return the number of output channels.

  virtual uint outChannels(void) const { return m_outChannels; }

  /// Return maximum input block size.

  virtual uint inMaxBlock(void) const { return m_inBlock; }
  
  /// Return maximum output block size.

  virtual uint outMaxBlock(void) const { return m_outBlock; }
  
  /// Get number of filled sample frames in the input sample buffer

  virtual int inFilledFrames(void) const = 0;
  
  /// Get number of filled sample frames in the output sample buffer

  virtual int outFilledFrames(void) const = 0;
  
  //@}
  
  /**@name I/O operations */
  
  //@{

  /** Read sample frames from the interface.

      This method reads 'frames' sample frames from the interface,
      and stores them into buffer pointed to by 'buffer'.

      Please note that one sample may take more than one byte, so
      'buffer' must point to a buffer of capable of holding 'frames'
      * N * M bytes, where N is the size of one sample and M is number of
      channels.
     
      The number of sample frames actually recorded is
      returned, or a negative value if the operation failed.
  */

  virtual int readFrames(void *buffer, uint frames) = 0;

  /** Write sample frames to the interface.

      This method writes 'frames' sample frames to the interface,
      reading them from buffer pointer to by 'buffer'. The number
      of sample frames actually written is returned, or a negative value if
      the operation failed. 
  */

  virtual int writeFrames(const void *buffer, uint frames) = 0;
  
  /** Wait until 'frames' # of frames can be read.
      
      Return value tells how many frames are filled as the call
      returns. Negative values indicate error of some sort.

  */
  
  virtual int waitInputFilled(uint frames, int timeOut = -1) = 0;

  /// Wait until output has only the number of frames specified filled.  

  virtual int waitOutputFilled(uint frames, int timeOut = -1) = 0;

  //@}

  /**@name Predicates */
  
  //@{

  ///

  virtual bool isOK(void) const = 0;

  /// True if End-Of-File condition was met.

  virtual bool isEOF(void) const = 0;

  //@}

protected:

  DI_IoMode            m_mode;
  DI_AudioSampleFormat m_inFormat;
  DI_AudioSampleFormat m_outFormat;
  uint                 m_inBufSize;
  uint                 m_outBufSize;
  uint                 m_inRate;
  uint                 m_outRate;
  uint                 m_inChannels;
  uint                 m_outChannels;
  uint                 m_inBlock;
  uint                 m_outBlock;
};

#endif  // DI_AUDIO_BASE_H
