/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

//
// DIVA Library. Copyright (C) Jarmo Hiipakka and Tommi Ilmonen, 1999-2001.
//
// $Log: di_audio_file.h,v $
// Revision 1.12  2004/11/30 13:51:47  tilmonen
// Copyright upodates...
//
// Revision 1.11  2004/11/30 13:14:38  tilmonen
// Copyright updates.
//
// Revision 1.10  2004/11/30 12:58:00  tilmonen
// Updated license text.
//
// Revision 1.9  2004/02/06 15:47:53  tilmonen
// Moved many data types from "ulong" to "uint".
//
// Revision 1.8  2002/09/16 06:28:36  tilmonen
// Bug fixes and minor improvements.
//
// Revision 1.7  2002/08/26 12:54:32  tilmonen
// Bug fixes.
//
// Revision 1.6  2002/08/19 12:45:59  tilmonen
// Merged old changes and fixed a few bugs shown by MipsPro compiler.
//
// Revision 1.5  2002/08/19 11:34:37  tilmonen
// Simplified the audio file class operations. Now much more usable.
//
// Revision 1.4  2002/08/09 07:52:37  tilmonen
// Fixes / additions.
//
// Revision 1.3  2001/07/27 08:23:26  tilmonen
// License text updates (=cosmetic changes).
//
// Revision 1.2  2001/06/06 10:34:59  tilmonen
// Updates for new glibc include strategies and libaudiofile-relaed fixes.
//
// Revision 1.1.1.1  2001/06/04 09:04:20  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#ifndef DI_AUDIO_FILE_H
#define DI_AUDIO_FILE_H

#include <audiofile.h>
#include "di_audio_base.h"

#include <stdlib.h> // NULL

/** Audio file formats. */
enum DI_AudioFileFormat
{
  DI_AFF_UNKNOWN = AF_FILE_UNKNOWN,
  DI_AFF_RAWDATA = AF_FILE_RAWDATA,
  DI_AFF_AIFFC = AF_FILE_AIFFC,
  DI_AFF_AIFF = AF_FILE_AIFF,
  DI_AFF_NEXTSND = AF_FILE_NEXTSND,
  DI_AFF_WAVE = AF_FILE_WAVE,
  // #if defined(LIBAUDIOFILE_MINOR_VERSION) && LIBAUDIOFILE_MINOR_VERSION == 2
  // #else
#if defined(AF_FILE_BICSF)
  DI_AFF_BICSF = AF_FILE_BICSF,
  DI_AFF_MPEG1BITSTREAM = AF_FILE_MPEG1BITSTREAM,
  DI_AFF_AVR = AF_FILE_AVR,
  DI_AFF_IFF_8SVX = AF_FILE_IFF_8SVX,
  DI_AFF_SAMPLEVISION = AF_FILE_SAMPLEVISION,
  DI_AFF_VOC = AF_FILE_VOC,
  DI_AFF_NIST_SPHERE = AF_FILE_NIST_SPHERE,
  DI_AFF_SOUNDFONT2 = AF_FILE_SOUNDFONT2
#endif
};

/** Audio file sample formats. */
enum DI_AudioFileSampleFormat
{
  DI_AFSF_TWOSCOMP = AF_SAMPFMT_TWOSCOMP,
  DI_AFSF_UNSIGNED = AF_SAMPFMT_UNSIGNED,
  DI_AFSF_FLOAT = AF_SAMPFMT_FLOAT,
  DI_AFSF_DOUBLE = AF_SAMPFMT_DOUBLE
};

/** Audio file byte order. */
enum DI_AudioFileByteOrder
{
  DI_AFBO_BIGENDIAN = AF_BYTEORDER_BIGENDIAN,
  DI_AFBO_LITTLEENDIAN = AF_BYTEORDER_LITTLEENDIAN
};


/** Audio file I/O.

    DI_AudioFile is declared in "di_audio_file.h" and defined in libdiio.

    @see DI_AudioFileThreaded

*/

class DI_AudioFile : public DI_AudioBase
{
public:

  /**@name Constructors */
  
  //@{

  /** Create a new audio file.
    
      This constructor creates a new audio file.
  */

  DI_AudioFile();
  
  ///

  virtual ~DI_AudioFile(void);

  //@}

  /**@name Access */

  //@{

  /// Get the name of an input interface.

  const char *inName(void) const
  { return (m_mode == DI_IO_INPUT) ? (const char *)m_name : NULL; }

  /// Get the name of an output interface.

  const char *outName(void) const
  { return (m_mode == DI_IO_OUTPUT) ? (const char *)m_name : NULL; }

  /// Get the label of an input interface.

  const char *inLabel(void) const
  { return (m_mode == DI_IO_INPUT) ? (const char *)m_name : NULL; }

  /// Get the label of an output interface.

  const char *outLabel(void) const
  { return (m_mode == DI_IO_OUTPUT) ? (const char *)m_name : NULL; }

  /// Get the number of filled sample frames in the input sample buffer

  int inFilledFrames(void) const { return 0; }
  
  /// Get the number of filled sample frames in the output sample buffer

  int outFilledFrames(void) const { return 0; }
  
  /// Get the audio file format.

  DI_AudioFileFormat getFileFormat() const;

  /// Set the audio file format.
  
  void setFileFormat(DI_AudioFileFormat format)
  { afInitFileFormat(m_fSetup,(int)format); }

  /// Set the audio file sampling rate.

  void setSampleRate(uint rate)
  { afInitRate(m_fSetup, AF_DEFAULT_TRACK, (double)rate); }

  /// Get the sampling rate
  uint getSampleRate() 
  { if(m_fHandle == AF_NULL_FILEHANDLE) return 0; 
  return (uint) afGetRate(m_fHandle, AF_DEFAULT_TRACK); }

  /// Set the number of channels in an audio file.

  void setChannels(uint channels)
  { afInitChannels(m_fSetup, AF_DEFAULT_TRACK, (int)channels); }

  /// Get the sample format of an audio file.

  DI_AudioFileSampleFormat getSampleFormat() const
  { return (DI_AudioFileSampleFormat)m_fileSFormat; }

  /// Set the sample format of an audio file.

  void setSampleFormat(DI_AudioFileSampleFormat format)
  { 
    afInitSampleFormat(m_fSetup,AF_DEFAULT_TRACK,(int)format,m_fileSWidth); 
    m_fileSFormat = format;
  }

  bool setFileSampleFormat(DI_AudioSampleFormat sampleFormat) 
  {
    DI_AudioFileSampleFormat afformat = DI_AFSF_TWOSCOMP;
    
    if(sampleFormat <= DI_ASF_INT32)
      afformat = DI_AFSF_TWOSCOMP;
    else if (sampleFormat == DI_ASF_FLOAT32)
      afformat = DI_AFSF_FLOAT;
    else if (sampleFormat == DI_ASF_FLOAT64)
      afformat = DI_AFSF_DOUBLE;
    else
      return false;

    setSampleFormat(afformat);
    setSampleWidth(DI_sampleWidth(sampleFormat) * 8);
    return true;
  }

  bool setUserSampleFormat(DI_AudioSampleFormat userFormat) 
  {
    int userSample = -1;
    int userSWidth = -1;
    
    switch(userFormat) {
    case DI_ASF_INT16:
      userSample = AF_SAMPFMT_TWOSCOMP;
      userSWidth = 16;
      break;
    case DI_ASF_INT32:
      userSample = AF_SAMPFMT_TWOSCOMP;
      userSWidth = 32;
      break;
    case DI_ASF_FLOAT32:
      userSample = AF_SAMPFMT_FLOAT;
      userSWidth = 32;
      break;
    case DI_ASF_FLOAT64:
      userSample = AF_SAMPFMT_DOUBLE;
      userSWidth = 64;
      break;
    default:
      return false;
    }
    
    afSetVirtualSampleFormat
      (m_fHandle,AF_DEFAULT_TRACK, userSample, userSWidth);
    return true;
  }

  /// Get the sample width of an integer sample format

  uint getSampleWidth() const { return (uint)m_fileSWidth; }

  /// Set the sample width of an integer sample format

  void setSampleWidth(uint width)
  { 
    afInitSampleFormat(m_fSetup,AF_DEFAULT_TRACK,m_fileSFormat,(int)width); 
    m_fileSWidth = width;
  }

  /// Get the byte order of the sample format.
  
  DI_AudioFileByteOrder getByteOrder() const
  { if(m_fHandle == AF_NULL_FILEHANDLE) return DI_AFBO_LITTLEENDIAN; 
  return (DI_AudioFileByteOrder)afGetByteOrder(m_fHandle,AF_DEFAULT_TRACK); }

  /// Set the byte order of the sample format.

  void setByteOrder(DI_AudioFileByteOrder order)
  { afInitByteOrder(m_fSetup, AF_DEFAULT_TRACK, (int)order); }
  
  //@}

  /**@name Operations */

  //@{

  /// Open the audio file.

  bool open(/// Audio file name.
	    const char *file,
	    /// IO mode: input or output
	    DI_IoMode mode,
	    /// Audio file sample format (the user format).
            DI_AudioSampleFormat userFormat,
	    /// Maximum block size.
	    ulong maxBlock);

  /** Close the file.

      This method closes the file and frees the resources used by the file.
      In case of an output file, this method flushes pending samples and
      writes any information that is needed to complete the file format.
  */

  bool close(void);

  //@}
  
  /**@name I/O operations */
  
  //@{

  /// Get the total number of sample frames in an audio file.

  long frameCount(void) const
  { if(m_fHandle == AF_NULL_FILEHANDLE) return -1; 
  return (long)afGetFrameCount(m_fHandle,AF_DEFAULT_TRACK); }

  /// Move the logical pointer to a specified sample frame.

  long seekFrame(long offset)
  { if(m_fHandle == AF_NULL_FILEHANDLE) return -1; 
  return (long)afSeekFrame(m_fHandle,AF_DEFAULT_TRACK,offset); }
  
  /// Get the location of the logical pointer in frames
  long currentFrame() const
  { if(m_fHandle == AF_NULL_FILEHANDLE) return -1; 
  return (long) afTellFrame((AFfilehandle) m_fHandle,AF_DEFAULT_TRACK); }

  /// Read sample frames from the file.

  int readFrames(void *buffer, uint frames);

  /// Write sample frames to the file.

  int writeFrames(const void *buffer, uint frames);

  /// Wait until 'frames' # of frames can be read.

  int waitInputFilled(uint frames, int timeOut = -1)
  { (void)frames; (void)timeOut; return 0; }

  /// Wait until output has only the number of frames specified filled.  

  int waitOutputFilled(uint frames, int timeOut = -1)
  { (void)frames; (void)timeOut; return 0; }

  //@}

  /**@name Predicates */
  
  //@{

  ///

  bool isOK(void) const;

  ///

  bool isEOF(void) const;

  //@}

  static double filePeak(const char * fileName, bool verbose);
  static int determineFormat(const char *formatName);
  static const char *formatStr(DI_AudioFileFormat);

private:

  DI_AudioFile(const DI_AudioFile &) {}

  char                 *m_name;
  long                  m_totFrames;
  // long                  m_readFrames;
  // void                 *m_tmpBuffer;
  // int                   m_tmpSWidth;
  
  AFfilesetup           m_fSetup;
  AFfilehandle          m_fHandle;

  int                   m_fileSFormat;
  int                   m_fileSWidth;
};

#endif  // DI_AUDIO_FILE_H
