/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */
#ifndef DI_AUDIO_FILE_THREADED_H
#define DI_AUDIO_FILE_THREADED_H

#include "di_audio_file.h"
#include "di_sleep.h"
#include "di_condition.h"
#include "di_thread.h"


/** Status of a threaded audio file. */
enum DI_AftStatus {
  DI_AFT_EMPTY,
  DI_AFT_OPENING,
  DI_AFT_READY,
  DI_AFT_READING,
  DI_AFT_WRITING,
  DI_AFT_SIZEOF
};

/** Threaded audio file. Experimental. 
    
    None of the methods should block if used correctly.

    In some cases you want to thread file access. In particular if
    the app requires low-latency situation this is necessary. 

    The read/write buffers must accumulate the exact multiples
    of the maximum block size. 

    \b Why ? Accessing hard-disk may take some time. This is due to
    the physical limitations of the device. It takes milliseconds -
    not microseconds - to access a file. To decrease the latency the
    file IO can be threaded. The read- or write thread takes care of
    buffering. As long as file-thread is not overloaded the host can
    read and write data with zero latency. The file thread will
    eventually deliver the data to/from the file in the background.

    The following example shows how to use this class:

\verbatim
  
  const uint blockSize = 1000;
  const uint buzSize = blockSize * 5; // Exact multiple.

  // Creates a threaded audio file object.
  DI_AudioFileThreaded aft;

  // Spans a thread to do the reading. 
  aft.startRead("test.wav",
		DI_ASF_FLOAT32,
		buzSize,
		0, 
		1000000,
		false);
  
  // Used before the first read/write to guarantee the file is ready
  aft.userPrepare();

  float * buf = new float[buzSize * aft.file().outChannels()];

  // Read audio
  aft.readFrames(buf, blockSize);

  // Prepare the file for shut-down.
  aft.requestEnd();

  // Make sure the file object has finished its work.  
  aft.waitEnd();
  delete [] buf;

\endverbatim

  @see DI_AudioFile

*/
class DI_AudioFileThreaded : public DI_Thread
{
public:
  DI_AudioFileThreaded();
  ~DI_AudioFileThreaded();

  bool startRead(const char *file, 
		 DI_AudioSampleFormat userFormat,
		 /// The disk-access block size in frames
		 ulong maxBlock,
		 long startFrame,
		 long endFrame,
		 bool loop,
		 /// Real-time priority
		 int priority = -1);

  bool startWrite(const char *file, 
		  DI_AudioSampleFormat userFormat,
		  uint sampleRate,
		  uint channels,
		  ulong maxBlock,
		  int priority = -1,
		  DI_AudioFileFormat fileFormat = DI_AFF_WAVE,
		  DI_AudioSampleFormat sampleFormat = DI_ASF_INT16,
		  DI_AudioFileByteOrder byteOrder = DI_AFBO_LITTLEENDIAN);
  
  bool userPrepare();

  /** Reads frames from the internal buffer. */
  int readFrames(void *buffer, ulong frames);

  /** Writes frames to the internal buffer. */
  int writeFrames(void *buffer, ulong frames);

  /** May block if called too early. */
  void requestEnd();
  
  /// Return the number fo frames the user has written or read.
  inline ulong userFrames() const { return m_userFrames; }

  bool loop() const { return m_loop; }
  
  /// Returns the audio file object.
  inline DI_AudioFile & file() { return m_file; }
protected:
  bool allocateBuffer(DI_AudioSampleFormat userFormat, 
		      ulong maxBlock, 
		      ulong ch);

  void childLoop();

  void readAudio();

  void writeAudio();

  void writeRemaining();

  char *       m_tmpBuffer;
  char *       m_fileName;

  /** The first two mutexes represent buffer mutexes. The third is
      buffer transition mutex. */

  DI_Mutex     m_mutex;

  DI_Condition m_cond;

  /// Size of one block.
  ulong                m_blockSize;
  /// Size of block in chars
  ulong                m_blockChars;
  /// Number of bytes in user sample
  ulong                m_userSWidth;
  /// User sample format
  DI_AudioSampleFormat m_userFormat;

  /// File format
  DI_AudioFileFormat   m_fileFormat;
  
  long                 m_startFrame, m_endFrame;
  /// Number frames processed by the user thread
  ulong                m_userFrames;
  /// Number of frames processed by the file thread
  ulong                m_threadFrames;
  /// Number of channels
  ulong                m_chans;
  
  /// Index of the file-thread buffer segment.
  uint                 m_threadBuffer;
  /// Number of buffer segments
  uint                 m_buffers;

  bool         m_loop;
  bool         m_end;
  bool         m_ok;

  ulong        m_fills, m_drains;

  /// Real-time priority level, negative if not used.
  int          m_priority;

  DI_IoMode    m_mode;
  DI_AftStatus m_status;
  DI_AudioFile m_file;

private:
  // Disabled
  DI_AudioFileThreaded(const DI_AudioFileThreaded &) {}
};

#endif
