/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

//
// DIVA Library. Copyright (C) Jarmo Hiipakka, 1999.
//
// $Log: di_numconv.h,v $
// Revision 1.11  2004/11/30 13:51:48  tilmonen
// Copyright upodates...
//
// Revision 1.10  2004/11/30 13:14:39  tilmonen
// Copyright updates.
//
// Revision 1.9  2004/11/30 12:58:02  tilmonen
// Updated license text.
//
// Revision 1.8  2002/08/26 10:56:39  tilmonen
// Bug fixes.
//
// Revision 1.7  2002/08/19 12:46:00  tilmonen
// Merged old changes and fixed a few bugs shown by MipsPro compiler.
//
// Revision 1.6  2002/05/22 12:42:46  tilmonen
// Changes due to port to Windows.
//
// Revision 1.5  2001/09/11 12:46:28  tilmonen
// Updates.
//
// Revision 1.4  2001/09/06 09:31:06  tilmonen
// Updates to make-system:
//
// Now the libraries never have explicit paths to them. Thus users can
// move the libs around and still get them working with proper
// LD_LIBRARY_PATH settings.
//
// Revision 1.3  2001/07/27 08:23:26  tilmonen
// License text updates (=cosmetic changes).
//
// Revision 1.2  2001/06/06 10:34:59  tilmonen
// Updates for new glibc include strategies and libaudiofile-relaed fixes.
//
// Revision 1.1.1.1  2001/06/04 09:04:21  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#ifndef DI_NUMCONV_H
#define DI_NUMCONV_H

#include <di_endian.h>
#include <di_types.h>

#include <netinet/in.h>
#include <string.h>

// 2^15 - 1
#define DI_ALSA_SSCALE 32767
// 2^31 - 1
#define DI_ALSA_ISCALE 2147483647

/** Fast numeric rounding and type conversion. 

    <B>These implementations trade accuracy for speed.</B> Typically
    the results are very close to correct.

    \b TFloat This is the floating point data type that will be
    converted.

    \b TInt This is the integer type to which the data will be
    converted.
*/

template <class TFloat, class TInt> class DI_Round
{
public:
  /** A number close to one (used in rounding). */
  static TFloat nearlyOne() { return 0.999999; }
  
  /** Round \b up to the next integer. */
  static inline TInt ceil(TFloat v) 
  { return v > 0 ? (TInt) (v + nearlyOne()) : (TInt) v; }

  /** Round \b down to the next integer. */
  static inline TInt floor(TFloat v) 
  { return v > 0 ? (TInt) v : (TInt) (v - nearlyOne()); }
};


inline void DI_ntoh4(void *data)
{
  u_long *d((u_long *)data);
  *d = ntohl(d[0]);
}

inline void DI_hton4(void *data)
{
  u_long *d((u_long *)data);
  *d = htonl(d[0]);
}

inline void DI_ntoh8(void *data)
{
#ifndef WORDS_BIGENDIAN
  u_long c, *d((u_long *)data);
  c = ntohl(d[0]);
  d[0] = ntohl(d[1]);
  d[1] = c;
#else
  (void)data;
#endif
}

inline void DI_hton8(void *data)
{
#ifndef WORDS_BIGENDIAN
  u_long c, *d((u_long *)data);
  c = htonl(d[0]);
  d[0] = htonl(d[1]);
  d[1] = c;
#else
  (void)data;
#endif
}

// double & int

inline void DI_double2Int(int *v1, const double *v2, const double scale,
			  const int offset, unsigned len)
{
  int *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (int) (*v2++ * scale) + offset;
}

inline void DI_int2Double(double *v1, const int *v2, const double scale,
			  const double offset, unsigned len)
{
  double *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (double) *v2++ * scale + offset;
}

inline void DI_ntohDouble2Int(int *v1, const double *v2,const double scale,
			      const int offset, unsigned len)
{
  int *sentinel = &v1[len];
  double t;
  
  while(v1 < sentinel)
  {
    t = *v2++;
    DI_ntoh8(&t);
    *v1++ = (int) (t * scale) + offset;
  }
}

inline void DI_ntohInt2Double(double *v1, const int *v2, const double scale,
			      const double offset, unsigned len)
{
  double *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (double)ntohl((u_long)(*v2++)) * scale + offset;
}

// float & int

inline void DI_float2Int(int *v1, const float *v2, const float scale,
			 const int offset, unsigned len)
{
  int *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (int) (*v2++ * scale) + offset;
}

inline void DI_int2Float(float *v1, const int *v2, const float scale,
			 const float offset, unsigned len)
{
  float *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (float) *v2++ * scale + offset;
}

inline void DI_ntohFloat2Int(int *v1, const float *v2, const float scale,
			     const int offset, unsigned len)
{
  int *sentinel = &v1[len];
  float t;
  
  while(v1 < sentinel)
  {
    t = *v2++;
    DI_ntoh4(&t);
    *v1++ = (int) (t * scale) + offset;
  }
}

inline void DI_ntohInt2Float(float *v1, const int *v2, const float scale,
			     const float offset, unsigned len)
{
  float *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (float)ntohl((u_long)(*v2++)) * scale + offset;
}

// double & short

inline void DI_double2Short(short *v1, const double *v2, const double scale,
			    const short offset, unsigned len)
{
  short *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (short) (*v2++ * scale) + offset;
}

inline void DI_short2Double(double *v1, const short *v2, const double scale,
			    const double offset, unsigned len)
{
  double *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (double) *v2++ * scale + offset;
}

inline void DI_ntohDouble2Short(short *v1, const double *v2,const double scale,
				const short offset, unsigned len)
{
  short *sentinel = &v1[len];
  double t;
  
  while(v1 < sentinel)
  {
    t = *v2++;
    DI_ntoh8(&t);
    *v1++ = (short) (t * scale) + offset;
  }
}

inline void DI_ntohShort2Double(double *v1,const short *v2,const double scale,
				const double offset,unsigned len)
{
  double *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (double)ntohs((u_short)(*v2++)) * scale + offset;
}

// float & short

inline void DI_float2Short(short *v1, const float *v2, const float scale,
			   const short offset, unsigned len)
{
  short *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (short) (*v2++ * scale) + offset;
}

inline void DI_short2Float(float *v1, const short *v2, const float scale,
			   const float offset, unsigned len)
{
  float *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (float) *v2++ * scale + offset;
}

inline void DI_ntohFloat2Short(short *v1, const float *v2, const float scale,
			       const short offset, unsigned len)
{
  short *sentinel = &v1[len];
  float t;
  
  while(v1 < sentinel)
  {
    t = *v2++;
    DI_ntoh4(&t);
    *v1++ = (short) (t * scale) + offset;
  }
}

inline void DI_ntohShort2Float(float *v1, const short *v2, const float scale,
			       const float offset, unsigned len)
{
  float *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (float)ntohs((u_short)(*v2++)) * scale + offset;
}

// double & char

inline void DI_double2Char(char *v1, const double *v2, const double scale,
			   const char offset, unsigned len)
{
  char *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (char) (*v2++ * scale) + offset;
}

inline void DI_char2Double(double *v1, const char *v2, const double scale,
			   const double offset, unsigned len)
{
  double *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (double) *v2++ * scale + offset;
}

// float & char

inline void DI_float2Char(char *v1, const float *v2, const float scale,
			  const char offset, unsigned len)
{
  char *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (char) (*v2++ * scale) + offset;
}

inline void DI_char2Float(float *v1, const char *v2, const float scale,
			  const float offset, unsigned len)
{
  float *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (float) *v2++ * scale + offset;
}

// int & int

inline void DI_int2Int(int *v1, const int *v2, int move, const int offset,
		       unsigned len)
{
  int *sentinel = &v1[len];

  if (move > 0)
    while(v1 < sentinel)
      *v1++ = ((*v2++) << move) + offset;
  else
  {
    move = -move;
    while(v1 < sentinel)
      *v1++ = ((*v2++) >> -move) + offset;
  }
}

// short & short

inline void DI_short2Short(short *v1, const short *v2, int move,
			   const short offset, unsigned len)
{
  short *sentinel = &v1[len];

  if (move > 0)
    while(v1 < sentinel)
      *v1++ = ((*v2++) << move) + offset;
  else
  {
    move = -move;
    while(v1 < sentinel)
      *v1++ = ((*v2++) >> -move) + offset;
  }
}

// char & char

inline void DI_char2Char(char *v1, const char *v2, int move,
			 const char offset,unsigned len)
{
  char *sentinel = &v1[len];

  if (move > 0)
    while(v1 < sentinel)
      *v1++ = ((*v2++) << move) + offset;
  else
  {
    move = -move;
    while(v1 < sentinel)
      *v1++ = ((*v2++) >> -move) + offset;
  }
}

// int & short

inline void DI_int2Short(short *v1, const int *v2, int move,
			 const short offset, unsigned len)
{
  short *sentinel = &v1[len];
  move = -move; // Can't move to the left
  
  while(v1 < sentinel)
    *v1++ = (short) ((*v2++) >> move) + offset;
}

inline void DI_ntohInt2Short(short *v1, const int *v2, int move,
			     const short offset, unsigned len)
{
  short *sentinel = &v1[len];
  move = -move; // Can't move to the left
  
  while(v1 < sentinel)
    *v1++ = (short) (ntohl((u_long)(*v2++)) >> move) + offset;
}

inline void DI_short2Int(int *v1, const short *v2, int move,
			 const int offset, unsigned len)
{
  int *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (((int) *v2++) << move) + offset;
}

inline void DI_ntohShort2Int(int *v1, const short *v2, int move,
			     const int offset, unsigned len)
{
  int *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (((int) ntohs((u_short)(*v2++))) << move) + offset;
}

// int & char

inline void DI_int2Char(char *v1, const int *v2, int move,
			const char offset, unsigned len)
{
  char *sentinel = &v1[len];
  move = -move; // Can't move to the left
  
  while(v1 < sentinel)
    *v1++ = (char) ((*v2++) >> move) + offset;
}

inline void DI_char2Int(int *v1, const char *v2, int move,
			const int offset, unsigned len)
{
  int *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (((int) *v2++) << move) + offset;
}

// short & char

inline void DI_short2Char(char *v1, const short *v2, int move,
			  const char offset, unsigned len)
{
  char *sentinel = &v1[len];
  move = -move; // Can't move to the left
  
  while(v1 < sentinel)
    *v1++ = (char) ((*v2++) >> move) + offset;
}

inline void DI_char2Short(short *v1, const char *v2, int move,
			  const short offset, unsigned len)
{
  short *sentinel = &v1[len];
  
  while(v1 < sentinel)
    *v1++ = (((short) *v2++) << move) + offset;
}

// float & double

inline void DI_float2Double(double *v1, const float *v2, unsigned len)
{
  double *sentinel = &v1[len];

  while(v1 < sentinel)
    *v1++ = (double) *v2++;
}

inline void DI_double2Float(float *v1, const double *v2, unsigned len)
{
  float *sentinel = &v1[len];

  while(v1 < sentinel)
    *v1++ = (float) *v2++;
}

inline void DI_ntohFloat2Double(double *v1, const float *v2, unsigned len)
{
  double *sentinel = &v1[len];
  float t;
  
  while(v1 < sentinel)
  {
    t = *v2++;
    DI_ntoh4(&t);
    *v1++ = (double) t;
  }
}

inline void DI_ntohDouble2Float(float *v1, const double *v2, unsigned len)
{
  float *sentinel = &v1[len];
  double t;

  while(v1 < sentinel)
  {
    t = *v2++;
    DI_ntoh8(&t);
    *v1++ = (float) t;
  }
}

// Network and host byte order change

inline void DI_htonInt(int *v1, const int *v2, unsigned len)
{
  int *sentinel = &v1[len];

  while(v1 < sentinel)
    *v1++ = (int)htonl((u_long)(*v2++));
}

inline void DI_ntohInt(int *v1, const int *v2, unsigned len)
{
  int *sentinel = &v1[len];

  while(v1 < sentinel)
    *v1++ = (int)ntohl((u_long)(*v2++));
}

inline void DI_htonShort(short *v1, const short *v2, unsigned len)
{
  short *sentinel = &v1[len];

  while(v1 < sentinel)
    *v1++ = (short)htons((u_short)(*v2++));
}

inline void DI_ntohShort(short *v1, const short *v2, unsigned len)
{
  short *sentinel = &v1[len];

  while(v1 < sentinel)
    *v1++ = (short)ntohs((u_short)(*v2++));
}

inline void DI_htonFloat(float *v1, const float *v2, unsigned len)
{
  memcpy(v1,v2,len*sizeof(float));

#ifndef WORDS_BIGENDIAN
  float *sentinel = &v1[len];

  while(v1 < sentinel)
    DI_hton4(v1++);
#endif  // WORDS_BIGENDIAN
}

inline void DI_ntohFloat(float *v1, const float *v2, unsigned len)
{
  memcpy(v1,v2,len*sizeof(float));

#ifndef WORDS_BIGENDIAN
  float *sentinel = &v1[len];

  while(v1 < sentinel)
    DI_ntoh4(v1++);
#endif  // WORDS_BIGENDIAN
}

inline void DI_htonDouble(double *v1, const double *v2, unsigned len)
{
  memcpy(v1,v2,len*sizeof(double));

#ifndef WORDS_BIGENDIAN
  double *sentinel = &v1[len];

  while(v1 < sentinel)
    DI_hton8(v1++);
#endif  // WORDS_BIGENDIAN
}

inline void DI_ntohDouble(double *v1, const double *v2, unsigned len)
{
  memcpy(v1,v2,len*sizeof(double));

#ifndef WORDS_BIGENDIAN
  double *sentinel = &v1[len];

  while(v1 < sentinel)
    DI_ntoh8(v1++);
#endif  // WORDS_BIGENDIAN
}


#endif  // DI_NUMCONV_H
