/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the DIVA IO library.
 *
 * Author(s): Jarmo Hiipakka and Tommi Ilmonen, 2000-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * DIVA - Digital Interactive Virtual Acoustics group 
 * Telecommunications Software and Multimedia Laboratory
 * Helsinki University of Technology
 * Finland
 *
 * http://www.tml.hut.fi/Research/DIVA/
 *
 * DIVA IO library is licensed under GNU Lesser General Public
 * License (LGPL). The LGPL conditions can be found in file "LGPL"
 * that is distributed with this source package or obtained from the
 * GNU organization (www.gnu.org).
 * 
 */

//
// DIVA Library. Copyright (C) Jarmo Hiipakka, 1999.
//
// $Log: di_virtual_audio.h,v $
// Revision 1.7  2004/11/30 13:51:48  tilmonen
// Copyright upodates...
//
// Revision 1.6  2004/11/30 13:14:40  tilmonen
// Copyright updates.
//
// Revision 1.5  2004/11/30 12:58:02  tilmonen
// Updated license text.
//
// Revision 1.4  2004/02/06 15:47:53  tilmonen
// Moved many data types from "ulong" to "uint".
//
// Revision 1.3  2002/05/22 12:42:47  tilmonen
// Changes due to port to Windows.
//
// Revision 1.2  2001/07/27 08:23:26  tilmonen
// License text updates (=cosmetic changes).
//
// Revision 1.1.1.1  2001/06/04 09:04:21  tilmonen
// Mustajuuri 4.6.2001, initial CVS cersion
//

#ifndef DI_VIRTUAL_AUDIO__H
#define DI_VIRTUAL_AUDIO__H

#include <stdlib.h>
#include <di_types.h>
#include <sys/ipc.h>
#include <sys/sem.h>

#include "di_audio_base.h"


/// Info structure for virtual audio interface.
struct DI_VaiInfo
{
  /// Sample buffer format.
  uint format;
  /// Sample buffer size.
  uint bufferSize;
  /// Number of channels.
  uint channels;

  /// Number of writers (currently 0 or 1)
  uint writers;

  /// Write index.
  uint write;
  /// Read index.
  uint read;

  /// Is at end. (End-Of-File)
  uint isEOF;
  /// Dummy for 64-bit alignment.
  uint dummy;
};



/** Virtual Audio Interface.

    DI_VirtualAudio is declared in "di_virtual_audio.h" and defined libdiio.

    {\large {\bf Introduction}}

    DI_VirtualAudio is a class that provides an interface that can be
    used to transmit audio data from one process to another.
    
*/

class DI_VirtualAudio : public DI_AudioBase
{
public:

  /**@name Constructors */

  //@{

  /** Create a new virtual audio port.

      This constructor creates a new virtual audio port. The port is
      unconnected and is made operational only by using the open method.
  */

  DI_VirtualAudio();

  ///

  ~DI_VirtualAudio(void);

  //@}

  /**@name Access */

  //@{


  /** Get the name of an input interface.

      Returns the name of the virtual audio interface. This name is set
      in the file $DIVAPATH/etc/config/DI_VirtualAudio.cfg
  */

  const char *inName(void) const
  { return (m_mode == DI_IO_INPUT) ? (const char *)m_name : NULL; }

  /// Get the name of an output interface.

  const char *outName(void) const
  { return (m_mode == DI_IO_OUTPUT) ? (const char *)m_name : NULL; }

  /// Get the label of an input interface.

  const char *inLabel(void) const
  { return (m_mode == DI_IO_INPUT) ? (const char *)m_name : NULL; }

  /// Get the label of an output interface.

  const char *outLabel(void) const
  { return (m_mode == DI_IO_OUTPUT) ? (const char *)m_name : NULL; }

  /// Return the input audio buffer size.

  uint inBufSize(void) const
  { return (m_info && (m_mode == DI_IO_INPUT)) ? m_info->bufferSize : 0; }

  /// Return the output audio buffer size.

  uint outBufSize(void) const
  { return (m_info && (m_mode == DI_IO_OUTPUT)) ? m_info->bufferSize : 0; }

  /// Return the number of input channels.

  uint inChannels(void) const
  { return (m_info && (m_mode == DI_IO_INPUT)) ? m_info->channels : 0; }

  /// Return the number of output channels.

  uint outChannels(void) const
  { return (m_info && (m_mode == DI_IO_OUTPUT)) ? m_info->channels : 0; }

  /// Get number of filled sample frames in the input sample buffer

  int inFilledFrames(void) const
  { return (m_mode == DI_IO_INPUT) ? filledFrames() : 0; }
  
  /// Get number of filled sample frames in the output sample buffer

  int outFilledFrames(void) const
  { return (m_mode == DI_IO_OUTPUT) ? filledFrames() : 0; }
  
  //@}
  
  /**@name Operations */
  
  //@{

  /** Open the virtual audio port.

      This method opens an actual audio port to a named virtual audio device.
      The audio port may be opened for recording or playback, but not for both.

      The 'bufferSize' parameter sets the audio queue buffer size
      (in sample frames) that  will be used as a temporary data storage
      during recording or playing. Buffer size is only affected when creating
      an input interface.

      If you create a virtual audio port for output, you may use the default
      arguments for buffersize and channel count. These values have already
      been set by the reader.

  */
  
  bool open(/// Virtual audio interface name.
	    const char *interface,
	    /// IO mode: input or output.
	    DI_IoMode mode,
	    /// Audio port sample format.
	    DI_AudioSampleFormat format, 
	    /// Audio port buffer size (sample frames).
	    uint bufferSize = 0,
	    /// Number of channels.
	    uint channels = 0);

  /// Close the virtual audio port.

  bool close(void);

  //@}
  
  /**@name I/O operations */
  
  //@{

  /** Read sample frames from the interface.

      This method reads 'frames' sample frames from the virtual interface,
      and stores them into buffer pointed to by 'buffer'.

      Please note that one sample may take more than one byte, so
      'buffer' must point to a buffer of capable of holding 'frames'
      * N * M bytes, where N is the size of one sample and M is number of
      channels.
     
      The number of sample frames actually recorded is
      returned, or a negative value if the operation failed.
  */

  int readFrames(void *buffer, uint frames);

  /** Write sample frames to the interface.

      This method writes 'frames' sample frames to the virtual interface,
      reading them from buffer pointer to by 'buffer'. The number
      of sample frames actually written is returned, or a negative value if
      the operation failed. 
  */

  int writeFrames(const void *buffer, uint frames);
  
  /** Wait until 'frames' # of frames can be read.
      
      Return value tells how many frames are filled as the call
      returns. Negative values indicate error of some sort.
  */
  
  int waitInputFilled(uint frames, int timeOut);

  /// Wait until output has only the number of frames specified filled.  

  int waitOutputFilled(uint frames, int timeOut);

  /// Send End-Of-File condition.

  void sendEOF(void) { if (m_info) m_info->isEOF = 1; }

  //@}

  /**@name Predicates */
  
  //@{

  ///

  bool isOK(void) const;

  /// True if End-Of-File condition was met.

  bool isEOF(void) const
  { return (m_info ? (m_info->isEOF && !filledFrames()) : true); }

  //@}

private:

  // Get a key from a name.
  key_t getKey(const char *name);

  /// Get number of filled sample frames in the sample buffer
  int filledFrames(void) const;

  /* Semaphore P (decrement, negative count) or
     V (increment, positive count) operation. */
  void semOp(short semNum, short count)
  {
    m_sop.sem_num = semNum;
    m_sop.sem_op = count;
    m_sop.sem_flg = 0;
    semop(m_semID,&m_sop,1);
  }

  char                 *m_name;
  void                 *m_bufBegin;
  short                *m_bufBeginS;
  int                  *m_bufBeginI;
  float                *m_bufBeginF;
  double               *m_bufBeginD;
  DI_VaiInfo           *m_info;
  
  int                   m_shmID;
  int                   m_semID;
  sembuf                m_sop;
};

#endif  // DI_VIRTUAL_AUDIO__H
