/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_ARRAY2D_H
#define MJ_ARRAY2D_H

#include <di_types.h>
#include <mj_allocator.h>


/** Simple class to hold information of a 2-D data matrix. */
template <class TElem> class MJ_Array2D
{
public:
  MJ_Array2D() : m_rows(0), m_cols(0), m_data(0) {}
  MJ_Array2D(const MJ_Array2D &that) 
  {
    m_rows = that.m_rows;
    m_cols = that.m_cols;
    
    uint total = m_rows * m_cols;
    
    if(total) {
      m_data = new TElem[total];
      for(uint i=0; i < total; i++)
	m_data[i] = that.m_data[i];
    }
    else
      m_data = 0;
  }

  virtual ~MJ_Array2D() { if(m_data) delete [] m_data; }
  bool resize(uint rows, uint cols, bool fatalOnFailure = true);

  void argCheck(uint row, uint col) const;
  
  TElem &get(uint row, uint col);
  TElem &getFast(uint row, uint col)
  { return m_data[row * m_cols + col]; }

  TElem *getRow(uint row) { argCheck(row, 0); return & m_data[row * m_cols]; }

  TElem *getLastRow() { return & m_data[(m_rows - 1) * m_cols]; }

  TElem *begin() { return m_data; }
  const TElem *begin() const { return m_data; }
  TElem *end() { return m_data + (m_rows * m_cols); }
  const TElem *end() const { return m_data + (m_rows * m_cols); }

  inline void setAll(const TElem &v)
  {
    const TElem *max = end();
    TElem *d = m_data;
    while(d < max) *d++ = v;
  }

  /** @name DANGEROUS methods. */
  //@{

  /// Buffer size in bytes
  inline uint dataSize() const { return m_rows * m_cols * sizeof(TElem); }

  /** Give an array to this object. Once this method is called the
      object will consider the data its own. This method DOES delete
      old array values. */
  inline void adoptData(TElem *data, uint rows, uint cols)
  { if(m_data) delete [] m_data; m_data = data; m_rows = rows; m_cols = cols; }
  
  /** Drop pointers to the array WITHOUT deleting the data. */
  inline void dropData() { m_rows = 0; m_cols = 0; m_data = 0; }
  
  inline void customResize(MJ_Allocator *allocator, uint rows, uint cols)
  {
    if(allocator) {
      if(rows && cols) {
	m_data = (TElem *) allocator->allocate(rows * cols * sizeof(TElem));
	if(m_data) {
	  m_rows = rows;
	  m_cols = cols;
	}
	else {
	  m_rows = 0;
	  m_cols = 0;
	}
      }
      else {
	m_data = 0;
	m_rows = 0;
	m_cols = 0;
      }
    }
    else {
      resize(rows, cols);
    }
  }

  //@}

  MJ_Array2D<TElem> &operator = (const MJ_Array2D &that) 
  { 
    uint total = that.m_rows * that.m_cols;

    if(total != m_rows * m_cols) {

      delete [] m_data;

      m_rows = that.m_rows;
      m_cols = that.m_cols;
      
      if(total) {
	m_data = new TElem[total];
      }
      else
	m_data = 0;
    }

    for(uint i=0; i < total; i++)
      m_data[i] = that.m_data[i];
    
    return *this; 
  }

  inline uint rows() const { return m_rows; }
  inline uint cols() const { return m_cols; }
  inline uint elements() const { return m_cols * m_rows; }
protected:
  uint m_rows, m_cols;

  TElem *m_data;
  
};


#endif
