/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_DSP_PLUGIN_FILTERS_H
#define MJ_DSP_PLUGIN_FILTERS_H

#include <mj_plugin_filter.h>
#include <mj_plugin_base.h>

#include <mj_logic.h>

#include <qptrlist.h>

/** A class that accepts all plugins. */
class MJ_DspPluginFilterAny : public MJ_PluginFilter
{
public:
  MJ_DspPluginFilterAny() {}
  virtual ~MJ_DspPluginFilterAny() {}
  
  virtual bool accept(MJ_PluginDbeBase *dbe);
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** A class that accepts plugins with no input channels. */
class MJ_DspPluginFilterNoInput : public MJ_PluginFilter
{
public:
  MJ_DspPluginFilterNoInput() {}
  virtual ~MJ_DspPluginFilterNoInput() {}
  
  virtual bool accept(MJ_PluginDbeBase *dbe);
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** A class that accepts plugins with no output channels. */
class MJ_DspPluginFilterNoOutput : public MJ_PluginFilter
{
public:
  MJ_DspPluginFilterNoOutput() {}
  virtual ~MJ_DspPluginFilterNoOutput() {}
  
  virtual bool accept(MJ_PluginDbeBase *dbe);
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** A class that accepts plugins that has given metafunctionality. */
class MJ_DspPluginFilterMetaFunc : public MJ_PluginFilter
{
public:
  MJ_DspPluginFilterMetaFunc(const char *metaFunc, bool negate = false);
  virtual ~MJ_DspPluginFilterMetaFunc();
  
  virtual bool accept(MJ_PluginDbeBase *dbe);
protected:

  char *m_str;
  bool  m_negate;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** A class that accepts plugins that has given parameter. */
class MJ_DspPluginFilterParam : public MJ_DspPluginFilterMetaFunc
{
public:
  MJ_DspPluginFilterParam(const char *paramStr, 
			  MJ_PLuginParameter param,
			  bool negate = false);
  virtual ~MJ_DspPluginFilterParam();
  
  virtual bool accept(MJ_PluginDbeBase *dbe);
protected:
  MJ_PLuginParameter m_param;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** A filter that combines multiple filters to one group. 

    Objects of this class must have at least one child filter before
    they can operate.

    \b Tip: You can build a tree with many multi-filters. By using
    different logical operators you can build arbitrary logical
    expressions.  */

class MJ_PluginFilterMulti : public MJ_PluginFilter
{
public:
  /// Creates a multifilter.
  MJ_PluginFilterMulti(
		       /// The logic to be used when combining results
		       MJ_Logic logic = MJ_LOGIC_AND, 
		       /// Whether autodelete list contents 
		       bool autoDelete = true);

  virtual ~MJ_PluginFilterMulti();
  
  virtual bool accept(MJ_PluginDbeBase *dbe);

  /** Adds a filter to the list. */
  void addFilter(MJ_PluginFilter *);

  /** Removes all child widgets. The children are deleted if
      autodeletion is enabled. */
  void clear() { m_filters.clear(); }

  /// Returns the logic currently used
  MJ_Logic logic() const { return m_logic; }
  /// Sets the logic to be used.
  void setLogic(MJ_Logic l) { m_logic = l; }

protected:
  
  QList<MJ_PluginFilter> m_filters;
  MJ_Logic               m_logic; 
};

#endif
