/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */


#ifndef MJ_SATELLITE_H
#define MJ_SATELLITE_H

#include <mj_undo_info.h>
#include <mj_message.h>

#include <qobject.h>
#include <qptrlist.h>
#include <qstringlist.h> 


class MJ_Base;

/** Satellite for MJ_ModuleDsp.
 
    The satellite provides a possiblity to use a common QObject

    <B>Q: Why ? </B>

    <B>A: </B> There is a real problem associated with the signal-slot
    mechanism and threading. The GUI objects must not access the
    signal-slot mechanism of the DSP object directly. Yet the
    ss-system is very handy method to distribute information between
    UGI obejcts. For example: When one DSP object has multiple GUIs
    open the GUIs can communicate changes via sending "change
    messsages" to each other. This is done in centralized fashion via
    the satellite object.

    Typically the GUI oobject will 1) Change some parameter of the
    DSP-object. 2) call dspobject->satellite()->change(). Other
    potential GUI objects will receive this message and can adjust
    themselves.

    While this class is currently only used in conjunction with
    MJ_ModuleDsp it is fit for other purposes as well.

    @see MJ_ModuleDsp */
class MJ_Satellite  : public QObject
{
  Q_OBJECT

public:

  MJ_Satellite(MJ_Base *);

  virtual ~MJ_Satellite();

  void addUndoInfo(MJ_UndoInfo *);

  void addRedoInfo(MJ_UndoInfo *);
  /// Get list of the undo items.
  QList<MJ_UndoInfo> &undos() { return m_undos; }

  /// The host object this satellite refers to
  MJ_Base *host() { return m_host; }

  bool undo(MJ_UndoInfo *);

  bool redo(MJ_UndoInfo *);

  /** The maximum number of undos kept in memory. The undo information
      needs to be removed at some stage to avoid infinite memory
      usage. This number limits the amount of undos in memory. */
  uint maxUndos() const { return m_maxUndos; }

  /// Sets the maximum number of undos (or redos) that will be saved.
  void setMaxUndos(uint m) 
  { m_maxUndos = m; limitUndos(m); limitRedos(m); }

  void incrementUiCount();

  void decrementUiCount();

  void pushMruFile(const QString &);

  const QStringList & mruFiles() { return m_mruFiles; }

  uint nofUndos() { return m_undos.count(); }
  uint nofRedos() { return m_redos.count(); }
  MJ_UndoInfo *undoInfo(uint i) { return m_undos.at(i); }
  MJ_UndoInfo *redoInfo(uint i) { return m_redos.at(i); }

public slots:

  void nextDelete();

  void modified(const char *);
  void ioModified(uint);

signals:

  void emitNextDelete();

  /// This signal is emitted by "modified".
  void emitModified(const char *);
  /// This signal is emitted as the IO setup is somehow modified.
  void emitIoModified(uint);

  /// Signals that a piece of undo information was added
  void emitUndoAdded(MJ_UndoInfo *);
  /// Signals that a piece of undo information was removed
  void emitUndoRemoved(MJ_UndoInfo *);

  /// Signals that a piece of redo information was added
  void emitRedoAdded(MJ_UndoInfo *);
  /// Signals that a piece of redo information was removed
  void emitRedoRemoved(MJ_UndoInfo *);

  /** This signal is emitted as the satellite dies. */
  void destroyed(QObject *);

  /** This signal is emitted as a new file is pushed to the MRU
      stack. */
  void mruChange();

protected:

  void limitUndos(uint);

  void limitRedos(uint);

  void undoToRedo(MJ_UndoInfo *);

  void redoToUndo(MJ_UndoInfo *);

  /** List of undo packets.
      
      @see MJ_UndoInfo 
   */
  QList<MJ_UndoInfo>  m_undos;

  /** List of redo packets.
      
      @see MJ_UndoInfo 
   */
  QList<MJ_UndoInfo>  m_redos;
  
  /** The maximum number of undo/redo operations to keep in memory. 
      
      Both lists may reach this size, resulting in 2 * m_maxUndos
      operations in memory.  */
  uint                m_maxUndos;
  
  /** The object whose satellite this is. The object should always
      outlive its satellite. */
  MJ_Base            *m_host;

  /** Keeps track of how many GUIs arre currently open. */

  int                 m_guiCount;

  /** List of most recently used files. */
  QStringList     m_mruFiles;
};

#endif
