/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_STRING_H
#define MJ_STRING_H

#include <string.h>

/* STRING CLASSES!!!! WHY ON EARTH!!!!!
   
   These are special string classes that should only be used in the
   moments of real need. They are are geared towards real-time
   usage. That is they do not do <B>ANY</B> memory management. Instead
   you can use them to allocate memory via stack or some other way.

   
*/

/** Base class for character (C-style) strings. */
class MJ_CStringBase
{
public:
  MJ_SCStringBase() {}
  virtual ~MJ_SCStringBase() {}

  virtual const char *ascii() const = 0;
  virtual unsigned length() const = 0;
  
};

inline bool operator == (const MJ_CStringBase &s1, const MJ_CStringBase &s2)
{
  return !strcmp(s1.ascii(), s2.ascii());
}

inline bool operator < (const MJ_CStringBase &s1, const MJ_CStringBase &s2)
{
  return strcmp(s1.ascii(), s2.ascii()) < 0;
}

inline bool operator > (const MJ_CStringBase &s1, const MJ_CStringBase &s2)
{
  return strcmp(s1.ascii(), s2.ascii()) > 0;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** String class with fixed string length. */
template <unsigned size>
class MJ_SCString : public MJ_CStringBase
{
public:
  MJ_SCString() : m_length(0) { m_data[0] = '\0'; }
  MJ_SCString(const char *s) { *this = s; }
  virtual ~MJ_SCString() {}
  
  virtual const char *ascii() const { return m_data; }
  virtual unsigned length() const { return m_length; }

  inline MJ_SCString & operator = (const char *s);
  
  char *data() { return m_data; }

protected:
  char m_data[size + 1];
  unsigned m_length;
};

/** Copies the string here. If the argument string is too long, then
    it not copied at all. */

inline MJ_SCString & MJ_SCString::operator = (const char *s)
{
  if(!s) {
    m_data[0] = '\0'; 
    m_length = 0;
  }
  else {
    l = strlen(s);
    if(l > size) {
      m_data[0] = '\0'; 
      m_length = 0;
    }
    else {
      m_length = l;
      strcpy(m_data, s);
    }
  }
  return *this;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Virtual/volatile string class. Only holds a pointer to an external
    string. */
class MJ_VCString : public MJ_CStringBase
{
public:
  MJ_VCString() : m_data(0), m_length(0) {}
  MJ_VCString(const char *s) : m_data(s), m_length(s ? strlen(s) : 0) {}
  virtual ~MJ_VCString() {}
  
  virtual const char *ascii() const { return m_data; }
  virtual unsigned length() const { return m_length; }

  inline MJ_VCString & operator = (const char *s)
  { m_data = s; m_length = s ? strlen(s) : 0; return *this; }
  
  char *data() { return m_data; }

protected:
  char *m_data;
  unsigned m_length;
};

#endif
