/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_MESSAGE_H
#define MJ_MESSAGE_H


#include <mj_config.h>
#include <mj_message_enums.h>

#include <string>

class MJ_Parameters;
class MJ_MessageCarrier;
class MJ_ModuleDsp;

template <class TElem> class MJ_MessageNumeric;

typedef MJ_MessageNumeric<double> MJ_MessageDouble;
typedef MJ_MessageNumeric<int> MJ_MessageInt;

class MJ_Message;
class MJ_MessageString;

class MJ_MessageMidiEvent;
class MJ_MessageMidiSysex;

class DI_DataBuffer2;

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Base class for messages. Messages are the generic way to pass any
    data between two objects. This class is inherited by a number of
    child classes that incorporate different data to the message.

    <B>Time stamps:</B> Messages are timestamped. The time stamp unit
    is master sampling rate. As long as the sample rates used in the
    system are below the master sampling rate this guarantees
    sample-accurate timestamps.

    @author Tommi Ilmonen

*/
class MJ_Message
{
public:

  typedef long stamp_t;

  /** Contruct a message. */
  MJ_Message(long time = -1) : m_timeStamp(time) 
  {
#ifdef MJ_MEM_DEBUG
    incrCount();
#endif
  }

  /** Construct a message. */
  MJ_Message(const MJ_Message &m)
    : m_timeStamp(m.m_timeStamp)
  {
#ifdef MJ_MEM_DEBUG
    incrCount();
#endif
  }

  /** Destruct a message. */
  virtual ~MJ_Message()
  {
#ifdef MJ_MEM_DEBUG
    decrCount();
#endif
  }

  /** @name Attempt conversion to numeric values */
  //@{
  virtual double toDouble(bool *ok = 0);
  virtual int    toInt(bool *ok = 0);
  virtual bool   toBool(bool *ok = 0);
  //@}

  virtual MJ_MessageType messageType() const;
  virtual MJ_Message *duplicate();
    
  virtual bool write2(DI_DataBuffer2 *b);
  virtual bool read2(DI_DataBuffer2 *b, int readFields);

  static MJ_Message *readFromBuffer(DI_DataBuffer2 *b);
  
  /** Describe the message contents. */
  virtual std::string describe();

  /** @name Time stamps. */ 
  //@{
  /// Message time stamp
  stamp_t timeStamp() const { return  m_timeStamp; }
  /// Set message time stamp
  void setTimeStamp(stamp_t time) { m_timeStamp = time; }
  //@}

  /** @name Cast to other message types. 

      If the object is of wrong type null is returned */
  //@{
  MJ_MessageDouble  *toMessageDouble();
  MJ_MessageInt     *toMessageInt();
  MJ_MessageString  *toMessageString();
  //@}

  static MJ_Message *createNew();

  /** The number of messages in the system. This value is valid only
      if macro MJ_MEM_DEBUG has been used during compilation. */
  static int msgCount() { return m_counter; }

  static void setDebug(bool debug) { m_debug = debug; }
protected:

  static void incrCount();
  static void decrCount();

  /// Event time stamp.
  stamp_t        m_timeStamp;

  static int     m_counter;

  static bool    m_debug;
};


#endif
