/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_ASR_H
#define MJ_ASR_H

#include <di_types.h>

/** Attack-sustain-release (ASR) interpolator. 
 *
 *
 * @author Tommi Ilmonen.
 *
 */
template <class T>
class MJ_Asr
{
public:
  
  enum State {
    ATTACK,
    SUSTAIN,
    RELEASE
  };
  
  MJ_Asr() { setTimes(0, (uint) -1, 0); rewind(); }
  ~MJ_Asr() {}

  /** 
   * Sets the time values of the ASR envelope.
   * 
   * @param attack Attack time in samples
   * @param sustain Sustain time in samples
   * @param release Release time in samples
   */

  void setTimes(uint attack, uint sustain, uint release)
  {
    m_attackTime = attack;
    m_sustainTime = sustain;
    m_releaseTime = release;
  }

  /** 
   * Rewinds the ASR envelope. Should be called after #setTimes.
   * 
   */

  void rewind() 
  { 

    if(m_attackTime != 0) {
      m_value = 0.0; 
      m_delta = 1.0 / m_attackTime;
      m_left = m_attackTime; 
      m_state = ATTACK;
    }
    else if(m_sustainTime != 0) {
      m_value = 1.0;
      m_delta = 0.0;
      m_left = m_sustainTime; 
      m_state = SUSTAIN;
    }
    else {
      m_left = (uint) -1;
      m_value = 0.0;
      m_delta = 0.0;
      m_state = RELEASE;
    }
  }

  float get() const { return m_value; }

  void enterSustain()
  {
    m_state = SUSTAIN;
    m_value = 1.0;
    m_delta = 0.0;
    m_left = m_sustainTime;
  }

  void enterRelease()
  {
    m_state = RELEASE;
    if(m_releaseTime) {
      m_delta = -1.0 / m_releaseTime;
      m_left = m_releaseTime;
    }
    else {
      m_delta = 0.0;
      m_value = 0.0;
      m_left = (uint) -1;
    }
  }

  void advance()
  {
    m_left--;
    m_value += m_delta;
    
    if(!m_left) {
      if(m_state == ATTACK)
	enterSustain();
      else if(m_state == SUSTAIN) {
	enterRelease();
      }
    }
  }

  void advanceN(uint n)
  {
    if(n < m_left) {
      m_left -= n;
      m_value += (n * m_delta);
    }
    else {
      // uint tmp = n - m_left;
      n -= m_left;

      if(m_state == ATTACK) {
	enterSustain();

	if(n >= m_left) {
	  n -= m_left;
	  enterRelease();
	}
      }
      else {
	enterRelease();
      }

      if(n > m_left)
	n = m_left - 1;

      m_left -= n;
      m_value += (n * m_delta);
    }
  }

  
  State state() const { return m_state; }
  uint left() const { return m_left; }

protected:
  State m_state;

  T    m_value;
  T    m_delta;

  uint m_left;
  uint m_attackTime;
  uint m_sustainTime;
  uint m_releaseTime;
};

#endif
