/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_DELAY_LINE_H
#define MJ_DELAY_LINE_H

#include <mj_ramp.h>
#include <mj_ringbuffer.h>

/** A delay line with fractional delay-time interpolation. 

    @see MJ_RingBuffer2

    @author Tommi Ilmonen
*/

template <class T> class MJ_DelayLine 
{
public:

  void prepare(unsigned maxdelay, const T & dfault)
  {
    m_delay.resize(maxdelay + 3);
    m_delay.reset(); 
    m_delay.setAll(dfault); 
  }

  void clear()
  { m_delaySamples.reset(0.0); m_delay.reset(); }

  void setDelay(double d) { m_delaySamples.reset(d); }
  void interpolateDelay(double d, unsigned n) 
  { m_delaySamples.setTarget(n, d, MJ_RAMP_LINEAR); }
 
  void put(const T &v) 
  {
    m_delaySamples.advanceLinear();
    m_delay.put(v);
  }
 
  T get() const 
  { return m_delay.getInterpolatedNewest1D(m_delaySamples.get()); }

  void process(T *target, const T *source, unsigned n);

protected:

  MJ_RingBuffer2<T> m_delay;
  MJ_Ramp<double>   m_delaySamples;
};

/** A delay line with fractional delay-time interpolation and gain control. 

    @see MJ_RingBuffer2
    @see MJ_Ramp

    @author Tommi Ilmonen
*/

template <class T> class MJ_DelayLine2
{
public:

  void prepare(unsigned maxdelay, const T & dfault)
  {
    m_delay.resize(maxdelay + 3);
    m_delay.reset(); 
    m_delay.setAll(dfault); 
  }

  void clear()
  { m_delaySamples.reset(0.0); m_delay.reset(); }

  void setDelay(double d) { m_delaySamples.reset(d); }
  void interpolateDelay(double d, unsigned n) 
  { m_delaySamples.setTarget(n, d, MJ_RAMP_LINEAR); }
 
  void setGain(double d) { m_gain.reset(d); }
  void interpolateGain(double d, unsigned n) 
  { m_gain.setTarget(n, d, MJ_RAMP_LINEAR); }

  double gainTarget() const { return m_gain.target(); }

  void put(const T &v) 
  {
    m_delaySamples.advanceLinear();
    m_delay.put(v);
  }
 
  T get() const 
  { return m_delay.getInterpolatedNewest1D(m_delaySamples.get()); }

  void process(T *target, const T *source, unsigned n);

protected:

  MJ_RingBuffer2<T> m_delay;
  MJ_Ramp<double>   m_delaySamples;
  MJ_Ramp<double>   m_gain;
};


#endif
