/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_FILTER_H
#define MJ_FILTER_H

#include <mj_config.h>

/** First order all-pole IIR filter. 
    
    Useful for simple low-pass or high-pass filter applications.  */
template <class TElem> class MJ_FilterAP1
{
public:
  MJ_FilterAP1() : m_a1(0.0) , m_b0(1.0), m_mem(0.0) {}
  ~MJ_FilterAP1() {} 

  /** Set filter coefficients. */
  void set(TElem a1, TElem b0) { m_a1 = a1; m_b0 = b0; }

  /** Coefficient a1 (the feedback coefficient). */
  inline TElem a1() const { return m_a1; }
  /** Coefficient b0 (amplification). */
  inline TElem b0() const { return m_b0; }

  /** Resets the memory area. */
  void resetMemory() { m_mem = 0.0; }

  void resetMemory(TElem outputValue) 
    { m_mem = outputValue / m_b0; }
  
  /// Put a single sample to the filter.
  void put(const TElem &s) { m_mem = s - m_mem * m_a1; }

  inline TElem putGetSample(TElem s)
  { put(s); return get(); }

  inline TElem putGet(TElem s)
  { put(s); return get(); }
  

  /** Put many samples in without calculating output. */
  void putMany(const float * samples, unsigned nofSamples);
  /** Put many samples in without calculating output. */
  void putMany(const double * samples, unsigned nofSamples);
  
  /** Put many samples in and calculate output. */
  void putMany(float * outsamples, const float * insamples, 
	       unsigned nofSamples);

  /** Put many samples, but raise each sample to the power of two
      before feeding it to the filter. */

  void putManySquared(const float * samples, unsigned nofSamples);

  /** Get current value. Note: The value will we calculated as this
      method is called -> You may want to cache the result in a local
      variable if you are using the value many times in the same
      function. */
  TElem get() const { return m_mem * m_b0; }

  /** Design a low pass filter. The -3db point will is in frequency
      cutOffFreq. The frequency is expressed in relation to Nyquist
      frequency - range [0, 1]. */
  void designLowPass(TElem cutOffFreq);

  /** Design a low pass filter with given cut-off frequency and
      sampling rate. */
  void designLowPass2(TElem cutOffFreq, TElem sr)
  { designLowPass((cutOffFreq * (TElem) 2.0) / sr); }

  /// Design a high-pass filter.
  void designHighPass(TElem cutOffFreq)
  { designLowPass(1.0 - cutOffFreq); m_a1 *= -1.0; }

  /** Multiply output gain by given factor. Can be used to set
      zero/Nyquist frequency gain. */
  void mulGain(TElem mul) { m_b0 *= mul; }
protected:
  TElem m_a1, m_b0, m_mem;
};

#endif
