/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_FILTER_IIR2_H
#define MJ_FILTER_IIR2_H

#include <mj_math.h>

/** Linear second order IIR filter. Rather optimized. 
 
    The filter coefficients are normalized like in Matlab - the
    coefficient a0 is assumed to be 1. As such it not stored or used
    in calculation. 

    @see MJ_FilterDFII MJ_FilterAllPole1
    @author Tommi Ilmonen
*/

template <class TElem> class MJ_FilterIIR2
{
public:
  /** Empty constructor. Constructor does \b not initialize any
      values. */
  MJ_FilterIIR2() {}

  /// Empty destructor
  ~MJ_FilterIIR2() {}

  /** @name Filter usage: */
  //@{

  /** Puts a sample to the filter. Changes the internal state, but
      does not calculate the output value. */
  inline void putSample(TElem s)
  {
    m_m2 = m_m1;
    m_m1 = m_m0;
    m_m0 = s - m_a1 * m_m1 - m_a2 * m_m2;
  }

  inline void put(TElem s) { putSample(s); }

  /** Get the current sample. Does not change the filter state, but
      causes computation. */
  inline TElem getSample() const
  {
    return m_b0 * m_m0 + m_b1 * m_m1 +  m_b2 * m_m2;
  }

  inline TElem get() const { return getSample() ; }

  /** Put sample and get output sample. */
  inline TElem putGetSample(TElem s)
  {
    m_m2 = m_m1;
    m_m1 = m_m0;
    m_m0 = s - m_a1 * m_m1 - m_a2 * m_m2;
    return m_b0 * m_m0 + m_b1 * m_m1 +  m_b2 * m_m2;
  }

  /** Put sample and get output sample on a foreign filter. The filter
      coeeficients of "this" filter are used on the memory of the
      other filter.*/
  inline TElem putGetSampleOther(MJ_FilterIIR2 *f, TElem s)
  {
    f->m_m2 = f->m_m1;
    f->m_m1 = f->m_m0;
    f->m_m0 = s - m_a1 * f->m_m1 - m_a2 * f->m_m2;
    return m_b0 * f->m_m0 + m_b1 * f->m_m1 +  m_b2 * f->m_m2;
  }

  void putSamples(const TElem *input, unsigned count);

  void putSamples(TElem * output, 
		  const TElem * input, 
		  unsigned count);

  void putGetSamplesInplace(TElem *inputOutput, unsigned count);

  /// Clear internal buffer
  inline void resetMemory()
  { m_m0 = 0.0; m_m1 = 0.0; m_m2 = 0.0; }

  //@}

  /** @name Filter design methods: */
  //@{

  /** Makes the filter have flat frequency response. */
  inline void flatResponse()
  {
    m_b0 = 1.0;
    m_b1 = 0.0;
    m_b2 = 0.0;
    m_a1 = 0.0;
    m_a2 = 0.0;    
  }

  /** Design a high-shelving EQ. See MJ_dspHighShelveEq.*/
  inline void designHighShelve1(TElem centerRad, 
				TElem gainDb, 
				TElem slope = 1.0)
  { 
    iir2HighShelveEq
      (centerRad, gainDb, slope, bptr(), aptr());
  }

  /** Design a high-shelving EQ. See iir2HighShelveEq.  */
  inline void designHighShelve2(TElem frequencyHz, 
				TElem sampleRate,
				TElem gainDb, 
				TElem slope = 1.0)
  { 
    designHighShelve1(MJ_2pi * frequencyHz / sampleRate, gainDb, slope);
  }

  /** Design a high-shelving EQ. See iir2LowShelveEq. */
  inline void designLowShelve1(TElem centerRad, 
			       TElem gainDb, 
			       TElem slope = 1.0)
  { 
    iir2LowShelveEq
      (centerRad, gainDb, slope, bptr(), aptr());
  }

  /** Design a high-shelving EQ. See iir2LowShelveEq. */
  inline void designLowShelve2(TElem frequencyHz, 
			       TElem sampleRate,
			       TElem gainDb, 
			       TElem slope = 1.0)
  { 
    designLowShelve1(MJ_2pi *frequencyHz / sampleRate, gainDb, slope);
  }

  /** Design a parametric EQ. This algorithm is not too flashy. It
      often creates numerically unstable filters. I have not removed
      this method since it may be of some use to someone... */
  
  void designParametric1(TElem boostAbs, 
			 TElem centerRad,
			 TElem bandWidthRad);

  void designParametric2(TElem boostDb, 
			 TElem centerHz,
			 TElem bandWidthOctaves,
			 TElem samplingRate);

  /** Design a parametric EQ. The algorithm is fairly ok. */

  void designParametricB1(TElem K,
			  TElem F,
			  TElem omega, 
			  TElem bw)
  { iir2ParametricEqB(K, F, omega, bw, bptr(), aptr()); }

  void designParametricB2(
			  /// Center frequency (Hz)
			  TElem hz, 
			  /// Sampling rate (Hz)
			  TElem sr,
			  /// Gain (decibels)
			  TElem gainDb,
			  /// Bandwidth (octaves)
			  TElem bw)
  { 
    if(MJ_abs(gainDb) < 0.005) flatResponse();
    else designParametricB1(MJ_pow(10.0, gainDb * 0.05), 
			    MJ_pow(10.0, gainDb * 0.025), 
			    hz / sr * MJ_2pi, bw); 
  }
  
  /** Design a parameteric filter that mimics analogue filters. 

      Algorithm by S. J. Orfanidis: "Digital Parametric Equalizer
      Design with Prescribed Nyquist Frequency Gain", J. Audio
      Eng. Soc., vol.45, p.444, June 1997.  
      
      The algorithm creates fairly ok frequency responses in most
      cases. The unfortunate exception: When the filter center
      frequency is high and the filter bandwidth is great the designed
      filter often has a rather poor curve shape. */

  void designParametricC1(
			  /// Center gain
			  TElem K,
			  /// Bandwidth gain
			  TElem F,
			  /// Band center radians/sample
			  TElem omega, 
			  /// Bandwidth in octaves
			  TElem bw)
  { iir2ParametricEqC(K, F, omega, bw, bptr(), aptr()); }

  /** Design a parameteric filter that mimics analogue filters. This
      is a utility method that uses designParametricC1 to do the dirty
      work.*/

  void designParametricC2(
			  /// Center frequency (Hz)
			  TElem hz, 
			  /// Sampling rate (Hz)
			  TElem sr,
			  /// Gain (decibels)
			  TElem gainDb,
			  /// Bandwidth (octaves)
			  TElem bw)
  { 
    if(MJ_abs(gainDb) < 0.005) flatResponse();
    else designParametricC1(MJ_pow(10.0, gainDb * 0.05), 
			    MJ_pow(10.0, gainDb * 0.025), 
			    hz / sr * MJ_2pi, bw); 
  }

  /// Design a low-pass filter
  void designLowPass1(TElem omega,
		      TElem Q)
  {
    iir2LowPassEq(omega, Q, &m_b0, &m_a1);
  }

  /// Design a low-pass filter
  void designLowPass2(
		      /// Cut-off frequency (Hz)
		      TElem hz, 
		      /// Sampling rate (Hz)
		      TElem sr,
		      TElem Q)
  {
    designLowPass1(hz / sr * MJ_2pi, Q);
  }

  /// Design a high-pass filter
  void designHighPass1(TElem omega,
		      TElem Q)
  {
    iir2HighPassEq(omega, Q, &m_b0, &m_a1);
  }

  /// Design a high-pass filter
  void designHighPass2(
		      /// Cut-off frequency (Hz)
		      TElem hz, 
		      /// Sampling rate (Hz)
		      TElem sr,
		      TElem Q)
  {
    designHighPass1(hz / sr * MJ_2pi, Q);
  }

  /// Design a band-pass filter
  void designBandPass1(
		       /// Center frequency (radians)
		       TElem omega, 
		       /// Q-value
		       TElem Q)
    { iir2BandPassEq(omega, Q, bptr(), aptr()); }

  /// Design a band-pass filter
  void designBandPass2(
		       /// Center frequency (Hz)
		       TElem hz, 
		       /// Sampling rate (Hz)
		       TElem sr,
		       /// Q-calue
		       TElem Q)
    { designBandPass1(hz / sr * MJ_2pi, Q); }

  /// Design a band-pass filter
  void designBandPassB1(
		       /// Center frequency (radians)
		       TElem omega, 
		       /// Bandwidth
		       TElem bw)
    { iir2BandPassEqB(omega, bw, bptr(), aptr()); }

  /// Design a band-pass filter
  void designBandPassB2(
			/// Center frequency (Hz)
			TElem hz, 
			/// Sampling rate (Hz)
			TElem sr,
			/// Bandwidth
			TElem bw)
    { designBandPassB1(hz / sr * MJ_2pi, bw); }

  /// Copy filter coefficients from the given filter.
  inline void copyCoefficients(const MJ_FilterIIR2 &iir)
  {
    m_b0 = iir.m_b0;
    m_b1 = iir.m_b1;
    m_b2 = iir.m_b2;

    m_a1 = iir.m_a1;
    m_a2 = iir.m_a2;
  }

  /// Multiply b-coefficients by factor scale.
  void mulGain(TElem scale)
  { m_b0 *= scale; m_b1 *= scale; m_b2 *= scale; }

  //@}

  /** @name Filter coefficient access: */
  //@{

  /// Filter coefficient b0
  inline TElem b0() const { return m_b0; }
  /// Filter coefficient b1
  inline TElem b1() const { return m_b1; }
  /// Filter coefficient b2
  inline TElem b2() const { return m_b2; }

  /// Returns pointer to the b-coefficient array (b0-b2)
  inline TElem *bptr() { return & m_b0; }

  /// Filter coefficient a1
  inline TElem a1() const { return m_a1; }
  /// Filter coefficient a2
  inline TElem a2() const { return m_a2; }

  /// Returns pointer to the a-coefficient array (a1-a2)
  inline TElem *aptr() { return & m_a1; }
  //@}

  void writeTo(char *);

  /** @name Miscellanous filter design functions. These are externally
   defined since someone might want to use them outside the
   MJ_FilterIIR2 class. */
  //@{

  static void iir2ParametricEq
    (TElem gr, TElem ga, TElem gb, TElem w0, TElem dw,
     TElem *b, TElem *a);
  
  static void iir2ParametricEqB
    (TElem K, TElem F, TElem omega, TElem bw, 
     TElem *b, TElem *a);
  
  static void iir2ParametricEqC
    (TElem K, TElem F, TElem omega, TElem bw, 
     TElem *b, TElem *a);
  
  static void iir2HighShelveEq
    (TElem rate, TElem gainDb, TElem slope,
     TElem *b, TElem *a);

  static void iir2LowShelveEq
    (TElem rate, TElem gainDb, TElem slope,
     TElem *b, TElem *a);

  static void iir2LowPassEq (TElem omega, TElem Q,
			     TElem *b, TElem *a);
  
  static void iir2HighPassEq (TElem omega, TElem Q,
			      TElem *b, TElem *a);

  static void iir2BandPassEq(TElem omega, TElem Q,
			     TElem *b, TElem *a);

  static void iir2BandPassEqB(TElem omega, TElem bw,
			      TElem *b, TElem *a);

  //@}

protected:
  TElem m_b0, m_b1, m_b2;
  TElem m_a1, m_a2;
  TElem m_m0, m_m1, m_m2;
};

#endif
