/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_LEVEL_METER_H
#define MJ_LEVEL_METER_H

/** A level meter. This class is completely inlined. 

    A level meter measures signal amplitude. It is assumed that the
    signal is always positive (take absolute value of your data,
    before feeding it here, if necessary).  */

template <class TElem> class MJ_LevelMeter
{
 public:
  /// Creates a new level meter (no further initializations needed)
  MJ_LevelMeter() {reset();}
  ~MJ_LevelMeter() {}

  /** Puts a new sample to the meter. */
  void put(
	   /// New sample (absolute value)
	   TElem vabs, 
	   /// Time to hold the peak value
	   unsigned holdTime) 
  {
    if(!--m_time) {
      m_time = holdTime;
      m_v1 = m_v2;
      m_v2 = 0.0;
    }
    
    if(vabs > m_v1) {
      m_time = holdTime;
      m_v2 = m_v1;
      m_v1 = vabs;
    }
    else if(vabs > m_v2) {
      m_v2 = vabs;
    }
  }

  /** Puts a new sample to the meter. */
  void put(
	   /// New sample
	   TElem value, 
	   /// Floor value
	   TElem vfloor, 
	   /// Time to hold the peak value
	   unsigned holdTime) 
  {
    if(!--m_time) {
      m_time = holdTime;
      m_v1 = m_v2;
      m_v2 = vfloor;
    }
    
    if(value > m_v1) {
      m_time = holdTime;
      m_v2 = m_v1;
      m_v1 = value;
    }
    else if(value > m_v2) {
      m_v2 = value;
    }
  }

  /// Resets level to zero.
  void reset() 
  { m_time = 1; m_v1 = m_v2 = 0.0; }

  /// Resets level to zero.
  void reset(TElem value) 
  { m_time = 1; m_v1 = m_v2 = value; }

  /// Current peak value
  const TElem &peak() const {return m_v1;}

protected:
  unsigned m_time;
  TElem m_v1, m_v2;
};

#endif
