/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of Mustajuuri LGPL modules.
 *
 * Author: Tommi Ilmonen, 1999-2004.
 * e-mail: firstname.lastname(at)hut.fi
 *
 * http://www.tml.hut.fi/~tilmonen/mustajuuri/
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. Extra clauses can be found from the file
 * MUSTAJUURI_LGPL and further explanations from the file
 * MUSTAJUURI_LICENSE. The license text can be found in file
 * "LGPL.txt".
 *
 * If you want a parallel license (for commercial reasons for example),
 * you should negotiate the matter with the author(s).
 *    
 */

#ifndef MJ_RANDOM_H
#define MJ_RANDOM_H

#include <mj_config.h>

#define MJ_MAX_UINT ((uint) 0xffffffff)

/** Random number generator. The internal base type is unsigned int. 

    Computationally efficient, but not such flash random
    numbers. Sufficiently good for many purposes tough.  */
class MJ_Random
{
  static unsigned m_randMul;

public:
  /** Creates a new random number generator. The seed is value is
      taken from internal seed counter. Initializing two random number
      generators successively does yield different sequences. */
  MJ_Random();
  MJ_Random(unsigned seed) : m_val(seed) {}
  ~MJ_Random() {}

  /** Returns a random 32-bit unsigned integer in range [0,
      max_unsigned].  This is the native method all other get*-methods
      turn to.  */
  inline unsigned get()
  { m_val = m_val * m_randMul + 1; return m_val; }
  
  /** Returns a 32-bit in integer range [min_int, max_int]. */
  inline int getInt() {return (int)get();}
  
  /** Returns a random double in range [0, 1]. */
  inline double getDouble0()
  { return  get() / (double) 0xffffffff; }
  
  /** Returns a random float in range [0, 1]. */
  inline float getFloat0()
  { return  get() / (float) 0xffffffff; }

  /** Returns a random double in range [-width/2, width/2] */
  double getDouble1(double width = 1.0)
  { return (((double) get()) / (double) MJ_MAX_UINT - 0.5) * width; }

  /** Returns a random float in range [-width/2, width/2] */
  float getFloat1(float width = 1.0)
  { return (((float) get()) / (float) MJ_MAX_UINT - 0.5f) * width; }

  /** Returns many samples in range [0, width] */
  void getManySamples0(MJ_Sample *data, 
		       unsigned count, 
		       MJ_Sample width = 1.0);

  /** Returns many samples in range [-width/2, width/2] */
  void getManySamples1(MJ_Sample *data, 
		       unsigned count, 
		       MJ_Sample width = 1.0);

  /** Gets a new seed value from the current system clock. 
      
      Works by calling gettimeofday and summing the second and
      microsecond part of the time and the incremental counter. 

      This method is not called by the constructor as it might take a
      CPU clock cycle too much.  */
  void getTimeSeed();

  void setSeed(unsigned seed) { m_val = seed; }

  static MJ_Random instance;
protected:
  /** Current seed value. */
  unsigned m_val;
};

#endif
