/* COPYRIGHT
 *
 * This file is part of FLUID
 *
 * Author: Tommi Ilmonen, Janne Kontkanen 1999-2002.
 * Tommi.Ilmonen@hut.fi, Janne.Kontkanen@hut.fi
 *
 * This file is licensed under the Lesser General Public License
 * (LGPL) - version 2.1. 
 *    
 */

#ifndef FLUID_VECTOR3T_HPP
#define FLUID_VECTOR3T_HPP

#include <FluidMath.hpp>
#include <FluidVector2.hpp>

#include <iostream>

namespace Fluid {

  template <class T>
  class Vector3T
  {
  public:
    T		x;										// x-component of the vector
    T		y;										// y-component of the vector
    T		z;										// z-component of the vector
    Vector3T()							   {}
    Vector3T(T xyz)		           { x = y = z = xyz; }
    Vector3T(T cx, T cy, T cz)		           { x = cx;	y = cy;		z = cz; }
    template <class S> Vector3T(const Vector3T<S>& v)		   { x = (T)v.x;	y = (T)v.y; z = (T)v.z; }
    template <class S> Vector3T(const Vector2T<S>& v, S az)		   { x = (T)v.x;	y = (T)v.y; z = az; }
    template <class S> Vector3T& operator=(const Vector3T<S>& v)	   { x = (T)v.x; y = (T)v.y; z = (T)v.z; return *this; }
    Vector3T&	clear		(void)				   { x = (T)(0);  y = (T)(0); z = (T)(0); return *this;	}
    Vector3T&	make		(T cx, T cy, T cz)  { x = (cx); y = (cy); z = (cz); return *this; }
    Vector3T&	make		(T xyz)			   { x = (xyz); y = (xyz); z = (xyz); return *this; }
    T * data() { return &x; }
    const T * data() const { return &x; }
    bool		operator==  (const Vector3T& src) const		   { return (x == src.x && y == src.y && z == src.z); }
    bool		operator!=  (const Vector3T& src) const		   { return !(x == src.x && y == src.y && z == src.z); }

    Vector3T      operator+	(const Vector3T& v) const { return Vector3T(x + v.x, y + v.y, z + v.z); }
    Vector3T      operator+	(T s)               const { return Vector3T(x + s, y + s, z + s); }
    Vector3T      operator-	(const Vector3T& v) const { return Vector3T(x - v.x, y - v.y, z - v.z); }
    Vector3T      operator-	(T s)               const { return Vector3T(x - s, y - s, z - s); }
    Vector3T      operator-	()                  const { return Vector3T(-x, -y, -z); }
    Vector3T      operator*	(T s) const               { return Vector3T(x * s, y * s, z * s); }
    Vector3T      operator/	(T s) const               { return Vector3T(x / s, y / s, z / s); }

    Vector3T&	operator+=	(const Vector3T& v)		   { x += v.x; y += v.y; z += v.z;  return *this; }
    Vector3T&	operator+=	(const T& v)		           { x += v; y += v; z += v;  return *this; }

    Vector3T&	operator-=	(const Vector3T& v)		   { x -= v.x; y -= v.y; z -= v.z;  return *this; }
    Vector3T&	operator-=	(const T& v)		           { x -= v; y -= v; z -= v;  return *this; }

    Vector3T&	operator*=	(T s)			   { x = (x*s), y = (y*s); z = (z*s); return *this; }
    Vector3T&	operator/=	(T s)			   { s = 1.0/s; x = (x*s), y = (y*s); z = (z*s); return *this; }

    bool		isOne		(void) const			   { return (x == 1.0f && y == 1.0f && z == 1.0f); }
    bool		isZero		(void) const			   { return (x == 0.0f && y == 0.0f && z == 0.0f); }
    bool          isFinite        (void) const { return finite(x) && finite(y) && finite(z); }
    double	length		(void) const			   { return Math::Sqrt(x*x+y*y+z*z); }
    double	lengthSqr	(void) const			   { return x*x+y*y+z*z; }
    Vector3T&	negate		(void)				   { x=-x; y=-y; z=-z; return *this; }
    Vector3T&	normalize	(double len = 1.0)		   { double l = length(); if (l!=0.0) *this *= (len/l); return *this; }
    Vector3T&	scale		(const Vector3T& v)		   { x *= v.x; y *= v.y; z *= v.z; return *this; }
    Vector3T&	descale		(const Vector3T& v)		   { x /= v.x, y /= v.y; z /= v.z; return *this; }
    Vector3T&	clampUnit	(void)				   { if(x <= (T)0.0) x = (T)0.0; else if(x >= (T)1.0) x = (T)1.0; if(y <= (T)0.0) y = (T)0.0; else if(y >= (T)1.0) y = (T)1.0; if(z <= (T)0.0) z = (T)0.0; else if(z >= (T)1.0) z = (T)1.0; return *this; }
    const	T&	operator[]	(int i) const	                   { return ((T*)this)[i]; }
    T&		operator[]	(int i)				   { return ((T*)this)[i]; }

    T&            get(int i)              			   { return ((T*)this)[i]; }
    void		set(int i, T v)              			   { ((T*)this)[i] = v; }

    T             maximum() const { T t = x > y ? x : y; return t > z ? t : z; }
    T             minimum() const { T t = x < y ? x : y; return t < z ? t : z; }

    template <class S>
    void copy(const S * data) { x = data[0]; y = data[1]; z = data[2]; }

    static Vector3T & cast(T * ptr) { return * (Vector3T *) ptr; }
    static const Vector3T & cast(const T * ptr) { return * (Vector3T *) ptr; }
    
    Vector2T<T> & vector2() { return * (Vector2T<T> *) this; }
    const Vector2T<T> & vector2() const { return * (Vector2T<T> *) this; }
    
  };

  typedef Vector3T<float> Vector3; 

  template <class T> 
  inline Fluid::Vector3T<T> operator* (T s, const Fluid::Vector3T<T>& v)
  { return v * s; }
  
} // namespace

template <class T>
inline T abs(Fluid::Vector3T<T> t)
{
  return t.length();
}

template <class S, class T>
inline T dot(const Fluid::Vector3T<S>& a, const Fluid::Vector3T<T>& b)
{
  return a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
}

template <class S, class T>
inline T dot2(const Fluid::Vector3T<S>& a, const Fluid::Vector2T<T>& b)
{
  return a[0]*b[0] + a[1]*b[1];
}

template <class S, class T>
inline T dot3(const Fluid::Vector3T<S>& a, const Fluid::Vector2T<T>& b)
{
  return a[0]*b[0] + a[1]*b[1] + a[2];
}

template <class T>
inline Fluid::Vector3T<T> operator - (const T & x, const Fluid::Vector3T<T> & t)
{
  return Fluid::Vector3T<T> (x - t.x, x - t.y, x - t.z);
}

template <class T>
inline Fluid::Vector3T<T> operator + (const T & x, const Fluid::Vector3T<T> & t)
{
  return Fluid::Vector3T<T> (x + t.x, x + t.y, x + t.z);
}

template <class T>
inline Fluid::Vector3T<T> cross(const Fluid::Vector3T<T>& a, const Fluid::Vector3T<T>& b)
{
  Fluid::Vector3T<T> v((a[1]*b[2])-(a[2]*b[1]),
		       -(a[0]*b[2])+(a[2]*b[0]),
		       (a[0]*b[1])-(a[1]*b[0]));
  return v;
}

template <class T>
inline std::ostream &operator<<(std::ostream &os, const Fluid::Vector3T<T> &t)
{
  os << t.x << ' ' << t.y << ' ' << t.z;
  return os;
}

template <class T>
inline std::istream &operator>>(std::istream &is, Fluid::Vector3T<T> &t)
{
  is >> t.x;
  is >> t.y;
  is >> t.z;
  return is;
}

using Fluid::operator *;

#endif

