/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_force_object.h> 

#include <vee_error.h>
#include <vee_renderer.h> 
#include <vee_scene.h> 
#include <vee_strs.h>
#include <vee_value.h> 

#include <typeinfo>

VEE_ForceObject::~VEE_ForceObject() 
{ 
  for(m_boxes.toFirst(); !m_boxes.isAtEnd(); m_boxes.advance()) {
    /* VEE_trace("VEE_ForceObject::~VEE_ForceObject # %p %p", 
       this, m_boxes.current()); */
    m_boxes.current()->forgetForce(this);
  }
}

/** Receive and process a value. 

    Parameters:

    <UL>
    
    <LI> "forcetype", int (VEE_ForceType)

    </UL>

*/
bool VEE_ForceObject::setValue(const char *id, VEE_Value *msg)
{
  if(VEE_Strs::equals(id, "forcetype")) {
    int v;

    bool ok = msg->getInt(&v);

    if(ok) {
      m_forceType = (VEE_ForceType) v;
    }
    return ok;
  }
  if(VEE_Strs::equals(id, "rendertype")) {
    int v;

    bool ok = msg->getInt(&v);

    if(ok) {
      m_renderAsForce = v == 1;
    }
    return ok;
  }
  else 
    return VEE_Particle2::setValue(id, msg);
}

VEE_Variable VEE_ForceObject::getVariable(const char *id) const
{
  VEE_Variable v;

  if(VEE_Strs::equals(id, "forcetype")) {
    v = (int) m_forceType;
  }
  else if(VEE_Strs::equals(id, "rendertype")) {
    v = (int) m_renderAsForce;
  }
  else 
    v = VEE_Particle2::getVariable(id);

  return v;
}

void VEE_ForceObject::render(VEE_Renderer *r)
{
  if(m_visible) {
    if(m_renderAsForce)
      r->renderForceObject(this);
    else
      r->renderParticle(this);
  }
}

void VEE_ForceObject::update(float dt)
{
  VEE_Vector3 oldLocation = m_location;

  VEE_Particle2::update(dt);

  m_moved = (oldLocation != m_location);
}

void VEE_ForceObject::unlinkBoxes()
{
  VEE_BoundingBox bbox;
  
  boundingBoxOfForce(&bbox);

  for(m_boxes.toFirst(); !m_boxes.isAtEnd(); m_boxes.advance()) {

    VEE_SceneBox * box = m_boxes.current();

    if(!bbox.intersects(box->boundingBox())) {
      box->forgetForce(this);
      m_boxes.remove(box);

      if(m_boxes.isAtEnd())
	return;
    }
  }
  
  // m_boxes.clear();
}

void VEE_ForceObject::linkBox(VEE_SceneBox *box)
{
  /* #ifndef VEE_FAST
  if(m_boxes.contains(box)) {
    for(m_boxes.toFirst(); !m_boxes.isAtEnd(); m_boxes.advance()) {
      VEE_trace("There is %p", m_boxes.current());
    }

    VEE_fatal("VEE_ForceObject::linkBox # already linked %p to %p, "
	      "%d forces %s", 
	      box, this, m_boxes.size(), typeid(*this).name());
  }
  #endif*/
  m_boxes.add(box);

  if(m_boxes.size() >= 20000)
    VEE_fatal("VEE_ForceObject::linkBox # m_boxes.size() >= 20000");
}

void VEE_ForceObject::unlinkBox(VEE_SceneBox *box)
{
#ifndef VEE_FAST
  if(!m_boxes.contains(box))
    VEE_fatal("VEE_ForceObject::unlinkBox # not linked %p %p", this, box);
#endif
  m_boxes.remove(box);
}


void VEE_ForceObject::updateSceneLocation(int frame)
{
  if(frame == m_lastUpdateFrame) return;

  // if(!m_moved) return;

  m_lastUpdateFrame = frame;

  unlinkBoxes();
  // m_sceneBox->addNewForce(this);

  VEE_BoundingBox bbox;
  boundingBoxOfForce(&bbox);

  m_sceneBox->addForceSomewhere(this, &bbox);
  // m_sceneBox->addForceHere(this, &bbox);
}

void VEE_calculateCubeBounds(VEE_BoundingBox *bb,
			     const VEE_Matrix3 &rotation,
			     const VEE_Vector3 &corner)
{
  VEE_Matrix3 rot = rotation;
 
  for(int i=0; i < 2; i++) {
    for(int j=0; j < 2; j++) {
      for(int k=0; k < 2; k++) {

      VEE_Vector3 c = corner;

      if(k) c[0] = - c[0];
      if(i) c[1] = - c[1];
      if(j) c[2] = - c[2];
      
      VEE_Vector3 worldCorner = rot * c;
      bb->expand(worldCorner);
      /* worldCorner = - worldCorner;
	 bb->expand(worldCorner); */
      }
    }
    
  }
}
