/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_object.h> 

#include <vee_error.h> 
#include <vee_math.h> 
#include <vee_renderer.h> 
#include <vee_scene.h> 
#include <vee_strs.h> 
#include <vee_values.h> 

#include <typeinfo>

#ifndef VEE_FAST

/** The number of VEE_Object instances floating around. 

    If the system runs with threads, then this variable may get random
    values even tough the system behaves correctly.
*/
long VEE_Object::m_count = 0;

#endif

VEE_Object::VEE_Object() 
  : m_sceneBox(0), 
    m_updateVelocity(true),
    m_age(0), 
    m_ageLimit(1e+10), 
    m_mass(1.0), 
    m_area(1.0),
    m_objectType(VEE_OBJECT_UNKNOWN),
	m_species(0),
    m_done(false),
    m_transient(false)
{ 
  addCheckCount(); 
}

VEE_Object::VEE_Object(VEE_SceneBox *box) 
  : m_sceneBox(0), 
    m_updateVelocity(true),
    m_age(0), 
    m_ageLimit(1e+10), 
    m_mass(1.0), 
    m_area(1.0),
    m_objectType(VEE_OBJECT_UNKNOWN),
	m_species(0),
    m_done(false),
    m_transient(false)
{ 
  if(box) {
    box->addObject(this);
    m_sceneBox = box;
  }
  addCheckCount(); 
}

VEE_Object::VEE_Object(const VEE_Object &o)
  : VEE_ParticleVisualInfo(o),
    m_sceneBox(o.m_sceneBox), 
    m_updateVelocity(o.m_updateVelocity),
    m_age(o.m_age), 
    m_ageLimit(o.m_ageLimit),
    m_mass(o.m_mass), 
    m_area(o.m_area),
    m_objectType(o.m_objectType),
	m_species(o.m_species),
    m_done(o.m_done),
    m_transient(o.m_transient)
{
  if(m_sceneBox) {
    VEE_SceneBox * box = m_sceneBox;
    m_sceneBox = 0;
    box->addObject(this);
  }

  addCheckCount();

  /* if(m_ageLimit > 10.0) {
    VEE_trace("VEE_Object::VEE_Object(const VEE_Object &o) # agelimit = %f",
	      m_ageLimit);
	      }*/
}

VEE_Object::VEE_Object
(const VEE_Vector3 &location,  const VEE_Vector3 &velocity, float ageLimit) 
  : VEE_ParticleVisualInfo(location, velocity),
    m_sceneBox(0), 
    m_updateVelocity(true),
    m_age(0.0), 
    m_ageLimit(ageLimit), 
    m_mass(1.0), 
    m_area(1.0),
	m_species(0),
    m_done(false),
    m_transient(false)
{ addCheckCount(); }

VEE_Object::VEE_Object
(const VEE_Vector3 &location, float ageLimit) 
  : VEE_ParticleVisualInfo(location, VEE_Vector3ZERO),
    m_age(0.0), 
    m_ageLimit(ageLimit), 
    m_mass(1.0), 
    m_area(1.0), 
	m_species(0),
    m_done(false),
    m_transient(false)
{ 
  addCheckCount(); 
}

/** Receive and process a value. 

    Parameters:

    <UL>
    
    <LI> "age", float, range: 0 to inf

    <LI> "agelimit", float, range: 0 to inf

    <LI> "location", Vector3

    <LI> "addlocation", Vector3

    <LI> "velocity", Vector3

    <LI> "addvelocity", Vector3

    <LI> "mass", float, range: -1 to 1

    <LI> "area", float, range: -1 to 1

    <LI> "color", Vector3

    <LI> "done", bool (int), range: true or false (1 or 0)

    </UL>

*/
bool VEE_Object::setValue(const char *id, VEE_Value *msg)
{
  bool ok = false;

  if(VEE_Strs::equals(id, VEE_Strs::age)) {
    float v;

    ok = msg->getFloat(&v);

    if(ok)
      m_age = v;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::agelimit)) {
    float v;

    ok = msg->getFloat(&v);

    // VEE_trace("VEE_Object::setValue # agelimit %f %d", v, (int) ok);

    if(ok)
      m_ageLimit = v;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::location)) {
    VEE_Vector3 v;

    ok = msg->getVector3(&v);

    if(ok)
      m_location = v;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::addlocation)) {
    VEE_Vector3 v;

    ok = msg->getVector3(&v);

    if(ok)
      m_location += v;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::velocity)) {
    VEE_Vector3 v;

    ok = msg->getVector3(&v);

    if(ok)
      m_velocity = v;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::addvelocity)) {
    VEE_Vector3 v;

    ok = msg->getVector3(&v);

    if(ok)
      m_velocity += v;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::mass)) {
    float v;

    ok = msg->getFloat(&v);

    if(ok)
      m_mass = v;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::area)) {
    float v;

    ok = msg->getFloat(&v);

    if(ok)
      m_area = v;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::color)) {
    VEE_ValueRgba *vrgb = dynamic_cast<VEE_ValueRgba *> (msg);

    if(vrgb)
      m_color = vrgb->getValue();

    else {
      VEE_Vector3 v;
      
      ok = msg->getVector3(&v);
      
      if(ok)
	m_color.set(v[0], v[1], v[2], m_color.alpha(), m_color.theta());
    }
  }
  else if(VEE_Strs::equals(id, "texture")) {
    int tmp;

    ok = msg->getInt(&tmp);

    if(ok)
      m_texture = tmp;
  }
  else if(VEE_Strs::equals(id, "size")) {
    float tmp;

    ok = msg->getFloat(&tmp);

    if(ok)
      m_size = tmp;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::done)) {
    int v;

    ok = msg->getInt(&v);

    if(ok)
      m_done = (v != 0);
  }
  else if(VEE_Strs::equals(id, "updatevelocity")) {
    int v;

    ok = msg->getInt(&v);

    if(ok)
      m_updateVelocity = (v != 0);
  }
  else if(strcmp(id, "rotation") == 0) {
    float tmp;

    ok = msg->getFloat(&tmp);
    if(ok)
      m_rotationRadians = tmp;
    
  }
  else if(strcmp(id, "omega") == 0) {
    float tmp;

    ok = msg->getFloat(&tmp);
    if(ok)
      m_omega = tmp;    
  }

  return ok;
}

VEE_Variable VEE_Object::getVariable(const char *id) const
{
  VEE_Variable v;

  if(VEE_Strs::equals(id, VEE_Strs::age))
    v = & m_age;
  else if(VEE_Strs::equals(id, VEE_Strs::agelimit))
    v = & m_ageLimit;
  else if(VEE_Strs::equals(id, VEE_Strs::color))
    v = m_color;
  else if(VEE_Strs::equals(id, VEE_Strs::location))
    v = & m_location;
  else if(VEE_Strs::equals(id, VEE_Strs::velocity))
    v = & m_velocity;
  else if(VEE_Strs::equals(id, VEE_Strs::mass))
    v = & m_mass;
  else if(VEE_Strs::equals(id, VEE_Strs::area))
    v = & m_area;
  else if(VEE_Strs::equals(id, "texture"))
    v = & m_texture;
  else if(VEE_Strs::equals(id, "size"))
    v = & m_size;
  else if(VEE_Strs::equals(id, "rotation"))
    v = & m_rotationRadians;
  else if(VEE_Strs::equals(id, "omega"))
    v = & m_omega;
  else if(VEE_Strs::equals(id, VEE_Strs::done))
    v = (int) m_done;
  else if(VEE_Strs::equals(id, "updatevelocity"))
    v = (int) m_updateVelocity;
  
  return v;
}

void VEE_Object::update(float dt)
{
  m_location += m_velocity * dt;
  m_age += dt;
  m_rotationRadians = VEE_checkRadianRange(m_rotationRadians + m_omega * dt);

  if(m_age > m_ageLimit)
    m_done = true;
}

void VEE_Object::updateVelocity(float dt)
{
  if(!m_updateVelocity) return;

  VEE_Vector3 grav, drag;

  grav.clear();
  drag = grav;
  
  if(VEE_abs(m_mass) > 1.0e-5) {
    sceneBox()->getForce(VEE_FORCE_GRAVITY, this, 
			 &grav, m_velocity, m_location);
  }

  if(VEE_abs(m_area) > 1.0e-5) {
    sceneBox()->getForce(VEE_FORCE_DRAG, this, 
			 &drag, m_velocity, m_location);
  }

#ifndef VEE_FAST
  if(!grav.isFinite())
    VEE_error("VEE_Object::updateVelocity # %s # Gravity is non-finite 0x%p", 
	      typeid(*this).name(), this);

  if(!drag.isFinite())
    VEE_error("VEE_Object::updateVelocity # %s # Drag is non-finite 0x%p", 
	      typeid(*this).name(), this);
#endif
  
  m_velocity += (m_mass * grav + m_area * drag) * dt;
}

void VEE_Object::collision(const VEE_Vector3 &)
{

}


const char * VEE_Object::className() const
{
  return typeid(*this).name();
}
