/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_operator_force_gradient.h>

#include <vee_error.h>
#include <vee_math.h>
#include <vee_object.h>
#include <vee_scene.h>
#include <vee_strs.h>
#include <vee_value.h>

VEE_OperatorForceGradient::VEE_OperatorForceGradient()
  : m_range(1.0f),
    m_scale(1.0f),
    m_forceType(VEE_FORCE_DRAG)
{
  // VEE_trace("VEE_OperatorForceGradient::VEE_OperatorForceGradient # %p", this);
}

VEE_OperatorForceGradient::~VEE_OperatorForceGradient()
{}

/** Receive and process a value. 

    Parameters:

    <UL>
    
    <LI> "range", float, greater than zero

    <LI> "scale", float, greater than zero

    <LI> "forcetype", int (VEE_ForceType), 0 to VEE_FORCE_SIZEOF - 1

    </UL>

*/
bool VEE_OperatorForceGradient::setValue(const char *id, VEE_Value *v)
{
  // VEE_trace("VEE_OperatorForceGradient::setValue # %s", id);

  bool ok = false;

  if(strcmp(id, VEE_Strs::range) == 0) {
    float tmp;
    ok = v->getFloat(&tmp);
    if(ok)
      m_range = VEE_max(tmp, 0.0f);
  }
  if(strcmp(id, "scale") == 0) {
    float tmp;
    ok = v->getFloat(&tmp);
    if(ok)
      m_scale = VEE_max(tmp, 0.0f);
  }
  else if(strcmp(id, "forcetype") == 0) {
    int tmp;

    ok = v->getInt(&tmp);
    if(ok)
      m_forceType = (VEE_ForceType) VEE_clamp(tmp, 0, VEE_FORCE_SIZEOF - 1);
  }

  return ok;
}

void VEE_OperatorForceGradient::operate(VEE_Object *o, float)
{
  VEE_SceneBox * scene = o->sceneBox();

  //VEE_trace("VEE_OperatorForceGradient::operate # %p %p %p", this, o, scene);

  if(!scene)
    return;

  VEE_Vector3 center, centerForce, loc, vel, force;

  center = o->location();
  vel = o->velocity();

  centerForce.clear();

  scene->getForce(m_forceType, o, & centerForce, vel, center);
  
  float gradientSum = 0.0f;

  uint i;

  for(i=0; i < 2; i++) {
    
    float offset = i ? m_range : -m_range;

    for(int j=0; j < 3; j++) {

      loc.clear();
      loc[j] = offset;

      loc += center;

      force.clear();

      scene->getForce(m_forceType, o, & force, vel, loc);

      gradientSum += (force - centerForce).length();
    }
  }

  gradientSum *= m_scale;
  
  /* VEE_trace("VEE_OperatorForceGradient::operate # %u curves",
     (uint) m_curves.size()); */

  for(i=0; i < m_curves.size(); i++) {

    Data * data = & m_curves[i];
    float val = data->m_curve.getValue(gradientSum);
    // bool ok = 
    o->setValueFloat(data->m_param.c_str(), val);

    /* VEE_trace("VEE_OperatorForceGradient::operate # Sent value \"%s\", %f %s",
       data->m_param.c_str(), val, ok ? "ok" : "failed");*/
  }
}

VEE_Operator * VEE_OperatorForceGradient::clone() const
{
  return new VEE_OperatorForceGradient(*this);
}
