/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_operator_list.h>

#include <vee_error.h>
#include <vee_object.h>
#include <vee_operator.h>
#include <vee_value.h>

#include <typeinfo>

VEE_OperatorList::VEE_OperatorList()
{}

VEE_OperatorList::~VEE_OperatorList()
{
  clear();
}

void VEE_OperatorList::appendOperator(VEE_Operator *o, const char * name)
{ 
  if(!o) VEE_fatal("VEE_OperatorList::appendOperator # NULL argument");
  
  Holder h;
  h.m_name = name ? name : "UNNAMED";
  h.m_operator = o;
  h.m_target = 0;

  m_operators.push_back(h); 
}
  
void VEE_OperatorList::appendOperator2
(VEE_Object *target, VEE_Operator *o, const char * name)
{
  if(!o) VEE_fatal("VEE_OperatorList::appendOperator # NULL argument");
  
  Holder h;
  h.m_name = name ? name : "UNNAMED";
  h.m_operator = o;
  h.m_target = target;

  m_operators.push_back(h);
}

/** Removes all operators with given name.  */

bool VEE_OperatorList::removeOperator(const char *name, bool alsoDelete)
{
  if(!name)
    return false;

  bool ok = false;

  // for(holderIter_t it = m_operators.begin(); it != m_operators.end(); ) {
  for(int i = m_operators.size(); --i >= 0; ) {

    holderIter_t it = m_operators.begin() + i;

    if((*it).m_name == name) {

// operators are now responsible for freeing themselves (operatorlist no longer owns them)
//      if(alsoDelete) delete (*it).m_operator;
      m_operators.erase(it);
      ok = true;
    }
  }
  return ok;
}

void VEE_OperatorList::operate(VEE_Object *obj, float dt)
{
  for(holderIter_t it = m_operators.begin(); it != m_operators.end(); ++it) {
    /* VEE_trace("VEE_OperatorList::operate # %s of %s", (*it).m_name.c_str(),
       typeid(*(*it).oper()).name()); */
    (*it).oper()->operate(obj, dt);
  }
}

void VEE_OperatorList::operate2(float dt)
{
  for(holderIter_t it = m_operators.begin(); it != m_operators.end(); ++it) {
    if((*it).m_target)
      (*it).oper()->operate((*it).m_target, dt);
  }
}

void VEE_OperatorList::clear()
{
// operators are now responsible for freeing themselves (operatorlist no longer owns them)
//  for(holderIter_t it = m_operators.begin(); it != m_operators.end(); ++it) {
//    delete  (*it).oper();
//  }
  m_operators.clear();
}

/** Sets the given value to all children with mathing name. Suppose we
    have a child named "joy". If we pass a value with name
    "joy/strength", then this value will be given to all operators
    with name "joy".

    @return This method returns true if at least one child object
    processed the value successfully and none failed at processing it.
*/
bool VEE_OperatorList::setValue(const char *name, VEE_Value *v)
{
  int slashAt = -1;

  for(int i=0; i < 1000; i++) {
    if(name[i] == '/') {
      slashAt = i;
      break;
    }
    if(name[i] == '\0') {
      return false;
    }
  }

  if(slashAt < 1)
    return false;

  int good = 0;
  int fail = 0;

  const char * name2 = name + slashAt + 1;
  
  for(holderIter_t it = m_operators.begin(); it != m_operators.end(); ++it) {

    if(strncmp((*it).m_name.c_str(), name, slashAt) == 0) {

      if(!(*it).oper()->setValue(name2, v)) {
	VEE_error("VEE_OperatorList::setValue # "
		  "Could not match id \"%s with %s\"",
		  name2, typeid(*(*it).oper()).name());
	fail++;
      }
      else 
	good++;
    }
  }

  if(!good)
    VEE_error("VEE_OperatorList::setValue # "
	      "Could not match id \"%s with %s\" (%d %d)",
	      name, typeid(*v).name(), fail, good);

  return good && !fail;
}

VEE_OperatorList & VEE_OperatorList::operator = (const VEE_OperatorList &l)
{
  clear();

  uint count = l.size();

  m_operators.resize(count);

  for(uint i=0; i < count; i++) {

    const Holder & source = l.m_operators[i];
    Holder & target = m_operators[i];
    
    target.m_name     = source.m_name;
    target.m_operator = source.m_operator->clone();
    target.m_target   = source.m_target;

  }

  return *this;
}
