
/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_orderer.h>

#include <vee_error.h>
#include <vee_force_object.h>
#include <vee_scene.h>

#include <typeinfo>

VEE_Orderer::VEE_Orderer()
{
  m_camPos.clear();
}

VEE_Orderer::~VEE_Orderer()
{}

/** Retrieve all objects from the scene (recursive). */
void VEE_Orderer::addScene(VEE_SceneBox *scene)
{
  VEE_SceneBox::iterator it = scene->begin();
  VEE_SceneBox::iterator sentinel = scene->end();
  
  while(it != sentinel) {
    VEE_Object * o = (*it);
    const VEE_Vector3 & loc = o->location();
    float distance = (m_camPos - loc).length();
    m_objects.insert(std::pair<float, VEE_Object *>(distance, o));
    it++;
  }

  if(scene->isSplit()) {
    for(uint i = 0; i < 8; i++) 
      addScene(scene->child(i));
  }
}

/** Clears the oredring information. */
void VEE_Orderer::clear()
{
  m_objects.clear();
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


VEE_OrdererCollector::VEE_OrdererCollector()
{}

VEE_OrdererCollector::~VEE_OrdererCollector()
{
  clear();
}

void VEE_OrdererCollector::addScene(VEE_SceneBox *scene)
{
  // VEE_Object **obj = & scene->object(0);
  // VEE_Object **sentinel =  obj + scene->numObjects();
  
  VEE_SceneBox::iterator it = scene->begin();
  VEE_SceneBox::iterator sentinel = scene->end();
  
  while(it != sentinel) {
    VEE_Object * o = (*it);

    int type = o->objectType();

    if(type == VEE_OBJECT_PARTICLE) {
      const VEE_Vector3 & loc = o->location();
      float distance = (m_camPos - loc).length();

      VEE_Particle * p = (VEE_Particle *) o;
      if(p->visible())
	m_objects.insert
	  (std::pair<float, VEE_ParticleVisualInfo>(distance, *o));
    }
    else if(type == VEE_OBJECT_FORCE) {
      
      VEE_ForceObject *force = (VEE_ForceObject *) o;

      if(!force->renderAsForce()) {	
	const VEE_Vector3 & loc = o->location();
	float distance = (m_camPos - loc).length();
	m_objects.insert(std::pair<float, VEE_ParticleVisualInfo>
			 (distance, *o));
      }
      else {

	VEE_Object *o2 = o->clone();
	
	VEE_ForceObject *fo = dynamic_cast<VEE_ForceObject *> (o2);
      
	if(!fo)
	  VEE_fatal("VEE_OrdererCollector::addScene # Not force %s (%s)",
		    typeid(*o).name(), typeid(*o2).name());
	
	if(fo->sceneBox())
	  fo->sceneBox()->removeChildObject(o2);
	fo->setSceneBox(0);
	fo->forgetLinks();
	
	m_forces.push_back(fo);
      }
    }
    it++;
  }

  if(scene->isSplit()) {
    for(uint i = 0; i < 8; i++) 
      addScene(scene->child(i));
  }
}

void VEE_OrdererCollector::clear()
{
  m_objects.clear();

  for(forceIterator it = forceBegin(); it != forceEnd(); it++) {
    VEE_ForceObject * force = getForce(it);
    /* VEE_trace("VEE_OrdererCollector::clear # About to delete %s %p",
       typeid(*force).name(), force); */
    delete force;
  }

  m_forces.clear();
}

