/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#ifndef VEE_ORDERER_H
#define VEE_ORDERER_H

#include <vee_object.h>

#include <map>
#include <list>

class VEE_ForceObject;

/** Collects and orders objects from a  scene. 

\code

VEE_Orderer collector;

collector.setCameraPosition(VEE_Vector3(0.1, 23.0, 0.0));
collector.clear();
collector.addScene(my_scene);

for(VEE_Orderer::iterator it = collector.begin();
    it != collector.end(); it++)
  doSomething(VEE_Orderer::geObject(it));

\endcode

*/

class VEE_Orderer
{
public:
  VEE_Orderer();
  ~VEE_Orderer();
  
  typedef std::multimap<float, VEE_Object *>::iterator iterator;

  /// Gets the first object
  iterator begin() { return m_objects.begin();}
  /// Gets the last object (one past the end)
  iterator end() { return m_objects.end(); }

  uint size() { return m_objects.size(); }

  /// Gets the object from the iterator
  static VEE_Object *getObject(iterator & it)
  { return (*it).second; }

  /// Set the camera position that will be used to evaluate distance
  void setCameraPosition(const VEE_Vector3 &p) { m_camPos = p; }

  void addScene(VEE_SceneBox *);
  void clear();

protected:
  std::multimap<float, VEE_Object *> m_objects;

  VEE_Vector3 m_camPos;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Gets all the child objects and sorts them. Ths class does not
    store pointers to the objects, but stores their visual information
    independently. Thus this class is useful for threaded rendering
    where this class can be used to collect the state of particle
    system in thread 1. The other threads can then render the state in
    other threads. */

class VEE_OrdererCollector
{
public:
  typedef std::multimap<float, VEE_ParticleVisualInfo>::iterator iterator;
  typedef std::list<VEE_ForceObject *>::iterator forceIterator;

  VEE_OrdererCollector();
  ~VEE_OrdererCollector();

  /// Gets the first object
  iterator begin() { return m_objects.begin();}
  /// Gets the last object (one past the end)
  iterator end() { return m_objects.end(); }

  uint size() { return m_objects.size(); }

  /// Gets the first object
  forceIterator forceBegin() { return m_forces.begin();}
  /// Gets the last object (one past the end)
  forceIterator forceEnd() { return m_forces.end(); }
  
  uint forceCount() { return m_forces.size(); }

  /// Gets the object from the iterator
  static VEE_ParticleVisualInfo & getObject(iterator & it)
  { return (*it).second; }

  /// Gets the force from the force iterator
  static VEE_ForceObject * getForce(forceIterator & it)
  { return (*it); }

  /// Set the camera position that will be used to evaluate distance
  void setCameraPosition(const VEE_Vector3 &p) { m_camPos = p; }

  void addScene(VEE_SceneBox *);
  void clear();

protected:

  std::multimap<float, VEE_ParticleVisualInfo> m_objects;
  std::list<VEE_ForceObject *> m_forces;
  
  VEE_Vector3 m_camPos;

private:
  // Disabled
  VEE_OrdererCollector(const VEE_OrdererCollector & ) {}
  VEE_OrdererCollector & operator = (const VEE_OrdererCollector & ) 
  { return * this; }
};

#endif
