/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_perlin_noise.h>

#include <vee_error.h>
// #include <vee_math.h>

// #include <Fluid/Vector3.hpp>
// #include <Fluid/Vector4.hpp>


float VEE_PerlinNoise::noise1d(float x, 
			       float octaveAttenuation, unsigned octaves)
{
  float weight = 1.0f, ws = 0.0f, r = 0.0f;

  for(unsigned i=0; i < octaves; i++) {
    r += noise(x) * weight;
    x *= 2.0f;
    ws += weight;
    weight *= octaveAttenuation;
  }

  return r / ws;
}

float VEE_PerlinNoise::noise2d(float x, float y, 
			       float octaveAttenuation, unsigned octaves)
{
  float weight = 1.0f, ws = 0.0f, r = 0.0f;

  for(unsigned i=0; i < octaves; i++) {
    r += noise(x, y) * weight;
    x *= 2.0f;
    y *= 2.0f;
    ws += weight;
    weight *= octaveAttenuation;
  }

  return r / ws;
}

float VEE_PerlinNoise::noise3d(float x, float y, float z, 
			       float octaveAttenuation, unsigned octaves)
{
  float weight = 1.0f, ws = 0.0f, r = 0.0f;

  for(unsigned i=0; i < octaves; i++) {
    r += noise(x, y, z) * weight;
    x *= 2.0f;
    y *= 2.0f;
    ws += weight;
    weight *= octaveAttenuation;
  }

  // VEE_trace("VEE_PerlinNoise::noise3d # %f %f %f -> %f", x, y, z, r);

  return r / ws;
}

/*
uint VEE_PerlinNoise::m_randMultipliers[VEE_PerlinNoise::MAX_OCTAVES + 1] = {
  134695621,
  134697601,
  18000259,
  13000159,
  180000173,
  16000129,
  12000239,
  14000087,
  17000161,
  19000357,
  20000269
};

uint VEE_PerlinNoise::m_randMultipliers2[VEE_PerlinNoise::MAX_OCTAVES + 1] = {
  15000151,
  134705027,
  134714179,
  120000121,
  130000097,
  140000171,
  154000097,
  163000091,
  170000267,
  190000169,
  203000177
};

inline double VEE_PerlinNoise::moduloRand(uint source, 
					  uint octave)
{
  uint mul = m_randMultipliers[octave];

  // source += m_randMultipliers[octave + 1];

  uint n = 5 + source & 0xf;

  for(uint i=0; i < n; i++) {
    source = source * mul + 1;
  }

  return ((double) source / (double) 0xFFFFFFFFu) - 0.5;
}

inline uint VEE_PerlinNoise::ra(uint source, uint mul, uint seed)
{
  source += seed;

  uint n = 5 + source & 0x7;
  
  for(uint i=0; i < n; i++) {
    source = source * mul + 1;
  }
  
  return source;
}


double VEE_PerlinNoise::octaveNoise1d(double x, 
				      double scale, 
				      uint octave,
				      Interpolation ip)
{
  double sx = x * scale;
  long ix1 = (long) floor(sx);
  long ix2 = ix1 + 1;

  double r1 = moduloRand(ix1, octave);
  double r2 = moduloRand(ix2, octave);

  double w2 = sx - ix1;
  double w1 = 1.0 - w2;

  if(ip == LINEAR)
    return w1 * r1 + w2 * r2;

  // Weight interpolation:

  w2 *= w2;
  w1 *= w1;

  double sum = w2 + w1;
  
  return (w1 * r1 + w2 * r2) / sum;
}

double VEE_PerlinNoise::noise1d(double x, 
				double octaveAttenuation, 
				uint octaves, 
				Interpolation ip)
{
  double r = 0.0;
  double scale = 1.0;
  double att = 1.0;

  if(octaves >= MAX_OCTAVES) octaves = MAX_OCTAVES - 1;

  for(uint i=0; i < octaves; i++) {
    r += octaveNoise1d(x, scale, i, ip) * att;
    att *= octaveAttenuation;
    scale *= 2.0;
  }

  return r;
}

double VEE_PerlinNoise::octaveNoise2d(double x, 
				      double y, 
				      double scale, 
				      uint octave,
				      Interpolation ip,
				      uint seed)
{
  double sx = x * scale;
  long ix1 = (long) floor(sx);
  long ix2 = ix1 + 1;

  double sy = y * scale;
  long iy1 = (long) floor(sy);
  long iy2 = iy1 + 1;

  uint m1 = m_randMultipliers[octave];
  // uint m2 = m_randMultipliers2[octave];

  // double r00 = moduloRand(ra(ix1, m1, seed) + ra(iy1, m2, seed), octave);
  //double r01 = moduloRand(ra(ix1, m1, seed) + ra(iy2, m2, seed), octave);
  //double r10 = moduloRand(ra(ix2, m1, seed) + ra(iy1, m2, seed), octave);
  //double r11 = moduloRand(ra(ix2, m1, seed) + ra(iy2, m2, seed), octave);

  double r00 = moduloRand(ra(ix1, m1, seed) + iy1, octave);
  double r01 = moduloRand(ra(ix1, m1, seed) + iy2, octave);
  double r10 = moduloRand(ra(ix2, m1, seed) + iy1, octave);
  double r11 = moduloRand(ra(ix2, m1, seed) + iy2, octave);

  double wx2 = sx - ix1;
  double wx1 = 1.0 - wx2;

  double wy2 = sy - iy1;
  double wy1 = 1.0 - wy2;

  double w00, w01, w10, w11;

  if(ip == LINEAR) {
    w00 = VEE_min(wx1, wy1);
    w01 = VEE_min(wx1, wy2);
    w10 = VEE_min(wx2, wy1);
    w11 = VEE_min(wx2, wy2);
  }
  else {
    const double ds = 1.3;

    double d00 = VEE_min(ds * (wx2 * wx2 + wy2 * wy2), 1.0);
    double d01 = VEE_min(ds * (wx2 * wx2 + wy1 * wy1), 1.0);
    double d10 = VEE_min(ds * (wx1 * wx1 + wy2 * wy2), 1.0);
    double d11 = VEE_min(ds * (wx1 * wx1 + wy1 * wy1), 1.0);

    w00 = cos(d00 * VEE_HALF_PI);
    w01 = cos(d01 * VEE_HALF_PI);
    w10 = cos(d10 * VEE_HALF_PI);
    w11 = cos(d11 * VEE_HALF_PI);

    w00 *= w00;
    w01 *= w01;
    w10 *= w10;
    w11 *= w11;

  }

  double sum = w00 + w01 + w10 + w11;

  return (r00 * w00 + r01 * w01 + r10 * w10 + r11 * w11) / sum;

}

double VEE_PerlinNoise::noise2d(double x, 
				double y, 
				double octaveAttenuation, 
				uint octaves, 
				Interpolation ip,
				uint seed)
{
  double r = 0.0;
  double scale = 1.0;
  double att = 1.0;

  if(octaves >= MAX_OCTAVES) octaves = MAX_OCTAVES - 1;

  for(uint i=0; i < octaves; i++) {
    r += octaveNoise2d(x, y, scale, i, ip, seed) * att;
    att *= octaveAttenuation;
    scale *= 2.0;
  }

  return r;
}


double VEE_PerlinNoise::octaveNoise3d(double x, 
				      double y, 
				      double z, 
				      double scale, 
				      uint octave,
				      Interpolation ip,
				      uint seed,
				      double range,
				      double beyondRange)
{
  typedef Fluid::Vector3T<double> DVector3;
  typedef Fluid::Vector3T<long>   LVector3;

  DVector3 v(x, y, z);
  DVector3 sv(v * scale);

  LVector3 l[2];
  uint i, j, k;

  for(i=0; i < 3; i++) {
    l[0][i] = (long) floor(sv[i]);
    l[1][i] = l[0][i] + 1;
  }

  uint m  = m_randMultipliers[octave];
  uint m2 = m_randMultipliers2[octave];

  // Calculate random values for each vertex:
  double r[2][2][2];

  for(i=0; i < 2; i++) {
    for(j=0; j < 2; j++) {
      for(k=0; k < 2; k++) {
	r[i][j][k] = moduloRand(ra(l[i][0], m, seed) +
				ra(l[j][1], m2, seed) +
				ra(l[k][2], m, seed), octave);
      }
    }
  }

  if(range > 0.0) {
    
    for(i=0; i < 2; i++) {
      for(j=0; j < 2; j++) {
	for(k=0; k < 2; k++) {
	  DVector3 loc;
	  loc[0] = l[i][0];
	  loc[1] = l[j][1];
	  loc[2] = l[k][2];
	  loc /= scale;
	  if(loc.length() > range)
	    r[i][j][k] = beyondRange;
	}  
      }
    }
  }


  // Calculate weight for each vertex;
  double w[2][2][2];

  DVector3 pw[2];

  for(i=0; i < 3; i++) {
    pw[1][i] = sv[i] - l[0][i];
    pw[0][i] = 1.0 - pw[1][i];
  }

  double wsum = 0.0;

  for(i=0; i < 2; i++) {
    for(j=0; j < 2; j++) {
      for(k=0; k < 2; k++) {
	double we = pw[i][0] * pw[j][1] * pw[k][2];
	if(ip == WEIGHT)
	  we *= we;
	wsum += we;
	w[i][j][k] = we;
      }
    }
  }

  double res = 0.0;
  
  for(i=0; i < 2; i++) {
    for(j=0; j < 2; j++) {
      for(k=0; k < 2; k++) {
	res += w[i][j][k] * r[i][j][k];
      }
    }
  }

  return res / wsum;

}

double VEE_PerlinNoise::noise3d(double x, 
				double y, 
				double z, 
				double octaveAttenuation, 
				uint octaves, 
				Interpolation ip,
				uint seed,
				double range,
				double beyondRange)
{
  double r = 0.0;
  double scale = 1.0;
  double att = 1.0;
  double sum = 0.0;

  if(octaves >= MAX_OCTAVES) octaves = MAX_OCTAVES - 1;

  for(uint i=0; i < octaves; i++) {
    sum += att;
    r += octaveNoise3d(x, y, z, scale, i, ip, seed, 
		       range, beyondRange) * att;
    att *= octaveAttenuation;
    scale *= 2.0;
  }

  return r / sum;
}

double VEE_PerlinNoise::octaveNoise4d(double x, 
				      double y, 
				      double z, 
				      double w, 
				      double scale, 
				      uint octave,
				      Interpolation ip,
				      uint seed)
{
  typedef Fluid::Vector4T<double> DVector4;
  typedef Fluid::Vector4T<long>   LVector4;

  DVector4 v(x, y, z, w);
  DVector4 sv(v * scale);

  LVector4 l[2];
  uint i, j, k, e;

  for(i=0; i < 4; i++) {
    l[0][i] = (long) floor(sv[i]);
    l[1][i] = l[0][i] + 1;
  }

  uint m  = m_randMultipliers[octave];
  uint m2 = m_randMultipliers2[octave];

  // Calculate random values for each vertex:
  double r[2][2][2][2];

  for(i=0; i < 2; i++) {
    for(j=0; j < 2; j++) {
      for(k=0; k < 2; k++) {
	for(e=0; e < 2; e++) {
	  r[i][j][k][e] = moduloRand(ra(l[i][0], m, seed) +
				     ra(l[j][1], m2, seed) +
				     ra(l[k][2], m, seed) +
				     ra(l[e][3], m2, seed), octave);
	}
      }
    }
  }

  // Calculate weight for each vertex;
  double W[2][2][2][2];

  DVector4 pw[2];

  for(i=0; i < 4; i++) {
    pw[1][i] = sv[i] - l[0][i];
    pw[0][i] = 1.0 - pw[1][i];
  }

  double wsum = 0.0;

  for(i=0; i < 2; i++) {
    for(j=0; j < 2; j++) {
      for(k=0; k < 2; k++) {
	for(e=0; e < 2; e++) {
	  double we = pw[i][0] * pw[j][1] * pw[k][2] * pw[e][3];
	  if(ip == WEIGHT)
	    we *= we;
	  wsum += we;
	  W[i][j][k][e] = we;
	}
      }
    }
  }

  double res = 0.0;
  
  for(i=0; i < 2; i++) {
    for(j=0; j < 2; j++) {
      for(k=0; k < 2; k++) {
	for(e=0; e < 2; e++) {
	  res += W[i][j][k][e] * r[i][j][k][e];
	}
      }
    }
  }

  return res / wsum;

}

double VEE_PerlinNoise::noise4d(double x, 
				double y, 
				double z, 
				double w, 
				double octaveAttenuation, 
				uint octaves, 
				Interpolation ip,
				uint seed)
{
  double r = 0.0;
  double scale = 1.0;
  double att = 1.0;
  double sum = 0.0;

  if(octaves >= MAX_OCTAVES) octaves = MAX_OCTAVES - 1;

  for(uint i=0; i < octaves; i++) {
    sum += att;
    r += octaveNoise4d(x, y, z, w, scale, i, ip, seed) * att;
    att *= octaveAttenuation;
    scale *= 2.0;
  }

  return r / sum;
}

void VEE_PerlinNoise::checkRand()
{
  for(uint i=0; i < 2; i++) {
    uint * muls = 0;

    if(i == 0)
      muls = m_randMultipliers;
    else
      muls = m_randMultipliers2;

    for(uint j=0; j < MAX_OCTAVES + 1; j++) {

      llong k, maxloops = 2000000000;
      // maxloops *= maxloops;

      uint tmp = 100;
      uint mul = muls[j];

      for(k = 0; k < 10000; k++)
	tmp = tmp * mul + 1;
      
      uint tmp2 = tmp;
      bool ok = true;
      
      VEE_trace("Checking randomizer %u %u %u", i, j, mul);

      for(k = 0; k < maxloops && ok; k++) {
	tmp = tmp * mul + 1;
	if(tmp == tmp2) {
	  VEE_trace("Randomizer %u %u wraps after %lf loops", i, j, 
		    (double) k);
	  ok = false;
	}
      }
      
    }
  }
}
*/
