/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */


#include <vee_algorithm.h> 
#include <vee_error.h> 
#include <vee_math.h>
#include <vee_polygon.h>
#include <vee_renderer.h> 
#include <vee_scene.h>
#include <vee_strs.h>
#include <vee_value.h>

#include <strings.h>

VEE_Polygon::VEE_Polygon()
  : m_maxTension(50.0f),
    m_stiffness(1.0),
    m_n(0),
    m_nTri(0),
    m_nLinks(0),
    m_meshColumns(0),
    m_meshRows(0),
    m_material(-1)
{}

VEE_Polygon::VEE_Polygon(uint num, const VEE_Vector3 *positions,
			 uint numTri, const uint *indices,
			 const VEE_Vector3 &velocity,
			 float ageLimit, 
			 int material,
			 VEE_Looks *looks)
  : VEE_Particle(VEE_Vector3ZERO, VEE_Vector3ZERO, ageLimit, looks),
    m_maxTension(50.0),
    m_stiffness(1.0), 
    m_n(num), 
    m_nTri(numTri), 
    m_material(material)
{
  // Allocate enough memory for all the possible links even
  // though as a default we only save the existing links.
  uint n = (uint) (num-1)*num;

  m_age = 0;
  m_distances.resize(n);
  m_tensions.resize(n);
  m_links.resize(n*2);

  m_vertexPositions.resize(num);
  memcpy(&* m_vertexPositions.begin(), positions, num*sizeof(VEE_Vector3));
  m_vertexTextureCoordinates.resize(num);
  m_indices.resize(numTri*3);
  memcpy(&* m_indices.begin(), indices, numTri*3*sizeof(uint));
  m_vertexVelocities.resize(num);
  m_gravForces.resize(num);
  m_dragForces.resize(num);

  // Set default links based on m_indices.
  resetLinks();
  initTensions();
  setVelocity(velocity);
  // Calculate the optimal distances for the defined links.
  resetDistances();

  m_location = m_vertexPositions[0];
  for(uint i=1; i<num; i++)
    m_location += m_vertexPositions[i];
  m_location /= num;

  m_meshColumns = 0;
  m_meshRows = 0;
}

VEE_Polygon::~VEE_Polygon()
{
}

/** Receive and process a value. 

    Parameters:

    <UL>
    
    <LI> "velocity", Vector3

    <LI> "addvelocity", Vector3

    <LI> "location", Vector3

    <LI> "addlocation", Vector3

    <LI> "agelimit", float, range: 0 to inf

    <LI> "material", int, range: 0 to inf

    <LI> "stiffness", float, range: 0 to inf

    </UL>

*/
bool VEE_Polygon::setValue(const char *id, VEE_Value *msg)
{
  bool ok = false;  

  if(VEE_Strs::equals(id, VEE_Strs::velocity)) {
    VEE_Vector3 v;

    ok = msg->getVector3(&v);

    if(ok)
      VEE_Algorithm::setAll(m_vertexVelocities.begin(), 
			    m_vertexVelocities.end(), 
			    v);
    return ok;

  }
  if(VEE_Strs::equals(id, VEE_Strs::addvelocity)) {
    VEE_Vector3 v;

    ok = msg->getVector3(&v);

    if(ok)
      VEE_Algorithm::accumulate(m_vertexVelocities.begin(), 
				m_vertexVelocities.end(), 
				v);
  }
  else if(VEE_Strs::equals(id, VEE_Strs::location)) {
    VEE_Vector3 v;
    
    ok = msg->getVector3(&v);

    if(ok)
      move(v - m_location);
  }
  else if(VEE_Strs::equals(id, VEE_Strs::addlocation)) {
    VEE_Vector3 v;
    
    ok = msg->getVector3(&v);

    if(ok)
      move(v);
  }
  else if(VEE_Strs::equals(id, VEE_Strs::stiffness)) {
    float v;

    ok = msg->getFloat(&v);

    if(ok)
      m_stiffness = v;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::transform)) {
    VEE_Matrix3 v;

    ok = msg->getMatrix3(&v);

    if(ok)
      transform(v);
  }
  else if(VEE_Strs::equals(id, VEE_Strs::material)) {
    int v;

    ok = msg->getInt(&v);

    if(ok)
      m_material = v;
  }
  else 
    return VEE_Particle::setValue(id, msg);
  
  return ok;
}

VEE_Polygon * VEE_Polygon::clone() const
{
  VEE_Polygon *p = new VEE_Polygon(*this);

  // VEE_trace("VEE_Polygon::clone # this = %p, new = %p", this, p);

  return p;
}

VEE_Polygon * VEE_Polygon::createNewMesh(const uint horizontalDivisions,
					 const uint verticalDivisions,
					 float width,
					 float height,
					 float ageLimit)
{
  double dx, dy, x, y;
  uint n = (horizontalDivisions+1)*(verticalDivisions+1)
    + (uint)((verticalDivisions+1)/2);
  uint k, i;

//  VEE_Vector3 vertices[n];
//  uint indices[(n-2)*3];
  std::vector<VEE_Vector3> vertices(n);
  std::vector<VEE_Vector2> textureCoordinates(n);

  std::vector<uint> indices((2*horizontalDivisions+1)*verticalDivisions*3);
  std::vector<uint> rowi(verticalDivisions+1);
  
  dx = 1.0/horizontalDivisions;
  dy = 1.0/verticalDivisions;
  y = 0.0;

  k = 0;
  for(uint j=0; j<verticalDivisions+1; j++) {
    rowi[j] = k;

    textureCoordinates[k] = VEE_Vector2(0.0, y);

    /* VEE_trace("textureCoordinates[%d] = %3.3f, %3.3f", k, 
       textureCoordinates[k][0], textureCoordinates[k][1]); */

    vertices[k++] = VEE_Vector3(0.0, 0.0, y * height);
    if(j%2 == 0) {
      x = dx;
      for(i=0; i<horizontalDivisions-1; i++, k++) {
	textureCoordinates[k] = VEE_Vector2(x, y);
	/* VEE_trace("textureCoordinates[%d] = %3.3f, %3.3f", k, 
	   textureCoordinates[k][0], textureCoordinates[k][1]); */
	vertices[k] = VEE_Vector3(x * width, 0.0, y * height);
	x+=dx;
      }
    }
    else {
      x = 0.5*dx;
      for(i=0; i<horizontalDivisions; i++, k++) {
	textureCoordinates[k] = VEE_Vector2(x, y);
	/* VEE_trace("textureCoordinates[%d] = %3.3f, %3.3f", k, 
	   textureCoordinates[k][0], textureCoordinates[k][1]); */
	vertices[k] = VEE_Vector3(x * width, 0.0, y * height);
	x+=dx;
      }
    }
    textureCoordinates[k] = VEE_Vector2(1.0, y);
    /* VEE_trace("textureCoordinates[%d] = %3.3f, %3.3f", k, 
       textureCoordinates[k][0], textureCoordinates[k][1]); */
    vertices[k++] = VEE_Vector3(width, 0.0f, (float) y * height);

    y+=dy;
  }

  k=0;
  for(uint j=0; j<verticalDivisions; j++) {
    if(j%2 == 0) {
      for(i=0; i<horizontalDivisions; i++) {
	indices[k*3] = rowi[j] + i;
	indices[k*3+1] = rowi[j] + i + 1;
	indices[k*3+2] = rowi[j+1] + i + 1;
	k++;
	indices[k*3] = rowi[j] + i;
	indices[k*3+1] = rowi[j+1] + i + 1;
	indices[k*3+2] = rowi[j+1] + i;
	k++;
      }
      indices[k*3] = rowi[j] + i;
      indices[k*3+1] = rowi[j+1] + i + 1;
      indices[k*3+2] = rowi[j+1] + i;
      k++;
    }
    else {
      for(i=0; i<horizontalDivisions; i++) {
	indices[k*3] = rowi[j] + i;
	indices[k*3+1] = rowi[j] + i + 1;
	indices[k*3+2] = rowi[j+1] + i;
	k++;
	indices[k*3] = rowi[j] + i + 1;
	indices[k*3+1] = rowi[j+1] + i + 1;
	indices[k*3+2] = rowi[j+1] + i;
	k++;
      }
      indices[k*3] = rowi[j] + i;
      indices[k*3+1] = rowi[j] + i + 1;
      indices[k*3+2] = rowi[j+1] + i;
      k++;
    }
  }

  VEE_Polygon *mesh = new VEE_Polygon(n, &* vertices.begin(), 
				      k, &* indices.begin(),
				      VEE_Vector3ZERO, ageLimit);
  mesh->setTextureCoordinates(textureCoordinates);
  mesh->m_meshColumns = horizontalDivisions;
  mesh->m_meshRows = verticalDivisions;

  return mesh;
}

VEE_Polygon * VEE_Polygon::createNewQuadMesh(const uint horizontalDivisions,
					     const uint verticalDivisions,
					     float width,
					     float height,
					     float ageLimit)
{
  double dx, dy, x, y;
  uint n = (horizontalDivisions+1)*(verticalDivisions+1);
  uint k, i, j;

  std::vector<VEE_Vector3> vertices(n);
  std::vector<VEE_Vector2> textureCoordinates(n);

  std::vector<uint> indices(2*horizontalDivisions*verticalDivisions*3);
  std::vector<uint> rowi(verticalDivisions+1);

  dx = 1.0/horizontalDivisions;
  dy = 1.0/verticalDivisions;
  y = 0.0;
  k = 0;
  for(j=0; j<verticalDivisions+1; j++) {
    rowi[j] = k;

    x = 0.0;
    for(i=0; i<horizontalDivisions+1; i++, k++) {
      textureCoordinates[k] = VEE_Vector2(x, y);
      vertices[k] = VEE_Vector3(x * width, 0.0, y * height);
      x+=dx;
    }
    y+=dy;
  }

  k = 0;
  for(j=0; j<verticalDivisions; j++) {
    for(i=0; i<horizontalDivisions; i++) {
      indices[k*3] = rowi[j] + i;
      indices[k*3+1] = rowi[j] + i + 1;
      indices[k*3+2] = rowi[j+1] + i;
      k++;
      indices[k*3] = rowi[j] + i + 1;
      indices[k*3+1] = rowi[j+1] + i + 1;
      indices[k*3+2] = rowi[j+1] + i;
      k++;
    }
  }

  VEE_Polygon *mesh = new VEE_Polygon(n, &*vertices.begin(), 
				      k, &* indices.begin(),
				      VEE_Vector3ZERO, ageLimit);
  mesh->setTextureCoordinates(textureCoordinates);
  mesh->m_meshColumns = horizontalDivisions;
  mesh->m_meshRows = verticalDivisions;
  
  return mesh;
}

void VEE_Polygon::resetLinks()
{
  uint a, b, l=0, n;
  bool linkExists = false;

  n = m_nTri*3;

  // Scan through all the triangles and save the links
  // (0-1, 1-2, 2-0), make sure there will be no
  // duplicates in the table.
  for(uint i=0; i<n; i+=3) {
    for(uint j=0, k=1; j<3; j++, k++) {
      if(k>2) {
	k=0;
      }

      // "a" will be the smaller index of the two
      a = m_indices[i+j];
      b = m_indices[i+k];
      if(b<a) {
	a = b;
	b = m_indices[i+j];
      }

      // does this link (a,b) already exist?
      for(uint m=0; m<l*2; m+=2)
	if((a==m_links[m]) && (b==m_links[m+1])) {
	  linkExists = true;
	  break;
	}
      
      if(!linkExists) {
	m_links[l] = a;
	m_links[l+1] = b;
	l+=2;
      }
      else
	linkExists = false;
    }
  }
  m_nLinks = l/2;
}

void VEE_Polygon::resetDistances()
{
  VEE_Vector3 d;

  for(uint i=0; i<m_nLinks; i++) {
    d = m_vertexPositions[m_links[i*2+1]]
      - m_vertexPositions[m_links[i*2]];
    m_distances[i] = d.length();
  }
}

void VEE_Polygon::setVelocity(const VEE_Vector3 &v)
{
  for(uint i=0; i<m_n; i++)
    m_vertexVelocities[i] = v;
}

void VEE_Polygon::initTensions()
{
  for(uint i=0; i<m_nLinks; i++)
    m_tensions[i] = 0.0f;
}

void VEE_Polygon::evaluateTensions()
{
  VEE_Vector3 d;
  float l;

  for(uint i=0; i<m_nLinks; i++) {
    d = m_vertexPositions[m_links[i*2+1]] - m_vertexPositions[m_links[i*2]];
    l = d.length() - m_distances[i];
    // Harmonic force F = -spring_constant * displacement
    m_tensions[i] = m_stiffness*l;
  }
}

void VEE_Polygon::updateVelocity(float dt)
{
  VEE_Vector3 f;

  bzero(&* m_gravForces.begin(), m_n*sizeof(VEE_Vector3));
  bzero(&* m_dragForces.begin(), m_n*sizeof(VEE_Vector3));

  // external forces
  if(VEE_abs(m_mass) > 1.0e-5) {
    sceneBox()->getForces(VEE_FORCE_GRAVITY, this,
			  &* m_gravForces.begin(), m_n,
			  &* m_vertexVelocities.begin(), 
			  &* m_vertexPositions.begin());
  }
  if(VEE_abs(m_area) > 1.0e-5) {
    sceneBox()->getForces(VEE_FORCE_DRAG, this,
			  &* m_dragForces.begin(), m_n,
			  &* m_vertexVelocities.begin(), 
			  &* m_vertexPositions.begin());
  }

  for(uint i=0; i<m_n; i++)
    m_vertexVelocities[i] += ( m_mass*m_gravForces[i] +
			       m_area*m_dragForces[i] ) * dt;

  // internal forces (tensions)
  evaluateTensions();

  for(uint i=0; i<m_nLinks; i++) {
    f = m_vertexPositions[m_links[i*2+1]] - m_vertexPositions[m_links[i*2]];
    f.normalize();

    f = f * 0.5 * m_tensions[i];
    m_vertexVelocities[m_links[i*2]] += f*dt; 
    m_vertexVelocities[m_links[i*2+1]] -= f*dt;
  }
}

void VEE_Polygon::update(float dt)
{
  for(uint i=0; i<m_n; i++)
    m_vertexPositions[i] += m_vertexVelocities[i]*dt;

  m_location = m_vertexPositions[0];
  for(uint i=1; i<m_n; i++)
    m_location += m_vertexPositions[i];
  m_location /= m_n;

  m_age += dt;

  if(m_age > m_ageLimit) {
    // VEE_trace("VEE_Polygon::update # done, age %f > %f", m_age, m_ageLimit);
    m_done = true;
  }
}


void VEE_Polygon::render(VEE_Renderer *r)
{
  if(m_visible)
    r->renderPolygon(this);
}

void VEE_Polygon::printTextureUV()
{
  for(uint i=0; i < m_n; i++)
    VEE_trace("VEE_Polygon::printTextureUV uv[%d] = %3.3f, %3.3f",
	      i, m_vertexTextureCoordinates[i][0], 
	      m_vertexTextureCoordinates[i][1]);
}

void VEE_Polygon::transform(const VEE_Matrix3 &m)
{
  VEE_Vector3 *vertexLoc = &* m_vertexPositions.begin();
  VEE_Vector3 *vertexVel = &* m_vertexVelocities.begin();

  m_location.clear();

  for(uint i=0; i < m_n; i++) {
    *vertexLoc = m * (*vertexLoc);
    *vertexVel = m * (*vertexVel);
    m_location += *vertexLoc;
    vertexLoc++;
    vertexVel++;
  }

  m_location /= m_n;
}

void VEE_Polygon::move(const VEE_Vector3 &v)
{
  VEE_Algorithm::accumulate(m_vertexPositions.begin(), 
			    m_vertexPositions.end(), 
			    v);
  m_location += v;
}
