/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#ifndef VEE_RANDOM_H
#define VEE_RANDOM_H

#include <vee_vector3.h>

/** Base class for random number generators.  

    @author Tommi Ilmonen
*/
class VEE_Random
{
public:
  VEE_Random() {}
  virtual ~VEE_Random() {}

  /** Create a new random number. */
  virtual float rand() = 0;
  virtual VEE_Random * clone() = 0;
};

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** Random numbers with uniform distribution between -1 and 1.  

    @author Tommi Ilmonen
*/
class VEE_RandomUniform : public VEE_Random
{
public:
  VEE_RandomUniform();
  virtual ~VEE_RandomUniform();

  virtual float rand();
  virtual VEE_Random * clone();

  inline float rand01() 
  { 
    unsigned tmp = m_val * m_randMul + 1;
    m_val = tmp;
    return (float) tmp * (1.0 / (float) ((unsigned) 0xffffffff));
  }

  inline float rand0X(float x) 
  { 
    unsigned tmp = m_val * m_randMul + 1;
    m_val = tmp;
    return (float) tmp * (x / (float) ((unsigned) 0xffffffff));
  }

  /// Random numbers between -1 and 1
  inline float rand11() 
  { 
    unsigned tmp = m_val * m_randMul + 1;
    m_val = tmp;
    return (float) tmp * (2.0f / (float) ((unsigned) 0xffffffff)) - 1.0f;
  }


  VEE_Vector3 randomVector11()
  {
    float xr = rand11();
    float yr = rand11();
    float zr = rand11();
    return VEE_Vector3(xr, yr, zr);
  }

  VEE_Vector3 randomVectorInSphere()
  {
    VEE_Vector3 tmp;

    do {
      tmp = randomVector11();
    } while(tmp.lengthSqr() > 1.0f);

    return tmp;
  }
  VEE_Vector3 randomVectorOnSphere()
  {
    VEE_Vector3 tmp = randomVectorInSphere();
    tmp.normalize();
    return tmp;
  }

  void setSeed(unsigned seed) { m_val = seed; }

protected:
  /// Current seed value.
  unsigned m_val;
  static const unsigned m_randMul = 134695621;  
};

extern VEE_RandomUniform VEE_rand;

#endif
