/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_random_force.h>

#include <vee_math.h>
#include <vee_perlin_noise.h>
#include <vee_value.h>

VEE_RandomForce::VEE_RandomForce()
  : m_octaveAttenuation(0.75),
    m_octaves(4)
{
  m_magnitude.make(1.0);
    
  m_locScale.make(1.0f);
}

VEE_RandomForce::~VEE_RandomForce()
{}

bool VEE_RandomForce::setValue(const char *id, VEE_Value *v)
{
  bool ok = VEE_ForceObject::setValue(id, v);

  if(ok) return true;

  if(strcmp(id, "magnitude") == 0) {
    float tmp;
    ok = v->getFloat(&tmp);
    if(ok)
      m_magnitude.make(tmp);
    else {
      VEE_Vector3 vtmp;

      ok = v->getVector3(&vtmp);
      if(ok)
	m_magnitude = vtmp;
    }
  }
  else if(strcmp(id, "attenuation") == 0) {
    float tmp;
    ok = v->getFloat(&tmp);
    if(ok)
      m_octaveAttenuation = tmp;
  }
  else if(strcmp(id, "octaves") == 0) {
    int tmp;
    ok = v->getInt(&tmp);
    if(ok)
      m_octaves = VEE_max(1, tmp);
  }
  else if(strcmp(id, "scale") == 0) {
    VEE_Vector3 tmp;
    ok = v->getVector3(&tmp);
    if(ok)
      m_locScale = tmp;
  }
  
  return ok;
}

VEE_Variable VEE_RandomForce::getVariable(const char *id) const
{
  VEE_Variable v = VEE_ForceObject::getVariable(id);

  if(v.isValid()) return v;

  if(strcmp(id, "magnitude") == 0) 
    v = & m_magnitude;
  else if(strcmp(id, "attenuation") == 0) 
    v = & m_octaveAttenuation;
  else if(strcmp(id, "octaves") == 0) 
    v = & m_octaves;
  else if(strcmp(id, "scale") == 0)
    v = & m_locScale;

  return v;
}

VEE_Object * VEE_RandomForce::clone() const
{
  return new VEE_RandomForce(*this);
}

void VEE_RandomForce::getForce(VEE_Vector3 *force,
			       const VEE_Vector3 &,
			       const VEE_Vector3 &location)
{
  VEE_Vector3 randloc(location);
  randloc.scale(m_locScale);

  const float e = 1.0e-2f; // epsilon
  const float onePerE = 1.0f / e;

  float rc = VEE_PerlinNoise::noise3d(randloc.x, randloc.y, randloc.z, 
				      m_octaveAttenuation, m_octaves);
  float rx = VEE_PerlinNoise::noise3d(randloc.x + e, randloc.y, randloc.z, 
				      m_octaveAttenuation, m_octaves);
  float ry = VEE_PerlinNoise::noise3d(randloc.x, randloc.y + e, randloc.z, 
				      m_octaveAttenuation, m_octaves);
  float rz = VEE_PerlinNoise::noise3d(randloc.x, randloc.y, randloc.z + e, 
				      m_octaveAttenuation, m_octaves);

  (*force) [0] = m_magnitude.x * (rc - rx) * onePerE;
  (*force) [1] = m_magnitude.y * (rc - ry) * onePerE;
  (*force) [2] = m_magnitude.z * (rc - rz) * onePerE;
}

void VEE_RandomForce::boundingBoxOfForce(VEE_BoundingBox *box)
{
  if(m_volume.ptr()) {
    m_volume.ptr()->getBoundingBox(box);
    box->move(m_location);
  }
  else {
    VEE_Vector3 tmp;
    tmp.make(1.0e+10);
    box->set(-tmp, tmp);
  }
}

VEE_Object * VEE_RandomForce::factory()
{
  return new VEE_RandomForce();
}
