/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_error.h>
#include <vee_misc.h>
#include <vee_scene.h>

#include <stdio.h>
#include <sys/time.h>

#include <algorithm>
#include <typeinfo>

#ifndef VEE_FAST
long VEE_SceneBox::m_count = 0;
#endif

/** Number of child-boxes in each scene box. 

    @see VEE_SceneBox
*/

/** Creates and uninitialized (thus unusable) VEE_SceneBox. */

VEE_SceneBox::VEE_SceneBox()
  : m_parent(0),
    m_children(0),
    m_frame(0)
{
#ifndef VEE_FAST
  m_count++;
#endif
  m_tvLast.getTimeOfDay();

  m_bbox.low().make(-10.0, -10.0, -10.0);
  m_bbox.high().make(10.0, 10.0, 10.0);
}

/** Creates an initialized VEE_SceneBox. */

VEE_SceneBox::VEE_SceneBox(VEE_SceneBox *parent, 
			   const VEE_Vector3 &low,
			   const VEE_Vector3 &high)
  : m_parent(parent),
    m_children(0),
    m_bbox(low, high),
    m_frame(0)
{
#ifndef VEE_FAST
  m_count++;
#endif
  m_tvLast.getTimeOfDay();
}

/** Deletes this box, also deleting any child boxes. */

VEE_SceneBox::~VEE_SceneBox()
{
#ifndef VEE_FAST
  m_count--;
  // VEE_trace("VEE_SceneBox::~VEE_SceneBox %p", this);
#endif
  VEE_SceneBox * tmp = m_children;

  m_children = 0;

  delete [] tmp;

  for(m_forces.toFirst(); !m_forces.isAtEnd(); m_forces.advance()) {
    
    VEE_ForceObject *fobj = m_forces.current();

    fobj->unlinkBox(this);

    if(m_parent) {

      if(!fobj->isLinked(m_parent)) {
	fobj->unlinkBoxes();
	VEE_BoundingBox bbox;
	fobj->boundingBoxOfForce(&bbox);
	m_parent->addForceHere(fobj, &bbox, this);
      }
    }
  }
}

/** Initialize a scene box. 

    @arg parent The parent (container) of this box.

    @arg low The location of the corner with the lowest x, y and z.

    @arg high The location of the corner with the highest x, y and z.
*/

void VEE_SceneBox::init(VEE_SceneBox *parent, 
			const VEE_Vector3 &low,
			const VEE_Vector3 &high)
{
  m_parent = parent;
  m_bbox.low() = low; 
  m_bbox.high() = high;
}

/** Adds an object to the box-graph. 
    
    The object may end up anywhere in the graph. 

    The scene-box field of the argument object is also set to the
    correct value.

    @return If the object is outside the largest scene box, then false
    is returned. Usually this method returns true.

    @see VEE_SceneBox::addChildObject
*/
bool VEE_SceneBox::addObject(VEE_Object *o)
{
#ifndef VEE_FAST
  if(!o)
    VEE_fatal("VEE_SceneBox::addObject # NULL argument");

  /*
  VEE_ForceObject *force = dynamic_cast<VEE_ForceObject *> (o);

  if(force) {
    VEE_fatal("VEE_SceneBox::addObject # Do not put forces here !!");
    }*/

#endif

  VEE_Vector3 *location = & o->location();

  if(m_bbox.contains(*location)) {
    addChildObject(o);
    return true;
  }
  else if(m_parent) {
    return m_parent->addObject(o);
  }
  else {
    o->setSceneBox(this);
    m_objects.add(o);
  }
  return true;
}
/** Adds an object to the box-graph. 
    
    The object may end up anywhere in the graph. 

    The scene-box field of the argument object is also set to the
    correct value.

    @return If the object is outside the largest scene box, then false
    is returned. Usually this method returns true.

    @see VEE_SceneBox::addChildObject
*/
bool VEE_SceneBox::addNewObject(VEE_Object *o)
{
  if(!o)
    VEE_fatal("VEE_SceneBox::addNewObject # NULL argument");

  if(o->sceneBox()) {
    VEE_error("VEE_SceneBox::addNewObject # "
	      "object %p (%s) is already in a scene", o, typeid(*o).name());
    return false;
  }

  VEE_ForceObject *force = dynamic_cast<VEE_ForceObject *> (o);

  if(force && force->getForceType() != VEE_FORCE_HIDDEN) {
    addNewForce(force);
    return true;
  }

  VEE_Vector3 *location = & o->location();

  if(m_bbox.contains(*location)) {
    addChildObject(o);
    return true;
  }
  else if(m_parent) {
    return m_parent->addObject(o);
  }
  else {
    o->setSceneBox(this);
    m_objects.add(o);
  }
  return true;
}

/** DO NOT CALL. This method is here only to make sure one does not
    accidentally add force objects with "addObject". */
 bool VEE_SceneBox::addObject(VEE_ForceObject *)
{
  VEE_fatal("VEE_SceneBox::addObject # DO NOT CALL");
  return false;
}

/** Use this method to add new force objects to the scene. */
bool VEE_SceneBox::addNewForce(VEE_ForceObject *o)
{
  VEE_BoundingBox bbox;

  o->boundingBoxOfForce(&bbox);
  addForceSomewhere(o, &bbox);
  // addForceHere(o, &bbox);

  addObject((VEE_Object *) o);

  return true;
}

/** Adds an object to either this box or below in the hierarchy. 

    The scene-box field of the argument object is also set to the
    correct value.
      
    It is assumed that the object is truly within this box. It is also
    assumed that the object is not a child of this before this call.

    The argument is not checked for correctness.

    @see VEE_SceneBox::addObject
*/

void VEE_SceneBox::addChildObject(VEE_Object *o)
{
  if(m_children) {
    VEE_Vector3 *location = & o->location();
    VEE_Vector3 midPoint(m_bbox.center());

    uint x = midPoint[0] > location->x ? 0 : 1;
    uint y = midPoint[1] > location->y ? 0 : 1;
    uint z = midPoint[2] > location->z ? 0 : 1;
    
    getChild(x, y, z).addChildObject(o);
  }
  else {
    // VEE_trace("VEE_SceneBox::addChildObject # %u %p", m_objects.size(), o);
    o->setSceneBox(this);
    m_objects.add(o);
  }
}

/** Update the location of an object in the VEE_SceneBox graph.
    
    The object must be a direct child of this box. 

    @return This method returns true if the location was modified.
*/
bool VEE_SceneBox::updateSceneLocation(VEE_Object *o)
{
  VEE_Vector3 location = o->location();

  if(m_bbox.contains(location) || !m_parent)
    return false;

  // VEE_trace("VEE_SceneBox::updateSceneLocation # Updating %p", o);

  removeChildObject(o);

  addObject(o);
  
  return true;
}


/** Removes an object (or force) anywhere down in the hierarchy. 
*/
void VEE_SceneBox::removeObject(VEE_Object *o)
{
	if(!o)
		return;
	
	// find root scenebox
	VEE_SceneBox *sbRoot=this;
	while(sbRoot->m_parent)
		sbRoot=sbRoot->m_parent;

	// remove force recursively
	sbRoot->removeObjectForce((VEE_ForceObject *)o);

	// remove object recursively.
	// VEE_ForceObjects include both a force and an object.
	sbRoot->removeObjectObject(o);
}

/// internal function to blindly remove an object
void VEE_SceneBox::removeObjectObject(VEE_Object *o)
{
	m_objects.remove(o);
	o->setSceneBox(0);

	if(m_children)
		for(uint i=0; i < VEE_SB_CHILDREN; i++)
			m_children[i].removeObjectObject(o);
}
void VEE_SceneBox::removeObjectForce(VEE_ForceObject *o)
{
	m_forces.remove(o);
	o->setSceneBox(0);

	if(m_children)
		for(uint i=0; i < VEE_SB_CHILDREN; i++)
			m_children[i].removeObjectForce(o);
}



/** Removes a direct child object. 

    If VEE is compiled with error reporting this method will both
    check the argument for sanity and check that the operation was
    successful. in error conditions warning values will be printed
    as necessary.
*/

void VEE_SceneBox::removeChildObject(VEE_Object *o)
{
  // Find the child object:

#ifndef VEE_FAST
  if(o->sceneBox() != this)
    VEE_fatal("VEE_SceneBox::removeChildObject o->sceneBox() != this");
  
  iterator it = std::find(m_objects.begin(), m_objects.end(), o);

  while(it == m_objects.end()) {
    VEE_error("VEE_SceneBox::removeChildObject failed %p", o);
    return;
  }
  
  o->setSceneBox(0);
#endif

  m_objects.remove(o);
}


void VEE_SceneBox::addForceHere(VEE_ForceObject *o, 
				const VEE_BoundingBox *bbox, 
				const VEE_SceneBox * caller)
{ 
  /* #ifndef VEE_FAST
  if(m_forces.contains(o)) {
    VEE_trace("VEE_SceneBox::addForceHere # %d forces", (int) m_forces.size());
    VEE_fatal("VEE_SceneBox::addForceHere # %p of type %s is already",
	      o, typeid(*o).name());
  }
  // VEE_trace("VEE_SceneBox::addForceHere # %p to %p by %p", o, this, caller);
  if(caller == this) {
    VEE_fatal("VEE_SceneBox::addForceHere # caller == this");
    
  }
  #endif */

  m_forces.add(o);
  o->linkBox(this);

  // VEE_trace("VEE_SceneBox::addForceHere # %u forces", m_forces.size());

  if(m_forces.size() >= 2000)
    VEE_fatal("VEE_SceneBox::addForceHere # m_forces.size() >= 2000");

  for(uint i=0; m_children && i < VEE_SB_CHILDREN; i++) {
    VEE_SceneBox *box = m_children + i;
    
    if(box != caller && box->m_bbox.intersects(*bbox))
      box->addForceHere(o, bbox, 0 /* this */);
  }

  // if(m_parent && m_parent != caller)
  /* if(m_parent && m_parent != caller && !m_bbox.contains(*bbox))
     m_parent->addForceHere(o, bbox, this);
  */
}

void VEE_SceneBox::addForceSomewhere(VEE_ForceObject *o, 
				     const VEE_BoundingBox *bbox, 
				     const VEE_SceneBox * caller)
{
  // if(m_bbox.intersects(*bbox) || !m_parent)
  if(m_bbox.contains(*bbox) || !m_parent)
    addForceHere(o, bbox, caller);
#ifndef VEE_FAST
  else if(m_parent == caller) 
    VEE_error("VEE_SceneBox::addForceSomewhere # this = %p, caller = %p",
	      this, caller);
#endif
  else if(m_parent)
    m_parent->addForceSomewhere(o, bbox, 0 /* this*/ );
}

void VEE_SceneBox::forgetForce(VEE_ForceObject *o) 
{
#ifndef VEE_FAST
  if(!m_forces.contains(o))
    VEE_fatal("VEE_SceneBox::forgetForce # %p of type %s is not here",
	      o, typeid(*o).name());
#endif

  m_forces.remove(o); 
}

/** Get all the child objects of this box recursively. */

void VEE_SceneBox::getObjects(ObjectStore * target)
{
  if(m_children) {
    for(uint i=0; i < VEE_SB_CHILDREN; i++) 
      m_children[i].getObjects(target);
  }

  iterator it = begin();
  iterator sentinel = end();

  while(it != sentinel) {
    target->add(*it);
    it++;
    // VEE_trace("VEE_SceneBox::getObjects # added %d objects", os);
  }
}

void VEE_SceneBox::getForces(std::set<VEE_ForceObject *> * target)
{
  for(m_forces.toFirst(); !m_forces.isAtEnd(); m_forces.advance()) {
    VEE_ForceObject *force = m_forces.current();
    target->insert(force);
  }

  for(uint i=0; i < VEE_SB_CHILDREN && m_children; i++)
    m_children[i].getForces(target);
}

/** VEE::SceneBox::getForce searches for all the forces of the
    specified force type and returns a single sum of these
    vectors, the force field, in VEE_Vector3 *force.
*/
void VEE_SceneBox::getForce(VEE_ForceType forceType,
			    const VEE_Object *object,
			    VEE_Vector3 *force,
			    const VEE_Vector3 &velocity,
			    const VEE_Vector3 &location)
{
  VEE_Vector3 f(VEE_Vector3ZERO);
  
  for(ForceStore::iterator it = m_forces.begin(); it != m_forces.end(); it++) {
    
    VEE_ForceObject *fojb = ForceStore::getObject(it);

    if(fojb->getForceType() == forceType && fojb != object ) {
      if(!fojb->done())
        fojb->getForce(&f, velocity, location);
      *force += f;
    
    }
  }

  /* for(m_forces.toFirst(); !m_forces.isAtEnd(); m_forces.advance()) {

    VEE_ForceObject *fojb = m_forces.current();

    if(fojb->getForceType() == forceType && fojb != object ) {
      fojb->getForce(&f, velocity, location);
      *force += f;
    }
    }*/
}

/** VEE_SceneBox::getForces calculates the magnitude and the
    direction of the force field of the specified type acting
    on the queried locations. Writes n number of resultant
    vectors to the structure VEE_Vector *force.
*/
uint VEE_SceneBox::getForces(VEE_ForceType forceType,
			     const VEE_Object *object,
			     VEE_Vector3 *force,
			     uint n,
			     const VEE_Vector3 *velocity,
			     const VEE_Vector3 *location)
{
  // VEE_trace("VEE_SceneBox::getForces 0x%p %d", this, n);

  uint p = getForcesDown
    (forceType, object, force, n, velocity, location);

  if(p < n) {

    if(!m_parent) {
      // VEE_fatal("VEE_SceneBox::getForces 0x%p %d %d", this, p, n);
      return p;
    }

    p += m_parent->getForces(forceType, 
			     object, 
			     force + p, 
			     n - p, 
			     velocity + p, 
			     location + p);
  }
  return p;
}

/** Gets a number of forces downwards (recursive). */
uint VEE_SceneBox::getForcesDown(VEE_ForceType forceType,
				 const VEE_Object *object,
				 VEE_Vector3 *force,
				 uint n,
				 const VEE_Vector3 *velocity,
				 const VEE_Vector3 *location)
{
  // VEE_trace("VEE_SceneBox::getForcesDown 0x%p %d", this, n);

  if(!m_children) {
    return getForcesNR(forceType, object, force, n, velocity, location);
  }
  else {
    VEE_Vector3 midPoint(m_bbox.center());

    int x, y, z;
    
    for(uint i=0; i < n;) {

      if(!m_bbox.contains(*location))
	return i;

      x = midPoint[0] > location->x ? 0 : 1;
      y = midPoint[1] > location->y ? 0 : 1;
      z = midPoint[2] > location->z ? 0 : 1;
      
      // Get n points at one blow:
      uint processed = getChild(x, y, z).getForcesDown(forceType, 
						       object, 
						       force, 
						       n-i,
						       velocity,
						       location);
      i += processed;

      force += processed;
      velocity += processed;
      location += processed;

#ifndef VEE_FAST
      if(!processed) VEE_fatal("VEE_SceneBox::getForcesDown 0x%p %d %d", 
			       this, i, n);
#endif
    }

    return n;
  }
}

/** Gets a number of forces (non-recursive). */
uint VEE_SceneBox::getForcesNR(VEE_ForceType forceType,
			       const VEE_Object *object,
			       VEE_Vector3 *force,
			       uint n,
			       const VEE_Vector3 *velocity,
			       const VEE_Vector3 *location)
{
  // VEE_trace("VEE_SceneBox::getForcesNR 0x%p %d", this, n);

  for(uint i=0; i < n; i++) {

    if(m_bbox.contains(*location) || !m_parent)
      getForce(forceType, object, force, *velocity, *location);
    else
      return i;
    
    force++;
    velocity++;
    location++;
  }
  return n;
}
    
/** Total number of child objects (recursive). */

uint VEE_SceneBox::totalChildren()
{
    uint sum = m_objects.size();

  if(isSplit()) {    
    for(int i=0; i < VEE_SB_CHILDREN; i++) {
      sum += m_children[i].totalChildren();
    }
  }

  return sum;
}

uint VEE_SceneBox::totalForces()
{
  std::set<VEE_ForceObject *> forces;
  getForces(&forces);
  return forces.size();
}

/** Updates the current box-graph. New branches are grown and shrunk
    as necessary. 

    This method uses VEE_SceneBox::trySplit and
    VEE_SceneBox::tryCollapse to do the real work.

    @arg splitTreshold Number of elements that are needed, before a
    paritular box should be split. If this (or any child box) has
    greater number of direct child objects, then the box is split.
    
    @arg collapseTreshold The number of objects that is the minimum
    before collapsing child boxes into this box.

    @arg allowRecursionLevels The number of levels that the tree is
    allowed to grow downdards. If this number is non-zero, then new
    levels can be added.

    @return This method returns true if this box or any child box was
    modified.
*/

bool VEE_SceneBox::updateTree(uint splitTreshold, 
			      uint collapseTreshold, 
			      int allowRecursionLevels)
{
  bool b1 = trySplit(splitTreshold, allowRecursionLevels);
  bool b2 = tryCollapse(collapseTreshold);

  return b1 || b2;
}

/** Tries to split the box to eight child boxes recursively. 
    
    @return This method returns true if the tree box or any child box
    was modified.

    @see VEE_SceneBox::updateTree.
*/

bool VEE_SceneBox::trySplit(uint treshold, int allowRecursionLevels)
{
  
  /* if(m_objects.size() > 50) 
     VEE_trace("VEE_SceneBox::trySplit %d %d %d", 
     m_objects.size(), treshold, allowRecursionLevels); */

  if(! --allowRecursionLevels) {
    if(!m_children)
      return false;      

    // Collapse the children:

    uint i;
    
    for(i=0; i < VEE_SB_CHILDREN; i++) {
      m_children[i].getObjects(& m_objects);
      // m_children[i].getForceObjects(& m_forces);
    }
    
    VEE_SceneBox::iterator it = begin();
    VEE_SceneBox::iterator sentinel = end();
  
    while(it != sentinel) {
      (*it)->setSceneBox(this);
      it++;
    }

    VEE_SceneBox *children = m_children;

    m_children = 0;

    delete [] children;

    return false;
  }

  // Try to split all children
  if(m_children) {
    bool change = false;
    
    for(int i = 0; i < VEE_SB_CHILDREN; i++) { 
      bool r = m_children[i].trySplit
	(treshold, allowRecursionLevels);
      
      change = change || r;
    }
    return change;
  }

  if(m_objects.size() < treshold) {
    return false;
  }

  /* if(m_objects.size() > 50) 
     VEE_trace("VEE_SceneBox::trySplit %d %d %d", 
     m_objects.size(), treshold, allowRecursionLevels);*/

  /* Split this. 
     
  1) Create child boxes.
  2) Fill them.
  3) Recursion.

  */

  // Init child boxes:
  int x, y, z, i;

  VEE_Vector3 midPoint(m_bbox.center());

  m_children = new VEE_SceneBox[VEE_SB_CHILDREN];

  for(x = 0; x < VEE_SB_SPLITS; x++ ) {

    float lowx, highx;

    if(x == 0) {
      lowx  = m_bbox.low()[0];
      highx = midPoint[0];
    }
    else {
      lowx  = midPoint[0];
      highx = m_bbox.high()[0];
    }

    for(y = 0; y < VEE_SB_SPLITS; y++ ) {
	
      float lowy, highy;

      if(y == 0) {
	lowy  = m_bbox.low()[1];
	highy = midPoint[1];
      }
      else {
	lowy  = midPoint[1];
	highy = m_bbox.high()[1];
      }

      for(z = 0; z < VEE_SB_SPLITS; z++ ) {
	  
	float lowz, highz;
	  
	if(z == 0) {
	  lowz  = m_bbox.low()[2];
	  highz = midPoint[2];
	}
	else {
	  lowz  = midPoint[2];
	  highz = m_bbox.high()[2];
	}

	getChild(x, y, z).init(this, 
			       VEE_Vector3(lowx, lowy, lowz), 
			       VEE_Vector3(highx, highy, highz));
      }  // z
    }  // y
  }  // x
    
  // Push all objects to the the child boxes:
  for(iterator it = begin(); it != end(); it++) {
    
    VEE_Object * o = *it;

    VEE_Vector3 *location = & o->location();

    x = midPoint[0] > location->x ? 0 : 1;
    y = midPoint[1] > location->y ? 0 : 1;
    z = midPoint[2] > location->z ? 0 : 1;

    getChild(x, y, z).addChildObject(o);
  }

  // No more direct child objects
  m_objects.clear();

  tryToSendForcesDown();

  // Now try to split the child boxes:
  for(i=0; i < VEE_SB_CHILDREN && allowRecursionLevels; i++)
    m_children[i].trySplit(treshold, allowRecursionLevels);
  
  return true;
}

/** Tries to collapse the child boxes of this box recursively.
    
    @return This method returns true if this box or any child box
    was modified.

    @see VEE_SceneBox::updateTree.
*/

bool VEE_SceneBox::tryCollapse(uint treshold)
{
  uint i;

  if(!m_children)
    return false;
  
  if(totalChildren() > treshold) {
    
    bool change = false;

    for(i = 0; i < VEE_SB_CHILDREN; i++) {
      bool r = m_children[i].tryCollapse(treshold);
      change = change || r;
    }

    return change;
  }
  /*
#ifndef VEE_FAST
  VEE_trace("VEE_SceneBox::tryCollapse # collapsing this: 0x%p %d %d", 
  this, totalChildren(), VEE_Object::count()); 
#endif
  */
  // Now collapse children:
  for(i=0; i < VEE_SB_CHILDREN; i++) {
    m_children[i].getObjects(& m_objects);
    // m_children[i].getForceObjects(& m_forces);
  }

  for(iterator it = begin(); it != end(); it++)
    (*it)->setSceneBox(this);
  
  VEE_SceneBox * tmp = m_children;

  m_children = 0;

  delete [] tmp;
  
  /* for(i = 0; i < m_forces.size(); i++)
     m_forces[i]->setSceneBox(this); */
  
  /*
#ifndef VEE_FAST
  VEE_trace("VEE_SceneBox::tryCollapse # collapsing DONE %d %d %d %d", 
	    m_objects.size(), m_forces.size(), 
	    totalChildren(), VEE_Object::count());
#endif
  */

  return true;
}

/** Writes out all the objects found inside the scenebox. */
bool VEE_SceneBox::write(std::ostream *s)
{
  if(!s) s = & std::cout;

  char buf[32];

  sprintf(buf, " 0x%p ", this);

  (*s) << "SceneBox " << buf
       << m_bbox.low()[0] << " " 
       << m_bbox.low()[1] << " " 
       << m_bbox.low()[2] << "  "
       << m_bbox.high()[0] << " " 
       << m_bbox.high()[1] << " " 
       << m_bbox.high()[2] << "\n";

  for(m_forces.toFirst(); !m_forces.isAtEnd(); m_forces.advance()) {

    VEE_ForceObject * fobj = m_forces.current();

    VEE_Vector3 *f = & fobj->location();
    
    sprintf(buf, "0x%p", fobj);

    (*s) << "force " << buf << "  "
	 <<  typeid(*fobj).name() << " located in: " 
	 << f->x << " " << f->y << " " << f->z << "\n";
  }

  if(m_children) {

    for(uint i=0; i < VEE_SB_CHILDREN; i++) {
      (*s) << "child box " << i << "\n";
      m_children[i].write(s);
    }
  }

  for(iterator it = begin(); it != end(); it++) {
    VEE_Object *obj = *it;
    
    VEE_Vector3 *p = & obj->location();
    
    (*s) << "child " << typeid(*obj).name() << " located in: " 
	 << p->x << " " << p->y << " " << p->z << "\n";
  }
  return true;
}


uint VEE_SceneBox::numChildBoxes()
{
  uint count = 0;

  if(m_children) {
    count += VEE_SB_CHILDREN;

    for(uint i=0; i < VEE_SB_CHILDREN; i++)
      count += m_children[i].numChildBoxes();
  }

  return count;
}

void VEE_SceneBox::updateVelocities(float dt)
{
  // uint i = m_forces.size();

  // while(i--) {
    // VEE_trace("VEE_SceneBox::updateVelocities %p", m_forces[i]);
    //m_forces[i]->updateVelocity(dt);
    // VEE_trace("VEE_SceneBox::updateVelocities %p done", m_forces[i]);
  //}
  if(m_children)
    for(uint i=0; i < VEE_SB_CHILDREN; i++)
      m_children[i].updateVelocities(dt);
  
  iterator sentinel = end();
  
  for(iterator it = begin(); it != sentinel; it++) 
    (*it)->updateVelocity(dt);
}

void VEE_SceneBox::updateR(float dt)
{
  if(m_children)
    for(uint i=0; i < VEE_SB_CHILDREN; i++)
      m_children[i].updateR(dt);

  // VEE_Object *last = 0;

  for(m_objects.toFirst(); !m_objects.isAtEnd(); ) {
    VEE_Object *obj = m_objects.current();

    if(!obj->done()) {
      obj->update(dt);
	} else {
      if(obj->transient()) {
        // VEE_trace("VEE_SceneBox::updateR obj %p deleted", obj);
      
        m_objects.removeCurrent();
        delete obj;

      }
    }

    if(!m_objects.isAtEnd())
      m_objects.advance();
  }
}

/** Updates the registering locations of all the objects
    and force objects. */

void VEE_SceneBox::updateScene(int frame)
{
  // VEE_trace("VEE_SceneBox::updateScene");
  // write(& std::cout); std::cout << std::endl;

  if(m_children) {
    for(uint i=0; i < VEE_SB_CHILDREN; i++) {
      m_children[i].updateScene(frame);
    }
  }
  else {

    VEE_Object * last = 0;

    for(m_objects.toFirst(); !m_objects.isAtEnd(); ) {
      // for(uint i=0; i < m_objects.size(); i++) {
      VEE_Object * obj = m_objects.current();
      if(obj != last)
	updateSceneLocation(obj);
      
      if(!m_objects.isAtEnd())
	m_objects.advance();

      last = obj;
    }
  }
  
  /* VEE_trace("VEE_SceneBox::updateScene 2");
     write(& std::cout); std::cout << std::endl;

  VEE_trace("VEE_SceneBox::updateScene 3");*/
#ifndef VEE_FAST
  uint count = 0;
#endif

  for(m_forces.toFirst(); !m_forces.isAtEnd(); m_forces.advance()) {
    
    VEE_ForceObject *fobj = m_forces.current();

    // VEE_trace("VEE_SceneBox::updateScene 4 0x%p", fobj);

    fobj->updateSceneLocation(frame);

    /* The contents of the m_forces may have changed in previous
       command. Lets make sure we do not iterate past the end of the
       storage area. */

    if(m_forces.isAtEnd()) return;

#ifndef VEE_FAST
    if(++count > 3000) 
      VEE_fatal("VEE_SceneBox::updateScene # Too many forces");
#endif

    // VEE_trace("VEE_SceneBox::updateScene 4 %s", typeid(*fobj).name());
  }
}

/** The main updating method. */

void VEE_SceneBox::update(uint splitTreshold, 
			  uint collapseTreshold, 
			  int allowRecursionLevels)
{
  VEE_Time tmp;

  tmp.getTimeOfDay();
  
  double dt = m_tvLast.secsTo(tmp);
  
  if(dt > 0.0) {

    m_frame++;

    m_tvLast = tmp;
    
    updateVelocities(dt);
    updateR(dt);
    updateScene(m_frame);
    updateTree(splitTreshold, collapseTreshold, allowRecursionLevels);
  }
}

void VEE_SceneBox::update(uint splitTreshold, 
			  uint collapseTreshold, 
			  int allowRecursionLevels,
			  float dt)
{
  m_frame++;

  // write(&std::cout); std::cout << std::endl;

  updateVelocities(dt);
  updateR(dt);

  // write(&std::cout); std::cout << std::endl;

  updateScene(m_frame);
  updateTree(splitTreshold, collapseTreshold, allowRecursionLevels);

  m_tvLast.getTimeOfDay();
}



void VEE_SceneBox::refreshTimeLastUpdate()
{
  m_tvLast.getTimeOfDay();
}

void VEE_SceneBox::tryToSendForcesDown()
{
  if(!m_children) return;

  VEE_BoundingBox bbox;

  for(m_forces.toFirst(); !m_forces.isAtEnd(); m_forces.advance()) {
    
    VEE_ForceObject *fobj = m_forces.current();
    
    fobj->boundingBoxOfForce(&bbox);
    
    for(uint i=0; i < VEE_SB_CHILDREN; i++) {

      if(m_children[i].m_bbox.intersects(bbox))
	m_children[i].addForceHere(fobj, &bbox, this);
    }
  }
}

