
/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_tail_particle.h>

#include <vee_error.h>
#include <vee_scene.h>
#include <vee_value.h>

#include <mj_ringbuffertmpl.hpp>

VEE_TailParticle::VEE_TailParticle()
  : m_nofPoints(0),
    // m_threshold(0.2),
    m_dtSum(0.0)
{
  resizeTail(3);
  m_objectType = VEE_OBJECT_TAIL_PARTICLE;
  m_prevLocation.clear();
}

VEE_TailParticle::~VEE_TailParticle()
{}

bool VEE_TailParticle::setValue(const char *id, VEE_Value *v)
{
  bool ok = VEE_Particle::setValue(id, v);

  if(ok) return true;

  if(strcmp(id, "tailsize") == 0) {
    int tmp;

    ok = v->getInt(&tmp);

    if(ok)
      resizeTail(VEE_max(2, tmp));
  }
  /* else if(strcmp(id, "threshold") == 0) {
    float tmp;

    ok = v->getFloat(&tmp);

    if(ok)
      m_threshold = tmp;
      }*/

  return ok;
}

VEE_Variable VEE_TailParticle::getVariable(const char *id) const
{
  VEE_Variable v = VEE_Particle::getVariable(id);

  if(v.type() != VEE_VAR_INVALID)
    return v;

  if(strcmp(id, "tailsize") == 0) {
    v = (int) m_nofPoints;
  }
  /* else if(strcmp(id, "threshold") == 0) {
    v = m_threshold;
    }*/

  return v;
}

void VEE_TailParticle::update(float dt)
{
  VEE_Particle::update(dt);

  m_dtSum += dt;

  const float dtThreshold = 0.02;

  if(m_dtSum < dtThreshold)
    return;

  Point * p;

  // p = & m_points.getNewest(1);
  
  if(m_dtSum < dtThreshold * 2.0 || !m_points.getSampleCount()) {
    /* if((m_prevLocation - m_location).length() < m_threshold)
       return;*/
    
    m_points.advance();
    p = & m_points.getNewest();

    p->m_location = m_location;
    p->m_velocity = m_velocity;
    p->m_color    = m_color;
    m_dtSum -= dtThreshold;
  }
  else {
    Point * pold = & m_points.getNewest();
    uint n = (uint) (m_dtSum / dtThreshold);

    for(uint i=0; i < n; i++) {
      m_points.advance();
      p = & m_points.getNewest();
      float newWeight = (float) (i + 1) / (float) n;
      float oldWeight = 1.0f - newWeight;

      p->m_location = m_location * newWeight + pold->m_location * oldWeight;
      p->m_velocity = m_velocity * newWeight + pold->m_velocity * oldWeight;
      p->m_color    = m_color.mulRgba(newWeight).
	addRgba(pold->m_color.mulRgba(oldWeight));
    }

    m_dtSum -= dtThreshold * n;
  }


  m_prevLocation = m_location;
}

VEE_Object *VEE_TailParticle::clone() const
{
  return new VEE_TailParticle(*this);
}

VEE_Object * VEE_TailParticle::factory()
{
  return new VEE_TailParticle();
}

void VEE_TailParticle::resizeTail(uint n)
{
  m_nofPoints = n;
  m_points.resize(n);

  Point p;
  p.m_location.clear();
  p.m_velocity.clear();
  p.m_color = VEE_Rgba::White;
  
  m_points.setAll(p);

  m_points.setSampleCount(0);
}

template class MJ_RingBufferDelay<VEE_TailParticle::Point>;

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


VEE_TailParticle2::VEE_TailParticle2()
  : m_springConstant(0.3),
    m_tailForceAttenuation(0.8)
{}

VEE_TailParticle2::~VEE_TailParticle2()
{}

bool VEE_TailParticle2::setValue(const char *id, VEE_Value *v)
{
  bool ok = VEE_TailParticle::setValue(id, v);

  if(ok) return true;

  if(strcmp(id, "springconstant") == 0) {
    float tmp;

    ok = v->getFloat(&tmp);

    if(ok)
      m_springConstant = tmp;
  }
  else if(strcmp(id, "forceattenuation") == 0) {
    float tmp;

    ok = v->getFloat(&tmp);

    if(ok)
      m_tailForceAttenuation = tmp;
  }

  return ok;
}

VEE_Variable VEE_TailParticle2::getVariable(const char *id) const
{
  VEE_Variable v = VEE_TailParticle::getVariable(id);

  if(v.type() != VEE_VAR_INVALID)
    return v;
  
  if(strcmp(id, "springconstant") == 0)
    v = m_springConstant;
  else if(strcmp(id, "forceattenuation") == 0)
    v = m_tailForceAttenuation;

  return v;
}

void VEE_TailParticle2::update(float dt)
{
  ulong count = m_points.getSampleCount();

  VEE_TailParticle::update(dt);
  
  if(count != m_points.getSampleCount()) {
    m_links.advance();
    
    float d = 
      (m_points.getNewest().location() - m_points.getNewest(1).location())
      .length();
  
    m_links.getNewest().m_distance = d;
  }
  
  uint n = pointsToRender();

  for(uint i=1; i < n; i++) {
    Point * leader = & m_points.getNewest(i);
    Point * hunter = & m_points.getNewest(i+1);
    
    {
      VEE_Vector3 grav, drag;
      
      grav.clear();
      drag = grav;
      
      if(VEE_abs(m_mass) > 1.0e-5) {
	sceneBox()->getForce(VEE_FORCE_GRAVITY, this, 
			     &grav, hunter->m_velocity, hunter->m_location);
      }
    
      if(VEE_abs(m_area) > 1.0e-5) {
	sceneBox()->getForce(VEE_FORCE_DRAG, this, 
			     &drag, hunter->m_velocity, hunter->m_location);
      }
      
      hunter->m_velocity += (m_mass * grav + m_area * drag) * 
	dt * m_tailForceAttenuation;
      
      hunter->m_velocity *= 0.7;
      
      hunter->m_location += hunter->m_velocity * dt;
    }

    VEE_Vector3 diff = leader->m_location - hunter->m_location;

    diff.normalize(m_links.getNewest(i).m_distance);

    // hunter->m_location = leader->m_location - diff;
    // hunter->m_velocity = diff;
  }
}
  
VEE_Object *VEE_TailParticle2::clone() const
{
  return new VEE_TailParticle2(*this);
}

VEE_Object * VEE_TailParticle2::factory()
{
  return new VEE_TailParticle2();
}

void VEE_TailParticle2::resizeTail(uint n)
{
  VEE_TailParticle::resizeTail(n);

  Link l;
  l.m_distance = 0.0;

  m_links.resize(n);
  m_links.setAll(l);

  m_links.setSampleCount(0);
}

template class MJ_RingBufferDelay<VEE_TailParticle2::Link>;

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VEE_TailParticle3::VEE_TailParticle3()
  : m_velocityScale(0.5)
{}

VEE_TailParticle3::~VEE_TailParticle3()
{}

bool VEE_TailParticle3::setValue(const char *id, VEE_Value *v)
{
  bool ok = VEE_TailParticle::setValue(id, v);

  if(ok) return true;

  if(strcmp(id, "tailvel") == 0) {
    float tmp;

    ok = v->getFloat(&tmp);

    if(ok)
      m_velocityScale = tmp;
  }
  
  return ok;
}

VEE_Variable VEE_TailParticle3::getVariable(const char *id) const
{
  VEE_Variable v = VEE_TailParticle::getVariable(id);

  if(v.type() != VEE_VAR_INVALID)
    return v;
  
  if(strcmp(id, "tailvel") == 0)
    v = m_velocityScale;

  return v;
}

void VEE_TailParticle3::update(float dt)
{
  // ulong count = m_points.getSampleCount();

  VEE_TailParticle::update(dt);

  /* if(count != m_points.getSampleCount()) {
    Point * p = & m_points.getNewest();
    p->m_velocity *= m_velocityScale;
    }*/
  
  uint n = pointsToRender();

  // VEE_trace("VEE_TailParticle3::update # %p %u", this, (uint) n);

  for(uint i=0; i < n; i++) {
    Point * p = & getDelayedPoint(i);
    p->location() += (p->velocity() * (dt * m_velocityScale));
  }
}
  
VEE_Object *VEE_TailParticle3::clone() const
{
  return new VEE_TailParticle3(*this);
}

VEE_Object * VEE_TailParticle3::factory()
{
  return new VEE_TailParticle3();
}

