/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_value_curve.h>

#include <vee_error.h>
#include <vee_object.h>
#include <vee_ptr.h>
#include <vee_strs.h>
#include <vee_value.h>

#include <mj_random.h>

VEE_ValueCurve::VEE_ValueCurve()
  : m_interval(1.0),
    m_untilNext(10.5)
{}

VEE_ValueCurve::~VEE_ValueCurve()
{}

/** Receive and process a value.

    Parameters:

    <UL>

    <LI> "rate", float, range: 1.0e-10 to inf

    <LI> "interval", float, range; 1.0e-3 to inf

    </UL>

*/

bool VEE_ValueCurve::setValue(const char *id, VEE_Value *v)
{
  bool ok = false;
  
  if(VEE_Strs::equals(id, VEE_Strs::rate)) {

    float tmp;

    ok = v->getFloat(&tmp);
    
    if(ok) {
      if(tmp < 1e-10) tmp = 1e-10;
      m_interval = 1.0f / tmp;
      if(m_untilNext > m_interval)
	m_untilNext = m_interval;
    }
  }
  else if(VEE_Strs::equals(id, "interval")) {

    float tmp;

    ok = v->getFloat(&tmp);
    
    if(ok) {
      if(tmp < 1e-3) tmp = 1e-3;
      m_interval = tmp;
    }
  }
  else if(VEE_Strs::equals(id, "trigger")) {

    m_curve.reset();
    m_untilNext = -1.0;

    ok = true;
  }

  return ok;
}

void VEE_ValueCurve::operate(VEE_Object *o, float dt)
{
  if(!m_curve.size()) {
    VEE_error("VEE_ValueCurve::operate # Empty curve");
    return;
  }

  if(m_untilNext < 0.0) {
    m_curve.move(dt);
    float v = m_curve.getValue2();
    o->setValueFloat(m_param.c_str(), v);

    if(m_curve.atEnd())
      m_untilNext = m_interval * 0.5 + 
	m_interval * 0.5 + MJ_Random::instance.getFloat0();
  }
  else if(m_untilNext < dt) {
    // VEE_trace("VEE_ValueCurve::operate # Triggering a curve");

    m_curve.reset(dt - m_untilNext);
    m_untilNext = -1.0;
  }
  else
    m_untilNext -= dt;
}

VEE_Operator * VEE_ValueCurve::clone() const
{
  return new VEE_ValueCurve(*this);
}



/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


VEE_ValueCurveMulti::VEE_ValueCurveMulti()
  : m_interval(1.0),
    m_untilNext(0.5)
{}

VEE_ValueCurveMulti::~VEE_ValueCurveMulti()
{}

/** Receive and process a value.

    Parameters:

    <UL>

    <LI> "rate", float, range: 1.0e-10 to inf

    <LI> "interval", float, range; 1.0e-3 to inf

    </UL>

*/

bool VEE_ValueCurveMulti::setValue(const char *id, VEE_Value *v)
{
  bool ok = false;
  
  if(VEE_Strs::equals(id, VEE_Strs::rate)) {

    float tmp;

    ok = v->getFloat(&tmp);
    
    if(ok) {
      if(tmp < 1e-10) tmp = 1e-10;
      m_interval = 1.0f / tmp;
      if(m_untilNext > m_interval)
	m_untilNext = m_interval;
    }
  }
  else if(VEE_Strs::equals(id, "interval")) {

    float tmp;

    ok = v->getFloat(&tmp);
    
    if(ok) {
      if(tmp < 1e-3) tmp = 1e-3;
      m_interval = tmp;
    }
  }

  return ok;
}

void VEE_ValueCurveMulti::operate(VEE_Object *o, float dt)
{
  if(!m_curves.size()) {
    VEE_error("VEE_ValueCurveMulti::operate # Empty curves");
    return;
  }
  
  if(!m_curves[0].m_curve.size()) {
    VEE_error("VEE_ValueCurveMulti::operate # Empty first curve");
    return;
  }

  if(m_untilNext < 0.0) {
    
    for(uint i=0; i < m_curves.size(); i++) {
      Data * data = & m_curves[i];
      data->m_curve.move(dt);
      float v = data->m_curve.getValue2();
      o->setValueFloat(data->m_param.c_str(), v);
      /* VEE_trace("VEE_ValueCurveMulti::operate # %s %f", 
	 data->m_param.c_str(), v); */
    }
    
    if(m_curves[0].m_curve.atEnd())
      m_untilNext = m_interval * 0.5 + 
	m_interval * 0.5 + MJ_Random::instance.getFloat0();
  }
  else if(m_untilNext < dt) {
    // VEE_trace("VEE_ValueCurveMulti::operate # Triggering a curve");

    for(uint i=0; i < m_curves.size(); i++)
      m_curves[i].m_curve.reset(dt - m_untilNext);
    m_untilNext = -1.0;
  }
  else
    m_untilNext -= dt;
}

VEE_Operator * VEE_ValueCurveMulti::clone() const
{
  return new VEE_ValueCurveMulti(*this);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


VEE_ValueCurve2::VEE_ValueCurve2()
  : m_scaling(VEE_TS_AGE_MATCH),
    m_offset(-1),
    m_time(0.0)
{}

VEE_ValueCurve2::~VEE_ValueCurve2()
{}

/** Receive and process a value.

    Parameters:


*/

bool VEE_ValueCurve2::setValue(const char *, VEE_Value *)
{
  bool ok = false;
  
  return ok;
}

void VEE_ValueCurve2::operate(VEE_Object *o, float dt)
{
  if(!(*m_curve).size()) {
    VEE_error("VEE_ValueCurve::operate # Empty curve");
    return;
  }

  m_time += dt;
  
  if(m_scaling == VEE_TS_PERIODIC) {
    // Check if we need to rewind
    if(m_time > (*m_curve).endTime())
      m_time -= (*m_curve).endTime();
  }

  float t = m_time;
  
  if(m_scaling == VEE_TS_AGE_MATCH)
    t *= (*m_curve).endTime() / o->ageLimit();

  float v = (*m_curve).getValueBisect(t);

  if(m_offset == -1) {
    const float * ptr = o->getVariable(m_param.c_str()).getFloatPtr();
    if(!ptr)
      m_offset = -2;
    else
      m_offset = VEE_Ptr::distance(ptr, o);
  }

  if(m_offset == -2)
    o->setValueFloat(m_param.c_str(), v);
  else
    * (float *) VEE_Ptr::offset(o, m_offset) = v;
}

VEE_Operator * VEE_ValueCurve2::clone() const
{
  return new VEE_ValueCurve2(*this);
}
