/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_wind.h>

#include <vee_bounding_box.h>
#include <vee_error.h>
#include <vee_math.h>
#include <vee_object.h>
// #include <vee_scene.h>
#include <vee_strs.h>
#include <vee_value.h>
#include <vee_values.h>
#include <vee_volume.h>

#include <mj_random.h>

static const float VEE_WIND_CALCULUS_FREQ = 100.0f;
static const float VEE_ONE_PER_WIND_CALCULUS_FREQ = 0.01f;

VEE_Wind::VEE_Wind(const VEE_Vector3 &dir,
		   float strength,
		   float gustiness,
		   float gustfreq)
  : m_direction(dir),
    m_strength(strength),
    m_gustStrength(gustiness),
    m_gustFreq(gustfreq),
    m_sinceLast(0.0)
{
  redesignFilter();
  m_gustFilter.resetMemory();
  
  m_forceType = VEE_FORCE_DRAG;
  recalculateWind();
}

VEE_Wind::~VEE_Wind()
{}

/** Receive and process a value. 

    Parameters:

    <UL>
    
    <LI> "direction", Vector3

    <LI> "strength", float

    <LI> "gustiness", float

    <LI> "gustfreq", float

    <LI> "volume/+", VEE_Value *, message to the volume object

    <LI> "volume", VEE_ValueVolume *

    </UL>

*/
bool VEE_Wind::setValue(const char *id, VEE_Value *msg)
{
  if(VEE_Strs::equals(id, VEE_Strs::direction)) {
    VEE_Vector3 v;

    bool ok = msg->getVector3(&v);

    if(ok) {
      m_direction = v;
      recalculateWind();
    }

    return ok;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::strength)) {
    float v;

    bool ok = msg->getFloat(&v);

    if(ok) {
      m_strength = v;
      recalculateWind();
    }

    return ok;
  }
  else if(VEE_Strs::equals(id, "gustiness")) {
    float v;

    bool ok = msg->getFloat(&v);

    if(ok) {
      m_gustStrength = VEE_clamp(v, 0.0f, 20.0f);
      recalculateWind();
    }

    return ok;
  }
  else if(VEE_Strs::equals(id, "gustfreq")) {
    float v;

    bool ok = msg->getFloat(&v);

    if(ok) {
      m_gustFreq = VEE_clamp(v, 0.1f, 200.0f);
      redesignFilter();
    }
    
    return ok;
  }
  else if(strcmp(VEE_Strs::volume, id) == 0) {

    VEE_ValueVolume *vvol = dynamic_cast<VEE_ValueVolume *> (msg);

    if(!vvol) delete vvol;
    m_volume = vvol->fabricate();

    return false;
  }
  else if(strncmp("volume/", id, 7) == 0) {

    if(m_volume.ptr())
      return m_volume.ptr()->setValue(id + 7, msg);

    return false;
  }
  else
    return VEE_ForceObject::setValue(id, msg);
}

VEE_Variable VEE_Wind::getVariable(const char *id) const
{
  VEE_Variable v;

  if(VEE_Strs::equals(id, VEE_Strs::direction)) {
    v = m_direction;
  }
  else if(VEE_Strs::equals(id, VEE_Strs::strength)) {
    v = m_strength;
  }
  else if(VEE_Strs::equals(id, "gustiness")) {
    v = m_gustStrength;
  }
  else if(VEE_Strs::equals(id, "gustfreq")) {
    v = m_gustFreq;
  }
  /* else if(strcmp(VEE_Strs::volume, id) == 0) {
     }*/
  /* else if(strncmp("volume/", id, 7) == 0) {

    if(m_volume.ptr())
      return m_volume.ptr()->setValue(id + 7, msg);

    return false;
    }*/
  else
    v = VEE_ForceObject::getVariable(id);

  return v;
}

VEE_Object * VEE_Wind::clone() const
{
  // VEE_trace("VEE_Wind::clone");
  return new VEE_Wind(*this);
}

void VEE_Wind::update(float dt)
{
  VEE_ForceObject::update(dt);

  for(m_sinceLast += dt; m_sinceLast > 0.0; 
      m_sinceLast -= VEE_ONE_PER_WIND_CALCULUS_FREQ) {

    m_gustFilter.put(MJ_Random::instance.getFloat1(2.0f));
  }
  recalculateWind();
}

void VEE_Wind::getForce(VEE_Vector3 *force,
			const VEE_Vector3 &vel,
			const VEE_Vector3 &loc)
{
  float strength;

  if(m_volume.ptr())
    strength = m_volume.ptr()->inside(loc - m_location);
  else strength = 1.0;
    
  *force = strength * (m_wind - vel);
}

void VEE_Wind::boundingBoxOfForce(VEE_BoundingBox *box)
{
  
  if(m_volume.ptr()) {
    m_volume.ptr()->getBoundingBox(box);
    box->move(m_location);
  }
  else {
    float v = -1e+12;
    box->low().make(v, v, v);
    box->high() = box->low() * -1.0;
  }
}

VEE_Object * VEE_Wind::factory()
{
  return new VEE_Wind;
}

void VEE_Wind::redesignFilter()
{ 
  m_gustFilter.designLowPass2(m_gustFreq, VEE_WIND_CALCULUS_FREQ);
}

void VEE_Wind::recalculateWind()
{
  m_direction.normalize();

  m_wind = m_direction * m_strength * 
    (1.0f + m_gustFilter.get() * m_gustStrength );
}
