
/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_magic_rain.h>

#include <vee_error.h>
#include <vee_looks_color_table.h>
#include <vee_math.h>
#include <vee_scene.h>
#include <vee_volume_disk.h>
#include <vee_strs.h>
#include <vee_value.h>


VEE_MagicRainDrop::VEE_MagicRainDrop(VEE_Vector3 location, 
				     VEE_Vector3 velocity,
				     float ageLimit, 
				     VEE_Looks *l,
				     float childSpeed,
				     float childAgeLimit,
				     float creationRate)
  : VEE_Particle(location, velocity, ageLimit, l),
    m_childSpeedMin(childSpeed),
    m_childSpeedMax(childSpeed),
    m_childAgeLimit(childAgeLimit),
    m_creationRate(creationRate),
    m_sinceLastCreated(0.0),
    m_childCount(0),
    m_rotation(VEE_Matrix3IDENTITY)
{}

bool VEE_MagicRainDrop::setValue(const char *id, VEE_Value *msg)
{
  bool ok = false;

  if(VEE_Strs::equals(id, VEE_Strs::rate)) {
    float v;
    ok = msg->getFloat(&v);
    if(ok) 
      m_creationRate = v;
  }
  else if(VEE_Strs::equals(id, "childspeedmin")) {
    float v;
    ok = msg->getFloat(&v);
    if(ok) 
      m_childSpeedMin = v;
  }
  else if(VEE_Strs::equals(id, "childspeedmax")) {
    float v;
    ok = msg->getFloat(&v);
    if(ok) 
      m_childSpeedMax = v;
  }
  else 
    return VEE_Particle::setValue(id, msg);

  return ok;
}

VEE_Object * VEE_MagicRainDrop::clone() const
{
  return new VEE_MagicRainDrop(*this);
}

void VEE_MagicRainDrop::update(float dt)
{
  VEE_Particle::update(dt);

  m_sinceLastCreated += dt;

  float creationInterval = 1.0 / m_creationRate;

  for(; m_sinceLastCreated > creationInterval; 
      m_sinceLastCreated -= creationInterval) {
    createNewParticle();
  }

  float test = m_rand.rand01() * m_sinceLastCreated;

  if(test > (m_creationRate * 0.5)) {
    createNewParticle();
    m_sinceLastCreated = 0.0;  
  }
}

void VEE_MagicRainDrop::createNewParticle()
{
  float speed;

  float speedDif = m_childSpeedMax - m_childSpeedMin;

  if(speedDif > 1e-3) {
    speed = m_childSpeedMin + m_rand.rand01() *  speedDif;
  }
  else 
    speed = m_childSpeedMax;

  float angle = m_rand.rand01() * VEE_TWO_PI;
    
  VEE_Vector3 vel(sinf(angle) * speed,
		   // m_velocity[1] + m_rand.rand() * m_childSpeed,
		   0,
		   cosf(angle) * speed);
  
  VEE_Particle *p = new VEE_Particle(m_location, 
				     vel, 
				     m_childAgeLimit, 
				     m_looks->clone());

  float initAge = m_childAgeLimit * 0.2 * m_rand.rand01();
  p->setAge(initAge);
  
  /* VEE_trace("VEE_MagicRainDrop::createNewParticle %f -> %f %f %f",
     angle, vel.x, vel.y, vel.z); */

  m_sceneBox->addObject(p);
  m_childCount++;
}  

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VEE_MagicRain::VEE_MagicRain(VEE_Vector3 location, 
			     VEE_Vector3 velocity,
			     float ageLimit, 
			     float radius,
			     float childSpeed,
			     float childRate)
  : VEE_Particle(location, velocity, ageLimit, 0),
    m_childSpeed(childSpeed),
    m_childRate(childRate),
    m_sinceLastCreated(0.0)
{
  m_volume = new VEE_VolumeDisk;
  m_volume->setRadius(radius);
  
  for(int i=0 ; i < 6 ; i++) {
    // m_colorTable.appendColor(VEE_Rgba::White);
    // m_colorTable.appendColor(VEE_Rgba::Black);
    m_colorTable.appendColor(VEE_Rgba(1.0, 1.0, 1.0, 0.4));
    m_colorTable.appendColor(VEE_Rgba(0.5, 0.5, 0.0, 0.4));
  }
  m_colorTable.appendColor(VEE_Rgba(1.0, 1.0, 0.2, 1.0));
  m_colorTable.appendColor(VEE_Rgba::BlackTransparent);

  m_updateVelocity = false;
}

VEE_MagicRain::~VEE_MagicRain()
{
  delete m_volume;
}

bool VEE_MagicRain::setValue(const char *id, VEE_Value *msg)
{
  bool ok = false;

  if(VEE_Strs::equals(id, "childspeed")) {
    float v;
    ok = msg->getFloat(&v);
    if(ok)
      m_childSpeed = v;
  }
  else if(VEE_Strs::equals(id, "childrate")) {
    float v;
    ok = msg->getFloat(&v);
    if(ok)
      m_childRate = v;
  }
  else return VEE_Particle::setValue(id, msg);
  
  return ok;
}

VEE_Object * VEE_MagicRain::clone() const
{
  VEE_trace("VEE_MagicRain::clone # %p", this);
  return new VEE_MagicRain(*this);
}

void VEE_MagicRain::render(VEE_Renderer *)
{}

void VEE_MagicRain::update(float dt)
{
  // VEE_trace("VEE_MagicRain::update %f %f", m_sinceLastCreated, m_childRate);

  VEE_Particle::update(dt);

  m_sinceLastCreated += dt;

  if(m_sinceLastCreated > (1.0 / m_childRate)) {

    //VEE_trace("VEE_MagicRain::update %f %f",m_sinceLastCreated,m_childRate);
    
    VEE_Vector3 vel;
    vel[0] = (m_rand.rand01() - 0.5f);
    vel[1] = (m_rand.rand01() - 0.5f);
    vel[2] = (m_rand.rand01() - 0.5f);

    vel.normalize();

    vel *= m_childSpeed;
  
    // VEE_trace("VEE_MagicRain::update # created new");
    
    VEE_MagicRainDrop *rainDrop = 
      new VEE_MagicRainDrop(m_volume->newRandomPoint() + m_location, 
			    vel, 
			    15.0, 
			    new VEE_LooksColorTable(&m_colorTable), 
			    1.0, 2.0, 50);

    m_sceneBox->addObject(rainDrop);
    
    m_sinceLastCreated = 0.0;
  }
}
