/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */
#include <Opcode.h>

#include <vee_opcode_collision.h>

#include <vee_bounding_box.h>
#include <vee_error.h>
#include <vee_scene.h>
#include <vee_value.h>

#include <vee_gl_renderer.h>
#include <vee_polygon.h>

#include <gee_group.h>
#include <gee_triangle_mesh.h>
#include <gee_world.h>

#include <OPC_MeshInterface.h>
#include <IceRay.h>
using namespace ICE;

VEE_OpcodeCollision::VEE_OpcodeCollision()
  : m_scale(1.0),
    m_bounce(1.0),
    m_drag(0.0)
{
  mMeshCommitted=false;
  m_objectType = VEE_OBJECT_COLLIDER;
  m_rotation.identity();
  setVelocityUpdate(false);
  m_active=true;
}

VEE_OpcodeCollision::~VEE_OpcodeCollision()
{}

bool VEE_OpcodeCollision::setValue(const char *id, VEE_Value *v)
{
  bool ok = VEE_Object::setValue(id, v);
  
  if(ok) return true;
  
  if(strcmp(id, "bounce") == 0) {
    float tmp;
    
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_bounce = tmp;
  }
  else if(strcmp(id, "drag") == 0) {
    float tmp;
    
    ok = v->getFloat(&tmp);
    
    if(ok)
      m_drag = tmp;
  }

  return ok;
}

VEE_Variable VEE_OpcodeCollision::getVariable(const char *id) const
{
  VEE_Variable v = VEE_Object::getVariable(id);

  if(v.type() != VEE_VAR_INVALID)
    return v;

  if(strcmp(id, "bounce") == 0)
    v = m_bounce;
  else if(strcmp(id, "drag") == 0)
    v = m_drag;
  
  return v;
}

void VEE_OpcodeCollision::update(float dt)
{
	if(mMeshCommitted)
	{
		VEE_Object::update(dt);

		m_dt = dt * 2.0;

		VEE_BoundingBox bb;
		VEE_Vector3 v;
		v.make(1.0e+10);
		bb.set(-v, v);

		m_opcMatrix.identity();
		m_opcMatrix.setRotation(m_rotation);

		VEE_SceneBox *scene = m_sceneBox;

		while(!scene->boundingBox().contains(bb) &&	scene->parent())
			scene = scene->parent();

		recursiveImpact(scene, &bb);
	}
}

void VEE_OpcodeCollision::render(VEE_Renderer *r)
{
	if(mMeshCommitted && m_visible)
	{
		VEE_GlRenderer *gr=(VEE_GlRenderer *)r;
		gr->renderOpcodeMesh(this);
	}
}

VEE_Object * VEE_OpcodeCollision::clone() const
{
  return new VEE_OpcodeCollision(*this);
}

/// Loads a model from a file.
bool VEE_OpcodeCollision::loadModel(const char *file)
{
  const char * fname = "VEE_OpcodeCollision::loadModel";

  GEE_World world;

  GEE_Group * model = new GEE_Group();

  world.addGeometry(model);

  model->loadModel(file);

  model->convertPolygonsToTriangles();

  for(GEE_Group::iterator it = model->begin(); it != model->end(); it++) {

    VEE_RefPtr<GEE_Geometry> * data = model->getGeometry(it);

    if(!data)
      continue;

    GEE_TriangleMesh * tm = dynamic_cast<GEE_TriangleMesh *> (data->ptr());

    if(!tm)
      continue;
    
    // Copy vertex locations and form remapping data
    std::map<int, int> vertexremap;

    for(GEE_TriangleMesh::vertexIterator vit = tm->beginVertex();
	vit != tm->endVertex(); vit++) {
      VEE_Vector3 loc = (*vit).second.m_location;
      int id = (*vit).first;

      int newId = m_vertices.size();
      vertexremap[id] = newId;

      m_vertices.push_back(loc);
    }

    // Construct triangles and remap the vertex ids
    
    for(GEE_TriangleMesh::triangleIterator tit = tm->beginTriangle();
	tit != tm->endTriangle(); tit++) {

      const GEE_TriangleIndices & trind = GEE_TriangleMesh::getTriangle(tit);
      Triangle tri;

      for(uint j=0; j < 3; j++)
	tri.m_indices[j] = vertexremap[trind.m_indices[j]];
      
      m_triangles.push_back(tri);

      // Calculate the normal for each triangle
      VEE_Vector3 v0 = m_vertices[tri.m_indices[0]];
      VEE_Vector3 v1 = m_vertices[tri.m_indices[1]];
      VEE_Vector3 v2 = m_vertices[tri.m_indices[2]];

      VEE_Vector3 n = cross(v1 - v0, v2 - v0);
      n.normalize();

      m_normals.push_back(n);
    }
  }

  /* The model has been loaded, now push it to the OPCODE.  */


  m_meshinterface.SetPointers
    ((ICEMATHS_API IndexedTriangle *) & m_triangles[0],
     (ICEMATHS_API ICE::Point *) & m_vertices[0]);

  m_meshinterface.SetNbTriangles(m_triangles.size());
  m_meshinterface.SetNbVertices(m_vertices.size());

  m_opcCreate.mIMesh = & m_meshinterface;

  m_opcModel.Build(m_opcCreate);

  m_rayCollider.SetCulling(true);
  m_rayCollider.SetClosestHit(true);
  m_rayCollider.SetDestination(& m_collisionFaces);

  const char * error = m_rayCollider.ValidateSettings();

  if(error) {
    VEE_error("%s # \"%s\"", fname, error);
    return false;
  }

  return true;
}

void VEE_OpcodeCollision::meshAddTri(VEE_Vector3 a,VEE_Vector3 b,VEE_Vector3 c)
{
	m_vertices.push_back(a);
	m_vertices.push_back(b);
	m_vertices.push_back(c);

	{
		Triangle tri=	{
							m_vertices.size()-3,	// a
							m_vertices.size()-2,	// b
							m_vertices.size()-1		// c
						};

		m_triangles.push_back(tri);

		{
			// Calculate the normal
			VEE_Vector3 v0 = m_vertices[tri.m_indices[0]];
			VEE_Vector3 v1 = m_vertices[tri.m_indices[1]];
			VEE_Vector3 v2 = m_vertices[tri.m_indices[2]];

			VEE_Vector3 n = cross(v1 - v0, v2 - v0);
			n.normalize();

			m_normals.push_back(n);
		}
	}
}

void VEE_OpcodeCollision::meshAddQuad(VEE_Vector3 a,VEE_Vector3 b,VEE_Vector3 c,VEE_Vector3 d)
{
	m_vertices.push_back(a);
	m_vertices.push_back(b);
	m_vertices.push_back(c);
	m_vertices.push_back(d);

	{
		Triangle tri=	{
							m_vertices.size()-4,	// a
							m_vertices.size()-3,	// b
							m_vertices.size()-2		// c
						};

		m_triangles.push_back(tri);

		{
			// Calculate the normal
			VEE_Vector3 v0 = m_vertices[tri.m_indices[0]];
			VEE_Vector3 v1 = m_vertices[tri.m_indices[1]];
			VEE_Vector3 v2 = m_vertices[tri.m_indices[2]];

			VEE_Vector3 n = cross(v1 - v0, v2 - v0);
			n.normalize();

			m_normals.push_back(n);
		}
	}

	{
		Triangle tri=	{
							m_vertices.size()-2,	// c
							m_vertices.size()-1,	// d
							m_vertices.size()-4		// a
						};

		m_triangles.push_back(tri);

		{
			// Calculate the normal
			VEE_Vector3 v0 = m_vertices[tri.m_indices[0]];
			VEE_Vector3 v1 = m_vertices[tri.m_indices[1]];
			VEE_Vector3 v2 = m_vertices[tri.m_indices[2]];

			VEE_Vector3 n = cross(v1 - v0, v2 - v0);
			n.normalize();

			m_normals.push_back(n);
		}
	}
}

void VEE_OpcodeCollision::meshClear(void)
{
	mMeshCommitted=false;

	m_vertices.clear();
	m_normals.clear();
	m_triangles.clear();
}

bool VEE_OpcodeCollision::meshCommit(void)
{
  const char * fname = "VEE_OpcodeCollision::meshCommit";

	if(!mMeshCommitted)
	{
		/* The model has been loaded, now push it to the OPCODE.  */


		m_meshinterface.SetPointers
		((ICEMATHS_API IndexedTriangle *) & m_triangles[0],
		 (ICEMATHS_API ICE::Point *) & m_vertices[0]);

		m_meshinterface.SetNbTriangles(m_triangles.size());
		m_meshinterface.SetNbVertices(m_vertices.size());

		m_opcCreate.mIMesh = & m_meshinterface;

		m_opcModel.Build(m_opcCreate);

// @@@ additional options?  see OPC_RayCollider.cpp for info
//		m_rayCollider.SetFirstContact(true);
//		m_rayCollider.SetTemporalCoherence(false);
		m_rayCollider.SetCulling(true);
		m_rayCollider.SetClosestHit(true);
		m_rayCollider.SetDestination(& m_collisionFaces);

		const char * error = m_rayCollider.ValidateSettings();

		if(error)
		{
			VEE_error("%s # \"%s\"", fname, error);
			return false;
		}
		
		mMeshCommitted=true;
	}

	return true;
}

bool VEE_OpcodeCollision::onCollide(VEE_Object *o)
{
	// default: affect object velocity
	return true;
}

void VEE_OpcodeCollision::recursiveImpact(VEE_SceneBox *scene, 
					  VEE_BoundingBox *bb)
{
	if(mMeshCommitted && m_active)
	{
		ICEMATHS_API Ray ray;

		for(VEE_SceneBox::iterator it = scene->begin(); it != scene->end(); it++)
		{

			VEE_Object * o = VEE_SceneBox::getObject(it);

			if(o == this)
				continue;    

			VEE_Vector3 l= o->location() - location();
			ICEMATHS_API ICE::Point loc = * ( ICEMATHS_API ICE::Point *) & l;

			VEE_Vector3 vel = o->velocity(); // Cache velocity to stack
			VEE_Vector3 dir = vel;
			float speed = dir.length();
			if(speed > 1.0e-8)
				dir *= (1.0f / speed);

			m_rayCollider.SetMaxDist(speed * m_dt * 2.0f + 0.1);

			ray.mOrig = loc;
			dir.normalize();
			ray.mDir = * ( ICEMATHS_API ICE::Point *) & dir;

			m_rayCollider.Collide(ray, m_opcModel, (const Matrix4x4 *) & m_opcMatrix);

			bool contact = m_rayCollider.GetContactStatus();

			if(!contact)
				continue;

//			if(!onCollide(o))
//				continue;

			const Opcode::CollisionFace * face = m_collisionFaces.GetFaces();

			// float distance = face->mDistance;

			const VEE_Vector3 normal = m_normals[face->mFaceID];

			float gamma = dot(normal, vel);

			VEE_Vector3 tangential = vel - gamma * normal;
			VEE_Vector3 normalBounce = - gamma * normal;

			VEE_Vector3 newVel = 
				tangential * (1.0f - m_drag) + normalBounce * m_bounce;

			o->collision(normal);
			o->setVelocity(newVel);
		}

		for(uint i=0; i < VEE_SB_CHILDREN && scene->children(); i++)
		{
			VEE_SceneBox *tmp = scene->child(i);
			if(bb->intersects(tmp->boundingBox()))
				recursiveImpact(tmp, bb);
		}
	}
}
