/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_collector_renderer.h>

#include <gee_group.h>
#include <gee_lines.h>
#include <gee_path.h>
#include <gee_particle_object.h>
#include <gee_polygon_mesh.h>
#include <gee_profile.h>
#include <gee_sphere.h>
#include <gee_triangle_mesh.h>

#include <vee_error.h>

#include <strings.h>

GEE_CollectorRenderer::GEE_CollectorRenderer()
  : m_stackIndex(0)
{
  bzero(m_stack, sizeof(m_stack));
  m_stack[0].identity();
  m_current.identity();
}

GEE_CollectorRenderer::~GEE_CollectorRenderer()
{}

void GEE_CollectorRenderer::renderGroupStructure(GEE_Group *d)
{
  m_solids.addGee(d, m_current);
}

void GEE_CollectorRenderer::renderLines3D(GEE_Lines3D *d)
{
  if(d->material().blend())
    m_transparencies.addGee(d, m_current);
  else
    m_solids.addGee(d, m_current);
}

void GEE_CollectorRenderer::renderPath3D(GEE_Path3D *d)
{
  if(d->material().blend())
    m_transparencies.addGee(d, m_current);
  else
    m_solids.addGee(d, m_current);
}

void GEE_CollectorRenderer::renderParticleObject(GEE_ParticleObject *po)
{
  /* VEE_trace("GEE_CollectorRenderer::renderParticleObject # %u",
     po->particleCount()); */

  /* First we calculate how much the particles should be enlarged or
     shrank. The calculation below calculates how much axis vectors
     are stretched and uses that value as the scale factor for this
     particle. */
  
  VEE_Vector3 origin(0, 0, 0);
  VEE_Vector3 axis[3] = { 
    VEE_Vector3_UNIT_X, 
    VEE_Vector3_UNIT_Y,
    VEE_Vector3_UNIT_Z
  };

  VEE_Matrix4 m = m_current;
  
  origin = m * origin;
  
  int i;

  for(i = 0; i < 3; i++)
    axis[i] = m * axis[i];
  
  float scale = 0.0f;

  for(i = 0; i < 3; i++)
    scale += (axis[i] - origin).length();
  
  scale *= 0.3333333f;

  // Now render the object:

  for(GEE_ParticleObject::iterator it = po->begin();
      it != po->end(); it++) {

    VEE_Object * obj = 
      GEE_ParticleObject::getParticleData(it).m_particle.ptr();

    if(obj)
      renderVeeObject(obj, 0, scale);
  }

  m_solids.addGee(po, m_current);

  /* VEE_Matrix4 m = m_current;
  VEE_trace("GEE_CollectorRenderer::renderParticleObject\n[ %f %f %f %f\n"
  "  %f %f %f %f\n  %f %f %f %f\n  %f %f %f %f]",
	    m[0][0], m[0][1], m[0][2], m[0][3], 
	    m[1][0], m[1][1], m[1][2], m[1][3],
	    m[2][0], m[2][1], m[2][2], m[2][3],
	    m[3][0], m[3][1], m[3][2], m[3][3]);
  */
  /* VEE_Vector3 l = m_current.column(3).vector3();
     VEE_trace("GEE_CollectorRenderer::renderParticleObject # %f %f %f",
     l.x, l.y, l.z); */
}

void GEE_CollectorRenderer::renderPolygonMesh(GEE_PolygonMesh *d)
{
  m_solids.addGee(d, m_current);
}

void GEE_CollectorRenderer::renderProfile(GEE_Profile *d)
{
  if(d->material().blend())
    m_transparencies.addGee(d, m_current);
  else
    m_solids.addGee(d, m_current);
}

void GEE_CollectorRenderer::renderSphere(GEE_Sphere *d)
{
  if(d->material().blend())
    m_transparencies.addGee(d, m_current);
  else
    m_solids.addGee(d, m_current);
}

void GEE_CollectorRenderer::renderTriangleMesh(GEE_TriangleMesh *d)
{
  if(d->material().blend())
    m_transparencies.addGee(d, m_current);
  else
    m_solids.addGee(d, m_current);
}

void GEE_CollectorRenderer::renderVeeObject(VEE_Object *d, 
					    const VEE_Matrix4 *,
					    float scale)
{
  // VEE_trace("GEE_CollectorRenderer::renderVeeObject # %p", d);
  m_transparencies.addVee(d, m_current, scale);
}

void GEE_CollectorRenderer::pushTransform(const VEE_Matrix4d &m)
{
  const VEE_Matrix4d & prev = m_stack[m_stackIndex];

  if(++m_stackIndex >= STACK_MAXIMUM) {
    VEE_error("GEE_CollectorRenderer::pushTransform # Stack overflow");
    m_stackIndex = STACK_MAXIMUM - 1;
    return;
  }

  m_stack[m_stackIndex] = prev * m;
  m_current.copy(m_stack[m_stackIndex].data());
}

void GEE_CollectorRenderer::pushTransform(const VEE_Matrix4 &m)
{
  const VEE_Matrix4d & prev = m_stack[m_stackIndex];

  if(++m_stackIndex >= STACK_MAXIMUM) {
    VEE_error("GEE_CollectorRenderer::pushTransform # Stack overflow");
    m_stackIndex = STACK_MAXIMUM - 1;
    return;
  }

  VEE_Matrix4d tmp(m.data());

  m_stack[m_stackIndex] = prev * tmp;
  m_current.copy(m_stack[m_stackIndex].data());
}

void GEE_CollectorRenderer::popTransform()
{
  if(!m_stackIndex) {
    VEE_error("GEE_CollectorRenderer::popTransform # Stack underflow");
    // m_stackIndex = 0;
    return;
  }

  m_stackIndex--;
}

void GEE_CollectorRenderer::enableTransforms(bool)
{}
  
bool GEE_CollectorRenderer::enableTransforms()
{
  return true;
}

void GEE_CollectorRenderer::clear()
{
  m_transparencies.clear();
  m_solids.clear();
}
