/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_erase_around_path.h>

#include <gee_io.h>
#include <gee_macros.h>
#include <gee_world.h>

#include <vee_error.h>

#include <typeinfo>

GEE_EraseAroundPath::GEE_EraseAroundPath(GEE_id_t pathId, 
					 bool deleteParts,
					 float radiusNorm,
					 float exaggerateRadius,
					 bool useNapra)
  : GEE_PathToX(pathId),
    m_deleteParts(deleteParts),
    m_exaggerateRadius(exaggerateRadius),
    m_radiusNorm(radiusNorm),
    m_useNapra(useNapra)
{}

GEE_EraseAroundPath::~GEE_EraseAroundPath()
{}

void GEE_EraseAroundPath::contained(GEE_Geometry *g)
{
  // VEE_trace("GEE_EraseAroundPath::contained # %s", typeid(*g).name());
  addEditedGeometry(g);
}

bool GEE_EraseAroundPath::act(GEE_World *w, const GEE_TimeStamp &time)
{
  clear();

  if(!GEE_PathToX::act(w, time))
    return false;

  if(locationCount()) {
    doWork(w);
  }

  return true;
}

static const ushort GEE_RadiusId   = GEE_CONST2('R', 'A');
static const ushort GEE_DelPartsId = GEE_CONST2('D', 'P');
static const ushort GEE_UseNapraId = GEE_CONST2('U', 'N');
static const ushort GEE_ExRadiusId = GEE_CONST2('E', 'R');
static const ushort GEE_RadiusNormId = GEE_CONST2('R', 'N');
static const ushort GEE_EndId      = GEE_CONST2('E', 'N');

bool GEE_EraseAroundPath::read(GEE_Io * io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != type())
    return false;
  
  readBasics(io);
  readPathId(io);

  int v = io->readInt32();

  if(v != 1)
    return false;

  ushort id = io->readInt16();

  while(id != GEE_EndId) {
    if(id == GEE_RadiusId)
      m_radius = io->readFloat32();
    else if(id == GEE_ExRadiusId)
      m_exaggerateRadius = io->readFloat32();
    else if(id == GEE_RadiusNormId)
      m_radiusNorm = io->readFloat32();
    else if(id == GEE_DelPartsId)
      m_deleteParts = io->readInt32();
    else if(id == GEE_UseNapraId)
      m_useNapra = io->readByte();
    else {
      VEE_error("GEE_EraseAroundPath::read # Bad id at %d (%c%c)", 
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }
    id = io->readInt16();
  }

  return true;
}

bool GEE_EraseAroundPath::write(GEE_Io * io)
{
  io->writeNullTerminatedString(type());

  writeBasics(io);
  writePathId(io);
  
  io->writeInt32(1); // version

  io->writeInt16(GEE_RadiusId);
  io->writeFloat32(m_radius);

  io->writeInt16(GEE_ExRadiusId);
  io->writeFloat32(m_exaggerateRadius);

  io->writeInt16(GEE_RadiusNormId);
  io->writeFloat32(m_radiusNorm);

  io->writeInt16(GEE_DelPartsId);
  io->writeInt32(m_deleteParts);

  io->writeInt16(GEE_UseNapraId);
  io->writeByte(m_useNapra);

  io->writeInt16(GEE_EndId);

  return true;
}

const char * GEE_EraseAroundPath::type() 
{ 
  return "GEE_EraseAroundPath"; 
}

GEE_Operator * GEE_EraseAroundPath::create()
{
  return new GEE_EraseAroundPath();
}

bool GEE_EraseAroundPath::handleNewVertex(GEE_Path3D::iterator &it)
{
  const GEE_PathVertex3D & pv = GEE_Path3D::getVertex(it);

  float radius = m_radius;

  VEE_Vector3 loc = pv.m_location;

  if(m_useNapra) {
    radius = 
      pv.calculateToolRadius(m_radius, m_radiusNorm, m_exaggerateRadius);
    loc = pv.calculateToolCenter();
  }
  addSphere(loc, radius);

  return true;
}

void GEE_EraseAroundPath::doWork(GEE_World *w)
{
  // VEE_trace("GEE_EraseAroundPath::doWork");
  GEE_Geometry * g;

  if(m_deleteParts)
    for(GEE_LayerGeometryIterator it(w, m_layer); (g = *it); it.nextVisible())
      g->eraseData(this);
  else {
    for(GEE_LayerGeometryIterator it(w, m_layer); (g = *it); ) {
      if(g->intersects(this))
	it.eraseAndNextVisible();
      else
	it.nextVisible();
    }
  }
}
