/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_gallery.h>

#include <gee_debug.h>
#include <gee_engine.h>
#include <gee_io.h>
#include <gee_macros.h>

#include <vee_error.h>

static bool m_debug;

GEE_Gallery::GEE_Gallery()
  : m_deleted(0)
{
  m_debug = GEE_Debug::debug("GEE_Gallery");
}

GEE_Gallery::~GEE_Gallery()
{}

/** Takes a snapshot of all visible things in the source
    world. Returns the id of the snapshot. */

GEE_id_t GEE_Gallery::takeSnapshot(GEE_World *w, 
				   /** Target layer (unless equals -1) */
				   GEE_id_t layer,
				   /** Only take selected objects. */
				   bool onlySelected)
{
  GEE_Geometry * g;

  GEE_World * w2 = new GEE_World();
  w2->setGarbageBin(m_deleted);

  w2->autoId();

  GEE_IdContainer::id_remap tmp;

  for(GEE_LayerGeometryIterator it(w, layer); (g = *it); ++it) {
    if(!onlySelected || g->isSelected())
      g->copy(w2, &tmp);
  }

  if(layer != -1)
    w2->allGeometriesToLayer(layer);

  for(GEE_LayerGeometryIterator it2(w2, layer); (g = *it2); ++it2) {
    g->setSelectionState(0);
    g->setGeneration(0);
  }

  m_worlds[w2->id()] = w2;

  return w2->id();
}

GEE_World * GEE_Gallery::item(GEE_id_t id)
{
  iterator it = m_worlds.find(id);

  if(it == m_worlds.end())
    return 0;

  return (*it).second.ptr();
}

VEE_RefPtr<GEE_World> * GEE_Gallery::item2(GEE_id_t id)
{
  iterator it = m_worlds.find(id);

  if(it == m_worlds.end())
    return 0;

  return & (*it).second;
}

void GEE_Gallery::clear()
{
  m_worlds.clear();
}

static const ushort GEE_ItemId   = GEE_CONST2('I', 'T');
static const ushort GEE_EndId    = GEE_CONST2('E', 'N');

bool GEE_Gallery::read(GEE_Io *io)
{
  GEE_FNAME = "GEE_Gallery::read";

  if(m_debug)
    VEE_trace(fname);

  clear();
  
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != "GEE_Gallery") {
    VEE_error("%s # Bad id (%s)", fname, str.c_str());
    return false;
  }

  int v = io->readInt32();

  if(v != 1)
    return false;

  ushort id = io->readInt16();

  while(id != GEE_EndId) {
    if(id == GEE_ItemId) {

      GEE_World * w = new GEE_World();
      w->setGarbageBin(m_deleted);

      bool ok2 = w->read(io);
      m_worlds[w->id()] = w;

      if(!ok2) {
	VEE_error("%s # Bad world", fname);
      }

    }
    else {
      VEE_error("%s # Bad id at %d (%c%c)", fname,
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }
  
    id = io->readInt16();
  }

  VEE_trace("%s # %d items", fname, (int) m_worlds.size());

  return true;
}

bool GEE_Gallery::write(GEE_Io *io)
{
  GEE_FNAME = "GEE_Gallery::write";

  if(m_debug)
    VEE_trace(fname);

  io->writeNullTerminatedString("GEE_Gallery");
  
  io->writeInt32(1); // version

  for(iterator it = m_worlds.begin(); it != m_worlds.end(); it++) {
    io->writeInt16(GEE_ItemId);
    (*it).second.ptr()->write(io);
  }

  io->writeInt16(GEE_EndId);

  return true;
}

/** Writes a particular gallery item alone into a gee file. The
    resulting gee file will contain nothing but that single gallery
    item. */

bool GEE_Gallery::writeItem(const char *file, GEE_id_t id)
{
  GEE_World * w = item(id);

  if(!w) {
    VEE_error("GEE_Gallery::writeItem # No item %d", (int) id);
    return false;
  }
  
  GEE_Engine tmp;
  tmp.gallery().takeSnapshot(w);
  return tmp.write(file);
}

void GEE_Gallery::collectIds(GEE_IdContainer::id_remap &idMap)
{
  for(iterator it = m_worlds.begin(); it != m_worlds.end(); it++) {
    (*it).second.ptr()->collectIds(idMap);
  }
}

void GEE_Gallery::remapIds(GEE_IdContainer::id_remap &idMap)
{
  container tmp;

  for(iterator it = m_worlds.begin(); it != m_worlds.end(); it++) {
    GEE_World * w = (*it).second.ptr();
    w->remapIds(idMap);
    tmp[w->id()] = (*it).second;

    VEE_trace("GEE_Gallery::remapIds # %d -> %d", 
	      (int) (*it).first, (int) w->id());
    w->dump();
  }

  m_worlds.swap(tmp);
}

/** Merges all items from another gallery. Does not modify the source
    gallery. */

void GEE_Gallery::merge(GEE_Gallery &g)
{
  m_worlds.insert(g.m_worlds.begin(), g.m_worlds.end());
}

GEE_Gallery::iterator GEE_Gallery::getIndex(uint index)
{
  iterator it = begin(); 
  uint i = 0;
  
  while(it != end() && i < index) {
    it++;
    i ++;
  }

  return it;
}

void GEE_Gallery::erase(iterator it) 
{ 
  m_worlds.erase(it); 
}

void GEE_Gallery::setGarbageBin(VEE_GarbageBin * bin)
{
  m_deleted = bin;

  for(iterator it = m_worlds.begin(); it != m_worlds.end(); it++)
    (*it).second.ptr()->setGarbageBin(bin);
}

GEE_id_t GEE_Gallery::itemId(int index)
{
  int i = 0;
  iterator it = m_worlds.begin();

  while(it != m_worlds.end() && i < index) {
    i++;
    it++;
  }

  return (it == m_worlds.end()) ? -1 : (*it).first;
}
