/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_group.h>

#include <gee_io.h>
#include <gee_macros.h>
#include <gee_polygon_mesh.h>
#include <gee_renderer.h>
#include <gee_world.h>

#include <gee_lwo.hpp>

#include <vee_bounding_box.h>
#include <vee_error.h>
#include <vee_vector2.h>


#include <fstream>
#include <list>
#include <map>
#include <vector>

std::vector<std::string> split_string(const std::string &str, 
				      const std::string &separators = " \n\t")
{
  std::vector<std::string> r;
  
  int i1 = 0;

  while(i1 + 1 < (int) str.size()) {

    i1 = str.find_first_not_of(separators, i1);
    int i2 = str.find_first_of(separators, i1 + 1);

    if(i2 < i1 || i2 >= (int) str.size()) {
      r.push_back(str.substr(i1, str.size() - i1));
      return r;
    }

    if(i1 != i2) {
      r.push_back(str.substr(i1, i2 - i1));
    }
    if(i1 == i2)
      i1++;
    else
      i1 = i2;
  }
  
  return r;
}

std::vector<std::string> split_string2(const std::string &str, 
				       const std::string &separators = " \n\t")
{
  std::vector<std::string> r;
  
  int i1 = 0;

  while(i1 < (int) str.size()) {
    // i1 = str.find_first_not_of(separators, i1);
    int i2 = str.find_first_of(separators, i1);

    if(i2 < i1 || i2 >= (int) str.size()) {
      r.push_back(str.substr(i1, str.size() + 1 - i1));
      return r;
    }

    if(i1 != i2) {
      r.push_back(str.substr(i1, i2 - i1));
    }
    else 
      r.push_back(std::string());

    i1 = i2 + 1;
    // printf("%d\n", i1);
  }
  
  return r;
}


GEE_Group::GEE_Group()
{}

/** Breaks the group.*/
GEE_Group::~GEE_Group()
{
  breakGroup();
}

GEE_IdContainer::id_container * GEE_Group::getExtraIds()
{
  return & m_children;
}

GEE_Geometry * GEE_Group::create()
{
  return new GEE_Group();
}

GEE_Geometry * GEE_Group::clone() const
{
  return new GEE_Group(*this);
}

GEE_Geometry * GEE_Group::copy(GEE_World *target, id_remap * remap) const
{
  GEE_Group * gr = new GEE_Group(*this);
  
  if(!gr) 
    return 0;

  gr->m_children.clear();

  id_remap::iterator it = remap->find(m_id);
  
  if(it == remap->end()) {
    gr->autoId();
    (*remap)[m_id] = gr->id();
  }
  else {
    gr->setId((*it).second);
  }

  target->addGeometry(gr);
  
  for(iterator it = m_children.begin(); it != m_children.end(); it++) {
    VEE_RefPtr<GEE_Geometry> * gptr = ((GEE_Group *) this)->getGeometry(it);

    if(!gptr) continue;

    GEE_Geometry * g = gptr->ptr()->copy(target, remap);

    g->setGroup(-1); // Needed since otherwise it cannot be added as a child

    gr->addChild(g->id());
  }
  
  return gr;
}

void GEE_Group::erase()
{
  if(!m_world)
    return;

  for(iterator it = m_children.begin(); it != m_children.end(); it++) {

    GEE_id_t c = getGeometryId(it);
    m_world->eraseGeometry(c);
  }
  
  m_children.clear();
}

bool GEE_Group::eraseData(GEE_Containment * co)
{
  if(!m_world)
    return false;

  for(iterator it = m_children.begin(); it != m_children.end(); it++) {
    VEE_RefPtr<GEE_Geometry> * geo = getGeometry(it);
    if(geo) geo->ptr()->eraseData(co);
  }

  return false;
}

bool GEE_Group::intersects(GEE_Containment * co)
{
  if(!m_world) {
    VEE_error("GEE_Group::intersects # m_world = NULL");
    return false;
  }

  VEE_trace("GEE_Group::intersects");

  for(iterator it = m_children.begin(); it != m_children.end(); it++) {
    VEE_RefPtr<GEE_Geometry> * geo = getGeometry(it);
    if(geo && geo->ptr()->intersects(co)) {
      VEE_trace("GEE_Group::intersects # Intersected");
      return true;
    }
  }
  return false;
}

void GEE_Group::selectVertices(GEE_Containment *co,  
			       uint mask, 
			       uint values)
{
  if(!m_world)
    return;

  for(iterator it = m_children.begin(); it != m_children.end(); it++) {
    VEE_RefPtr<GEE_Geometry> * geo = getGeometry(it);
    if(geo)
      geo->ptr()->selectVertices(co, mask, values);
  }
}

void GEE_Group::selectFaces(GEE_Containment *co,  
			    uint mask, 
			    uint values)
{
  if(!m_world)
    return;

  for(iterator it = m_children.begin(); it != m_children.end(); it++) {
    VEE_RefPtr<GEE_Geometry> * geo = getGeometry(it);
    if(geo)
      geo->ptr()->selectFaces(co, mask, values);
  }
}

bool GEE_Group::colorData(GEE_ReColor *rc)
{
  if(!m_world)
    return false;

  bool hit = false;

  for(iterator it = m_children.begin(); it != m_children.end(); it++) {
    VEE_RefPtr<GEE_Geometry> * geo = getGeometry(it);
    if(geo)
      hit = hit || geo->ptr()->colorData(rc);
  }
  
  return hit;
}

bool GEE_Group::moveData(GEE_Mover *m)
{
  if(!m_world)
    return false;

  bool hit = false;

  for(iterator it = m_children.begin(); it != m_children.end(); it++) {
    VEE_RefPtr<GEE_Geometry> * geo = getGeometry(it);
    if(geo) {
      bool hit2 = geo->ptr()->moveData(m);
      hit = hit || hit2;
    }
  }  
  return hit;
}

void GEE_Group::render(GEE_Renderer *r)
{
  r->pushTransform(transformation());

  r->renderGroupStructure(this);

  if(r->enableTransforms()) {

    for(iterator it = begin(); it != end(); it++) {
      VEE_RefPtr<GEE_Geometry> * d = getGeometry(it);
      if(d)
	d->ptr()->render(r);
    }
  }

  r->popTransform();
}

bool GEE_Group::getBoundingBox(VEE_BoundingBox *box)
{
  if(!size())
    return false;

  // VEE_trace("GEE_Group::getBoundingBox");

  bool ok = false;

  for(iterator it = begin(); it != end(); it++) {    

    // VEE_trace("GEE_Group::getBoundingBox # %d", (int) (*it));

    VEE_BoundingBox box2;

    VEE_RefPtr<GEE_Geometry> * d = getGeometry(it);
    if(!d)
      continue;

    GEE_Geometry * g = d->ptr();
    g->getBoundingBox(&box2);
    
    VEE_Matrix4 m(g->transformation().data());
    box2 = m * box2;

    if(!ok) {
      *box = box2;
      ok = true;
    }
    else 
      box->expand(box2);
  }

  // VEE_trace("GEE_Group::getBoundingBox # EXIT");

  return ok;
}

bool GEE_Group::move(const VEE_Vector3 & offset)
{
  bool ok = true;

  for(iterator it = begin(); it != end(); it++) {
    VEE_RefPtr<GEE_Geometry> * d = getGeometry(it);
    if(d)
      d->ptr()->move(offset);
  }

  return ok;
}

static const ushort GEE_IdId       = GEE_CONST2('I', 'D');
static const ushort GEE_EndId      = GEE_CONST2('E', 'N');

bool GEE_Group::read(GEE_Io * io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != "GEE_Group")
    return false;
  
  readBasics(io);

  int v = io->readInt32();

  if(v != 1)
    return false;

  ushort id = io->readInt16();

  while(id != GEE_EndId) {
    if(id == GEE_IdId)
      m_children.insert(io->readInt64());
    else {
      VEE_error("GEE_Group::read # Bad id at %d (%c%c)", 
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }
    id = io->readInt16();
  }
  
  return true;
}

bool GEE_Group::write(GEE_Io * io)
{
  io->writeNullTerminatedString("GEE_Group");

  writeBasics(io);

  io->writeInt32(1); // version

  for(iterator it = begin(); it != end(); it++) {
    io->writeInt16(GEE_IdId);
    io->writeInt64(*it);
  }
  
  io->writeInt16(GEE_EndId);

  return true;
}

bool GEE_Group::finalizeRead()
{
  id_container tmp;
  tmp.swap(m_children);

  for(iterator it = tmp.begin(); it != tmp.end(); it++)
    addChild(*it);
  
  return true;
}

void GEE_Group::setSelectionState(uint s)
{
  GEE_Geometry::setSelectionState(s);

  for(iterator it = m_children.begin(); it != m_children.end(); it++) {
    GEE_Geometry * g = m_world->geometry(*it);
    if(g)
      g->setSelectionState(s);
  }
}

/** Sets the material of the group and all children. */

void GEE_Group::setMaterial(const GEE_Material &m)
{
  GEE_Geometry::setMaterial(m);

  if(!m_world && m_children.size()) {
    VEE_error("GEE_Group::setMaterial # m_world = NULL");
    return;
  }

  for(iterator it = m_children.begin(); it != m_children.end(); it++) {
    GEE_Geometry * g = m_world->geometry(*it);
    if(g)
      g->setMaterial(m);
  }
}

void GEE_Group::clear()
{
  m_children.clear();
}

void GEE_Group::addChild(GEE_id_t id)
{
  GEE_FNAME = "GEE_Group::addChild";

  if(!m_world) {
    VEE_error("%s # m_world = NULL", fname);
    return;
  }

  if(id == m_id) {
    VEE_error("%s # id = m_id", fname);
    return;
  }
  
  GEE_Geometry * g = m_world->geometry(id);

  if(!g) {
    VEE_error("%s # Geometry %d does not exist", fname, (int) id);
    return;
  }

  if(g->group() >= 0 && g->group() != m_id) {
    VEE_error("%s # Geometry %d belongs to another group", fname, (int) id);
    return;
  }

  m_children.insert(id);

  g->setGroup(m_id);
  g->setVisible(false);
}

void GEE_Group::breakGroup()
{
  GEE_FNAME = "GEE_Group::breakGroup";

  if(!m_world) {
    VEE_error("%s # m_world = NULL", fname);
    return;
  }

  VEE_Matrix4d m;
  getWorldTransformation(& m);

  for(iterator it = m_children.begin(); it != m_children.end(); it++) {

    GEE_id_t c = getGeometryId(it);

    GEE_Geometry * g = m_world->geometry(c);
    
    if(!g) {
      VEE_error("%s # Geometry %d does not exist", fname, (int) c);
      return;
    }
    else {
      g->setGroup(-1);
      g->setVisible(visible());
      g->setTransformation(m * g->transformation());
    }
  }

  m_children.clear();
}

VEE_RefPtr<GEE_Geometry> * GEE_Group::getGeometry(iterator & it)
{
  if(!m_world) {
    VEE_error("GEE_Group::getGeometry # m_world = NULL");
    return 0;
  }

  GEE_id_t id = getGeometryId(it);
  
  return m_world->geometry2(id);
}

bool GEE_Group::loadModel(const char *filename, bool verbose)
{
  const char * fname = "GEE_Group::loadModel";

  std::vector<std::string> parts = split_string(filename, ".");

  if(parts.size() < 2) {
    if(verbose)
      VEE_error("%s # filename does not include suffix", fname);
    return false;
  }

  const std::string & suffix = parts[parts.size() - 1];

  if(suffix == "obj")
    return loadWavefrontObj(filename, verbose);
  else if(suffix == "lwo")
    return loadLightwaveObj(filename, verbose);

  if(verbose)
    VEE_error("%s # unrecognized suffix %s", fname, suffix.c_str());

  return false;
}

/// Helper for GEE_Group::loadLightwaveObj
class TrimeshStore
{
public:
  TrimeshStore() : m_mesh(new GEE_TriangleMesh) { }
  GEE_TriangleMesh * m_mesh;
};

bool GEE_Group::loadLightwaveObj(const char *filename, bool verbose)
{
  using namespace std;

  GEE_FNAME = "GEE_Group::loadLightwaveObj";

  if(!m_world) {
    VEE_error("%s # m_world = NULL", fname);
    return false;
  }

  Mesh lwo(filename);

  if(!lwo.getTriangleCount())
    return false;

  const vector<Mesh::Vertex>   & lvertices  = lwo.getVertices();
  const vector<Mesh::Triangle> & ltriangles = lwo.getTriangles();
  // const vector<Mesh::Edge>     & ledges     = lwo.getEdges();
  const vector<Mesh::Surface>  & lsurfaces  = lwo.getSurfaces();

  map<string, TrimeshStore> meshes;

  // Convert triangle and vertex data from LWO
  for(uint t=0; t < ltriangles.size(); t++) {
    const Mesh::Triangle & tri = ltriangles[t];
    
    const string & material = lsurfaces[tri.m_surface].m_name;

    GEE_TriangleMesh * trimesh = meshes[material].m_mesh;
    GEE_TriangleIndices gtri;

    for(int i=0; i < 3; i++) {

      int vertexId = tri.m_v[i];
      gtri.m_indices[i] = vertexId;

      if(!trimesh->getVertex(vertexId)) {
	const Mesh::Vertex & lv = lvertices[vertexId];

	GEE_PolygonVertex gv;
	gv.m_location.copy(& lv.m_position[0]);
	gv.m_normal.copy(& lv.m_normal[0]);
	gv.m_uv.copy(& lv.m_uv[0]);
	gv.m_rgba.copy(& lv.m_diffuse[0]);

	trimesh->setVertex(vertexId, gv);
      }

    }

    trimesh->addTriangle(gtri);
  }

  // Set up the material names and store meshes to this group:
  
  for(map<string, TrimeshStore>::iterator it = meshes.begin();
      it != meshes.end(); it++) {

    GEE_TriangleMesh * trimesh = (*it).second.m_mesh;
    trimesh->setMaterialName((*it).first);
    trimesh->setGeneration(1);
    trimesh->setVisible(false);

    addChild(m_world->addGeometry(trimesh));
  }

  return true;
}

/// Helper for GEE_Group::loadWavefrontObj
class ObjVertex 
{
public:
  ObjVertex() { v = vn = vt = -1; }
  int v, vn, vt;
};

bool GEE_Group::loadWavefrontObj(const char *filename, bool verbose)
{
  const char * fname = "GEE_Group::loadWavefrontObj";

  using namespace std;

  clear();

  if(!m_world) {
    VEE_error("%s # m_world = NULL", fname);
    return false;
  }

  ifstream file;
  
  file.open(filename, ifstream::in);

  if(!file.good()) {
    if(verbose)
      VEE_error("%s # Failed to open file %s", fname, filename);

    return false;
  }
  else if(verbose)
    VEE_trace("%s # Loading model %s", fname, filename);


  const uint len = 1024;
  
  char buf[len];

  string id;

  vector<VEE_Vector3> vertices;
  vector<VEE_Vector3> normals;
  vector<VEE_Vector2> uvs;

  map<string, list<list<ObjVertex> > > model;

  string mtl;

  while(file.good()) {

    file >> id;

    if(id[0] == '#' || !id.size()) {
      // Eat up comments
      file.getline(buf, len);
      continue;
    }
    
    if(id == "v" || id == "vt" || id == "vn") {
      // Vertex-related stuff
      file.getline(buf, len);

      float x = 0.0, y = 0.0, z = 0.0, w = 0.0;
      sscanf(buf, "%f %f %f %f", &x, &y, &z, &w);
      
      if(id == "v") {
	vertices.push_back(VEE_Vector3(x, y, z));
	/* if(vertices.size() < 10)
	   VEE_trace("x = %f, y = %f, z = %f", x, y, z); */
      }
      if(id == "vn")
	normals.push_back(VEE_Vector3(x, y, z));
      if(id == "vt")
	uvs.push_back(VEE_Vector2(x, y));
    }
    else if(id == "f") {
      file.getline(buf, len);

      vector<string> vertices = split_string(buf);

      // VEE_trace("%s -> %d vertices", buf, (int) vertices.size());

      list<ObjVertex> polygon;

      for(uint vert = 0; vert < vertices.size(); vert++) {
	vector<string> params = split_string2(vertices[vert], "/");
	
	ObjVertex ov;

	if(params.size() >= 1 && params[0].size())
	  sscanf(params[0].c_str(), "%d", &ov.v);
	if(params.size() >= 2 && params[1].size())
	  sscanf(params[1].c_str(), "%d", &ov.vt);
	if(params.size() >= 3 && params[2].size())
	  sscanf(params[2].c_str(), "%d", &ov.vn);
	
	polygon.push_back(ov);

	/* VEE_trace("poly v = %d, vt = %d, vn =  %d (%s %s)", 
	   ov.v, ov.vt, ov.vn, 
	   params[0].c_str(), vertices[vert].c_str()); */
      }

      model[mtl].push_back(polygon);
    }
    else if(id == "usemtl") {
      file >> mtl;
    }
    else
      file.getline(buf, len); // Eat unknown stuff
  }

  if(verbose)
    VEE_trace("%s # About to parse the model (%d materials)", 
	      fname, (int) model.size());

  uint polygonCount = 0;
  uint vertexCount = 0;

  // Iterate over each material:
  for(map<string, list<list<ObjVertex> > >::iterator it = model.begin();
      it != model.end(); it++) {
    
    GEE_PolygonMesh * pm = new GEE_PolygonMesh();

    pm->setMaterialName((*it).first);

    list<list<ObjVertex> > * polygons = & (*it).second;

    if(verbose)
      VEE_trace("%s # material \"%s\" with %d polygons",
		fname, (*it).first.c_str(), (int) polygons->size());

    // Iterate over each polygon
    for(list<list<ObjVertex> >::iterator pit = polygons->begin();
	pit != polygons->end(); pit++) {

      GEE_PolygonIndices pind;

      // Iterate over each vertex:
      for(list<ObjVertex>::iterator vit = (*pit).begin();
	  vit != (*pit).end(); vit++) {

	GEE_PolygonVertex pv;

	ObjVertex ov = (*vit);
	
	/* VEE_trace("poly v = %d, vt = %d, vn =  %d", 
	   ov.v, ov.vt, ov.vn); */

	if(ov.v > 0)
	  pv.m_location = vertices[ov.v - 1];
	if(ov.vt > 0)
	  pv.m_uv = uvs[ov.vt - 1];
	if(ov.vn > 0)
	  pv.m_normal = normals[ov.vn - 1];

	int pvi = pm->addVertex(pv);

	pind.addVertex(pvi);
	vertexCount++;
      }
      pm->addPolygon(pind);
      polygonCount++;
    }
    
    pm->setGeneration(1);
    pm->setVisible(false);
    addChild(m_world->addGeometry(pm));
    // addChild(pm);
  }

  if(verbose)
    VEE_trace("%s # Loaded %s with %u polygons and %u vertices",
	      fname, filename, polygonCount, vertexCount);

  return true;
}

void GEE_Group::convertPolygonsToTriangles()
{
  if(!m_world) {
    VEE_error("GEE_Group::getGeometry # m_world = NULL");
    return;
  }

  for(iterator it = begin(); it != end(); it++) {
    VEE_RefPtr<GEE_Geometry> * d = getGeometry(it);

    if(!d)
      continue;

    GEE_PolygonMesh * pm = dynamic_cast<GEE_PolygonMesh *> (d->ptr());
    GEE_Group * g = dynamic_cast<GEE_Group *> (d->ptr());

    if(pm) {
      GEE_TriangleMesh * tm = pm->triangleMesh();
      addChild(m_world->addGeometry(tm));
    }
    if(g)
      g->convertPolygonsToTriangles();
  }
}
