/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_lines.h>

#include <gee_containment.h>
#include <gee_group.h>
#include <gee_io.h>
#include <gee_macros.h>
#include <gee_mover.h>
#include <gee_recolor.h>
#include <gee_renderer.h>

#include <vee_error.h>

#include <typeinfo>

static const int GEE_LineVertex3DId = GEE_CONST('L', 'N', 'V', 'X');

bool GEE_LineVertex3D::read(GEE_Io *io)
{
  if(io->readInt32() != GEE_LineVertex3DId)
    return false;
  
  int v = io->readInt32();

  if(v != 1)
    return false;

  m_location  = io->readVector3();
  m_rgba      = io->readVector4();
  m_mixture   = io->readFloat32();
  m_continues = io->readByte();

  return true;
}

bool GEE_LineVertex3D::write(GEE_Io *io)
{
  io->writeInt32(GEE_LineVertex3DId);
  io->writeInt32(1); // Version  
  
  io->writeVector3(m_location);
  io->writeVector4(m_rgba);
  io->writeFloat32(m_mixture);
  io->writeByte(m_continues);
  
  return true;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

GEE_Lines3D::GEE_Lines3D()
  : m_lineWidth(1.0)
{
  m_bbox.clear();
}

GEE_Lines3D::~GEE_Lines3D()
{}

bool GEE_Lines3D::eraseData(GEE_Containment *co)
{
  bool erasedPrevious = false;

  bool hit = false;

  VEE_Matrix4 m;
  getWorldTransformation(&m);

  if(!intersectsBounded(co, &m_bbox, m))
    return false;

  for(iterator it = m_vertices.begin(); it != m_vertices.end(); ) {
    GEE_LineVertex3D & lv = getVertex(it);

    if(co->contains(m * lv.m_location)) {
      if(it == m_vertices.begin()) {
	m_vertices.erase(it);
	it = m_vertices.begin();
      }
      else {
	/* Here we do a difficult thing: If the preceding line was cut
	   off we need to erase that line segment as well.  There are
	   all sorts of special cases that we need to handle. */

	iterator tmp = it;
	tmp--;
	m_vertices.erase(it);

	if(tmp == m_vertices.begin()) {
	  it = tmp;
	  it++;
	  m_vertices.erase(tmp);
	}
	else {

	  iterator tmp2 = tmp;
	  tmp2--;

	  if(!(*tmp2).second.m_continues) {

	    m_vertices.erase(tmp);
	    // if(tmp2 != m_vertices.end())
	    tmp2++; // locks here ?
	    it = tmp2;
	  }
	  else {
	    getVertex(tmp).m_continues = false;
	    tmp++;
	    it = tmp;
	  }
	}
      }
      erasedPrevious = true;
      hit = true;
    }
    else {
      if(erasedPrevious && !lv.m_continues) {
	iterator tmp = it;
	tmp++;
	m_vertices.erase(it);
	it = tmp;
      }
      else {
	it++;
	erasedPrevious = false;
      }
    }
  }  

  if(hit) {
    incrementGeneration();
    co->contained(this);
  }

  return hit;
}

bool GEE_Lines3D::intersects(GEE_Containment *co)
{
  VEE_Matrix4 m;
  getWorldTransformation(&m);

  if(!intersectsBounded(co, &m_bbox, m))
    return false;

  for(iterator it = begin(); it != end(); it++) {
    GEE_LineVertex3D & lv = getVertex(it);

    if(co->contains(m * lv.m_location))
      return true;
  }
  
  return false;
}

bool GEE_Lines3D::colorData(GEE_ReColor * rc)
{  
  bool vertexColors = m_material.useVertexColors();

  VEE_Matrix4 m;
  getWorldTransformation(&m);

  if(!intersectsBounded(rc, &m_bbox, m)) {
    // VEE_trace("GEE_Lines3D::colorData # culled a recolor");
    return false;
  }

  bool hit = false;

  for(iterator it = m_vertices.begin(); it != m_vertices.end(); it++) {

    GEE_LineVertex3D & pv = (*it).second;
 
    VEE_Vector4 color = hit ? pv.m_rgba : m_material.ambient();
    
    bool change = rc->recolor(m * pv.m_location, & color);

    if(change) {

      if(!vertexColors) {
	m_materialName.resize(0);
	m_material.setUseVertexColors(true);
	// Put the ambient color to all the vertices.
	VEE_Vector4 rgba = m_material.ambient();
	for(iterator it2 = m_vertices.begin(); it2 != m_vertices.end(); it2++){
	  (*it2).second.m_rgba = rgba;
	}
	
	// m_material.setColor(VEE_Vector4(rc->basicColor().data()));
	
	/*VEE_trace("GEE_Lines3D::colorData # Switching to vertex colors %d",
	  (int) m_material.useVertexColors()); */
	vertexColors = true;
      }
      hit = true;
      pv.m_rgba = color;
    }
  }

  // VEE_trace("GEE_Lines3D::colorData # DONE %d", (int) m_vertices.size());

  if(hit) {
    incrementGeneration();
    rc->recolored(this);
  }

  return hit;
}

bool GEE_Lines3D::moveData(GEE_Mover *m)
{
  bool hit = false;

  VEE_Matrix4 ma, mi;
  bool ok ;
  getWorldTransformations(&ma, & mi, & ok);

  if(!ok) {
    VEE_error("GEE_Lines3D::moveData # Bad inverse matrix");
    return false;
  }

  if(!intersectsBounded(m, &m_bbox, ma)) {
    // VEE_trace("GEE_Lines3D::moveData # culled a mover");
    return false;
  }

  for(iterator it = m_vertices.begin(); it != m_vertices.end(); it++) {
    GEE_LineVertex3D & pd = (*it).second;
    VEE_Vector3 tmp = ma * pd.m_location;
    if(m->move(&tmp)) {
      hit = true;
      tmp = mi * tmp;
      pd.m_location = tmp;
      m_bbox.expand(tmp);
    }
  }

  if(hit) {
    incrementGeneration();
    m->moved(this);
  }

  return false;
}

void GEE_Lines3D::setMaterial(const GEE_Material &m)
{
  VEE_trace("GEE_Lines3D::setMaterial");
  m_material = m;
  incrementGeneration();
}

GEE_Geometry * GEE_Lines3D::create()
{
  return new GEE_Lines3D();
}

GEE_Geometry * GEE_Lines3D::clone() const
{
  return new GEE_Lines3D(*this);
}

bool GEE_Lines3D::getBoundingBox(VEE_BoundingBox *bb)
{
  *bb = m_bbox;
  return true;
}

void GEE_Lines3D::render(GEE_Renderer *r)
{
  // VEE_trace("GEE_Lines3D::render # %p %s", r, typeid(*r).name());
  r->pushTransform(transformation());
  r->renderLines3D(this);
  r->popTransform();
}

bool GEE_Lines3D::read(GEE_Io *io)
{
  m_vertices.clear();

  std::string id;
  io->readNullTerminatedStdString(id);

  if(id != "GEE_Lines3D") {
    VEE_error("GEE_Lines3D::read # Bad type (%s)", id.c_str());
    return false;
  }

  readBasics(io);

  int v = io->readInt32();
  if(v != 1)
    return false;
  
  m_lineWidth = io->readFloat32();

  uint vertices = io->readInt32();

  GEE_LineVertex3D vertex;

  for(uint i=0; i < vertices; i++) {
    int id = io->readInt32();
    if(vertex.read(io))
      addVertex(id, vertex);
  }

  VEE_trace("GEE_Lines3D::read # EXIT (%d vertices)", 
	    (int) m_vertices.size());

  return true;
}

bool GEE_Lines3D::write(GEE_Io *io)
{
  io->writeNullTerminatedString("GEE_Lines3D");

  writeBasics(io);

  io->writeInt32(1); // version
  io->writeFloat32(m_lineWidth);
  
  io->writeInt32(m_vertices.size());

  for(iterator it = m_vertices.begin(); it != m_vertices.end(); it++) {
    io->writeInt32((*it).first);
    (*it).second.write(io);
  }
  
  return true;
}

void GEE_Lines3D::addVertex(int index, const GEE_LineVertex3D &vertex)
{ 
  if(!m_vertices.size())
    m_bbox.set(vertex.m_location);
  else
    m_bbox.expand(vertex.m_location);

  m_vertices[index] = vertex;
  m_generation++;
}
