#include <gee_lwo.hpp>

#include <float.h>

#include <vee_error.h>

namespace VECTORMATRIX
{


  //==============================================================================================
  //==============================================================================================

  // matrix inversion and determinant code from Graphics Gems I
  static inline T det2x2( T m00, T m10, T m01, T m11 )
  {
    return m00 * m11 - m01 * m10;
  }

  //==============================================================================================
  //==============================================================================================

  static inline T det3x3( T m00, T m10, T m20, T m01, T m11, T m21, T m02, T m12, T m22 )
  {
    return m00 * det2x2( m11, m21, m12, m22 )
      - m01 * det2x2( m10, m20, m12, m22 )
      + m02 * det2x2( m10, m20, m11, m21 );
  }

  //==============================================================================================
  //==============================================================================================






  //==============================================================================================
  //==============================================================================================

  bool Matrix2::invert()
  {
    T d = det();
    if( d == T(0) ) return false;
    d = T(1) / d;
    // inv(A) =       [ a   -a   ]
    //                        1   [  11   01 ]
    //                      ----- [-a    a   ]
    //                      det A [  10   00 ]
    //
    T t = matrix[0][0]; matrix[0][0] = matrix[1][1]; matrix[1][1] = t;
    matrix[0][1] = -matrix[0][1];
    matrix[1][0] = -matrix[1][0];
    *this *= d;
    return true;
  }

  //==============================================================================================
  //==============================================================================================

  void Matrix2::rotate( T radians, MultiplicationOrder o )
  {
    T c = COS( radians );
    T s = SIN( radians );
    if( o == LEFT_MULTIPLY )
      s = -s; //transposes the matrix
    Matrix2 m( c, s,
	       -s, c );
    multiply( m, o );
  }

  //==============================================================================================
  //==============================================================================================





  //==============================================================================================
  //==============================================================================================

  bool Matrix3::invert()
  {
    T det00 = matrix[1][1]*matrix[2][2] - matrix[2][1]*matrix[1][2];
    T det01 = matrix[2][0]*matrix[1][2] - matrix[1][0]*matrix[2][2];
    T det02 = matrix[1][0]*matrix[2][1] - matrix[2][0]*matrix[1][1];

    T det = matrix[0][0]*det00 + matrix[0][1]*det01 + matrix[0][2]*det02;
    if( det == T(0) ) return false; //singular, leave the matrix unmodified and return false
    det = T(1) / det;

    Matrix3 t;
    t[0][0] = det * det00;
    t[1][0] = det * det01;
    t[2][0] = det * det02;
    t[0][1] = det * (matrix[2][1]*matrix[0][2] - matrix[0][1]*matrix[2][2]);
    t[1][1] = det * (matrix[0][0]*matrix[2][2] - matrix[2][0]*matrix[0][2]);
    t[2][1] = det * (matrix[2][0]*matrix[0][1] - matrix[0][0]*matrix[2][1]);
    t[0][2] = det * (matrix[0][1]*matrix[1][2] - matrix[1][1]*matrix[0][2]);
    t[1][2] = det * (matrix[1][0]*matrix[0][2] - matrix[0][0]*matrix[1][2]);
    t[2][2] = det * (matrix[0][0]*matrix[1][1] - matrix[1][0]*matrix[0][1]);
    *this = t;
    return true;
  }

  //==============================================================================================
  //==============================================================================================

  void Matrix3::rotate( T radians, const Vector3& aboutThis, MultiplicationOrder o )
  {
    T c = COS( radians );
    T t = T(1) - c;
    T s = SIN( radians );
    if( o == LEFT_MULTIPLY )
      s = -s; //transposes the matrix
    Vector3 vn( aboutThis );
    if( !vn.normalize( ) )
      return; //in case the about vector is zero, do nothing
    Matrix3 m( t*vn.x*vn.x+c,          t*vn.x*vn.y+s*vn.z, t*vn.x*vn.z-s*vn.y,
	       t*vn.x*vn.y-s*vn.z, t*vn.y*vn.y+c,      t*vn.y*vn.z+s*vn.x,
	       t*vn.x*vn.z+s*vn.y, t*vn.y*vn.z-s*vn.x, t*vn.z*vn.z+c           );
    multiply( m, o );
  }

  //==============================================================================================
  //==============================================================================================





  //==============================================================================================
  //==============================================================================================

  static void adjoint( Matrix4& m )
  {
    T m00 = m[0][0], m01 = m[0][1], m02 = m[0][2], m03 = m[0][3];
    T m10 = m[1][0], m11 = m[1][1], m12 = m[1][2], m13 = m[1][3];
    T m20 = m[2][0], m21 = m[2][1], m22 = m[2][2], m23 = m[2][3];
    T m30 = m[3][0], m31 = m[3][1], m32 = m[3][2], m33 = m[3][3];

    // row column labeling reversed since we transpose rows and columns
    m[0][0]  =  det3x3( m11, m21, m31, m12, m22, m32, m13, m23, m33 );
    m[1][0]  = -det3x3( m10, m20, m30, m12, m22, m32, m13, m23, m33 );
    m[2][0]  =  det3x3( m10, m20, m30, m11, m21, m31, m13, m23, m33 );
    m[3][0]  = -det3x3( m10, m20, m30, m11, m21, m31, m12, m22, m32 );

    m[0][1]  = -det3x3( m01, m21, m31, m02, m22, m32, m03, m23, m33 );
    m[1][1]  =  det3x3( m00, m20, m30, m02, m22, m32, m03, m23, m33 );
    m[2][1]  = -det3x3( m00, m20, m30, m01, m21, m31, m03, m23, m33 );
    m[3][1]  =  det3x3( m00, m20, m30, m01, m21, m31, m02, m22, m32 );

    m[0][2]  =  det3x3( m01, m11, m31, m02, m12, m32, m03, m13, m33 );
    m[1][2]  = -det3x3( m00, m10, m30, m02, m12, m32, m03, m13, m33 );
    m[2][2]  =  det3x3( m00, m10, m30, m01, m11, m31, m03, m13, m33 );
    m[3][2]  = -det3x3( m00, m10, m30, m01, m11, m31, m02, m12, m32 );

    m[0][3]  = -det3x3( m01, m11, m21, m02, m12, m22, m03, m13, m23 );
    m[1][3]  =  det3x3( m00, m10, m20, m02, m12, m22, m03, m13, m23 );
    m[2][3]  = -det3x3( m00, m10, m20, m01, m11, m21, m03, m13, m23 );
    m[3][3]  =  det3x3( m00, m10, m20, m01, m11, m21, m02, m12, m22 );
  }

  //==============================================================================================
  //==============================================================================================

  bool Matrix4::invert()
  {
    T d = det( );
    if( d == T(0) ) return false;
    adjoint( *this );
    *this *= (T(1) / d);
    return true;
  }

  //==============================================================================================
  //==============================================================================================

  T Matrix4::det() const
  {
    T m00 = matrix[0][0], m01 = matrix[0][1], m02 = matrix[0][2], m03 = matrix[0][3];
    T m10 = matrix[1][0], m11 = matrix[1][1], m12 = matrix[1][2], m13 = matrix[1][3];
    T m20 = matrix[2][0], m21 = matrix[2][1], m22 = matrix[2][2], m23 = matrix[2][3];
    T m30 = matrix[3][0], m31 = matrix[3][1], m32 = matrix[3][2], m33 = matrix[3][3];

    return m00 * det3x3( m11, m21, m31, m12, m22, m32, m13, m23, m33 ) -
      m01 * det3x3( m10, m20, m30, m12, m22, m32, m13, m23, m33 ) +
      m02 * det3x3( m10, m20, m30, m11, m21, m31, m13, m23, m33 ) -
      m03 * det3x3( m10, m20, m30, m11, m21, m31, m12, m22, m32 );
  }

  //==============================================================================================
  //==============================================================================================

  void Matrix4::rotate( T radians, const Vector3& aboutThis, MultiplicationOrder o )
  {
    T c = COS( radians );
    T t = T(1) - c;
    T s = SIN( radians );
    if( o == LEFT_MULTIPLY )
      s = -s; //transposes the matrix
    Vector3 vn( aboutThis );
    if( !vn.normalize( ) )
      return; //in case the about vector is zero, do nothing
    Matrix4 m( t*vn.x*vn.x+c,          t*vn.x*vn.y+s*vn.z, t*vn.x*vn.z-s*vn.y, T(0),
	       t*vn.x*vn.y-s*vn.z, t*vn.y*vn.y+c,      t*vn.y*vn.z+s*vn.x, T(0),
	       t*vn.x*vn.z+s*vn.y, t*vn.y*vn.z-s*vn.x, t*vn.z*vn.z+c,      T(0),
	       T(0),               T(0),               T(0),               T(1) );
    multiply( m, o );
  }

  //==============================================================================================
  //==============================================================================================





  //==============================================================================================
  //==============================================================================================

  bool Matrix2x3::invert()
  {
    //translation = -(invert(rotation) * translation)
    //rotation = invert(rotation)

    Matrix2 rotation( getRotation() );
    if( !rotation.invert( ) )
      return false;

    Vector2 translation( matrix[0][2], matrix[1][2] );
    translation.multiply( rotation, LEFT_MULTIPLY );

    set( rotation[0][0], rotation[0][1], -translation[0],
	 rotation[1][0], rotation[1][1], -translation[1] );
    return true;
  }

  //==============================================================================================
  //==============================================================================================

  const Matrix2x3 operator*( const Matrix2x3& m1, const Matrix2x3& m2 )
  {
    // dest.rotation = m1.rotation * m2.rotation
    // dest.translation = m1.rotation * m2.translation + m1.translation

    Matrix2 rotation1( m1.getRotation() ), rotation2( m2.getRotation() );
    Vector2 translation1( m1[0][2], m1[1][2] ), translation2( m2[0][2], m2[1][2] );

    rotation2.multiply( rotation1, LEFT_MULTIPLY );
    translation2.multiply( rotation1, LEFT_MULTIPLY );
    translation2 += translation1;

    return Matrix2x3( rotation2[0][0], rotation2[0][1], translation2[0],
		      rotation2[1][0], rotation2[1][1], translation2[1] );
  }

  //==============================================================================================
  //==============================================================================================

  void Matrix2x3::rotate( T radians )
  {
    T c = COS( radians );
    T s = SIN( radians );
    //always a left multiply
    Matrix2x3 m( c, -s, T(0),
		 s,  c, T(0) );
    multiply( m );
  }

  //==============================================================================================
  //==============================================================================================






  //==============================================================================================
  //==============================================================================================

  bool Matrix3x2::invert()
  {
    //translation = -(translation * invert(rotation))
    //rotation = invert(rotation)

    Matrix2& rotation = (Matrix2&)(*this);  //works because upperleft corner of Matrix3x2 is Matrix2x2
    if( !rotation.invert( ) )
      return false;

    Vector2& translation = (*this)[2];
    translation.multiply( rotation, RIGHT_MULTIPLY );
    translation.negate( );
    return true;
  }

  //==============================================================================================
  //==============================================================================================

  const Matrix3x2 operator*( const Matrix3x2& m1, const Matrix3x2& m2 )
  {
    // dest.rotation = m1.rotation * m2.rotation
    // dest.translation = m1.translation * m2.rotation + m2.translation

    Matrix3x2 t( m1 );
    Matrix2& rotation1 = (Matrix2&)t;                               //works because upperleft corner of Matrix3x2 is Matrix2x2
    Vector2& translation1 = t[2];
    const Matrix2& rotation2 = (const Matrix2&)m2;  //works because upperleft corner of Matrix3x2 is Matrix2x2
    const Vector2& translation2 = m2[2];

    rotation1.multiply( rotation2, RIGHT_MULTIPLY );
    translation1.multiply( rotation2, RIGHT_MULTIPLY );
    translation1 += translation2;
    return t;
  }

  //==============================================================================================
  //==============================================================================================

  void Matrix3x2::rotate( T radians )
  {
    T c = COS( radians );
    T s = SIN( radians );
    //always a right multiply
    Matrix3x2 m(   c,   s,
		   -s,   c,
		   T(0), T(0) );
    multiply( m );
  }

  //==============================================================================================
  //==============================================================================================






  //==============================================================================================
  //==============================================================================================

  bool Matrix3x4::invert()
  {
    //translation = -(invert(rotation) * translation)
    //rotation = invert(rotation)

    Matrix3 rotation( getRotation() );
    if( !rotation.invert( ) )
      return false;

    Vector3 translation( matrix[0][3], matrix[1][3], matrix[2][3] );
    translation.multiply( rotation, LEFT_MULTIPLY );

    set( rotation[0][0], rotation[0][1], rotation[0][2], -translation[0],
	 rotation[1][0], rotation[1][1], rotation[1][2], -translation[1],
	 rotation[2][0], rotation[2][1], rotation[2][2], -translation[2] );
    return true;
  }

  //==============================================================================================
  //==============================================================================================

  const Matrix3x4 operator*( const Matrix3x4& m1, const Matrix3x4& m2 )
  {
    // dest.rotation = m1.rotation * m2.rotation
    // dest.translation = m1.rotation * m2.translation + m1.translation

    Matrix3 rotation1( m1.getRotation() ), rotation2( m2.getRotation() );
    Vector3 translation1( m1[0][3], m1[1][3], m1[2][3] ), translation2( m2[0][3], m2[1][3], m2[2][3] );

    rotation2.multiply( rotation1, LEFT_MULTIPLY );
    translation2.multiply( rotation1, LEFT_MULTIPLY );
    translation2 += translation1;

    return Matrix3x4( rotation2[0][0], rotation2[0][1], rotation2[0][2], translation2[0],
		      rotation2[1][0], rotation2[1][1], rotation2[1][2], translation2[1],
		      rotation2[2][0], rotation2[2][1], rotation2[2][2], translation2[2] );
  }

  //==============================================================================================
  //==============================================================================================

  void Matrix3x4::rotate( T radians, const Vector3& aboutThis )
  {
    //always a left multiply
    T c = COS( radians );
    T s = -SIN( radians );
    T t = T(1) - c;
    Vector3 vn( aboutThis );
    if( !vn.normalize( ) )
      return; //in case the about vector is zero, do nothing
    Matrix3x4 m( t*vn.x*vn.x+c,          t*vn.x*vn.y+s*vn.z, t*vn.x*vn.z-s*vn.y, T(0),
		 t*vn.x*vn.y-s*vn.z, t*vn.y*vn.y+c,      t*vn.y*vn.z+s*vn.x, T(0),
		 t*vn.x*vn.z+s*vn.y, t*vn.y*vn.z-s*vn.x, t*vn.z*vn.z+c,      T(0) );
    multiply( m );
  }

  //==============================================================================================
  //==============================================================================================






  //==============================================================================================
  //==============================================================================================

  bool Matrix4x3::invert()
  {
    //translation = -(translation * invert(rotation))
    //rotation = invert(rotation)

    Matrix3& rotation = (Matrix3&)(*this);  //works because upperleft corner of Matrix4x3 is Matrix3x3
    if( !rotation.invert( ) )
      return false;

    Vector3& translation = (*this)[3];
    translation.multiply( rotation, RIGHT_MULTIPLY );
    translation.negate( );
    return true;
  }

  //==============================================================================================
  //==============================================================================================

  const Matrix4x3 operator*( const Matrix4x3& m1, const Matrix4x3& m2 )
  {
    // dest.rotation = m1.rotation * m2.rotation
    // dest.translation = m1.translation * m2.rotation + m2.translation

    Matrix4x3 t( m1 );
    Matrix3& rotation1 = (Matrix3&)t;                               //works because upperleft corner of Matrix4x3 is Matrix3x3
    Vector3& translation1 = t[3];
    const Matrix3& rotation2 = (const Matrix3&)m2;  //works because upperleft corner of Matrix4x3 is Matrix3x3
    const Vector3& translation2 = m2[3];

    rotation1.multiply( rotation2, RIGHT_MULTIPLY );
    translation1.multiply( rotation2, RIGHT_MULTIPLY );
    translation1 += translation2;
    return t;
  }

  //==============================================================================================
  //==============================================================================================

  void Matrix4x3::rotate( T radians, const Vector3& aboutThis )
  {
    //always a right multiply
    T c = COS( radians );
    T s = SIN( radians );
    T t = T(1) - c;
    Vector3 vn( aboutThis );
    if( !vn.normalize( ) )
      return; //in case the about vector is zero, do nothing
    Matrix4x3 m( t*vn.x*vn.x+c,          t*vn.x*vn.y+s*vn.z, t*vn.x*vn.z-s*vn.y,
		 t*vn.x*vn.y-s*vn.z, t*vn.y*vn.y+c,      t*vn.y*vn.z+s*vn.x,
		 t*vn.x*vn.z+s*vn.y, t*vn.y*vn.z-s*vn.x, t*vn.z*vn.z+c,    
		 T(0),               T(0),               T(0)              );
    multiply( m );
  }

  //==============================================================================================
  //==============================================================================================





  //==============================================================================================
  //==============================================================================================

  const Vector2   Vector2Factory::randInSquare( T edgeLength )
  {
    T e = edgeLength * T(0.5);
    return Vector2( e*RAND(), e*RAND() );
  }

  //==============================================================================================
  //==============================================================================================

  const Vector2   Vector2Factory::randInCircle( T radius )
  {
    //distribution is uniform when random vectors inside a unit circle are selected
    Vector2 t;
    T d;
    do
      {
	t.set( RAND(), RAND() );
	d = dot( t, t );
      }
    while( d > T(1) );
    t *= radius;
    return t;
  }

  //==============================================================================================
  //==============================================================================================

  const Vector2   Vector2Factory::randOnCircle( T radius )
  {
    //distribution is uniform when random vectors inside a unit circle are normalized
    Vector2 t;
    T d;
    do
      {
	t.set( RAND(), RAND() );
	d = dot( t, t );
      }
    while( d == T(0) || d > T(1) );

    t.normalize();
    t *= radius;
    return t;
  }

  //==============================================================================================
  //==============================================================================================

  const Vector2 Vector2Factory::perpendicular( const Vector2& v )
  {
    return Vector2( -v.y, v.x );
  }

  //==============================================================================================
  //==============================================================================================




  //==============================================================================================
  //==============================================================================================

  const Vector3   Vector3Factory::randInCube( T edgeLength )
  {
    T e = edgeLength * T(0.5);
    return Vector3( e*RAND(), e*RAND(), e*RAND() );
  }

  //==============================================================================================
  //==============================================================================================

  const Vector3   Vector3Factory::randInSphere( T radius )
  {
    //distribution is uniform when random vectors inside a unit sphere are selected
    Vector3 t;
    T d;
    do
      {
	t.set( RAND(), RAND(), RAND() );
	d = dot( t, t );
      }
    while( d > T(1) );
    t *= radius;
    return t;
  }

  //==============================================================================================
  //==============================================================================================

  const Vector3   Vector3Factory::randOnSphere( T radius )
  {
    //distribution is uniform when random vectors inside a unit sphere are normalized
    Vector3 t;
    T d;
    do
      {
	t.set( RAND(), RAND(), RAND() );
	d = dot( t, t );
      }
    while( d == T(0) || d > T(1) );

    t.normalize();
    t *= radius;
    return t;
  }

  //==============================================================================================
  //==============================================================================================

  const Vector3 Vector3Factory::perpendicular( const Vector3& v )
  {
    Vector3 t;
    //select the shortest of projections of axes on v (the closest to perpendicular to v),
    // and project it to the plane defined by v
    if( ABS(v.x) < ABS(v.y) )
      {       //x < y
	if( ABS(v.x) < ABS(v.z) )
	  {       // x < y && x < z
	    t.x = T(1) - v.x * v.x;
	    t.y = -v.x * v.y;
	    t.z = -v.x * v.z;
	    return t;
	  }
      }
    else
      {       //y <= x
	if( ABS(v.y) < ABS(v.z) )
	  {       // y <= x && y < z
	    t.x = -v.y * v.x;
	    t.y = T(1) - v.y * v.y;
	    t.z = -v.y * v.z;
	    return t;
	  }
      }

    //z <= x && z <= y
    t.x = -v.z * v.x;
    t.y = -v.z * v.y;
    t.z = T(1) - v.z * v.z;
    return t;
  }

  //==============================================================================================
  //==============================================================================================




  //==============================================================================================
  //==============================================================================================

  const Vector4   Vector4Factory::randInCube( T edgeLength )
  {
    T e = edgeLength * T(0.5);
    return Vector4( e*RAND(), e*RAND(), e*RAND(), e*RAND() );
  }

  //==============================================================================================
  //==============================================================================================

  const Vector4   Vector4Factory::randInSphere( T radius )
  {
    //distribution is uniform when random vectors inside a unit sphere are selected
    Vector4 t;
    T d;
    do
      {
	t.set( RAND(), RAND(), RAND(), RAND() );
	d = dot( t, t );
      }
    while( d > T(1) );
    t *= radius;
    return t;
  }

  //==============================================================================================
  //==============================================================================================

  const Vector4   Vector4Factory::randOnSphere( T radius )
  {
    //distribution is uniform when random vectors inside a unit sphere are normalized
    Vector4 t;
    T d;
    do
      {
	t.set( RAND(), RAND(), RAND(), RAND() );
	d = dot( t, t );
      }
    while( d == T(0) || d > T(1) );

    t.normalize();
    t *= radius;
    return t;
  }

  //==============================================================================================
  //==============================================================================================



  /*
  //==============================================================================================
  //==============================================================================================

  Matrix3 MatrixFactory::rotationX( T radians, MultiplicationOrder h )
  {
  T c = T(cos( radians ));
  T s = T(sin( radians ));
  if( h == LEFT_MULTIPLY ) s = -s;        //transposes the matrix
  return Matrix3( T(1), T(0), T(0),
  T(0),    c,    s,
  T(0),   -s,    c );
  }

  //==============================================================================================
  //==============================================================================================

  Matrix3 MatrixFactory::rotationY( T radians, MultiplicationOrder h )
  {
  T c = T(cos( radians ));
  T s = T(sin( radians ));
  if( h == LEFT_MULTIPLY ) s = -s;        //transposes the matrix
  return Matrix3(    c, T(0),   -s,
  T(0), T(1), T(0),
  s, T(0),    c );
  }

  //==============================================================================================
  //==============================================================================================

  Matrix3 MatrixFactory::rotationZ( T radians, MultiplicationOrder h )
  {
  T c = T(cos( radians ));
  T s = T(sin( radians ));
  if( h == LEFT_MULTIPLY ) s = -s;        //transposes the matrix
  return Matrix3(    c,    s, T(0),
  -s,    c, T(0),
  T(0), T(0), T(1) );
  }

  //==============================================================================================
  //==============================================================================================

  Matrix3 MatrixFactory::rotationAbout( const Vector3& v, T radians, MultiplicationOrder h )
  {
  T c = T(cos( radians ));
  T s = T(sin( radians ));
  T t = T(1) - c;
  Vector3 vn( v );
  bool r = vn.normalize( );
  ASSERT( r );    //input vector is a zero vector
  Matrix3 m( t*vn.x*vn.x+c,          t*vn.x*vn.y+s*vn.z, t*vn.x*vn.z-s*vn.y,
  t*vn.x*vn.y-s*vn.z, t*vn.y*vn.y+c,      t*vn.y*vn.z+s*vn.x,
  t*vn.x*vn.z+s*vn.y, t*vn.y*vn.z-s*vn.x, t*vn.z*vn.z+c           );
  if( h == LEFT_MULTIPLY )
  m.transpose();
  return m;
  }
  */
  //==============================================================================================
  //==============================================================================================

  const Matrix3 MatrixFactory::target( const Vector3& pos, const Vector3& tgt, const Vector3& up, MultiplicationOrder o )
  {
    // Advanced Animation and Rendering Techniques, Watt & Watt (p. 9)
    Matrix3 m;
    // z is the normalized direction from pos to tgt
    m[2] = tgt;
    m[2] -= pos;
    if( !m[2].normalize() )
      m[2].set( T(0), T(0), T(1) );   //z was a zero vector
    // y (up) is pointing towards the up vector, but is perpendicular to z
    m[1] = up - project( up, m[2] );        //y = up - up's projection on z
    m[1].normalize( );
    // x is the cross product between y and z
    m[0] = cross(m[1], m[2]);
    //      if( righthanded )
    //              m[2].negate();
    if( o == LEFT_MULTIPLY )
      m.transpose();
    return m;
  }

  //==============================================================================================
  //==============================================================================================

  //The (left, bottom, znear) and (right, top, znear) parameters specify the points on the
  //near clipping plane that are mapped to the lower-left and upper-right corners of the window,
  //respectively, assuming that the eye is located at (0, 0, 0)
  //this maps the view frustum to a box with dimensions [-1,1]
  //this corresponds to OpenGL projection matrix format
  const Matrix4 MatrixFactory::frustum( T left, T right, T bottom, T top, T zNear, T zFar, MultiplicationOrder o )
  {
    T ooRL = T(1)/(right-left), ooTB = T(1)/(top-bottom), ooFN = T(1)/(zFar-zNear), ZNear2 = T(2)*zNear;
    Matrix4 frustum( ZNear2 * ooRL,          T(0),                   T(0),                  T(0),
		     T(0),                   ZNear2 * ooTB,          T(0),                  T(0),
		     -(right + left) * ooRL, -(top + bottom) * ooTB,  (zFar + zNear) * ooFN, T(1),
		     T(0),                   T(0),                  -ZNear2 * zFar * ooFN,  T(0) );
    //      if( righthanded )
    //              frustum[2].negate();
    if( o == LEFT_MULTIPLY )
      frustum.transpose();
    return frustum;
  }

  //==============================================================================================
  //==============================================================================================

  //this maps the view frustum to a box with dimensions [-1,1]
  //this corresponds to OpenGL projection matrix format
  //FOV is given in radians
  const Matrix4 MatrixFactory::frustum( T horizontalFOV, T verticalFOV, T zNear, T zFar, MultiplicationOrder o )
  {
    T w = T(1)/TAN(horizontalFOV * T(0.5)), h = T(1)/TAN(verticalFOV * T(0.5)), ooFN = T(1)/(zFar - zNear);
    Matrix4 frustum(   w, T(0),  T(0),                       T(0),
		       T(0),    h,  T(0),                       T(0),
		       T(0), T(0),  (zFar + zNear) * ooFN,      T(1),
		       T(0), T(0), -T(2) * zNear * zFar * ooFN, T(0) );
    //      if( righthanded )
    //              frustum[2].negate();
    if( o == LEFT_MULTIPLY )
      frustum.transpose();
    return frustum;
  }

  //==============================================================================================
  //==============================================================================================

  //The (left, bottom, near) and (right, top, near) parameters specify the points on the
  //near clipping plane that are mapped to the lower-left and upper-right corners of the window,
  //respectively, assuming that the eye is located at (0, 0, 0)
  //this maps the view frustum to a box with dimensions [-1,1]
  //this corresponds to OpenGL projection matrix format
  const Matrix4 MatrixFactory::ortho( T left, T right, T bottom, T top, T zNear, T zFar, MultiplicationOrder o )
  {
    T ooRL = T(1)/(right - left), ooTB = T(1)/(top - bottom), ooFN = T(1)/(zFar - zNear);
    Matrix4 ortho(  T(2) * ooRL,            T(0),                    T(0),                  T(0),
		    T(0),                   T(2) * ooTB,             T(0),                  T(0),
		    T(0),                   T(0),                    T(2) * ooFN,           T(0),
		    -(right + left) * ooRL, -(top + bottom) * ooTB,  -(zFar + zNear) * ooFN, T(1) );
    //      if( righthanded )
    //              ortho[2].negate();
    if( o == LEFT_MULTIPLY )
      ortho.transpose();
    return ortho;
  }

  //==============================================================================================
  //==============================================================================================

  //this maps the view frustum to a box with dimensions [-1,1] for x- and y-axis, and [0,1] for z-axis
  //this corresponds to Direct3D projection matrix format
  //FOV is given in radians
  const Matrix4 MatrixFactory::frustum01( T horizontalFOV, T verticalFOV, T zNear, T zFar, MultiplicationOrder o )
  {
    T w = T(1)/TAN(horizontalFOV * T(0.5)), h = T(1)/TAN(verticalFOV * T(0.5)), ooFN = T(1)/(zFar - zNear);
    Matrix4 frustum(   w, T(0),  T(0),                T(0),
		       T(0),    h,  T(0),                T(0),
		       T(0), T(0),  zFar * ooFN,         T(1),
		       T(0), T(0), -zFar * zNear * ooFN, T(0) );
    //      if( righthanded )
    //              frustum[2].negate();
    if( o == LEFT_MULTIPLY )
      frustum.transpose();
    return frustum;
  }

  //==============================================================================================
  //==============================================================================================

};      //namespace VECTORMATRIX

#include <gee_lwo.hpp>
#include <fstream>

//==============================================================
//==============================================================

bool fileExists( const char *name )
{
  assert( name );

  FILE* pF;
  if( (pF = fopen( name, "rb" )) )
    {
      fclose( pF );
      return true;
    }
  return false;
}

//==============================================================
//==============================================================

char *fileLoad( const char *name )
{
  assert( name );

  char* p;
  FILE* pF;
  if( (pF = fopen( name, "rb" )) )
    {
      fseek( pF, 0L, SEEK_END );
      unsigned int l = ftell( pF );
      fseek( pF, 0L, SEEK_SET );

      p = new char[l];

      if( (fread( p, sizeof(char), l, pF )) != l)
	VEE_error( "fileLoad - error while reading file '%s'",name );

      fclose( pF );
    }
  else
    VEE_error( "fileLoad - couldn't open file '%s'",name );

  return p;
}

//==============================================================
//==============================================================

void fileSave( const char *name, const void *s, unsigned int l )
{
  assert( name );
  assert( s );

  FILE *pF;
  if( (pF = fopen( name, "wb" ) ) )
    {
      fwrite( s, sizeof(char), l, pF );
      fclose( pF );
    }
  else
    VEE_error( "fileSave - couldn't open file '%s'",name );
}

//==============================================================
//==============================================================

unsigned int fileSize( const char *name )
{
  assert( name );

  unsigned int l;
  FILE* pF;
  if( (pF = fopen( name, "rb" )) )
    {
      fseek( pF, 0L, SEEK_END );
      l = ftell( pF );
      fclose( pF );
    }
  else
    VEE_error( "fileSize - couldn't open file '%s'",name );

  return l;
}

//==============================================================
//==============================================================

void fileFree( char *data )
{
  assert(data);
  delete[] data;
}

//==============================================================
//==============================================================
// #include <float.h>

//=======================================================================
//=======================================================================

#define CONST( a,b,c,d )        ((a << 24) | (b << 16) | (c << 8) | d)

//=======================================================================
//=======================================================================

void Mesh::loadLW5( )
{
  int chunklen,chunkname,subchunkname,subchunklen;
  int i,number,vtable[256];
  Vector3 p;
  float r;
  Vector4 color,spec;
  unsigned char *chunks, *chunkend,*subchunkend,*vertexPointer = NULL;

  chunks = m_readptr;

  m_boundingBoxMin.set(FLT_MAX,FLT_MAX,FLT_MAX);
  m_boundingBoxMax.set(-FLT_MAX,-FLT_MAX,-FLT_MAX);
  m_boundingRadius = 0.0f;

  // read file and find out surface,vertex and polygon counts
  m_nextchunkptr = chunks;
  while(m_nextchunkptr < m_fileend)
    {
      m_readptr = m_nextchunkptr;
      chunkname = readID4();
      chunklen = readU4();
      chunkend = m_readptr + chunklen;
      m_nextchunkptr = chunkend;
      if( chunklen & 1 ) m_nextchunkptr++;    //if size is odd, add 1 padding

      switch(chunkname)
	{
	case CONST('P','N','T','S'):
	  vertexPointer = m_readptr;
	  m_vertices.resize( chunklen / 12 );
	  for(i=0;i<m_vertices.size();i++)
	    {
	      p = readVEC12();
	      m_vertices[i].m_position = p;
	      m_vertices[i].m_normal.set( 0, 0, 0 );
	      m_vertices[i].m_triangleUsing = -1;
	      if( p.x > m_boundingBoxMax.x ) m_boundingBoxMax.x = p.x;
	      if( p.x < m_boundingBoxMin.x ) m_boundingBoxMin.x = p.x;
	      if( p.y > m_boundingBoxMax.y ) m_boundingBoxMax.y = p.y;
	      if( p.y < m_boundingBoxMin.y ) m_boundingBoxMin.y = p.y;
	      if( p.z > m_boundingBoxMax.z ) m_boundingBoxMax.z = p.z;
	      if( p.z < m_boundingBoxMin.z ) m_boundingBoxMin.z = p.z;

	      r = dot(p, p);
	      if(r > m_boundingRadius) m_boundingRadius = r;
	    }
	  break;
	case CONST('S','R','F','S'):
	  while(m_readptr < chunkend)
	    {
	      Surface s;
	      s.m_name = readS0();
	      s.m_polygons = 0;
	      m_surfaces.push_back( s );
	    }
	  break;
	default:
	  break;
	}
    }

  m_nextchunkptr = chunks;
  while(m_nextchunkptr < m_fileend)
    {
      m_readptr = m_nextchunkptr;
      chunkname = readID4();
      chunklen = readU4();
      chunkend = m_readptr + chunklen;
      m_nextchunkptr = chunkend;
      if( chunklen & 1 ) m_nextchunkptr++;    //if size is odd, add 1 padding

      switch(chunkname)
	{
	case CONST('P','O','L','S'):
	  while(m_readptr < chunkend)
	    {
	      int vert = readU2();
	      if(vert >= 3)
		{
		  for(i=0;i<vert;i++)
		    {
		      vtable[i] = readU2();
		    }
		}
	      else
		{
		  m_readptr += 2*vert;
		}

	      int surf = readI2();
	      if(vert >= 3)
		{
		  Triangle t;
		  t.m_surface = abs(surf) - 1;
		  assert( t.m_surface < m_surfaces.size() );
		  m_surfaces[t.m_surface].m_polygons += vert-2;
		  t.m_n[0] = -1;
		  t.m_n[1] = -1;
		  t.m_n[2] = -1;

		  Vector3 e1 = m_vertices[vtable[1]].m_position;
		  e1 -= m_vertices[vtable[0]].m_position;
		  Vector3 e2 = m_vertices[vtable[vert-1]].m_position;
		  e2 -= m_vertices[vtable[0]].m_position;
		  Vector3 n = cross(e1,e2);
		  n.normalize();

		  for(i=0;i<vert-2;i++)
		    {
		      t.m_v[0] = vtable[0];
		      t.m_v[1] = vtable[i+1];
		      t.m_v[2] = vtable[i+2];
		      t.m_normal = n;
                                                
		      m_vertices[t.m_v[0]].m_triangleUsing = (m_triangles.size()<<2)|0;
		      m_vertices[t.m_v[1]].m_triangleUsing = (m_triangles.size()<<2)|1;
		      m_vertices[t.m_v[2]].m_triangleUsing = (m_triangles.size()<<2)|2;

		      e1 = m_vertices[t.m_v[1]].m_position;
		      e1 -= m_vertices[t.m_v[0]].m_position;
		      e2 = m_vertices[t.m_v[2]].m_position;
		      e2 -= m_vertices[t.m_v[0]].m_position;
		      e1 = cross(e1,e2);
		      t.m_area = 0.5f * e1.length();
		      m_triangles.push_back( t );
		    }
		  for(i=0;i<vert;i++)
		    {
		      m_vertices[vtable[i]].m_normal += n;
		    }
		}

	      if(surf < 0)    //surface number, if <0 detail polygons follow
		{
		  for(i=0;i<readU2();i++) //number of detail polygons
		    {
		      m_readptr += 2 * readU2(); //number of vertices
		    }
		}
	    }
	  break;
	case CONST('S','U','R','F'):
	  for(number=0;number<m_surfaces.size();number++)
	    {
	      if(!strcmp((char *)m_readptr,m_surfaces[number].m_name.c_str()))
		goto foundName;
	    }
	  assert(0);              //surface not found
	  break;
	foundName:
	  readS0();

	  m_nextsubchunkptr = m_readptr;
	  while(m_nextsubchunkptr < chunkend)
	    {
	      m_readptr = m_nextsubchunkptr;
	      subchunkname = readID4();
	      subchunklen = readU2();
	      subchunkend = m_readptr + subchunklen;
	      m_nextsubchunkptr = subchunkend;
	      if( subchunklen & 1 ) m_nextsubchunkptr++;      //if size is odd, add 1 padding

	      switch(subchunkname)
		{
		case CONST('C','O','L','R'):
		  m_surfaces[number].m_color.x = readU1();
		  m_surfaces[number].m_color.y = readU1();
		  m_surfaces[number].m_color.z = readU1();
		  m_surfaces[number].m_color.w = 255;
		  break;
		}
	    }
	  break;
	default:
	  break;
	}
    }

  m_boundingRadius = float(sqrt( m_boundingRadius ));
  Vector3 up(0,1,0);
  for(i=0;i<m_vertices.size();i++)
    {
      m_vertices[i].m_normal.normalize();
      m_vertices[i].m_t1 = cross(up, m_vertices[i].m_normal);
      m_vertices[i].m_t2 = cross(m_vertices[i].m_normal, m_vertices[i].m_t1);
      if( (float)fabs(dot(up, m_vertices[i].m_normal)) > 0.999f ) m_vertices[i].m_reliable = false;
      else m_vertices[i].m_reliable = true;
    }
}

//=======================================================================
//=======================================================================

#undef CONST


// #include "Mesh.h"
// #include <float.h>

//=======================================================================
//=======================================================================

#define CONST( a,b,c,d )        ((a << 24) | (b << 16) | (c << 8) | d)

//=======================================================================
//=======================================================================

struct PolyToTri
{
  int                     m_firstTri;
  int                     m_triCount;
};

struct VMADStruct
{
  void operator=(const VMADStruct &v)
  {
    m_oldVertex = v.m_oldVertex;
    m_newVertex = v.m_newVertex;
    memcpy(m_value,v.m_value,sizeof(float)*16);
  }
  int                     m_oldVertex;
  int                     m_newVertex;
  float           m_value[16];
};

void Mesh::loadLW6( )
{
  int chunklen,chunkname,subchunkname,subchunklen;
  int i;
  Vector3 p;
  float r;
  Vector4 color,spec;
  unsigned char *chunks, *chunkend,*subchunkend;
  int PNTS = 0, POLS = 0, VMAP = 0, TAGS = 0, PTAG = 0, VMAD = 0, SURF = 0;
  int lastPnts = 0;
  std::vector<const char *> tags;
  std::vector<const char *> vmaps;
  std::vector<PolyToTri> polygons;
  std::vector<VMADStruct> vmads;

  m_boundingBoxMin.set(FLT_MAX,FLT_MAX,FLT_MAX);
  m_boundingBoxMax.set(-FLT_MAX,-FLT_MAX,-FLT_MAX);
  m_boundingRadius = 0.0f;

  chunks = m_readptr;

  m_nextchunkptr = chunks;
  while(m_nextchunkptr < m_fileend)
    {
      m_readptr = m_nextchunkptr;
      chunkname = readID4();
      chunklen = readU4();
      chunkend = m_readptr + chunklen;
      m_nextchunkptr = chunkend;
      if( chunklen & 1 ) m_nextchunkptr++;    //if size is odd, add 1 padding

      switch(chunkname)
	{
	case CONST('P','N','T','S'):
	  VEE_trace("PNTS %d vertices\n",chunklen/12);
	  //                      if( PNTS == 0 ) //consider only the first PNTS chunk
	  {
	    lastPnts = m_vertices.size();
	    m_vertices.resize( lastPnts + chunklen / 12 );
	    for(i=lastPnts;i<m_vertices.size();++i)
	      {
		p = readVEC12();
		m_vertices[i].m_position = p;
		m_vertices[i].m_normal.set(0,0,0);
		m_vertices[i].m_triangleUsing = -1;
		if( p.x > m_boundingBoxMax.x ) m_boundingBoxMax.x = p.x;
		if( p.x < m_boundingBoxMin.x ) m_boundingBoxMin.x = p.x;
		if( p.y > m_boundingBoxMax.y ) m_boundingBoxMax.y = p.y;
		if( p.y < m_boundingBoxMin.y ) m_boundingBoxMin.y = p.y;
		if( p.z > m_boundingBoxMax.z ) m_boundingBoxMax.z = p.z;
		if( p.z < m_boundingBoxMin.z ) m_boundingBoxMin.z = p.z;

		r = dot(p, p);
		if(r > m_boundingRadius) m_boundingRadius = r;
	      }
	  }
	  PNTS++;
	  break;
	case CONST('V','M','A','P'):
	  VEE_trace("VMAP");
	  if( PNTS == 0 ) VEE_error("invalid LWO2 file: VMAP chunk before PNTS");
	  //                      if( PNTS == 1 ) //consider only VMAPS associated to the first PNTS chunk
	  {
	    unsigned int type = readID4();  //PICK,WGHT,MNVW,TXUV,RGB,RGBA,MORF,SPOT
	    unsigned int dimension = readU2();
	    const char *name = readS0();
	    vmaps.push_back( name );        //store the name of VMAP in case a VMAD is associated with it
	    static float val[16];   //we assume max dimension to be 16
	    int read = 0;
	    while( m_readptr < chunkend )
	      {
		++read;
		int vi = readVX() + lastPnts;
		for(i=0;i<dimension;++i)
		  val[i] = readF4();
		switch( type )
		  {
		  case CONST('T','X','U','V'):
		    if( dimension == 2 )    //just in case..
		      m_vertices[vi].m_uv.set(val[0], 1.0f-val[1]);
		    break;
		  case CONST('R','G','B',' '):
		    m_vertices[vi].m_diffuse.set( val[0],val[1],val[2], 1 );
		    break;
		  case CONST('R','G','B','A'):
		    m_vertices[vi].m_diffuse.set( val[0],val[1],val[2],val[3] );
		    break;
		  }
	      }
	    VEE_trace(" %c%c%c%c %s %d",((type>>24)&0xff),((type>>16)&0xff),((type>>8)&0xff),(type&0xff),name,read);
	  }
	  VEE_trace("\n");
	  VMAP++;
	  break;
	case CONST('V','M','A','D'):
	  VEE_trace("VMAD");
	  if( PNTS == 0 || POLS == 0 ) VEE_error("invalid LWO2 file: VMAD chunk before PNTS or POLS");
	  //                      if( PNTS == 1 && POLS == 1 ) //VMADS associated to the first PNTS and POLS
	  {
	    //TODO test this
	    unsigned int type = readID4();  //PICK,WGHT,MNVW,TXUV,RGB,RGBA,MORF,SPOT
	    unsigned int dimension = readU2();
	    const char *name = readS0();    //VMAD can be associated with the VMAP with the same name
	    int vmap = -1;
	    int read = 0;
	    for(i=0;i<vmaps.size();++i)
	      {
		if( !strcmp(name,vmaps[i]) )
		  {
		    vmap = i;
		    break;
		  }
	      }
	    if( vmap != -1 )        //corresponding VMAP found
	      {
		//TODO is it correct to clear vmads here?
		vmads.clear();
		static float val[16];   //we assume max dimension to be 16
		while( m_readptr < chunkend )
		  {
		    ++read;
		    int vi = readVX() + lastPnts;
		    int pi = readVX();
		    for(i=0;i<dimension;++i)
		      val[i] = readF4();
		    //                                              VEE_trace(" %d/%f",vi,val[0]);

		    //create new vertex
		    if( type == CONST('T','X','U','V') ||
			type == CONST('R','G','B',' ') ||
			type == CONST('R','G','B','A') )
		      {
			//check whether there's already new vertex with the same attributes
			int nv = -1;
			for(i=0;i<vmads.size();++i)
			  {
			    if( vmads[i].m_oldVertex == vi && !memcmp(val,vmads[i].m_value,dimension*sizeof(float)) )
			      {
				nv = vmads[i].m_newVertex;
				break;
			      }
			  }
			if( nv == -1 )
			  {       //vertex doesn't exist yet
			    static Vertex ver;
			    ver = m_vertices[vi];
			    ver.m_triangleUsing = -1;
			    switch( type )
			      {
			      case CONST('T','X','U','V'):
				if( dimension == 2 )    //just in case..
				  ver.m_uv.set(val[0], 1.0f-val[1]);
				break;
			      case CONST('R','G','B',' '):
				ver.m_diffuse.set( val[0],val[1],val[2], 1 );
				break;
			      case CONST('R','G','B','A'):
				ver.m_diffuse.set( val[0],val[1],val[2],val[3] );
				break;
			      }

			    nv = m_vertices.size();
			    m_vertices.push_back( ver );

			    static VMADStruct vmad;
			    vmad.m_oldVertex = vi;
			    vmad.m_newVertex = nv;
			    memcpy( vmad.m_value, val, dimension*sizeof(float) );
			    vmads.push_back( vmad );
			  }

			//remap polygons using the old vertex
			int base = polygons[pi].m_firstTri;
			for(i=0;i<polygons[pi].m_triCount;++i)
			  {
			    for(int j=0;j<3;++j)
			      {
				if( m_triangles[base + i].m_v[j] == vi )
				  m_triangles[base + i].m_v[j] = nv;
			      }
			  }
		      }
		  }
	      }
	    else
	      {
		//TODO corresponding VMAP not found
	      }
	    //else: each polygon has discontinuous mapping
	    VEE_trace(" %c%c%c%c %s %d",((type>>24)&0xff),((type>>16)&0xff),((type>>8)&0xff),(type&0xff),name,read);
	  }
	  VEE_trace("\n");
	  VMAD++;
	  break;
	case CONST('P','O','L','S'):
	  VEE_trace("POLS");
	  if( PNTS == 0 ) VEE_error("invalid LWO2 file: POLS chunk before PNTS");
	  //                      if( PNTS == 1 ) //POLS associated only with the first PNTS
	  {
	    unsigned int type = readID4();  //FACE,CURV,PTCH,MBAL,BONE
	    int read = 0;
	    polygons.clear();
	    static int vtable[1023];        //max number of vertices is 1023
	    while( m_readptr < chunkend )
	      {
		++read;
		unsigned int vert = readU2() & 0x3f;    //6 msb are flags, we don't need them
		for(i=0;i<vert;++i)
		  {
		    vtable[i] = readVX() + lastPnts;
		  }
		if( type == CONST('F','A','C','E') )
		  {
		    PolyToTri poly;
		    poly.m_firstTri = m_triangles.size();
		    poly.m_triCount = MAX((int)vert - 2, 0);
		    polygons.push_back( poly );

		    if( vert >= 3 )
		      {
			Triangle t;
			t.m_surface = -1;
			t.m_n[0] = -1;
			t.m_n[1] = -1;
			t.m_n[2] = -1;

			//normal is cross product of the first and the last edges
			Vector3 e1 = m_vertices[vtable[1]].m_position;
			e1 -= m_vertices[vtable[0]].m_position;
			Vector3 e2 = m_vertices[vtable[vert-1]].m_position;
			e2 -= m_vertices[vtable[0]].m_position;
			Vector3 n = cross(e1, e2);
			n.normalize();

			for(i=0;i<vert-2;i++)
			  {
			    t.m_v[0] = vtable[0];
			    t.m_v[1] = vtable[i+1];
			    t.m_v[2] = vtable[i+2];
			    t.m_normal = n;
                                                                
			    e1 = m_vertices[t.m_v[1]].m_position;
			    e1 -= m_vertices[t.m_v[0]].m_position;
			    e2 = m_vertices[t.m_v[2]].m_position;
			    e2 -= m_vertices[t.m_v[0]].m_position;
			    e1 = cross(e1, e2);
			    t.m_area = 0.5f * e1.length();
			    m_triangles.push_back( t );
			  }
			for(i=0;i<vert;i++)
			  {
			    m_vertices[vtable[i]].m_normal += n;
			  }
		      }
		  }
	      }
	    VEE_trace(" %c%c%c%c %d",((type>>24)&0xff),((type>>16)&0xff),((type>>8)&0xff),(type&0xff),read);
	  }
	  VEE_trace("\n");
	  POLS++;
	  break;
	case CONST('T','A','G','S'):
	  VEE_trace("TAGS");
	  //                      if( TAGS == 0 ) //consider only the first TAGS chunk
	  {
	    while( m_readptr < chunkend )
	      {
		VEE_trace(" %s",m_readptr);
		tags.push_back( (const char *)m_readptr );
		readS0();
	      }
	  }
	  VEE_trace("\n");
	  TAGS++;
	  break;
	case CONST('P','T','A','G'):
	  VEE_trace("PTAG");
	  if( TAGS == 0 || POLS == 0 ) VEE_error("invalid LWO2 file: PTAG chunk before TAGS or POLS");
	  //                      if( TAGS == 1 && POLS == 1 ) //PTAGs associated with the first TAGS and POLS chunk
	  {
	    unsigned int type = readID4();  //SURF,PART,SMGP
	    int read = 0;
	    while( m_readptr < chunkend )
	      {
		++read;
		int polyindex = readVX();
		if( polyindex < 0 || polyindex > polygons.size() - 1 )
		  VEE_error("invalid LWO2 file: PTAG polygon index invalid (%d)",polyindex);
		int tagindex = readU2();
		if( tagindex < 0 || tagindex > tags.size() - 1 )
		  VEE_error("invalid LWO2 file: PTAG polygon tag invalid (%d)",tagindex);
		int base = polygons[polyindex].m_firstTri;
		for(i=0;i<polygons[polyindex].m_triCount;++i)
		  {
		    if( m_triangles[base + i].m_surface != -1 )
		      VEE_trace(" duplicate PTAG %d", polyindex);
		    m_triangles[base + i].m_surface = tagindex;
		  }
	      }
	    VEE_trace(" %c%c%c%c %d",((type>>24)&0xff),((type>>16)&0xff),((type>>8)&0xff),(type&0xff),read);
	  }
	  VEE_trace("\n");
	  PTAG++;
	  break;
	case CONST('S','U','R','F'):
	  VEE_trace("SURF %s",m_readptr);
	  {
	    int tagindex = -1;
	    for(i=0;i<tags.size();++i)
	      {
		if( !strcmp((const char*)m_readptr, tags[i]) )
		  {
		    tagindex = i;
		    break;
		  }
	      }
	    if( tagindex == -1 ) VEE_error("invalid LWO2 file: surface %s not found from the TAGS list",m_readptr);
	    if( m_surfaces.size() < tagindex + 1 ) m_surfaces.resize( tagindex + 1 );
	    m_surfaces[tagindex].m_name = (const char *)m_readptr;
	    readS0();       //read the name
	    readS0();       //read the source surface (we're not going to support that)

	    m_nextsubchunkptr = m_readptr;
	    while(m_nextsubchunkptr < chunkend)
	      {
		m_readptr = m_nextsubchunkptr;
		subchunkname = readID4();
		subchunklen = readU2();
		subchunkend = m_readptr + subchunklen;
		m_nextsubchunkptr = subchunkend;
		if( subchunklen & 1 ) m_nextsubchunkptr++;      //if size is odd, add 1 padding

		switch(subchunkname)
		  {
		  case CONST('C','O','L','R'):
		    m_surfaces[tagindex].m_color = Vector4(readCOL12(), 1);
		    break;
		  }
	      }
	  }
	  VEE_trace("\n");
	  SURF++;
	  break;
	default:
	  break;
	}
    }

  m_boundingRadius = float(sqrt( m_boundingRadius ));

  //calculate how many triangles are using each surface
  //set triangle using info
  for(i=0;i<m_triangles.size();++i)
    {
      if( m_triangles[i].m_surface >= 0 )
	++m_surfaces[m_triangles[i].m_surface].m_polygons;
      else
	VEE_error("invalid LWO2 file: not every polygon has a surface");

      m_vertices[m_triangles[i].m_v[0]].m_triangleUsing = (i<<2) | 0;
      m_vertices[m_triangles[i].m_v[1]].m_triangleUsing = (i<<2) | 1;
      m_vertices[m_triangles[i].m_v[2]].m_triangleUsing = (i<<2) | 2;
    }

  //normalize vertex normals and calculate initial tangent spaces
  Vector3 up(0,1,0);
  for(i=0;i<m_vertices.size();i++)
    {
      m_vertices[i].m_normal.normalize();
      m_vertices[i].m_t1 = cross(up, m_vertices[i].m_normal);
      m_vertices[i].m_t2 = cross(m_vertices[i].m_normal, m_vertices[i].m_t1);
      if( (float)fabs(dot(up, m_vertices[i].m_normal)) > 0.999f ) m_vertices[i].m_reliable = false;
      else m_vertices[i].m_reliable = true;
    }
}

//=======================================================================
//=======================================================================

void Mesh::saveLW6()
{
  //calculate the file size
  unsigned int filesize = 0;
  filesize += 12 * m_vertices.size();
  filesize += (2 + 12 + 6) * m_triangles.size();
  filesize += 512 * m_surfaces.size();
  filesize += 1024;       //other stuff (hope that this is enough!)

  m_fileptr = new unsigned char[filesize];
  m_writeptr = m_fileptr;
  assert( !((unsigned int)m_writeptr & 3) );      //memory address must be aligned

  unsigned char* bookmarkChunksize;
  unsigned int chunksize;

  //write file header
  writeID4( CONST('F','O','R','M') );
  unsigned char* bookmarkFilesize = m_writeptr;
  m_writeptr += 4;
  writeID4( CONST('L','W','O','2') );

  //make one layer
  writeID4( CONST('L','A','Y','R') );
  bookmarkChunksize = m_writeptr;
  m_writeptr += 4;
  writeU2( 0 );   //layer number 0
  writeU2( 0 );   //flags
  writeVEC12( Vector3(0,0,0) );   //pivot
  writeS0( "Layer0" );    //layer name
  writeU2( -1 );  //no parent
  chunksize = (unsigned int)(m_writeptr - bookmarkChunksize) - 4;
  SWAP( m_writeptr, bookmarkChunksize );
  writeU4( chunksize );
  SWAP( m_writeptr, bookmarkChunksize );
  if( (unsigned int)m_writeptr & 1 )
    *m_writeptr++ = 0;      //add padding if the chunk ends at odd address

  //write vertices
  writeID4( CONST('P','N','T','S') );
  bookmarkChunksize = m_writeptr;
  m_writeptr += 4;
  int i;
  for(i=0;i<m_vertices.size();i++)
    writeVEC12( m_vertices[i].m_position );
  chunksize = (unsigned int)(m_writeptr - bookmarkChunksize) - 4;
  SWAP( m_writeptr, bookmarkChunksize );
  writeU4( chunksize );
  SWAP( m_writeptr, bookmarkChunksize );

  //write polygons
  writeID4( CONST('P','O','L','S') );
  bookmarkChunksize = m_writeptr;
  m_writeptr += 4;
  writeID4( CONST('F','A','C','E') );
  for(i=0;i<m_triangles.size();i++)
    {
      writeU2( 3 );   //write number of indices per polygon
      writeVX( m_triangles[i].m_v[0] );
      writeVX( m_triangles[i].m_v[1] );
      writeVX( m_triangles[i].m_v[2] );
    }
  chunksize = (unsigned int)(m_writeptr - bookmarkChunksize) - 4;
  SWAP( m_writeptr, bookmarkChunksize );
  writeU4( chunksize );
  SWAP( m_writeptr, bookmarkChunksize );

  //write tag list
  writeID4( CONST('T','A','G','S') );
  bookmarkChunksize = m_writeptr;
  m_writeptr += 4;
  if( m_surfaces.size() )
    {
      for(i=0;i<m_surfaces.size();i++)
	writeS0( m_surfaces[i].m_name.c_str() );
    }
  else
    writeS0( "Default" );
  chunksize = (unsigned int)(m_writeptr - bookmarkChunksize) - 4;
  SWAP( m_writeptr, bookmarkChunksize );
  writeU4( chunksize );
  SWAP( m_writeptr, bookmarkChunksize );

  //write polygon tags
  writeID4( CONST('P','T','A','G') );
  bookmarkChunksize = m_writeptr;
  m_writeptr += 4;
  writeID4( CONST('S','U','R','F') );
  for(i=0;i<m_triangles.size();i++)
    {
      writeVX( (unsigned int)i );
      if( m_surfaces.size() )
	{
	  int s = m_triangles[i].m_surface;
	  assert( s >= 0 && s < m_surfaces.size() );
	  writeU2( s );
	}
      else
	writeU2( 0 );
    }
  chunksize = (unsigned int)(m_writeptr - bookmarkChunksize) - 4;
  SWAP( m_writeptr, bookmarkChunksize );
  writeU4( chunksize );
  SWAP( m_writeptr, bookmarkChunksize );

  //write surfaces
  if( m_surfaces.size() )
    {
      for(i=0;i<m_surfaces.size();i++)
	{
	  writeID4( CONST('S','U','R','F') );
	  bookmarkChunksize = m_writeptr;
	  m_writeptr += 4;
	  writeS0( m_surfaces[i].m_name.c_str() );
	  writeS0( "" );
	  writeID4( CONST('C','O','L','R') );
	  writeU2( 14 );
	  writeCOL12( Vector3( m_surfaces[i].m_color.x, m_surfaces[i].m_color.y, m_surfaces[i].m_color.z ) );
	  writeU2( 0 );
	  chunksize = (unsigned int)(m_writeptr - bookmarkChunksize) - 4;
	  SWAP( m_writeptr, bookmarkChunksize );
	  writeU4( chunksize );
	  SWAP( m_writeptr, bookmarkChunksize );
	}
    }
  else
    {
      writeID4( CONST('S','U','R','F') );
      bookmarkChunksize = m_writeptr;
      m_writeptr += 4;
      writeS0( "Default" );
      writeS0( "" );
      writeID4( CONST('C','O','L','R') );
      writeU2( 14 );
      writeCOL12( Vector3( 0.5f, 0.5f, 0.5f ) );
      writeU2( 0 );
      chunksize = (unsigned int)(m_writeptr - bookmarkChunksize) - 4;
      SWAP( m_writeptr, bookmarkChunksize );
      writeU4( chunksize );
      SWAP( m_writeptr, bookmarkChunksize );
    }

  //patch the filesize
  filesize = (unsigned int)(m_writeptr - bookmarkFilesize) - 4;
  SWAP( m_writeptr, bookmarkFilesize );
  writeU4( filesize );
  SWAP( m_writeptr, bookmarkFilesize );
}

//=======================================================================
//=======================================================================

unsigned int Mesh::readID4()
{
  return readU4();
}
int Mesh::readI1()
{
  signed char a = (signed char)readU1();
  return a;
}
int Mesh::readI2()
{
  signed short a = (signed short)readU2();
  return a;
}
int Mesh::readI4()
{
  signed int a = (signed int)readU4();
  return a;
}
unsigned char Mesh::readU1()
{
  unsigned char a = *m_readptr++;
  return a;
}
unsigned short Mesh::readU2()
{
  unsigned char a,b;
  a = *m_readptr++;
  b = *m_readptr++;
  return (unsigned short)((a<<8) + b);
}
unsigned int Mesh::readU4()
{
  unsigned char a,b,c,d;
  a = *m_readptr++;
  b = *m_readptr++;
  c = *m_readptr++;
  d = *m_readptr++;
  return (a<<24) + (b<<16) + (c<<8) + d;
}
float Mesh::readF4()
{
  unsigned int a = readU4();
  float *b = (float *)&a;
  return *b;
}
const char *Mesh::readS0()
{
  unsigned char *ptr = m_readptr;
  int l = strlen( (const char*)m_readptr ) + 1;   //string length + terminating 0
  if( l & 1 ) ++m_readptr;        //if odd, add 1 padding
  m_readptr += l;
  return (const char*)ptr;
}
unsigned int Mesh::readVX()
{
  unsigned char a = *m_readptr++;
  if( a == 0xff ) //if first is 0xff, 4 byte index is used, otherwise 2 bytes
    {
      unsigned char b = *m_readptr++;
      unsigned char c = *m_readptr++;
      unsigned char d = *m_readptr++;
      return (b<<16) + (c<<8) + d;
    }
  unsigned char b = *m_readptr++;
  return (a << 8 ) + b;
}
Vector3 Mesh::readCOL12()
{
  return readVEC12();
}
Vector3 Mesh::readVEC12()
{
  Vector3 a;
  a.x = readF4();
  a.y = readF4();
  a.z = readF4();
  return a;
}
//=======================================================================
//=======================================================================

void Mesh::writeID4( unsigned int i )
{
  writeU4( i );
}
void Mesh::writeI1( signed char i )
{
  writeU1( (unsigned char)i );
}
void Mesh::writeI2( signed short i )
{
  writeU2( (unsigned short)i );
}
void Mesh::writeI4( int i )
{
  writeU4( (unsigned int)i );
}
void Mesh::writeU1( unsigned char i )
{
  *m_writeptr++ = i;
}
void Mesh::writeU2( unsigned short i )
{
  *m_writeptr++ = (unsigned char)((i >> 8) & 0xff);
  *m_writeptr++ = (unsigned char)(i & 0xff);
}
void Mesh::writeU4( unsigned int i )
{
  *m_writeptr++ = (unsigned char)((i >> 24) & 0xff);
  *m_writeptr++ = (unsigned char)((i >> 16) & 0xff);
  *m_writeptr++ = (unsigned char)((i >> 8) & 0xff);
  *m_writeptr++ = (unsigned char)(i & 0xff);
}
void Mesh::writeF4( float f )
{
  unsigned int* a = (unsigned int*)&f;
  writeU4( *a );
}
void Mesh::writeS0( const char* s )
{
  assert( s );
  int l = strlen( s ) + 1;        //string length + terminating 0
  strcpy( (char*)m_writeptr, s );
  m_writeptr += l;
  if( l & 1 )
    *m_writeptr++ = 0;      //if odd, add 1 padding
}
void Mesh::writeVX( unsigned int i )
{
  // If the index value is less than 65,280 (0xFF00), then the index is written as an
  // unsigned two-byte integer. Otherwise the index is written as an unsigned four byte
  // integer with bits 24-31 set.
  if( i < 0xff00 )
    writeU2( (unsigned short)i );
  else
    writeU4( i | 0xff000000 );
}
void Mesh::writeCOL12( const Vector3& c )
{
  writeVEC12( c );
}
void Mesh::writeVEC12( const Vector3& v )
{
  writeF4( v.x );
  writeF4( v.y );
  writeF4( v.z );
}

//=======================================================================
//=======================================================================

#undef CONST


// #include "Mesh.h"
#include <deque>
// #include <float.h>

// #define USE_HL_SET

//=======================================================================
//=======================================================================

#define CONST( a,b,c,d )        ((a << 24) | (b << 16) | (c << 8) | d)

//=======================================================================
//=======================================================================

unsigned char *Mesh::m_fileptr = NULL;
unsigned char *Mesh::m_nextchunkptr = NULL;
unsigned char *Mesh::m_nextsubchunkptr = NULL;
unsigned char *Mesh::m_fileend = NULL;
unsigned char *Mesh::m_readptr = NULL;
unsigned char *Mesh::m_writeptr = NULL;

//=======================================================================
//=======================================================================

Mesh::Mesh()
{
  m_ambiguousNeighbours = 0;
  m_adjacencyCalculated = false;
  m_filename = "";
}

//=======================================================================
//=======================================================================

Mesh::Mesh( int vertices, int triangles )
{
  assert(vertices > 0);
  assert(triangles > 0);

  m_ambiguousNeighbours = 0;
  m_adjacencyCalculated = false;
  m_filename = "";
  m_vertices.resize(vertices);
  m_triangles.resize(triangles);
}

//=======================================================================
//=======================================================================

Mesh::Mesh( const char *fileName )
{
  load( fileName );
}

//=======================================================================
//=======================================================================

Mesh::Mesh( const Mesh &m )
{
  copy( m );
}

//=======================================================================
//=======================================================================

Mesh::~Mesh()
{
}

//=======================================================================
//=======================================================================

void Mesh::copy( const Mesh& m )
{
  m_vertices = m.m_vertices;
  m_triangles = m.m_triangles;
  m_edges = m.m_edges;
  m_surfaces = m.m_surfaces;
  m_filename = m.m_filename;

  m_boundingBoxMin = m.m_boundingBoxMin;
  m_boundingBoxMax = m.m_boundingBoxMax;
  m_boundingRadius = m.m_boundingRadius;
  m_ambiguousNeighbours = m.m_ambiguousNeighbours;
  m_adjacencyCalculated = m.m_adjacencyCalculated;
}

//=======================================================================
//=======================================================================

void Mesh::clear( int newVertices, int newTriangles )
{
  m_vertices.resize( newVertices );
  m_triangles.resize( newTriangles );
  m_edges.clear();
  //don't clear m_surfaces
  m_filename = "";

  m_boundingBoxMin.set(0,0,0);
  m_boundingBoxMax.set(0,0,0);
  m_boundingRadius = 0.0f;
  m_ambiguousNeighbours = 0;
  m_adjacencyCalculated = false;
}

//=======================================================================
//=======================================================================

static char *loadFile( const char *name )
{
  FILE* pF;
  char* p = NULL;

  assert( name );//, "Phile::load: Philename not specified" );

  pF = fopen( name, "rb" );
  if( pF )
    {
      fseek( pF, 0L, SEEK_END );
      unsigned int l = ftell( pF );
      fseek( pF, 0L, SEEK_SET );

      p = new char[l];

      if( (fread( p, sizeof(char), l, pF )) != l)
	VEE_error( "Phile::load - error while reading file '%s'",name );

      fclose( pF );
    }
  else
    VEE_error( "Phile::load - couldn't open file '%s'",name );
  return p;
}

//=======================================================================
//=======================================================================

void Mesh::load( const char* fileName )
{
  VEE_trace("loading %s\n",fileName);

  m_vertices.clear();
  m_triangles.clear();
  m_edges.clear();
  m_surfaces.clear();

  m_ambiguousNeighbours = 0;
  m_adjacencyCalculated = false;

  m_filename = fileName;
  m_fileptr = (unsigned char *)loadFile(fileName);

  m_readptr = m_fileptr;

  if(readID4() != CONST('F','O','R','M') )
    VEE_error("not a valid IFF file");

  m_fileend = m_fileptr + readU4();

  unsigned int i = readID4();
  if(i == CONST('L','W','O','B')) loadLW5( );
  else if(i == CONST('L','W','O','2')) loadLW6( );
  else VEE_error("not a LWO file");

  delete[] (char*)m_fileptr;

  VEE_trace("polys %d vertices %d surfaces %d\n",m_triangles.size(),m_vertices.size(),m_surfaces.size());
}

//=======================================================================
//=======================================================================

void Mesh::save( const char* fileName )
{
  saveLW6();
  //m_fileptr points to the beginning of the buffer, m_writeptr to the end

  FILE *pF;
  if( (pF = fopen( fileName, "wb" ) ) )
    {
      fwrite( m_fileptr, sizeof(char), (unsigned int)(m_writeptr - m_fileptr), pF );
      fclose( pF );
    }
  else
    VEE_error( "Mesh::save - couldn't open file '%s'",fileName );

  delete[] m_fileptr;
}

//=======================================================================
//=======================================================================

// The HL Set was removed (no need).

#include <set>

//=======================================================================
//=======================================================================

void Mesh::calculateAdjacency()
{
  if( m_adjacencyCalculated ) return;

  VEE_trace("calculating polygon adjacency\n");

#ifdef USE_HL_SET
  static Set<Edge> edgeset;
  edgeset.removeAll();
#else
  static std::set<Edge> edgeset;
  edgeset.clear();
#endif

  //clear triangleUsing (this is essential in case there are unused vertices!)
  int i;
  for(i=0;i<m_vertices.size();i++)
    m_vertices[i].m_triangleUsing = -1;

  //create a set of edges + calculate neighbourhood information
  //O(T log E) with stl, near O(T) with HL::Set (uses hash table)
  m_ambiguousNeighbours = 0;
  for(i=0;i<m_triangles.size();++i)
    {
      for(int j=0;j<3;j++)
	{
	  //set triangleUsing
	  m_vertices[m_triangles[i].m_v[j]].m_triangleUsing = (i<<2) | j;
        
	  m_triangles[i].m_n[j] = -1;

	  int k = nextMod3(j);
	  Edge edge;
	  edge.m_v[0] = MIN(m_triangles[i].m_v[j],m_triangles[i].m_v[k]);
	  edge.m_v[1] = MAX(m_triangles[i].m_v[j],m_triangles[i].m_v[k]);
	  edge.m_t[0] = i;
	  edge.m_t[1] = -1;
#ifdef USE_HL_SET
	  Edge* setedge = (Edge*)edgeset.get( edge );
	  if( setedge )
#else
	    std::pair<std::set<Edge>::iterator,bool> ret = edgeset.insert( edge );  //O(log E)
	  if( !ret.second )
#endif
	    {       //already exists, check if it already has 2 triangles
#ifdef USE_HL_SET
	      if( setedge->m_t[1] != -1 )
		++m_ambiguousNeighbours;
#else
	      if( (*ret.first).m_t[1] != -1 )
		++m_ambiguousNeighbours;
#endif
	      else
		{       //neighbour found, update information
#ifdef USE_HL_SET
		  setedge->m_t[1] = i;
		  int n = setedge->m_t[0];
#else
		  // (*ret.first).m_t[1] = i; RRR
		  int n = (*ret.first).m_t[0];
#endif
		  int ne = -1;
		  if( edge.m_v[0] == m_triangles[n].m_v[0] )
		    {
		      if( edge.m_v[1] == m_triangles[n].m_v[1] ) ne = 0;
		      else if( edge.m_v[1] == m_triangles[n].m_v[2] ) ne = 2;
		    }
		  else if( edge.m_v[0] == m_triangles[n].m_v[1] )
		    {
		      if( edge.m_v[1] == m_triangles[n].m_v[0] ) ne = 0;
		      else if( edge.m_v[1] == m_triangles[n].m_v[2] ) ne = 1;
		    }
		  else if( edge.m_v[0] == m_triangles[n].m_v[2] )
		    {
		      if( edge.m_v[1] == m_triangles[n].m_v[0] ) ne = 2;
		      else if( edge.m_v[1] == m_triangles[n].m_v[1] ) ne = 1;
		    }
		  m_triangles[i].m_n[j] = n;
		  m_triangles[i].m_nedge[j] = (char)ne;
		  m_triangles[n].m_n[ne] = i;
		  m_triangles[n].m_nedge[ne] = (char)j;
		}
	    }
	  else edgeset.insert( edge );
	}
    }
  if( m_ambiguousNeighbours )
    VEE_trace("edges shared by more than 2 triangles: %d\n",m_ambiguousNeighbours);

  //copy edges to vector
  //O( E )
#ifdef USE_HL_SET
  edgeset.getArray( m_edges );
#else
  i=0;
  int interior = 0,boundary = 0;
  m_edges.resize( edgeset.size() );
  std::set<Edge>::iterator it = edgeset.begin();
  while( it != edgeset.end() )
    {
      m_edges[i] = (*it);
      if( m_edges[i].m_t[1] == -1 ) ++boundary;
      else ++interior;
      ++it;
      ++i;
    }
  VEE_trace("edges %d interior edges %d boundary edges %d\n",m_edges.size(),interior,boundary);
#endif
  m_adjacencyCalculated = true;
}

//=======================================================================
//=======================================================================

void Mesh::recalculateAdjacency()
{
  m_adjacencyCalculated = false;
  calculateAdjacency();
}

//=======================================================================
//=======================================================================

void Mesh::setParametrization()
{
  for(int t=0;t<m_triangles.size();t++)
    {
      m_triangles[t].m_uv[0].set(0,0);
      m_triangles[t].m_uv[1].set(1,0);
      m_triangles[t].m_uv[2].set(0,1);
      m_triangles[t].m_baseTriangle = t;
    }
}

//=======================================================================
//=======================================================================

void Mesh::calculateNormals()
{
  //recalculate triangle and vertex normals
  int i;
  for(i=0;i<m_vertices.size();i++)
    m_vertices[i].m_normal.set(0,0,0);
  float area;
  for(int t=0;t<m_triangles.size();t++)
    {
      Vector3 e1 = m_vertices[m_triangles[t].m_v[1]].m_position;
      e1 -= m_vertices[m_triangles[t].m_v[0]].m_position;
      Vector3 e2 = m_vertices[m_triangles[t].m_v[2]].m_position;
      e2 -= m_vertices[m_triangles[t].m_v[0]].m_position;
      Vector3 n = cross( e1, e2 );
      area = n.length();
      m_triangles[t].m_area = 0.5f * area;
      n *= 1.0f/area;
      m_triangles[t].m_normal = n;

      m_vertices[m_triangles[t].m_v[0]].m_normal += n;
      m_vertices[m_triangles[t].m_v[1]].m_normal += n;
      m_vertices[m_triangles[t].m_v[2]].m_normal += n;
    }
  for(i=0;i<m_vertices.size();i++)
    m_vertices[i].m_normal.normalize();
}

//=======================================================================
//=======================================================================

void Mesh::subdivide( int n, Mesh::DisplacementFunc* displacement )
{
  if( n == 0 ) return;
  assert(n > 0);

  calculateAdjacency();
  if( m_ambiguousNeighbours )
    VEE_error("can't subdivide: '%s' contains %d edges which are shared by more than 2 polygons",m_filename.c_str(),m_ambiguousNeighbours);

  //copy the original vertices to the input (displacement functions may need the original data)
  static std::vector<Vertex>              vertices1;
  static std::vector<Triangle>    triangles1;
  static std::vector<Edge>                edges1;
  static std::vector<Vertex>              vertices2;
  static std::vector<Triangle>    triangles2;
  static std::vector<Edge>                edges2;
  std::vector<Vertex>                             *vIn,*vOut;     //these must be pointers, otherwise SWAP copies the arrays
  std::vector<Triangle>                   *tIn,*tOut;
  std::vector<Edge>                               *eIn,*eOut;

  vertices1 = m_vertices;
  triangles1 = m_triangles;
  edges1 = m_edges;

  vIn = &vertices1;
  tIn = &triangles1;
  eIn = &edges1;
  vOut = &vertices2;
  tOut = &triangles2;
  eOut = &edges2;

  //t(n+1) = t(n) * 4
  //e(n+1) = e(n) * 2 + t(n) * 3
  //v(n+1) = v(n) + e(n)

  //t(n) = t(0) * 2^2n   , n>=0
  //e(n) = e(0) * 2^n + t(0) * 3 * 2^(n-1) * (2^n - 1)   , n>=0
  //v(n) = v(0) + e(0) * (2^n - 1) + t(0) * 3 * (2^2n / 6 - 2^(n-1) + 1)  , n > 0

  //calculate the number of resulting vertices, triangles and edges
  int ct = m_triangles.size(),cv = m_vertices.size(),ce = m_edges.size();
  cv += ce * ((1<<n)-1);
  if(n > 1) cv += ct * 3 * ((1<<(2*n)) / 6 + 1 - (1<<(n-1)));
  ce *= (1<<n);
  ce += ct * 3 * (1<<(n-1)) * ((1<<n) - 1);
  ct *= (1<<(2*n));
  //      VEE_trace("subdivision reserve      v:%d t:%d e:%d\n",cv,ct,ce);

  //reserve enough room for all arrays (TODO other array (n-1), other n)
  vertices1.reserve( cv );
  vertices2.reserve( cv );
  triangles1.reserve( ct );
  triangles2.reserve( ct );
  edges1.reserve( ce );
  edges2.reserve( ce );

  VEE_trace("subdivision before       v:%d t:%d e:%d\n",m_vertices.size(),m_triangles.size(),m_edges.size());

  for(int step=0;step<n;step++)
    {
      if( displacement )
	displacement( this, *vIn, *tIn, *eIn, step );

      vOut->resize( vIn->size() + eIn->size() );
      tOut->resize( tIn->size() * 4 );
      eOut->resize( eIn->size() * 2 + tIn->size() * 3 );
      int vo = 0, to = 0, eo = 0;

      //move even vertices
      for(int v=0;v<vIn->size();v++)
	{
	  if( (*vIn)[v].m_triangleUsing == -1 )
	    {       //not used
	      vo++;
	      continue;
	    }

	  int spoly = (*vIn)[v].m_triangleUsing >> 2,cpoly,npoly,lB = -1,rB = -1;
	  int sedge = (*vIn)[v].m_triangleUsing & 3,cedge,nedge,cv;
	  assert( sedge >= 0 && sedge < 3 );
	  assert( spoly >= 0 && spoly < tIn->size() );
	  int valence = 0;
	  bool boundary = false;

	  Vertex sum;
	  sum.null();

	  //loop right
	  npoly = spoly;
	  nedge = sedge;
	  do
	    {
	      cpoly = npoly;
	      cedge = nedge;

	      //loop code
	      valence++;
	      cv = (*tIn)[cpoly].m_v[nextMod3(cedge)];
	      sum.add( (*vIn)[cv] );

	      npoly = (*tIn)[cpoly].m_n[cedge];
	      if( npoly == -1 )
		{
		  boundary = true;
		  rB = cv;
		  break;
		}
	      nedge = nextMod3((*tIn)[cpoly].m_nedge[cedge]);
	    } while( npoly != spoly );
	  if( boundary )
	    {
	      npoly = spoly;
	      nedge = prevMod3(sedge);
	      //loop left
	      do
		{
		  cpoly = npoly;
		  cedge = nedge;

		  //loop code
		  valence++;
		  cv = (*tIn)[cpoly].m_v[cedge];
		  sum.add( (*vIn)[cv] );

		  npoly = (*tIn)[cpoly].m_n[cedge];
		  if( npoly == -1 )
		    {
		      lB = cv;
		      break;
		    }
		  nedge = prevMod3((*tIn)[cpoly].m_nedge[cedge]);
		} while( npoly != spoly );

	      (*vOut)[vo].null();
	      (*vOut)[vo].mad( (*vIn)[v], 0.75f );
	      (*vOut)[vo].mad( (*vIn)[lB], 0.125f );
	      (*vOut)[vo].mad( (*vIn)[rB], 0.125f );
	      (*vOut)[vo].normalize();
	      (*vOut)[vo].m_triangleUsing = ((spoly * 4 + sedge) << 2) | sedge;
	      vo++;
	    }
	  else    //no boundary
	    {
	      float beta;
	      if( valence == 3 )
		beta = 3.0f / 16.0f;
	      else
		beta = 3.0f / (float(valence) * 8.0f);

	      (*vOut)[vo].null();
	      (*vOut)[vo].mad( (*vIn)[v], (1.0f - float(valence) * beta) );
	      (*vOut)[vo].mad( sum, beta );
	      (*vOut)[vo].normalize();
	      (*vOut)[vo].m_triangleUsing = ((spoly * 4 + sedge) << 2) | sedge;
	      vo++;
	    }
	}
      assert( vo == vIn->size() );

      for(int t=0;t<tIn->size();t++)
	{
	  //calculate edge midpoints
	  int nv[3];
	  for(int e=0;e<3;e++)
	    {
	      int i1,i2,i3,a3,nbour,nedge;

	      i1 = (*tIn)[t].m_v[ e ];
	      i2 = (*tIn)[t].m_v[ nextMod3(e) ];
	      i3 = (*tIn)[t].m_v[ prevMod3(e) ];

	      nv[e] = vo;

	      nbour = (*tIn)[t].m_n[e];
	      if( nbour != -1 )
		{
		  nedge = (*tIn)[t].m_nedge[e];
		  if( nbour < t )
		    {
		      //edge is split already
		      nv[e] = (*tOut)[ nbour * 4 + nedge ].m_v[ nextMod3(nedge) ];
		    }
		  else
		    {
		      //new vertex
		      a3 = (*tIn)[ nbour ].m_v[ prevMod3(nedge) ];

		      (*vOut)[vo].null();
		      (*vOut)[vo].mad( (*vIn)[i1], 0.375f );
		      (*vOut)[vo].mad( (*vIn)[i2], 0.375f );
		      (*vOut)[vo].mad( (*vIn)[i3], 0.125f );
		      (*vOut)[vo].mad( (*vIn)[a3], 0.125f );
		      (*vOut)[vo].normalize();
		      (*vOut)[vo].m_triangleUsing = ((t * 4 + e) << 2) | nextMod3(e);
		      vo++;

		      (*eOut)[eo].m_t[0] = t * 4 + e;
		      (*eOut)[eo].m_t[1] = nbour * 4 + nextMod3(nedge);
		      (*eOut)[eo].m_v[0] = i1;
		      (*eOut)[eo].m_v[1] = nv[e];
		      eo++;

		      (*eOut)[eo].m_t[0] = t * 4 + nextMod3(e);
		      (*eOut)[eo].m_t[1] = nbour * 4 + nedge;
		      (*eOut)[eo].m_v[0] = nv[e];
		      (*eOut)[eo].m_v[1] = i2;
		      eo++;
		    }
		}
	      else //boundary
		{
		  (*vOut)[vo].null();
		  (*vOut)[vo].mad( (*vIn)[i1], 0.5f );
		  (*vOut)[vo].mad( (*vIn)[i2], 0.5f );
		  (*vOut)[vo].normalize();
		  (*vOut)[vo].m_triangleUsing = ((t * 4 + e) << 2) | nextMod3(e);
		  vo++;

		  (*eOut)[eo].m_t[0] = t * 4 + e;
		  (*eOut)[eo].m_t[1] = -1;
		  (*eOut)[eo].m_v[0] = i1;
		  (*eOut)[eo].m_v[1] = nv[e];
		  eo++;

		  (*eOut)[eo].m_t[0] = t * 4 + nextMod3(e);
		  (*eOut)[eo].m_t[1] = -1;
		  (*eOut)[eo].m_v[0] = nv[e];
		  (*eOut)[eo].m_v[1] = i2;
		  eo++;
		}
	    }

	  //make new triangles
	  int ct = to;

	  (*tOut)[to].m_v[0] = (*tIn)[t].m_v[0];
	  (*tOut)[to].m_v[1] = nv[0];
	  (*tOut)[to].m_v[2] = nv[2];
	  (*tOut)[to].m_n[0] = (*tIn)[t].m_n[0] != -1 ? (*tIn)[t].m_n[0] * 4 + nextMod3((*tIn)[t].m_nedge[0]) : -1;
	  (*tOut)[to].m_n[1] = ct + 3;
	  (*tOut)[to].m_n[2] = (*tIn)[t].m_n[2] != -1 ? (*tIn)[t].m_n[2] * 4 + (*tIn)[t].m_nedge[2]: -1;
	  (*tOut)[to].m_uv[0] = (*tIn)[t].m_uv[0];
	  (*tOut)[to].m_uv[1] = 0.5f * ((*tIn)[t].m_uv[1] + (*tIn)[t].m_uv[0]);
	  (*tOut)[to].m_uv[2] = 0.5f * ((*tIn)[t].m_uv[2] + (*tIn)[t].m_uv[0]);
	  (*tOut)[to].m_nedge[0] = (*tIn)[t].m_nedge[0];
	  (*tOut)[to].m_nedge[1] = 0;
	  (*tOut)[to].m_nedge[2] = (*tIn)[t].m_nedge[2];
	  (*tOut)[to].m_surface = (*tIn)[t].m_surface;
	  (*tOut)[to].m_baseTriangle = (*tIn)[t].m_baseTriangle;
	  to++;

	  (*tOut)[to].m_v[0] = nv[0];
	  (*tOut)[to].m_v[1] = (*tIn)[t].m_v[1];
	  (*tOut)[to].m_v[2] = nv[1];
	  (*tOut)[to].m_n[0] = (*tIn)[t].m_n[0] != -1 ? (*tIn)[t].m_n[0] * 4 + (*tIn)[t].m_nedge[0]: -1;
	  (*tOut)[to].m_n[1] = (*tIn)[t].m_n[1] != -1 ? (*tIn)[t].m_n[1] * 4 + nextMod3((*tIn)[t].m_nedge[1]) : -1;
	  (*tOut)[to].m_n[2] = ct + 3;
	  (*tOut)[to].m_uv[0] = 0.5f * ((*tIn)[t].m_uv[0] + (*tIn)[t].m_uv[1]);
	  (*tOut)[to].m_uv[1] = (*tIn)[t].m_uv[1];
	  (*tOut)[to].m_uv[2] = 0.5f * ((*tIn)[t].m_uv[2] + (*tIn)[t].m_uv[1]);
	  (*tOut)[to].m_nedge[0] = (*tIn)[t].m_nedge[0];
	  (*tOut)[to].m_nedge[1] = (*tIn)[t].m_nedge[1];
	  (*tOut)[to].m_nedge[2] = 1;
	  (*tOut)[to].m_surface = (*tIn)[t].m_surface;
	  (*tOut)[to].m_baseTriangle = (*tIn)[t].m_baseTriangle;
	  to++;

	  (*tOut)[to].m_v[0] = nv[2];
	  (*tOut)[to].m_v[1] = nv[1];
	  (*tOut)[to].m_v[2] = (*tIn)[t].m_v[2];
	  (*tOut)[to].m_n[0] = ct + 3;
	  (*tOut)[to].m_n[1] = (*tIn)[t].m_n[1] != -1 ? (*tIn)[t].m_n[1] * 4 + (*tIn)[t].m_nedge[1]: -1;
	  (*tOut)[to].m_n[2] = (*tIn)[t].m_n[2] != -1 ? (*tIn)[t].m_n[2] * 4 + nextMod3((*tIn)[t].m_nedge[2]) : -1;
	  (*tOut)[to].m_uv[0] = 0.5f * ((*tIn)[t].m_uv[0] + (*tIn)[t].m_uv[2]);
	  (*tOut)[to].m_uv[1] = 0.5f * ((*tIn)[t].m_uv[1] + (*tIn)[t].m_uv[2]);
	  (*tOut)[to].m_uv[2] = (*tIn)[t].m_uv[2];
	  (*tOut)[to].m_nedge[0] = 2;
	  (*tOut)[to].m_nedge[1] = (*tIn)[t].m_nedge[1];
	  (*tOut)[to].m_nedge[2] = (*tIn)[t].m_nedge[2];
	  (*tOut)[to].m_surface = (*tIn)[t].m_surface;
	  (*tOut)[to].m_baseTriangle = (*tIn)[t].m_baseTriangle;
	  to++;

	  (*tOut)[to].m_v[0] = nv[2];
	  (*tOut)[to].m_v[1] = nv[0];
	  (*tOut)[to].m_v[2] = nv[1];
	  (*tOut)[to].m_n[0] = ct;
	  (*tOut)[to].m_n[1] = ct + 1;
	  (*tOut)[to].m_n[2] = ct + 2;
	  (*tOut)[to].m_uv[0] = 0.5f * ((*tIn)[t].m_uv[2] + (*tIn)[t].m_uv[0]);
	  (*tOut)[to].m_uv[1] = 0.5f * ((*tIn)[t].m_uv[0] + (*tIn)[t].m_uv[1]);
	  (*tOut)[to].m_uv[2] = 0.5f * ((*tIn)[t].m_uv[2] + (*tIn)[t].m_uv[1]);
	  (*tOut)[to].m_nedge[0] = 1;
	  (*tOut)[to].m_nedge[1] = 2;
	  (*tOut)[to].m_nedge[2] = 0;
	  (*tOut)[to].m_surface = (*tIn)[t].m_surface;
	  (*tOut)[to].m_baseTriangle = (*tIn)[t].m_baseTriangle;
	  to++;

	  (*eOut)[eo].m_t[0] = ct;
	  (*eOut)[eo].m_t[1] = ct + 3;
	  (*eOut)[eo].m_v[0] = nv[0];
	  (*eOut)[eo].m_v[1] = nv[2];
	  eo++;

	  (*eOut)[eo].m_t[0] = ct + 1;
	  (*eOut)[eo].m_t[1] = ct + 3;
	  (*eOut)[eo].m_v[0] = nv[1];
	  (*eOut)[eo].m_v[1] = nv[0];
	  eo++;

	  (*eOut)[eo].m_t[0] = ct + 2;
	  (*eOut)[eo].m_t[1] = ct + 3;
	  (*eOut)[eo].m_v[0] = nv[2];
	  (*eOut)[eo].m_v[1] = nv[1];
	  eo++;
	}
      assert( vo == vOut->size() );
      assert( to == tOut->size() );
      assert( eo == eOut->size() );

      SWAP(vIn,vOut);
      SWAP(tIn,tOut);
      SWAP(eIn,eOut);

      VEE_trace("subdivision after step %d v:%d t:%d e:%d\n",(step+1),vIn->size(),tIn->size(),eIn->size());
    }

  m_vertices = *vIn;
  m_triangles = *tIn;
  m_edges = *eIn;
}

//=======================================================================
//=======================================================================

const std::vector<Mesh::ventry> &Mesh::get1Neighbourhood( int ver, int &valence, bool &boundary, int &left, int &right ) const
{
  assert( ver >= 0 && ver < m_vertices.size() );
  static std::vector<ventry> vertexlist;
  valence = 0;
  boundary = false;
  vertexlist.clear();
  if( m_vertices[ver].m_triangleUsing == -1 ) return vertexlist;  //not used
  if( m_ambiguousNeighbours ) VEE_error("can't traverse neighbourhood: '%s' contains %d edges which are shared by more than 2 polygons",m_filename.c_str(),m_ambiguousNeighbours);

  int spoly = m_vertices[ver].m_triangleUsing >> 2,cpoly,npoly;
  int sedge = m_vertices[ver].m_triangleUsing & 3,cedge,nedge;
  assert( sedge >= 0 && sedge < 3 );
  assert( spoly >= 0 && spoly < m_triangles.size() );
  int i;
  ventry ve;

  //these loops count on that there are no neighbour ambiguities (edges with >2 polygons)
  // if this condition doesn't hold, these may loop infinitely
  //loop right
  npoly = spoly;
  nedge = sedge;
  i = 0;
  do
    {
      cpoly = npoly;
      cedge = nedge;

      valence++;
      ve.m_vertex = m_triangles[cpoly].m_v[nextMod3(cedge)];
      ve.m_i = i;
      vertexlist.push_back( ve );
      i++;

      npoly = m_triangles[cpoly].m_n[cedge];
      if( npoly == -1 )
	{
	  boundary = true;
	  right = ve.m_vertex;
	  break;
	}
      nedge = nextMod3(m_triangles[cpoly].m_nedge[cedge]);
    } while( npoly != spoly );
  if( boundary )
    {
      i = -1;
      npoly = spoly;
      nedge = prevMod3(sedge);
      //loop left
      do
	{
	  cpoly = npoly;
	  cedge = nedge;

	  valence++;
	  ve.m_vertex = m_triangles[cpoly].m_v[cedge];
	  ve.m_i = i;
	  vertexlist.push_back( ve );
	  i--;

	  npoly = m_triangles[cpoly].m_n[cedge];
	  if( npoly == -1 )
	    {
	      left = ve.m_vertex;
	      break;
	    }
	  nedge = prevMod3(m_triangles[cpoly].m_nedge[cedge]);
	} while( npoly != spoly );
    }
  return vertexlist;
}

//=======================================================================
//=======================================================================

const std::vector<int>& Mesh::get1NeighbourhoodPolygons( int ver ) const
{
  assert( ver >= 0 && ver < m_vertices.size() );
  static std::vector<int> trianglelist;
  trianglelist.clear();
  if( m_vertices[ver].m_triangleUsing == -1 ) return trianglelist;        //not used
  if( m_ambiguousNeighbours ) VEE_error("can't traverse neighbourhood: '%s' contains %d edges which are shared by more than 2 polygons",m_filename.c_str(),m_ambiguousNeighbours);

  int spoly = m_vertices[ver].m_triangleUsing >> 2,cpoly,npoly;
  int sedge = m_vertices[ver].m_triangleUsing & 3,cedge,nedge;
  assert( sedge >= 0 && sedge < 3 );
  assert( spoly >= 0 && spoly < m_triangles.size() );

  //these loops count on that there are no neighbour ambiguities (edges with >2 polygons)
  // if this condition doesn't hold, these may loop infinitely
  //loop right
  npoly = spoly;
  nedge = sedge;
  bool boundary = false;
  do
    {
      cpoly = npoly;
      cedge = nedge;

      trianglelist.push_back( cpoly );

      npoly = m_triangles[cpoly].m_n[cedge];
      if( npoly == -1 )
	{
	  boundary = true;
	  //                      right = ve.m_vertex;
	  break;
	}
      nedge = nextMod3(m_triangles[cpoly].m_nedge[cedge]);
    } while( npoly != spoly );
  if( boundary )
    {
      npoly = spoly;
      nedge = prevMod3(sedge);
      //loop left
      do
	{
	  cpoly = npoly;
	  cedge = nedge;

	  trianglelist.push_back( cpoly );

	  npoly = m_triangles[cpoly].m_n[cedge];
	  if( npoly == -1 )
	    {
	      //                              left = ve.m_vertex;
	      break;
	    }
	  nedge = prevMod3(m_triangles[cpoly].m_nedge[cedge]);
	} while( npoly != spoly );
    }
  return trianglelist;
}

//=======================================================================
//=======================================================================

void Mesh::calculateCurvature( )
{
  calculateAdjacency();

  if( m_ambiguousNeighbours )
    VEE_error("can't calculate curvature: '%s' contains %d edges which are shared by more than 2 polygons",m_filename.c_str(),m_ambiguousNeighbours);
  VEE_trace("calculating curvature\n");

  //max curvature must be bigger than this
  float planarThreshold = 0.01f;
  //max curvature must be this many times bigger than min curvature
  float ratioThreshold = 2.0f;

  Vector3 a0,a11,a12,a20,a21,a22, Fu,Fv,Fuu,Fvv,Fuv, pi, n;
  float E,F,G,L,M,N,e1,e2,K,H,t;
  int valence, i, left, right;
  bool boundary;
  int reliable = 0;
  for(int ver=0;ver<m_vertices.size();ver++)
    {
      if( m_vertices[ver].m_triangleUsing == -1 ) continue;   //not used
      m_vertices[ver].m_reliable = true;

      const std::vector<ventry> &vertexlist = get1Neighbourhood( ver, valence, boundary, left, right );
      a11.set(0,0,0);
      a12.set(0,0,0);
      a20.set(0,0,0);
      a21.set(0,0,0);
      a22.set(0,0,0);
      assert(valence > 0);
      float ang,oov = 1.0f / (float)valence;
      for(i=0;i<vertexlist.size();i++)
	{
	  pi = m_vertices[vertexlist[i].m_vertex].m_position;
	  ang = 2.0f * PI * float(vertexlist[i].m_i) * oov;
	  a11 += pi * float(sin( ang ));
	  a12 += pi * float(cos( ang ));
	  a20 += pi;
	  a21 += pi * float(sin( 2.0f * ang ));
	  a22 += pi * float(cos( 2.0f * ang ));
	}
      a11 *= 2.0f * oov;
      a12 *= 2.0f * oov;
      a21 *= 2.0f * oov;
      a22 *= 2.0f * oov;
      a20 *= oov;
      a20 -= m_vertices[ver].m_position;

      //second order approximation of the surface
      //F(u,v) = a0 + a11 * v + a12 * u + a20 * (u^2 + v^2) + a21 * 2uv + a22 * (u^2 - v^2)
      //Fu = a12 + a20 * 2u + a21 * 2v + a22 * 2u
      //Fv = a11 + a20 * 2v + a21 * 2u - a22 * 2v
      //Fuu = 2 * a20 + 2 * a22
      //Fvv = 2 * a20 - 2 * a22
      //Fuv = Fvu = 2 * a21

      Fu = a12;       //at u,v = (0,0)
      Fv = a11;       //at u,v = (0,0)
      Fuu = 2.0f * a20 + 2.0f * a22;
      Fvv = 2.0f * a20 - 2.0f * a22;
      Fuv = 2.0f * a21;

      n = cross(Fv, Fu);
      t = dot(n, n);
      if( t == 0.0f ) m_vertices[ver].m_reliable = false;
      else n *= OOSQRT(t);    //normalize
      E = dot(Fu, Fu);
      F = dot(Fv, Fu);
      G = dot(Fv, Fv);
      L = dot(Fuu, n);
      M = dot(Fuv, n);
      N = dot(Fvv, n);

      //first fundamental form B = |E F|  second fundamental form D = |L M|
      //                           |F M|                              |M N|
      //principal curvatures k are the solutions of the eigenvalue problem (D - kB)x = 0 <=>
      // (L - kE)du + (M - kF)dv = 0
      // (M - kF)du + (N - kG)dv = 0
      //principal direction vk is given by vk = Fu * du + Fv * dv
      //principal curvature directions are perpendicular

      //The Gaussian Curvature at a point is the product of principal curvatures, K = k1 * k2
      //The Mean Curvature at a point is the average of principal curvatures, H = (k1 + k2) / 2
      //A point on a surface is called elliptic if K >0. The surface looks locally convex at an elliptic point. 
      //A point on a surface is called hyperbolic if K <0. The surface looks locally saddle shaped at a hyperbolic point. 
      //A point on the surface is called parabolic if K=0 and H!=0. At a parabolic point, one of the principal curvatures is zero. The points on a cylinder are examples of parabolic points. 
      //A point on the surface is called planar if K=0 and H=0. At a planar point, both principal curvatures are zero. 

      //solve
      t = E * G - F * F;
      if( t == 0.0f ) m_vertices[ver].m_reliable = false;
      else t = 1.0f / t;
      K = (L * N - M * M) * t;        //Gaussian curvature (det(D)/det(B))
      H = 0.5f * (G * L + E * N - 2.0f * F * M) * t;  //mean curvature
      t = H * H - K;
      if( t < 0.0f ) m_vertices[ver].m_reliable = false;
      else t = SQRT( t );
      e1 = H + t;
      e2 = H - t;
      float maxCurvature,minCurvature;
      if( fabs(e1) > fabs(e2) )
	{
	  maxCurvature = e1;
	  minCurvature = e2;
	  m_vertices[ver].m_t1 = Fu * (M - e1 * F) - Fv * (L - e1 * E);
	}
      else
	{
	  maxCurvature = e2;
	  minCurvature = e1;
	  m_vertices[ver].m_t1 = Fu * (N - e2 * G) - Fv * (M - e2 * F);
	}
      m_vertices[ver].m_t1.normalize();
      m_vertices[ver].m_normal = n;

      //consider only approx. parabolic curvature as reliable
      //(on other kinds of points directions are not defined)
      if( fabs(maxCurvature) < planarThreshold ) m_vertices[ver].m_reliable = false;
      if( fabs(maxCurvature) < ratioThreshold * fabs(minCurvature) ) m_vertices[ver].m_reliable = false;


      /*              //TODO negate when necessary (all similar directions should point the same way)
		      Vector3 tt1;
		      if( fabs(n.y) > 0.1f ) tt1 = n * Vector3(0,1,0);
		      else if( fabs(n.z) > 0.1f ) tt1 = n * Vector3(0,0,1);
		      else if( fabs(n.x) > 0.1f ) tt1 = n * Vector3(1,0,0);
		      Vector3 tt2 = tt1 * n;
		      float d1;
		      d1 = m_vertices[ver].m_t1 | tt1;
		      if( fabs(d1) < 0.714 ) d1 = m_vertices[ver].m_t1 | tt2;
		      if( d1 < 0.0f ) m_vertices[ver].m_t1.negate();
      */
      if( m_vertices[ver].m_t1.y < 0.0f )     //according to y-axis, produces singularities on top and bottom of object
	m_vertices[ver].m_t1.negate();

      m_vertices[ver].m_t2 = cross( n, m_vertices[ver].m_t1 );

      if( m_vertices[ver].m_reliable ) ++reliable;
    }
  VEE_trace("vertex curvature: reliable %d unreliable %d\n",reliable, m_vertices.size() - reliable);
}

//=======================================================================
//=======================================================================

void Mesh::interpolateTangents( float stddev )
{
  VEE_trace("interpolating tangents\n");

  assert( stddev > 0.0f );

  int valence, i, left, right, v, nv;
  float g,m;
  Vector3 t,tn;
  bool boundary;

  std::vector<int> dist;
  dist.resize( m_vertices.size() );

  m = 1.0f / ( -2.0f * SQR(stddev) );

  std::deque<int> q;
  //for each vertex interpolate tangent from reliable tangents
  //using gaussian radial basis functions (g = exp(-r^2/2stddev^2))
  //distance: each edge is 1 unit
  //O( V * (V + E) )
  int ver;
  for(ver=0;ver<m_vertices.size();ver++)
    {
      if( m_vertices[ver].m_triangleUsing == -1 ) continue;   //not used
      //              if( m_vertices[ver].m_reliable ) continue;      //reliable don't need to be handled

      //clear distances
      for(v=0;v<m_vertices.size();v++)
	{
	  dist[v] = -1;
	}

      t.set(0,0,0);

      //calculate distances to reliable vertices
      assert(q.empty());
      q.push_back( ver );
      dist[ver] = 0;
      while( !q.empty() )
	{
	  v = q.front();
	  q.pop_front();

	  if( m_vertices[v].m_reliable )
	    {       //reliable tangent found
	      g = (float)exp( SQR(float(dist[v])) * m );

	      //project tangent to this vertex's tangent plane + add it
	      float d = dot(m_vertices[v].m_t1, m_vertices[ver].m_normal);
	      tn = m_vertices[v].m_t1;
	      tn -= d * m_vertices[ver].m_normal;
	      tn.normalize();
	      t += g * tn;
	    }

	  //push 1-neighbourhood vertices not visited yet
	  const std::vector<ventry> &vertexlist = get1Neighbourhood( v, valence, boundary, left, right );
	  for(i=0;i<vertexlist.size();++i)
	    {
	      nv = vertexlist[i].m_vertex;
	      if( dist[nv] == -1 )
		{       //not visited yet
		  dist[nv] = dist[v] + 1;
		  q.push_back( nv );
		}
	    }
	}

      t.normalize();
      m_vertices[ver].m_t2 = t;       //store new tangent (so that old tangents can be used in calculations)
    }

  //calculate second tangent
  for(ver=0;ver<m_vertices.size();++ver)
    {
      m_vertices[ver].m_t1 = m_vertices[ver].m_t2;
      m_vertices[ver].m_t2 = cross( m_vertices[ver].m_normal, m_vertices[ver].m_t1 );
    }
}

//=======================================================================
//=======================================================================

struct dvVertex
{
  int             m_oldVertex;
  int             m_newVertex;
  int             m_surface;
};

void Mesh::singleSurfaceVertices()
{
  VEE_trace("singleSurfaceVertices ");
  int i;
  std::vector<dvVertex> remap;
  std::vector<int> vsurf;

  vsurf.resize(m_vertices.size());
  for(i=0;i<m_vertices.size();i++) vsurf[i] = -1;

  int newVertex = m_vertices.size(),vi,surf;
  for(i=0;i<m_triangles.size();i++)
    {
      surf = m_triangles[i].m_surface;
      for(int j=0;j<3;j++)
	{
	  vi = m_triangles[i].m_v[j];
	  if( vsurf[vi] == -1 || vsurf[vi] == surf )
	    {       //surface is the same, everything is fine
	      vsurf[vi] = surf;
	    }
	  else    //surface is different, we have to remap
	    {
	      int found = -1;
	      for(int k=0;k<remap.size();k++)
		{
		  if( remap[k].m_oldVertex == vi && remap[k].m_surface == surf )
		    {
		      found = k;
		      break;
		    }
		}
	      if( found >= 0 )
		{       //found already remapped vertex with correct surface
		  vi = remap[found].m_newVertex;
		}
	      else    //no remapped vertex exist yet, create it now
		{
		  //create new vertex
		  m_vertices.push_back( m_vertices[vi] );

		  static dvVertex v;
		  v.m_oldVertex = vi;
		  v.m_surface = surf;
		  v.m_newVertex = newVertex++;
		  remap.push_back( v );
		  vi = v.m_newVertex;
		}
	      //make the triangle to use the new vertex
	      m_triangles[i].m_v[j] = vi;
	    }
	}
    }

  //update vertices' triangle usage
  for(i=0;i<m_triangles.size();i++)
    {
      surf = m_triangles[i].m_surface;
      for(int j=0;j<3;j++)
	{
	  m_vertices[m_triangles[i].m_v[j]].m_triangleUsing = (i<<2) | j;
	}
    }

  VEE_trace("duplicated %d vertices\n",remap.size());
}

//=======================================================================
//=======================================================================

void Mesh::silhouette( std::vector<int> &lines, const Matrix4x3 &objectToWorldMatrix, const Vector3 &cameraInWorld )
{
  Matrix4x3 worldToObject( objectToWorldMatrix );
  worldToObject.invert();
  Vector3 c( cameraInWorld );
  c *= worldToObject;

  lines.clear();
  for(int i=0;i<getEdgeCount();++i)
    {
      const Edge &e = m_edges[i];
      const Vertex &v1 = m_vertices[ e.m_v[0] ];
      if( e.m_t[1] != -1 )
	{
	  Vector3 p( c );
	  p -= v1.m_position;
	  if( dot(p, m_triangles[e.m_t[0]].m_normal) * dot(p, m_triangles[e.m_t[1]].m_normal) <= 0.0f )
	    {
	      lines.push_back( e.m_v[0] );
	      lines.push_back( e.m_v[1] );
	    }
	}
      else    //boundary
	{
	  lines.push_back( e.m_v[0] );
	  lines.push_back( e.m_v[1] );
	}
    }
}

//=======================================================================
//=======================================================================

void Mesh::calculateBoundingBox()
{
  m_boundingBoxMin.set(FLT_MAX,FLT_MAX,FLT_MAX);
  m_boundingBoxMax.set(-FLT_MAX,-FLT_MAX,-FLT_MAX);
  m_boundingRadius = 0.0f;
  for(int i=0;i<m_vertices.size();i++)
    {
      Vector3 v = m_vertices[i].m_position;
      if( v.x > m_boundingBoxMax.x ) m_boundingBoxMax.x = v.x;
      if( v.y > m_boundingBoxMax.y ) m_boundingBoxMax.y = v.y;
      if( v.z > m_boundingBoxMax.z ) m_boundingBoxMax.z = v.z;
      if( v.x < m_boundingBoxMin.x ) m_boundingBoxMin.x = v.x;
      if( v.y < m_boundingBoxMin.y ) m_boundingBoxMin.y = v.y;
      if( v.z < m_boundingBoxMin.z ) m_boundingBoxMin.z = v.z;

      if( dot(v,v) > m_boundingRadius ) m_boundingRadius = dot(v,v);
    }
  m_boundingRadius = SQRT(m_boundingRadius);
}

//=======================================================================
//=======================================================================

void Mesh::duplicateVertices()
{
  std::vector<Vertex> newv;
  newv.resize( m_triangles.size() * 3);
  int v = 0;
  for(int i=0;i<m_triangles.size();i++)
    {
      Triangle& t = m_triangles[i];
      for(int j=0;j<3;j++)
	{
	  newv[v] = m_vertices[t.m_v[j]];
	  t.m_v[j] = v;
	  v++;
	}
    }
  m_vertices = newv;

  m_adjacencyCalculated = false;
}

//=======================================================================
//=======================================================================

#undef CONST
