/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#ifndef GEE_MATERIAL_HPP
#define GEE_MATERIAL_HPP

#include <vee_vector4.h>

#include <string>

class GEE_Io;

enum GEE_MaterialFlag {
    /// Use blending (use alpha channel)
  GEE_MAT_BLEND = 0x1,
  /// Use lighting
  GEE_MAT_LIGHTING = 0x2,
  /// Smooth shading
  GEE_MAT_SMOOTH = 0x4,
  /// Two sided object (do not cull back faces)
  GEE_MAT_TWO_SIDED = 0x8,
  /// Each vertex has a separate color
  GEE_MAT_VERTEX_COLORS = 0x10
};

class GEE_Material
{
public:

  GEE_Material(float shine = 0.5, 
	       const VEE_Vector4 & color = VEE_Vector4ONE, 
	       int flags = GEE_MAT_LIGHTING | GEE_MAT_SMOOTH | GEE_MAT_TWO_SIDED);
  ~GEE_Material();

  bool read(GEE_Io *);
  bool write(GEE_Io *);
  
  float shine() const { return m_shine; }
  void setShine(float s) { m_shine = s; }

  int flags() const { return m_flags; }
  void setFlags(int f) { m_flags = f; }

  /// Sets all color components at once.
  void setColor(const VEE_Vector4 & c) 
  { m_ambient = m_diffuse = m_specular = c; }

  const VEE_Vector4 & ambient() const { return m_ambient; }
  void setAmbient(const VEE_Vector4 & color)  { m_ambient = color; }

  const VEE_Vector4 & diffuse() const { return m_diffuse; }
  void setDiffuse(const VEE_Vector4 & color)  { m_diffuse = color; }

  const VEE_Vector4 & specular() const { return m_specular; }
  void setSpecular(const VEE_Vector4 & color) { m_specular = color; }
  
  float burn() const { return m_burn; }
  void setBurn(float b) { m_burn = b; }

  bool useVertexColors() const 
  { return (m_flags & GEE_MAT_VERTEX_COLORS) != 0; }
  void setUseVertexColors(bool useColors)
  {
    if(useColors) m_flags |= GEE_MAT_VERTEX_COLORS;
    else m_flags &= ~(GEE_MAT_VERTEX_COLORS);
  }

  void setTwoSided(bool twoSided) 
  {
    if(twoSided) m_flags |= GEE_MAT_TWO_SIDED;
    else m_flags &= ~(GEE_MAT_TWO_SIDED);
  }

  void enableBlend(bool blend) 
  {
    if(blend) m_flags |= GEE_MAT_BLEND;
    else m_flags &= ~(GEE_MAT_BLEND);
  }

  bool blend() const { return m_flags & GEE_MAT_BLEND; }

  const std::string & texture() const { return m_texture; }
  void setTexture(const std::string &t) { m_texture = t; }

  bool operator == (const GEE_Material & that);

protected:
  float       m_shine;
  /// Colors
  VEE_Vector4 m_ambient, m_diffuse, m_specular;

  float m_burn;

  int         m_flags;

  std::string m_texture;
};


#endif

