/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#ifndef GEE_PARAM_WRITER_H
#define GEE_PARAM_WRITER_H

#include <di_types.h>

#include <gee_config.h>

#include <vee_matrix3.h>
#include <vee_matrix4.h>
#include <vee_vector3.h>
#include <vee_vector4.h>

#include <map>

class GEE_Io;

/** A Utility class for reading and writing parameters to/from a
    stream. 

    This class uses a mighty brutal approach: The client gives
    pointers to the parameters that need IO
    management. GEE_ParamWriter computes offset values to those
    variables from "this"-pointer. 

    \b Important: The parameters must be relative to
    "this"-pointer. If they are not then the offset values will get
    broken whenever objects are copied.

    This class is best used by either inheriting it or using it as a
    member variable in some class.

    @author Tommi Ilmonen.
*/

class GEE_ParamWriter
{
public:
  GEE_ParamWriter();
  virtual ~GEE_ParamWriter();

  bool writeParameters(GEE_Io *);
  bool readParameters(GEE_Io *);

  void clear();
  
  enum Type {
    FLOAT,
    INT,
    ID,
    BOOL,
    MATRIX3,
    MATRIX4,
    MATRIX4D,
    VECTOR3,
    VECTOR4,
    STLSTRING    
  };

  void addMemberFloat(ushort id, const float *x) { addMember(FLOAT, id, x); }
  void addMemberInt(ushort id, const int *x) { addMember(INT, id, x); }
  void addMemberUInt(ushort id, const unsigned *x) { addMember(INT, id, x); }
  void addMemberId(ushort id, const GEE_id_t *x) { addMember(ID, id, x); }
  void addMemberBool(ushort id, const bool *x) { addMember(BOOL, id, x); }
  void addMemberVector3(ushort id, const VEE_Vector3 *x)
  { addMember(VECTOR3, id, x); }
  void addMemberVector4(ushort id, const VEE_Vector4 *x)
  { addMember(VECTOR4, id, x); }
  void addMemberMatrix3(ushort id, const VEE_Matrix3 *x)
  { addMember(MATRIX3, id, x); }
  void addMemberMatrix4(ushort id, const VEE_Matrix4 *x)
  { addMember(MATRIX4, id, x); }
  void addMemberMatrix4d(ushort id, const VEE_Matrix4d *x)
  { addMember(MATRIX4D, id, x); }

  void addMember(Type, ushort, const void *);

private:

  class Item
  {
  public:
    int  m_offset;
    Type m_type;
  };

  typedef std::map<ushort, Item> container;

  container m_params;
};

#endif
