/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_path_to_lines.h>

#include <gee_debug.h>
#include <gee_io.h>
#include <gee_macros.h>
#include <gee_path_to_mesh.h>
#include <gee_world.h>

#include <vee_error.h>

static bool m_debug = false;

GEE_PathToLines::GEE_PathToLines(GEE_id_t pathId, 
				 GEE_id_t lineId)
  : GEE_PathToX(pathId),
    m_lineId(lineId),
    m_lineWidth(1.0),
    m_napraRadius(1.0),
    m_useNapra(false)
{
  addIdMember(&m_lineId);
  m_debug = GEE_Debug::debug("GEE_PathToLines");

  m_writer.addMemberFloat(GEE_CONST2('W', 'I'), & m_lineWidth);
  m_writer.addMemberFloat(GEE_CONST2('R', 'A'), & m_napraRadius);
  m_writer.addMemberId   (GEE_CONST2('L', 'I'), & m_lineId);
  m_writer.addMemberBool (GEE_CONST2('U', 'N'), & m_useNapra);
}

GEE_PathToLines::~GEE_PathToLines()
{}

bool GEE_PathToLines::start(GEE_World *w, const GEE_TimeStamp &ts)
{
  GEE_FNAME = "GEE_PathToLines::start";

  if(m_debug) VEE_trace(fname);
  
  if(!GEE_PathToX::start(w, ts))
    return false;
  
  GEE_Lines3D * lines = new GEE_Lines3D();
  lines->setMaterial(m_material);
  lines->setWidth(m_lineWidth);

  lines->setId(m_lineId);
  lines->setLayer(m_layer);
  m_lines = lines;

  m_lineId = w->addGeometry2(m_lines);

  if(m_debug) VEE_trace("%s # EXIT OK", fname);

  return true;
}

bool GEE_PathToLines::stop(GEE_World *w, const GEE_TimeStamp &ts)
{
  if(m_debug) VEE_trace("GEE_PathToLines::stop");

  GEE_PathToX::stop(w, ts);

  return true;
}

/* static const ushort GEE_WidthId    = GEE_CONST2('W', 'I');
   static const ushort GEE_LineId     = GEE_CONST2('L', 'I');
   static const ushort GEE_EndId      = GEE_CONST2('E', 'N'); */

bool GEE_PathToLines::read(GEE_Io * io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != type())
    return false;
  
  readBasics(io);
  readPathId(io);

  int v = io->readInt32();

  if(v != 1)
    return false;

  m_writer.readParameters(io);
  /* ushort id = io->readInt16();

  while(id != GEE_EndId) {
    if(id == GEE_WidthId)
      m_lineWidth = io->readFloat32();
    else if(id == GEE_LineId)
      m_lineId = io->readInt64();
    else {
      VEE_error("GEE_PathToLines::read # Bad id at %d (%c%c)", 
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }
    id = io->readInt16();
    }*/

  bool ok = m_material.read(io);
  
  return ok;
}

bool GEE_PathToLines::write(GEE_Io * io)
{  
  io->writeNullTerminatedString(type());

  writeBasics(io);
  writePathId(io);

  io->writeInt32(1); // version

  /* io->writeInt16(GEE_LineId);
  io->writeInt64(m_lineId);

  io->writeInt16(GEE_WidthId);
  io->writeFloat32(m_lineWidth);
  
  io->writeInt16(GEE_EndId);
  */

  m_writer.writeParameters(io);

  m_material.write(io);

  return true;
}

GEE_Operator * GEE_PathToLines::create()
{
  return new GEE_PathToLines;
}

bool GEE_PathToLines::handleNewVertex(GEE_Path3D::iterator &it)
{
  GEE_Lines3D * lines = 
    dynamic_cast<GEE_Lines3D *> (m_lines.ptr());
  
  if(!lines) {
    VEE_error("GEE_PathToLines::handleNewVertex # lines = NULL");
    return false;
  }

  const GEE_PathVertex3D &pv = GEE_Path3D::getVertex(it);

  VEE_Vector3 loc;

  if(m_useNapra) {
    VEE_Matrix4 m = GEE_PathToMesh::createMatrix(pv, m_napraRadius);
    loc = m * VEE_Vector3(0.0, 0.5, 0.0);
  }
  else
    loc = pv.m_location;

  lines->addVertex(GEE_LineVertex3D(loc, m_material.ambient()));

  return true;
}
