/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_path_to_mesh.h>

#include <gee_debug.h>
#include <gee_io.h>
#include <gee_macros.h>
#include <gee_path.h>
#include <gee_triangle_mesh.h>
#include <gee_world.h>

#include <vee_error.h>

static bool m_debug;

GEE_PathToMesh::GEE_PathToMesh(GEE_id_t pathId,
			       GEE_id_t meshId)
  : GEE_PathToX(pathId),
    m_meshId(meshId),
    m_scale(1.0),
    m_radius(1.0),
    m_threshold(0.01),
    m_align(GEE_ALIGN_MOTION),
    m_withNapra(false),
    m_checkNapra(true),
    m_specialFinger(-1)
{
  addIdMember(&m_meshId);
  m_debug = GEE_Debug::debug("GEE_PathToMesh");
}

GEE_PathToMesh::~GEE_PathToMesh()
{}

bool GEE_PathToMesh::start(GEE_World *w, const GEE_TimeStamp &ts)
{
  const char * fname = "GEE_PathToMesh::start";

  if(!GEE_PathToX::start(w, ts))
    return false;  
  
  // int i;

  GEE_TriangleMesh * mesh = new GEE_TriangleMesh;
  m_mesh = mesh;
  mesh->setMaterial(m_material);
  mesh->setId(m_meshId);
  mesh->setLayer(m_layer);
  mesh->attachEditor();

  m_meshId = w->addGeometry2(m_mesh);
  m_prevLoc.clear();
  m_newSegment = true;
  m_segmentCount = 0;

  if(m_checkNapra) {

    GEE_Path3D * path = dynamic_cast<GEE_Path3D *> (w->geometry(pathId()));
  
    if(path) {
      m_withNapra = path->activeNapra();
    }
    else
      m_withNapra = false;
  }

  if(m_debug)
    VEE_trace("%s # EXIT OK", fname);

  VEE_trace("%s # m_withNapra = %d", fname, (int) m_withNapra);

  return true;
}

bool GEE_PathToMesh::stop(GEE_World *, const GEE_TimeStamp &)
{
  if(m_debug)
    VEE_trace("GEE_PathToMesh::stop");

  GEE_Geometry * g = m_mesh.ptr();
  g->detachEditor();

  return true;
}

static const ushort GEE_AlignId    = GEE_CONST2('A', 'L');
static const ushort GEE_MeshId     = GEE_CONST2('M', 'I');
static const ushort GEE_ScaleId    = GEE_CONST2('S', 'C');
static const ushort GEE_RadiusId   = GEE_CONST2('R', 'A');
static const ushort GEE_ThresholdId  = GEE_CONST2('T', 'H');
static const ushort GEE_CheckNapraId = GEE_CONST2('C', 'N');
static const ushort GEE_WithNapraId  = GEE_CONST2('W', 'N');
static const ushort GEE_FingerId     = GEE_CONST2('F', 'I');
static const ushort GEE_EndId        = GEE_CONST2('E', 'N');

bool GEE_PathToMesh::read(GEE_Io * io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != type())
    return false;
  
  readBasics(io);
  readPathId(io);

  int v = io->readInt32();

  if(v != 1)
    return false;

  ushort id = io->readInt16();

  while(id != GEE_EndId) {
    if(id == GEE_MeshId)
      m_meshId = io->readInt64();
    else if(id == GEE_ScaleId)
      m_scale = io->readFloat32();
    else if(id == GEE_RadiusId)
      m_radius = io->readFloat32();
    else if(id == GEE_ThresholdId)
      m_threshold = io->readFloat32();
    else if(id == GEE_AlignId)
      m_align = io->readInt32();
    else if(id == GEE_WithNapraId)
      m_withNapra = io->readByte();
    else if(id == GEE_CheckNapraId)
      m_checkNapra = io->readByte();
    else if(id == GEE_FingerId) {
      m_specialFinger = (int) io->readByte();
      /* Work-around for MIPSpro compilers - thay do not make "char to
	 int" conversion properly... */
      if(m_specialFinger > 4)
	m_specialFinger = -1;
    }
    else {
      VEE_error("GEE_PathToMesh::read # Bad id at %d (%c%c)", 
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }
    id = io->readInt16();
  }

  bool ok = m_material.read(io);
  ok = ok && m_profile.read(io);

  VEE_trace("GEE_PathToMesh::read # %d %d %d", 
	    (int) m_withNapra, (int) m_checkNapra, m_specialFinger);
  
  return ok;
}

bool GEE_PathToMesh::write(GEE_Io * io)
{  
  io->writeNullTerminatedString(type());
  writeBasics(io);
  writePathId(io);

  io->writeInt32(1); // version
  
  io->writeInt16(GEE_MeshId);
  io->writeInt64(m_meshId);

  io->writeInt16(GEE_ScaleId);
  io->writeFloat32(m_scale);

  io->writeInt16(GEE_RadiusId);
  io->writeFloat32(m_radius);

  io->writeInt16(GEE_ThresholdId);
  io->writeFloat32(m_threshold);

  io->writeInt16(GEE_AlignId);
  io->writeInt32(m_align);

  io->writeInt16(GEE_WithNapraId);
  io->writeByte(m_withNapra);

  io->writeInt16(GEE_CheckNapraId);
  io->writeByte(m_checkNapra);

  io->writeInt16(GEE_FingerId);
  io->writeByte(m_specialFinger);

  io->writeInt16(GEE_EndId);

  m_material.write(io);
  m_profile.write(io);

  VEE_trace("GEE_PathToMesh::write # %d %d", 
	    (int) m_withNapra, (int) m_checkNapra);

  return true;
}

void GEE_PathToMesh::normalizeProfile(GEE_Profile * profile)
{
  // GEE_FNAME = "GEE_PathToMesh::normalizeProfile";

  profile->normalize(VEE_Vector3(1, 1, 0));
}


GEE_TriangleMesh * GEE_PathToMesh::triMesh()
{ 
  return dynamic_cast<GEE_TriangleMesh *> (m_mesh.ptr()); 
}

GEE_Operator * GEE_PathToMesh::create()
{
  return new GEE_PathToMesh;
}

VEE_Matrix4 GEE_PathToMesh::createMatrix(const GEE_PathVertex3D & tv,
					 float scale)
{
  VEE_Vector3 edge1 = tv.m_tips[1] - tv.m_tips[0];
  VEE_Vector3 edge2 = tv.m_tips[2] - tv.m_tips[0];

  VEE_Vector3 v = edge2; // Horizontal
  edge2 *= 0.5f; // In the following arithmetic we use half-edge value
  VEE_Vector3 u = edge1 - edge2; // vertical
  VEE_Vector3 w = cross(v, u); // z (just in case)
  w.normalize((u.length() + v.length()) * 0.5f);
  
  VEE_Matrix4 m, m2;
  m.identity();
  m.setColumn3(0, v * scale);
  m.setColumn3(1, u * scale);
  m.setColumn3(2, w * scale);
  m.setColumn3(3, (tv.m_tips[0] + (u + v) * 0.5f) * scale);
  m[3][0] = 0;
  m[3][1] = 0;
  
  /* m[1].copy3(u.data()); */
  
  m2.setRotation(tv.m_rotation);
  m2.setColumn3(3, tv.m_location);
  m2[3].make(0, 0, 0, 1);

  // m2[3][0] = m2[3][1] = m2[3][2] = 0.0;

  return (m2 * m);
}


bool  GEE_PathToMesh::handleNewVertex(GEE_Path3D::iterator &it)
{
  GEE_FNAME = "GEE_PathToMesh::handleNewVertex";

  GEE_PathVertex3D & tv = GEE_Path3D::getVertex(it);

  if(m_debug) {
    const VEE_Vector3 l = tv.m_location;
    VEE_trace("%s # New vertex %f %f %f", fname, l.x, l.y, l.z);
  }

  GEE_TriangleMesh * mesh = triMesh();

  if(!mesh) {
    VEE_error("%s # mesh = NULL", fname);
    return false;
  }

  if(m_newSegment)
    m_segmentCount++;

  int i, verticesBefore = mesh->vertexCounter();
  // int triangleCounter = mesh->triangleCount();

  if(!m_withNapra)
    setVerticesWithoutNapra(tv, mesh);
  else 
    setVerticesWithNapra(tv, mesh);

  // Now we proceed to build triangles out of the vertices:

  if(m_debug)
    VEE_trace("%s # Building triangles from %d %d (%d)", fname,
	      verticesBefore, (int) m_profile.size(), 
	      (int) processedPathVertices());

  const VEE_Vector3 & tloc = tv.m_location;

  if(processedPathVertices() == 0) {
    m_prevLoc = tloc;
    return true; // Not relevant in the first vertex -> no triangles to create
  }

  int segment = 0;
  int verticesAfter = mesh->vertexCounter();
  int profileVertices = m_profile.size();

  for(i = 0; i < profileVertices && m_newSegment; i++) {

    GEE_ProfileVertex2D & pv = m_profile.getVertex(i);

    if(!pv.m_continues)
      continue;

    uint lowerVertex = verticesAfter - 2 * profileVertices;
      
    uint vl1 = lowerVertex + i;
    uint vl2 = lowerVertex + (i + 1) % profileVertices;
    uint vh1 = vl1 + profileVertices;
    uint vh2 = vl2 + profileVertices;

    uint tricount = m_segmentCount * 2; // processedPathVertices() * 2;
    uint base = 2000000 * segment + tricount;

    GEE_TriangleIndices t1(vl1, vl2, vh1, true);
    GEE_TriangleIndices t2(vh1, vl2, vh2, false);

    mesh->addTriangle(t1, base);
    mesh->addTriangle(t2, base + 1);
    // Now connect the previous triangles to these new ones...
    
    if(m_segmentCount >= 2) {
      GEE_TriangleIndices * old = 
	mesh->getTriangle(tricount + 1 - profileVertices * 2);
      
      if(old)
	old->m_continues = true;
    }

    segment++;
  }

  if(m_newSegment) {
    for(i = verticesBefore; i < verticesAfter; i++) {
      if(!mesh->normalizeVertex(i, false))
	VEE_error("%s # Failed to normalize vertex %d", fname, (int) i);
    }
  }
  else {
    for(i = verticesAfter - profileVertices; i < verticesAfter; i++) {
      mesh->normalizeVertex(i, false);
    }
  }

  // Check if we should keep the current tube segment:
  if((m_dir.length() > m_threshold)) {
    m_newSegment = true;
    m_prevLoc = tloc;
  }
  else
    m_newSegment = false;
  
  return true;
}

void GEE_PathToMesh::setVerticesWithoutNapra(GEE_PathVertex3D & pv,
					     GEE_TriangleMesh * mesh)
{
  int verticesBefore = mesh->vertexCounter();

  VEE_Vector3 tloc = pv.m_location;
  VEE_Matrix3 trot = pv.m_rotation;

  VEE_Vector3 dir = tloc - m_prevLoc;
  m_dir = dir;

  VEE_Vector3 u, v;

  if(m_align == GEE_ALIGN_MOTION) {

    VEE_Vector3 localUp = trot * VEE_Vector3(0, 1, 0);
    
    u = cross(localUp, dir);
    v = cross(u, dir);
  }
  else {
    u = trot.column(0);
    v = trot.column(1);
  }

  u.normalize(m_scale);
  v.normalize(m_scale);
    
  int i, profileVertices = m_profile.size();

  GEE_PolygonVertex vertex3d;

  vertex3d.m_rgba = mesh->material().ambient();

  for(i = 0; i < profileVertices; i++) {

    GEE_ProfileVertex2D & pv = m_profile.getVertex(i);
    /* VEE_Vector3 pl(pv.m_location.x, pv.m_location.y, 0.0f);
       VEE_Vector3 pn(pv.m_normal.x, pv.m_normal.y, 0.0f);

       vertex3d.m_location = trot * pl + tloc;
       vertex3d.m_normal   = trot * pn;*/

    vertex3d.m_location = pv.m_location.x * u + pv.m_location.y * v + tloc;
    vertex3d.m_normal   = pv.m_normal.x * u + pv.m_normal.y * v;

    vertex3d.m_uv.make(vertex3d.m_location.x, vertex3d.m_location.y);

    /* {
       const VEE_Vector3 v = vertex3d.m_location;
       VEE_trace("GEE_PathToMesh::setVerticesWithoutNapra # %d %f %f %f", 
       i, v.x, v.y, v.z);
       }*/

    if(m_newSegment)
      mesh->addVertex(vertex3d);
    else {
      GEE_PolygonVertex * old = 
	mesh->getVertex(verticesBefore - profileVertices + i);
      old->m_location = vertex3d.m_location;
      old->m_normal   = vertex3d.m_normal;
      old->m_uv       = vertex3d.m_uv;
      old->m_rgba     = vertex3d.m_rgba;
      mesh->vertexWasMoved(vertex3d.m_location);
    }
  }
}

/** Thumb, index and middle finger vectors in this order. */

static VEE_Vector3 __tiplocs[3] = {
  VEE_Vector3(-0.5, -0.5, 0.0),
  VEE_Vector3(0.0, +0.5, 0.0),
  VEE_Vector3(+0.5, -0.5, 0.0)
};

void GEE_PathToMesh::setVerticesWithNapra(GEE_PathVertex3D & pv, 
					  GEE_TriangleMesh * mesh)
{
  // const char * fname = "GEE_PathToMesh::setVerticesWithNapra";

  // VEE_trace(fname);

  if(m_specialFinger >= 0) {

    VEE_Matrix4 m = createMatrix(pv, m_radius);

    VEE_Vector3 loc = m * __tiplocs[m_specialFinger];
    GEE_PathVertex3D nv(pv);
    nv.m_location = loc;
    
    setVerticesWithoutNapra(nv, mesh);
    return;
  }

  VEE_Matrix4 m = createMatrix(pv, m_scale);

  int verticesBefore = mesh->vertexCounter();

  VEE_Vector3 dir = pv.m_location - m_prevLoc;
  m_dir = dir;
  
  int i, profileVertices = m_profile.size();

  GEE_PolygonVertex vertex3d;

  vertex3d.m_rgba = mesh->material().ambient();
  vertex3d.m_normal.make(1, 0, 0);

  for(i = 0; i < profileVertices; i++) {

    GEE_ProfileVertex2D & pv = m_profile.getVertex(i);
    VEE_Vector2 & vloc = pv.m_location;
    vertex3d.m_location = m * VEE_Vector3(vloc.x, vloc.y, 0.0);

    // vertex3d.m_normal   = pv.m_normal.x * u + pv.m_normal.y * v;

    vertex3d.m_uv.make(vertex3d.m_location.x, vertex3d.m_location.y);

    if(m_newSegment)
      mesh->addVertex(vertex3d);
    else {
      GEE_PolygonVertex * old = 
	mesh->getVertex(verticesBefore - profileVertices + i);
      old->m_location = vertex3d.m_location;
      old->m_normal   = vertex3d.m_normal;
      old->m_uv       = vertex3d.m_uv;
      old->m_rgba     = vertex3d.m_rgba;
      mesh->vertexWasMoved(vertex3d.m_location);
    }

    // VEE_trace("%s # %d", fname, verticesBefore - profileVertices + i);
  }
}

