/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_path_to_particles.h>

#include <gee_io.h>
#include <gee_macros.h>
#include <gee_path_to_mesh.h>
#include <gee_world.h>

#include <vee_error.h>
#include <vee_math.h>
#include <vee_particle.h>
//#include <vee_quaternion.h>
#include <vee_time.h>

GEE_PathToParticles::GEE_PathToParticles(GEE_id_t pathId, 
					 GEE_id_t particleId,
					 GEE_id_t headPathId)
  : GEE_PathToX(pathId, headPathId),
    m_particleId(particleId),
    m_prevVertex(0,0,0),
    m_rate(10.0),
    m_radius(0.0),
    m_exaggerateRadius(0.0),
    m_radiusNorm(1.0),
    m_toolCenterWeight(0.0),
    m_randomize(0.0f),
    m_useNapra(false),
    m_modulateParticleSize(false),
    m_alignType(VEE_ALIGN_BILLBOARD),
    m_colorMix(0.5),
    m_version((pathId >= 0) ? 2 : 1)
{
  addIdMember(&m_particleId);  
  m_particle.m_particle = new VEE_Particle();

  // Fabricate a random seed:
  static uint counter = 0;
  VEE_Time tmp;
  tmp.getTimeOfDay();
  m_randomSeed = tmp.usecs() + counter++;

  m_writer.addMemberFloat(GEE_CONST2('H', 'Z'), & m_rate);
  m_writer.addMemberFloat(GEE_CONST2('E', 'R'), & m_exaggerateRadius);
  m_writer.addMemberFloat(GEE_CONST2('R', 'N'), & m_radiusNorm);
  m_writer.addMemberFloat(GEE_CONST2('C', 'W'), & m_toolCenterWeight);
  m_writer.addMemberId   (GEE_CONST2('P', 'I'), & m_particleId);
  m_writer.addMemberFloat(GEE_CONST2('R', 'A'), & m_radius);
  m_writer.addMemberUInt (GEE_CONST2('R', 'S'), & m_randomSeed);
  m_writer.addMemberBool (GEE_CONST2('U', 'N'), & m_useNapra);
  m_writer.addMemberBool (GEE_CONST2('M', 'S'), & m_modulateParticleSize);
  m_writer.addMemberFloat(GEE_CONST2('R', 'R'), & m_randomize);
  m_writer.addMemberFloat(GEE_CONST2('C', 'M'), & m_colorMix);
  m_writer.addMemberInt  (GEE_CONST2('A', 'L'), & m_alignType);
  m_writer.addMemberInt  (GEE_CONST2('V', 'E'), & m_version);
}

GEE_PathToParticles::~GEE_PathToParticles()
{}

bool GEE_PathToParticles::start(GEE_World *w, const GEE_TimeStamp &ts)
{
  // GEE_FNAME = "GEE_PathToParticles::start";

  if(!GEE_PathToX::start(w, ts))
    return false;  

  GEE_ParticleObject * particles = new GEE_ParticleObject();
  m_particles = particles;
  particles->setId(m_particleId);
  particles->setLayer(m_layer);

  m_particleId = w->addGeometry2(m_particles);
  m_remaining = 0.0;
  m_prevTime  = 0.0;
  m_prevVertex.clear();
  particles->attachEditor();

  /* m_rand.setSeed(m_randomSeed);

  VEE_trace("%s # %p %f", fname, m_particle.m_particle.ptr(), m_rand.rand11());
  */
  m_rand.setSeed(m_randomSeed);

  return true;
}


bool GEE_PathToParticles::stop(GEE_World *w, const GEE_TimeStamp &ts)
{
  // VEE_trace("GEE_PathToParticles::stop");

  GEE_PathToX::stop(w, ts);

  m_particles.ptr()->detachEditor();

  /*
  GEE_ParticleObject * po = 
    dynamic_cast<GEE_ParticleObject *> (m_particles.ptr());

  for(GEE_ParticleObject::iterator it = po->begin();
      it != po->end(); it++)
    VEE_trace("GEE_PathToParticles::stop # %p %p", po, (*it).m_particle.ptr());
  */

  return true;
}

bool GEE_PathToParticles::read(GEE_Io * io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != type())
    return false;
  
  readBasics(io);
  readPathId(io);

  int v = io->readInt32();

  if(v > 2)
    return false;

  m_toolCenterWeight = 1.0f; // Backwards compatibility

  m_writer.readParameters(io);
  
  return m_particle.read(io);
}

bool GEE_PathToParticles::write(GEE_Io * io)
{
  io->writeNullTerminatedString(type());

  writeBasics(io);
  writePathId(io);

  io->writeInt32(1); // file-format version 

  m_writer.writeParameters(io);

  return m_particle.write(io);
}

GEE_Operator * GEE_PathToParticles::create()
{
  return new GEE_PathToParticles;
}

void GEE_PathToParticles::setTextureName(const std::string &name)
{
  m_particle.setTexture(name, GEE_World::textureId(name));
}

void GEE_PathToParticles::alignParticle(GEE_ParticleData * particle,
					const GEE_PathVertex3D & pv,
					const VEE_Vector3 & headLoc,
					VEE_MotionAlign alignType,
					float baseSize,
					bool adjustSizeOfParticle,
					float * particleSize)
{
  VEE_Matrix4 m = GEE_PathToMesh::createMatrix(pv, baseSize);
  
  // Transformed origin etc:
  VEE_Vector3 to = m.column(3).vector3();
  VEE_Vector3 hand = pv.m_location;
  
  particle->setLocation(hand);
  
  VEE_Vector3 tu[3]; // Transformed unit axis vectors
  VEE_Matrix3 rot; // Difference between transformed unit axis vectors and origin
  
  for(uint i = 0; i < 3; i++) {
    VEE_Vector3 tmp;
    tmp.clear();
    tmp[i] = 1.0f;
    
    tu[i] = m * tmp;
    rot[i] = tu[i] - to;
  }
  
  if(adjustSizeOfParticle || particleSize) {

    float size = (rot[0].length() + rot[1].length()) * 0.5;
    
    if(particleSize)
      * particleSize = size;

    if(adjustSizeOfParticle)
      particle->setSize(size);
  }
  
  if(alignType == VEE_ALIGN_BILLBOARD) {
    
    // Head to hand vector
    VEE_Vector3 hth = headLoc - hand;
    
    VEE_Matrix3 rot;
    rot.rotateY(-atan2(hth.x, hth.z));
    
    VEE_Vector3 v = rot * (pv.m_rotation * VEE_Vector3_UNIT_Z);
    
    float angle = -atan2(v.x, v.y);
    
    particle->setRotation(angle);
  }
  else if(alignType == VEE_ALIGN_MOTION_TOTAL) {
    
    VEE_Vector3 v = pv.m_rotation * VEE_Vector3_UNIT_Y;
    particle->setVelocity(v);
    particle->setRotation(0.0);
  }
  else if(alignType == VEE_ALIGN_FIXED) {
    
    /* Now we fix the rotation matrix. The original rotation
       matrix rotates x, y and z in non-uniform fashion. Here are
       the things we know:
       
       1) All column vectors are of random length, nut non-zero
       
       2) X and Y vectors are usually not perpendicular
       
    */

    rot[2] = cross(rot[0], rot[1]);
    rot[1] = cross(rot[2], rot[0]);

    rot[0].normalize();
    rot[1].normalize();
    rot[2].normalize();

    rot.transpose();

    VEE_Vector3 axis;
    float radians;
    rot.getRotateAroundAxis(axis, radians);

    particle->setVelocity(axis);
    particle->setRotation(radians);
  }
}


bool GEE_PathToParticles::handleNewVertex(GEE_Path3D::iterator &it)
{
  return createParticles(it, it);
}

bool GEE_PathToParticles::handleNewVertices(GEE_Path3D::iterator & hand, 
					    GEE_Path3D::iterator & head)
{
  return createParticles(hand, head);
}

bool GEE_PathToParticles::createParticles(GEE_Path3D::iterator & it,
					  GEE_Path3D::iterator & head)
{
  // VEE_trace("GEE_PathToParticles::handleNewVertex");


  GEE_ParticleObject * particles = 
    dynamic_cast<GEE_ParticleObject *> (m_particles.ptr());

  if(!particles) {
    VEE_error("GEE_PathToParticles::handleNewVertex # particles = NULL");
    return false;
  }
  
  const GEE_PathVertex3D &pv  = GEE_Path3D::getVertex(it);
  const GEE_PathVertex3D &pv2 = GEE_Path3D::getVertex(head);

  const VEE_Vector3 & headLoc = pv2.m_location;

  float time = GEE_Path3D::getTime(it);

  float radius;
  VEE_Vector3 newLoc;

  if(m_useNapra) {
    radius = 
      pv.calculateToolRadius(m_radius, m_radiusNorm, m_exaggerateRadius);

    if(m_toolCenterWeight != 0.0f)
      newLoc = pv.calculateToolCenter() * m_toolCenterWeight + 
	pv.m_location * (1.0f - m_toolCenterWeight);
    else
      newLoc = pv.m_location;
  }
  else {
    newLoc = pv.m_location;
    radius = m_radius;
  }

  if(!processedPathVertices()) {
    m_prevTime = time;
    m_prevVertex = newLoc;
    return true;
  }

  /*
  VEE_trace("GEE_PathToParticles::handleNewVertex # %f %f %f",
	    newLoc.x, newLoc.y, newLoc.z);
  */

  // Now calculate where the next particle will be located.
  float interval = 1.0 / m_rate;
  float dt = time - m_prevTime;

  /* VEE_trace("GEE_PathToParticles::handleNewVertex # time: %f %f", 
     dt, m_remaining); */

  GEE_PathVertex3D tmp(pv);

  while(m_remaining < dt) {

    float rel = m_remaining / dt;
    
    tmp.m_location = m_prevVertex * (1.0f - rel) + newLoc * rel;
    
    if(m_version == 2) {
      float tmpSize;

      alignParticle(& m_particle, tmp, headLoc, 
		    (VEE_MotionAlign) m_alignType, m_particle.m_size, 
		    false, & tmpSize);

      /* Here we adjust the size of the particle directly, bypassing
	 GEE_ParticleData member functions. This is done to prevent
	 the GEE_ParticleData::m_size variable from being
	 overwritten. */

      if(m_modulateParticleSize)
	m_particle.m_particle.ptr()->setSize(tmpSize);
    }
    else
      m_particle.setLocation(tmp.m_location);
      
    /* Update color information, our clever trick is to update only
       the particle color, not color of the GEE_ParticleData object we
       have. */

    VEE_Rgba & baseColor = m_particle.m_color;

    VEE_Vector3 & color = VEE_Vector3::cast(baseColor.data());
    
    VEE_Vector3 c = 
      pv2.calculateLinearMix(color, m_colorMix, VEE_Vector3(0, 0, 0),
			     0.0, 2.0);
    
    VEE_Rgba tmp(c.x, c.y, c.z, baseColor.alpha(), baseColor.theta());
    
    m_particle.m_particle.ptr()->setColor(tmp);

    /* These random numbers must be calculated even if they are not
       used since backwards compatibility requires that we get the
       same random numbers as we got when any GEE file was saved. */
    VEE_Vector3 rv1 = m_rand.randomVectorInSphere();
    VEE_Vector3 rv2 = m_rand.randomVectorInSphere();
    float rf1 = (m_randomize != 0.0f) ? m_rand.rand11() : 0.0f;

    VEE_Vector3 loc = m_particle.m_location + rv1 * radius;

    m_particle.setLocation(loc);
    
    // if(m_randomize != 0.0f) {
    if(0) {

      if(m_alignType == VEE_ALIGN_BILLBOARD) {
	
	float rot = m_particle.m_rotation + rf1 * m_randomize * VEE_PI;
	m_particle.setRotation(rot);
      }
      else if(m_alignType == VEE_ALIGN_MOTION_TOTAL) {
	VEE_Vector3 newVel = (1.0f - m_randomize) * m_particle.m_velocity + 
	  m_randomize * rv2;
	newVel.normalize();
	m_particle.setVelocity(newVel);
      }
      else if(VEE_ALIGN_FIXED) {

	// Create a random rotation matrix.

	VEE_Matrix3 randrot;

	randrot[0] = m_rand.randomVectorOnSphere();
	randrot[1] = m_rand.randomVectorOnSphere();
	randrot[2] = cross(randrot[0], randrot[1]);

	while(randrot[2].lengthSqr() < 1.0e-6) {
	  randrot[1] = m_rand.randomVectorOnSphere();
	  randrot[2] = cross(randrot[0], randrot[1]);
	}

	randrot[1] = cross(randrot[2], randrot[0]);

/*	// @@@ this is commented out because the Quaternion class wasn't included with the source distribution.  Seems to pretty much work without it..


	// Now interpolate between this random rotation and original

	VEE_Quaternion q1, q2;
	q1.fromAxisAngle(m_particle.m_velocity,
			 m_particle.m_rotation);
	q2.fromRotationMatrix(randrot);

	q1 = VEE_Quaternion::slerp(m_randomize, q1, q2);
	
	VEE_Vector3 axis;
	float angle;

	// Push the new rotation back to the particle:
	
	q1.toAxisAngle(axis, angle);

	m_particle.setVelocity(axis);
	m_particle.setRotation(angle);
*/
      }

    }
    
    particles->addParticle(m_particle);
    
    m_remaining += interval;
    
    /* VEE_trace("GEE_PathToParticles::handleNewVertex # %f %f %f (%f %f %f) ",
       loc.x, loc.y, loc.z, newLoc.x, newLoc.y, newLoc.z); */
  }
  m_remaining -= dt;

  m_prevTime = time;
  m_prevVertex = newLoc;
  
  return true;
}
