/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_profile.h>

#include <gee_io.h>
#include <gee_macros.h>
#include <gee_renderer.h>

#include <vee_bounding_box.h>
#include <vee_error.h>
#include <vee_math.h>

static int GEE_ProfileVertex2DId = GEE_CONST('P', 'F', 'V', 'X');

bool GEE_ProfileVertex2D::read(GEE_Io * io)
{
  if(io->readInt32() != GEE_ProfileVertex2DId)
    return false;
  
  int v = io->readInt32();
  
  m_location = io->readVector2();
  m_normal   = io->readVector2();
  m_continues = io->readByte();

  return v == 1;
}

bool GEE_ProfileVertex2D::write(GEE_Io * io)
{
  io->writeInt32(GEE_ProfileVertex2DId);
  io->writeInt32(1); // version

  io->writeVector2(m_location);
  io->writeVector2(m_normal);
  io->writeByte(m_continues);

  return true;
}


/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

GEE_Profile::GEE_Profile()
{}

GEE_Profile::~GEE_Profile()
{}

void GEE_Profile::render(GEE_Renderer *r)
{
  r->renderProfile(this);
}

bool GEE_Profile::getBoundingBox(VEE_BoundingBox *bb)
{
  if(!m_vertices.size())
    bb->clear();
  else {
    VEE_Vector2 & v = m_vertices[0].m_location;
    bb->set(VEE_Vector3(v.x, v.y, 0.0f));
    
    for(uint i=1; i < m_vertices.size(); i++) {
      VEE_Vector2 & v = m_vertices[i].m_location;
      bb->expand(VEE_Vector3(v.x, v.y, 0.0f));
    }
  }

  return true;
}

bool GEE_Profile::read(GEE_Io * io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != "GEE_Profile")
    return false;
  
  readBasics(io);

  int v = io->readInt32();

  if(v != 1)
    return false;
  
  int count = io->readInt32();

  if(count > 1000000 || count < 0)
    VEE_error("GEE_Profile::read # About to load %d vertices", count);

  m_vertices.resize(count);

  for(iterator it = begin(); it != end(); it++)
    (*it).read(io);

  return true;
}

bool GEE_Profile::write(GEE_Io * io)
{
  io->writeNullTerminatedString("GEE_Profile");
  writeBasics(io);
  io->writeInt32(1); // version

  io->writeInt32(m_vertices.size());
  
  for(iterator it = begin(); it != end(); it++)
    (*it).write(io);
  
  return true;  
}

void GEE_Profile::clear()
{
  m_vertices.clear();
}

/** Generates a circle with desired number of segments. */

void GEE_Profile::generateCircle(float radius, int segments)
{
  clear();

  for(int i = 0; i < segments; i++) {
    float angle = VEE_TWO_PI * (float) i / (float) segments;
    float sa = sin(angle);
    float ca = cos(angle);

    VEE_Vector2 location(sa * radius, ca * radius), normal(sa, ca);
    
    addVertex(GEE_ProfileVertex2D(location, normal));
  }
}

void GEE_Profile::generateBar(float height, float topWidth)
{
  clear();

  VEE_Vector2 n(1.0, 0.0);
  addVertex(GEE_ProfileVertex2D(VEE_Vector2(0.0, -height), n, true));
  addVertex(GEE_ProfileVertex2D(VEE_Vector2(0.0,  height), n, false));
  
  if(topWidth <= 0.0)
    return;

  n.make(0.0, 1.0);
  addVertex(GEE_ProfileVertex2D(VEE_Vector2(-topWidth, height), n, true));
  addVertex(GEE_ProfileVertex2D(VEE_Vector2(0.0, height), n, true));
  addVertex(GEE_ProfileVertex2D(VEE_Vector2(topWidth,  height), n, false));

  n.make(0.0, -1.0);
  addVertex(GEE_ProfileVertex2D(VEE_Vector2(topWidth, -height), n, true));
  addVertex(GEE_ProfileVertex2D(VEE_Vector2(0.0, -height), n, true));
  addVertex(GEE_ProfileVertex2D(VEE_Vector2(-topWidth,-height), n, false));
}  

void GEE_Profile::generateCross(float radius)
{
  clear();
  
  VEE_Vector2 n(1.0, 0.0);

  addVertex(GEE_ProfileVertex2D(VEE_Vector2(0.0, -radius), n, true));
  addVertex(GEE_ProfileVertex2D(VEE_Vector2(0.0,  radius), n, false));

  n.make(0, 1);

  addVertex(GEE_ProfileVertex2D(VEE_Vector2(-radius, 0), n, true));
  addVertex(GEE_ProfileVertex2D(VEE_Vector2( radius, 0), n, false));
}

void GEE_Profile::normalize(const VEE_Vector3 & span)
{
  VEE_BoundingBox bb;
  getBoundingBox(&bb);
  
  VEE_Vector3 bspan = bb.span();
  VEE_Vector2 center = bb.center().vector2();
  
  VEE_Vector2 scale;

  uint i;
  
  for(i = 0; i < 2; i++)
    scale[i] = (bspan[i] > 1.0e-5) ? (span[i] / bspan[i]) : 1.0f;
  
  for(i = 0; i < size(); i++) {

    GEE_ProfileVertex2D & vertex = getVertex(i);

    vertex.m_location -= center;
    vertex.m_location.scale(scale);
    
    /* VEE_trace("%s # Vertex %d: %f %f", fname, i, 
       vertex.m_location.x, vertex.m_location.y); */
  }
}


uint GEE_Profile::segments() const
{
  uint count = 0;

  for(container::const_iterator it = m_vertices.begin(); 
      it != m_vertices.end(); it++)
    if((*it).m_continues)
      count++;

  return count;
}
