/* -*- C++ -*- */

/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#ifndef GEE_TRANSPARENCY_LIST_H
#define GEE_TRANSPARENCY_LIST_H

#include <vee_vector3.h>
#include <vee_matrix4.h>

#include <map>

class GEE_Geometry;

class VEE_Object;
class VEE_SceneBox;

/** A class that is used to store a list of trasparent objects
    (pointers only). The objects are sorted by distance and ready to
    be rendered back-to-front. 

    @author Tommi Ilmonen
*/

class GEE_TransparencyList
{
public:
  enum Type {
    UNKNOWN,
    VEE_OBJ,
    GEE_DATA
  };

  class Item {
  public:
    Item(Type t = UNKNOWN, void * ptr = 0)
      : m_type(t), m_ptr(ptr) {}

    Item(GEE_Geometry * d, const VEE_Matrix4 &m) 
      : m_type(GEE_DATA), m_ptr(d), m_transformation(m)
    {}

    Item(VEE_Object * d, const VEE_Matrix4 &m, float scale) 
      : m_type(VEE_OBJ), m_ptr(d), m_transformation(m), m_scale(scale)
    {}

    Type type() const { return m_type; }
    void * ptr() { return m_ptr; }

    VEE_Matrix4 &       transformation()         { return m_transformation; }
    const VEE_Matrix4 & transformation() const   { return m_transformation; }
    void setTransformation(const VEE_Matrix4 &t) { m_transformation = t; }

    float scale() const { return m_scale; }

  protected:
    Type   m_type;
    void * m_ptr;
    VEE_Matrix4 m_transformation;
    float  m_scale;
  };

  typedef std::multimap<float, Item> container;
  typedef container::iterator iterator;

  GEE_TransparencyList();
  ~GEE_TransparencyList();

  void addVee(VEE_Object *, const VEE_Matrix4 &, float scale = 0.0f);
  void addVeeScene(VEE_SceneBox *, const VEE_Matrix4 &);
  void addGee(GEE_Geometry *, const VEE_Matrix4 &);

  iterator begin()  { return m_items.begin(); }
  iterator end()    { return m_items.end(); }
  uint size() const { return m_items.size(); }

  void setCameraLocation(const VEE_Vector3 &l) { m_camPos = l; }

  void clear();

  static Item & getItem(iterator & it) { return (*it).second; }

protected:

  container m_items;
  /** Caches the latest insert location. This is used to speed up the
      positioning of particles into the map. It is assumed that
      successive particles are close to each other. */
  iterator  m_latest;

  VEE_Vector3 m_camPos;
};

#endif
