/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_turn_around_path.h>

#include <gee_io.h>
#include <gee_macros.h>
#include <gee_world.h>

#include <vee_error.h>
#include <vee_math.h>

GEE_TurnAroundPath::GEE_TurnAroundPath(GEE_id_t pathId, 
				       GEE_id_t geoId,
				       float speed)
  : GEE_PathToX(pathId),
    m_geoId(geoId),
    m_steps(0),
    m_speed(speed)
{
  addIdMember(&m_geoId);
}

GEE_TurnAroundPath::~GEE_TurnAroundPath()
{}
  
bool GEE_TurnAroundPath::start(GEE_World *w, const GEE_TimeStamp &now)
{
  GEE_PathToX::start(w, now);
  m_steps = 0;

  return true;
}

bool GEE_TurnAroundPath::act(GEE_World *w, const GEE_TimeStamp &now)
{
  GEE_PathToX::act(w, now);

  if(m_steps) doTurn(w);

  return true;
}

static const ushort GEE_GeoId      = GEE_CONST2('G', 'I');
static const ushort GEE_OmegaId    = GEE_CONST2('O', 'M');
static const ushort GEE_EndId      = GEE_CONST2('E', 'N');

bool GEE_TurnAroundPath::read(GEE_Io *io)
{
  std::string str;
  io->readNullTerminatedStdString(str);
  if(str != type())
    return false;
  
  readBasics(io);
  readPathId(io);

  int v = io->readInt32();

  if(v != 1)
    return false;

  ushort id = io->readInt16();

  while(id != GEE_EndId) {
    if(id == GEE_GeoId)
      m_geoId = io->readInt64();
    else if(id == GEE_OmegaId)
      m_speed = io->readFloat32();
    else {
      VEE_error("GEE_TurnAroundPath::read # Bad id at %d (%c%c)", 
		(int) io->position(), (uchar) (id >> 8), (uchar) id);
      return false;
    }
    id = io->readInt16();
  }

  return true;

}

bool GEE_TurnAroundPath::write(GEE_Io *io)
{
  io->writeNullTerminatedString(type());

  writeBasics(io);
  writePathId(io);
  
  io->writeInt32(1); // version

  io->writeInt16(GEE_GeoId);
  io->writeInt64(m_geoId);

  io->writeInt16(GEE_OmegaId);
  io->writeFloat32(m_speed);

  io->writeInt16(GEE_EndId);

  return true;
}

const char * GEE_TurnAroundPath::type()
{
  return "GEE_TurnAroundPath";
}

GEE_Operator * GEE_TurnAroundPath::create()
{
  return new GEE_TurnAroundPath();
}

bool GEE_TurnAroundPath::handleNewVertex(GEE_Path3D::iterator &)
{
  m_steps++;
  return true;
}

void GEE_TurnAroundPath::doTurn(GEE_World *w)
{
  double angle = m_steps * VEE_PI * m_speed;
  
  GEE_Geometry * g = w->geometry(m_geoId);

  if(!g)
    return;

  VEE_BoundingBox bb;

  g->getBoundingBox(&bb);
  VEE_Vector3 center = bb.center();

  VEE_Matrix4d m, m2;
  m.identity();
  m2 = m;

  m.setColumn3(3, -center);

  VEE_Matrix3d rot;
  rot.rotateY(angle);
  m2.setRotation(rot);

  m = m2 * m;

  m2.identity();
  m2.setColumn3(3, center);

  m = m2 * m;

  g->setTransformation(m * g->transformation());

  m_steps = 0;
}

