/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <gee_gl_material.h>

#include <vee_error.h>
#include <gee_gl_renderer.h>


GEE_GlMaterial::GEE_GlMaterial(float shine, 
			       const VEE_Rgba & color,
			       GLuint textureId,
			       int flags)
  : m_textureId(textureId),
    m_shine(shine),
    m_ambient(color), 
    m_diffuse(color), 
    m_specular(color),
    m_uvScale(1.0f),
    m_flags(flags)
{}

GEE_GlMaterial::GEE_GlMaterial()
  : m_textureId((GLuint)-1),
    m_shine(0.1),
    m_ambient(VEE_Rgba::Gray),
    m_diffuse(VEE_Rgba::Gray),
    m_specular(VEE_Rgba::Gray),
    m_uvScale(1.0f),
    m_flags(GEE_MAT_LIGHTING | GEE_MAT_SMOOTH)
{}

GEE_GlMaterial::~GEE_GlMaterial()
{}

void GEE_GlMaterial::applyGl(GEE_GlRenderer *renderer) const
{
  renderer->applyGlLighting(m_flags & GEE_MAT_LIGHTING);
  renderer->applyGlBlend(m_flags & GEE_MAT_BLEND);
  renderer->applyGlSmooth(m_flags & GEE_MAT_SMOOTH);
  renderer->applyGlTwoSided(m_flags & GEE_MAT_TWO_SIDED);
  renderer->applyGlTexture2D(m_textureId);
  
  float shine = m_shine * 128.0f;
  if(shine > 128.0f) shine = 128.0f;
  glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, & shine);

  if(m_flags & GEE_MAT_LIGHTING) {
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, 1);
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, m_ambient.data());
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, m_diffuse.data());
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, m_specular.data());
  }
  else
    glColor4fv(m_ambient.data());

  if(useVertexColors())
    glEnable(GL_COLOR_MATERIAL);
  else
    glDisable(GL_COLOR_MATERIAL);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

/** The default material. Usually just fairly normal gray. */
GEE_GlMaterial GEE_GlMaterialBank::m_default
(70.0, VEE_Rgba(0.7, 0.7, 0.7, 1.0));

GEE_GlMaterialBank::GEE_GlMaterialBank()
{
  addUsualColors();
}

GEE_GlMaterialBank::~GEE_GlMaterialBank()
{}

void GEE_GlMaterialBank::addMaterial(const std::string &name, 
				     const GEE_GlMaterial &material)
{
  // trace("GEE_GlMaterialBank::addMaterial # \"%s\"", name.c_str());
  m_materials[name] = material;
}

/** Gets a material by name. If the material is not found, then the
    default material is returned. */

const GEE_GlMaterial & GEE_GlMaterialBank::getMaterial(const std::string &s)
{
  const std::string *s2 = m_forceMaterial.size() ? & m_forceMaterial : & s;

  std::map<std::string, GEE_GlMaterial>::iterator it = m_materials.find(*s2);

  if(it != m_materials.end()) {
    // trace("GEE_GlMaterialBank::getMaterial # \"%s\"", s2.c_str());
    return (*it).second;
  }

  /* VEE_trace("GEE_GlMaterialBank::getMaterial # No material %s",
     s.c_str()); */

  return m_default;
}

void GEE_GlMaterialBank::addUsualColors()
{
  addMaterial("black", GEE_GlMaterial(0.5, VEE_Rgba::Black));
  addMaterial("gray",  GEE_GlMaterial(0.5, VEE_Rgba::Gray));
  addMaterial("white", GEE_GlMaterial(0.5, VEE_Rgba::White));

  addMaterial("red",   GEE_GlMaterial(0.5, VEE_Rgba::Red));
  addMaterial("green", GEE_GlMaterial(0.5, VEE_Rgba::Green));
  addMaterial("blue",  GEE_GlMaterial(0.5, VEE_Rgba::Blue));

  addMaterial("red-nolight",   
	      GEE_GlMaterial(0.5, VEE_Rgba::Red, (GLuint) -1, 
			     GEE_MAT_SMOOTH));
  addMaterial("green-nolight",   
	      GEE_GlMaterial(0.5, VEE_Rgba::Green, (GLuint) -1, 
			     GEE_MAT_SMOOTH));
  addMaterial("blue-nolight",   
	      GEE_GlMaterial(0.5, VEE_Rgba::Blue, (GLuint) -1, 
			     GEE_MAT_SMOOTH));
}
