/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */


#include <vee_gl_sl_renderer.h>

#include <vee_error.h>

#include <ConfigReader.hpp>
#include <errno.h>

const char * VEE_SlParamNames[] = {
  "particle_size",
  "campos",
  "windowscale",
  "depthparams",
  "age",
  "color",
  "additive",
  "location",
  "velocity",
  "corner_uv",
  "tex",
  "depth_map"
};

char * VEE_loadFile(const char * fileName)
{
  
  FILE *file = fopen(fileName, "rb");

  if(!file) {
    VEE_error("VEE_loadFile # Unable to open \"%s\" (%s)", 
	      fileName, strerror(errno));
    errno = 0;
    return false;
  }
  
  fseek(file,0,SEEK_END);

  int size = ftell(file);

  char * res = 0;

  if(size) {
    fseek(file,0,SEEK_SET);
    res = (char *) malloc(size + 1);
    fread(res, 1, size, file) == 1;
    res[size] = '\0';
  }

  fclose(file);

  return res;
}

VEE_GlSlDriver::VEE_GlSlDriver()
  : m_vertexShader(0),
    m_fragmentShader(0),
    m_program(0)
{}

VEE_GlSlDriver::~VEE_GlSlDriver()
{
  deleteShaders();
}

bool VEE_GlSlDriver::loadShadersFromFile(const char * fileName)
{
  const char * fname = "VEE_GlSlDriver::loadShadersFromFile";

  VEE_trace("%s # \"%s\"", fname, fileName);

  deleteShaders();

  int i;
  GLint status;
  GLenum e;

  std::string file(fileName ? fileName : "");
  
  char * vert = VEE_loadFile((file + ".vert").c_str());
  char * frag = VEE_loadFile((file + ".frag").c_str());

  if(!vert && !frag)
    return false;
  
  if(vert) {
    m_vertexShader = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
    glShaderSourceARB(m_vertexShader, 1, (const GLcharARB**) & vert, NULL);
    free(vert);
    VEE_trace("%s # Loaded vertex shader source", fname);
  }
  if(frag) {
    m_fragmentShader = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
    glShaderSourceARB(m_fragmentShader, 1, (const GLcharARB**) & frag, NULL);  
    free(frag);
    VEE_trace("%s # Loaded fragment shader source", fname);
  }

  GLhandleARB shaders[2];
  shaders[0] = m_vertexShader;
  shaders[1] = m_fragmentShader;
  
  for(i=0; i < 2; i++) {

    GLhandleARB sh = shaders[i];

    if(!sh)
      continue;
  
    glCompileShaderARB(sh);
    
    e = glGetError();
    
    status = 0;

    glGetObjectParameterivARB(sh,
			      GL_OBJECT_COMPILE_STATUS_ARB, 
			      &status);

    std::string log = infoLog(sh);

    if(!status) {
      VEE_error("%s # %p # Shader %d / %s Compilation ERROR, log:\n%s", 
		fname, this, i, fileName, log.c_str());

      deleteShaders();
      return false;
    }
    
    if(log.size())
      VEE_trace("%s # Shader %d compilation log:\n%s\n", 
		fname, i, log.c_str());
  }
  
  m_program = glCreateProgramObjectARB();
  glAttachObjectARB(m_program, m_vertexShader);
  glAttachObjectARB(m_program, m_fragmentShader);

  glLinkProgramARB(m_program);

  std::string log = infoLog(m_program);
  
  if(log.size())
    VEE_trace("%s # Shader link log:\n%s\n", fname, log.c_str());
  
  glGetObjectParameterivARB(m_program,
			    GL_OBJECT_LINK_STATUS_ARB, 
			    &status);

  for(i = 0 ; i < VEE_SLP_SIZEOF; i++) {

    int tmp = glGetUniformLocationARB(m_program, VEE_SlParamNames[i]);
    m_params[i] = tmp;
    
    if(tmp < 0)
      VEE_error("%s # No uniform named %s", fname, VEE_SlParamNames[i]);
    else
      VEE_trace("%s # Uniform %s -> %d", fname, VEE_SlParamNames[i], tmp);
  }

  /* if(m_params[VEE_SLP_TEX] >= 0)
    glUniform1iARB(m_params[VEE_SLP_TEX], 0);

  if(m_params[VEE_SLP_DEPTH_MAP] >= 0)
  glUniform1iARB(m_params[VEE_SLP_DEPTH_MAP], 1);*/

  glGetError();

  return true;
}

void VEE_GlSlDriver::deleteShaders()
{
  if(m_vertexShader) {

    if(m_program)
      glDetachObjectARB(m_program, m_vertexShader);
      
    glDeleteObjectARB(m_vertexShader);
    m_vertexShader = 0;
  }

  if(m_fragmentShader) {
    
    if(m_program)
      glDetachObjectARB(m_program, m_fragmentShader);

    glDeleteObjectARB(m_fragmentShader);
    m_vertexShader = 0;
  }
  
  if(m_program) {
    glDeleteObjectARB(m_program);
    m_program = 0;
  }
}

bool VEE_GlSlDriver::bindProgram(VEE_GlSlRenderer *renderer)
{
  glUseProgramObjectARB(m_program);
  
  GLenum e = glGetError();

  if(e) {
    VEE_error("VEE_GlSlDriver::bindProgram %p # GL ERROR %s", 
	      this, gluErrorString(e));
    return false;
  }

  if(m_params[VEE_SLP_WINDOW_SIZE] >= 0) {

#ifdef GL_TEXTURE1_ARB
    glActiveTexture(GL_TEXTURE1_ARB); 
    renderer->depthMap().use();
    // glEnable(GL_TEXTURE_2D);
    glActiveTexture(GL_TEXTURE0_ARB); 
#endif    

    float texw = renderer->depthMap().width();
    float texh = renderer->depthMap().height();

    glUniform3fARB(m_params[VEE_SLP_WINDOW_SIZE], 
		   1.0f / texw,
		   1.0f / texh,
		   1.0f / renderer->farDistance());
  }
  if(m_params[VEE_SLP_DEPTH_PARAMS] >= 0) {
    float range = renderer->farDistance() - renderer->nearDistance();

    glUniform3fARB(m_params[VEE_SLP_DEPTH_PARAMS], 
		   renderer->nearDistance(),
		   1.0 / range,
		   0.1 / range);
  }

  if(m_params[VEE_SLP_TEX] >= 0)
    glUniform1iARB(m_params[VEE_SLP_TEX], 0);

  if(m_params[VEE_SLP_DEPTH_MAP] >= 0)
    glUniform1iARB(m_params[VEE_SLP_DEPTH_MAP], 1);

  /* if(VEE_GlSlRenderer::debug()) 
     VEE_trace("VEE_GlSlRenderer::bindProgram # OK"); */
  
  return true;
}

/** Issues the commands that will render a particle using quads. All
    relevant texture modes, rendering modes etc. are assumed to be
    already on. 
    
    You can override this method in your own classes to handle the
    rendering in some other way.
*/

void VEE_GlSlDriver::renderParticle(VEE_Particle *p,
				    VEE_GlSlRenderer *r)
{
  // if(m_params[VEE_SLP_PARTICLE_SIZE] >= 0) 
  {
    // Use texture coordinate 1 to transfer the particle size to the shader.
    float size = 0.3 * r->farDistance() / (p->size() * r->particleSize());
    glMultiTexCoord1f(1, size);
    // glUniform1fARB(m_params[VEE_SLP_PARTICLE_SIZE], size);
    // VEE_trace("VEE_GlSlDriver::renderParticle # size = %f", size);
  }

  if(!m_vertexShader) {
    r->VEE_GlRenderer::renderParticle(p);
    return;
  }

  VEE_Vector3 pos = p->location();
  const VEE_Vector3 relPos = r->camPos() - pos;
  float dist = relPos.length();
  
  // Skip objects that are not in the range.
  if(dist > r->farDistance() || dist <= r->nearDistance())
    return;
  
  // Set color:
  { 
    VEE_Rgba color = r->transformColor(p->color());
    
    if(dist < r->fadeDistance()) {
      float dfade = (dist - r->nearDistance()) / r->fadeRange();
      color.setAlpha(color.alpha() * dfade);
    }
    
    glColor4fv(color.data());

    if(m_params[VEE_SLP_ADDITIVE])
      glUniform1fARB(m_params[VEE_SLP_ADDITIVE], color.theta());
  }

  // Set other stuff:
  if(m_params[VEE_SLP_AGE] >= 0)
    glUniform1fARB(m_params[VEE_SLP_AGE], p->age());

  if(m_params[VEE_SLP_VELOCITY] >= 0)
    glUniform3fvARB(m_params[VEE_SLP_VELOCITY], 1, p->velocity().data());
  
  glUniform4fARB(m_params[VEE_SLP_CORNER_UV], 0.0, 0.0, 0.0, 0.0);
  glVertex3fv(pos.data());

  glUniform4fARB(m_params[VEE_SLP_CORNER_UV], 1.0, 0.0, 1.0, 0.0);
  glVertex3fv(pos.data());

  glUniform4fARB(m_params[VEE_SLP_CORNER_UV], 1.0, 1.0, 1.0, 1.0);
  glVertex3fv(pos.data());

  glUniform4fARB(m_params[VEE_SLP_CORNER_UV], 0.0, 1.0, 0.0, 1.0);
  glVertex3fv(pos.data());
}

std::string VEE_GlSlDriver::infoLog(GLhandleARB obj)
{
  if(!obj) return "";

  GLint infologLength = 0;
  GLsizei charsWritten  = 0;
  GLcharARB *infoLog;
  
  glGetObjectParameterivARB(obj, GL_OBJECT_INFO_LOG_LENGTH_ARB,
			    &infologLength);
  
  if(infologLength <= 0)
    return "";
  
  infoLog = (GLcharARB *) malloc(infologLength);
  
  glGetInfoLogARB(obj, infologLength, &charsWritten, infoLog);

  std::string res = (char *) infoLog;
  
  free(infoLog);

  return res;
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////


bool VEE_GlSlRenderer::m_debug = true;

VEE_GlSlRenderer::VEE_GlSlRenderer(VEE_SceneBox *box)
  : VEE_GlRenderer(box),
    m_currentDriver(0),
    m_useShaders(true)
{
  m_useDepthMap = true;
  m_nearDistance = 0.5f;
  m_farDistance = 200.0f;
}

VEE_GlSlRenderer::~VEE_GlSlRenderer()
{}

void VEE_GlSlRenderer::initializeGL()
{
  if(m_debug) VEE_trace("VEE_GlSlRenderer::initializeGL");

  VEE_GlRenderer::initializeGL();

  loadShaders();
}

void VEE_GlSlRenderer::endGlMode()
{
  VEE_GlRenderer::endGlMode();

  disableSl();
}

void VEE_GlSlRenderer::renderParticle(VEE_Particle *p)
{
  VEE_RenderClass * rc = p->renderClass();

  int shader = rc ? rc->shader() : 0;
  
  if(shader <= 0 || !m_useShaders) {
    if(m_currentShader)
      disableSl();
    
    VEE_GlRenderer::renderParticleVisual(p);
    return;
  }
  else if(!m_currentShader)
    enableSl();

  // Change shaders if necessary:

  if(shader != m_currentShader) {
    driver_iter dit = m_drivers.find(shader);
    if(dit == m_drivers.end()) {
      VEE_error("VEE_GlSlRenderer::renderParticle # No shader %d", 
		shader);
      return;
    }
    
    newMode(RenderNothing);

    m_currentShader = shader;
    (*dit).second.ptr()->bindProgram(this);
    m_currentDriver = (*dit).second.ptr();
  }

  // Now do the rendering:
  
  newMode(RenderQuads);
  
  m_currentDriver->renderParticle(p, this);
}

void VEE_GlSlRenderer::disableSl()
{
  newMode(RenderNothing);  

  glUseProgramObjectARB(0);
  m_currentShader = 0;

#ifdef GL_TEXTURE1_ARB
  glActiveTexture(GL_TEXTURE1_ARB);
  glDisable(GL_TEXTURE_2D);
  glActiveTexture(GL_TEXTURE0_ARB);
#endif   
}

void VEE_GlSlRenderer::enableSl()
{
}

/** This method loads all requested shaders. If the compilation of a
    particular shader fails, then that shader is removed from the
    shader table. */

void VEE_GlSlRenderer::loadShaders()
{  
  for(std::map<std::string, int>::iterator it = m_shaderNames.begin();
      it != m_shaderNames.end(); ) {

    std::string sh = (*it).first;

    VEE_GlSlDriver * driver = new VEE_GlSlDriver();

    if(!driver->loadShadersFromFile(sh.c_str())) {
      delete driver;
      std::map<std::string, int>::iterator tmp = it;
      tmp++;
      m_shaderNames.erase(it);
      it = tmp;
    }
    else {
      int id = (*it).second;

      m_drivers[id] = driver;
      
      VEE_trace("VEE_GlSlRenderer::loadShaders # Added shader %s (%d)", 
		sh.c_str(), id);
      it++;
    }
  }

  if(m_drivers.size())
    m_applyDepthMap = true;
}
