/* COPYRIGHT
 *
 * This file is part of the Visual Effects Engine - VEE
 *
 * Read the "VEE-LICENSE" file for the license.
 *
 * Authors & Copyright:   
 *
 * Tommi Ilmonen, Tuukka Heikura, Marko Myllymaa and 
 * Janne Kontkanen 2001-2004
 *
 * Additional copyrights: Tekes 2003-2004
 *
 * firstname.lastname@hut.fi
 *
 */

#include <vee_gl_texture.h>

#include <vee_error.h>
#include <vee_image.h>

#include <di_types.h>

#ifdef __APPLE__
# include <OpenGL/glu.h>
#else
# include <GL/glu.h>
#endif

#include <stdio.h>

VEE_GlTexture2D::VEE_GlTexture2D()
  : m_glId((GLuint)-1),
    m_width(0),
    m_height(0)
{}

VEE_GlTexture2D::~VEE_GlTexture2D()
{
  // unallocateId();
}

bool VEE_GlTexture2D::genTexture1D(const uchar * data, uint n, 
				   GLenum format, bool mipmaps)
{
  const char * fname = "VEE_GlTexture2D::genTexture1D";

  allocateId();
  glBindTexture(GL_TEXTURE_1D, m_glId);
  
  if(mipmaps) {
    GLint r = gluBuild1DMipmaps
      (GL_TEXTURE_1D, format, n, format, GL_UNSIGNED_BYTE, data);
    if(r) {
      VEE_error("%s # Error building mipmaps %s", 
		fname, gluErrorString(r));
      return false;
    }
  }
  else 
    glTexImage1D(GL_TEXTURE_1D, 0, format, n, 0, format, GL_UNSIGNED_BYTE,
		 data);
  
  glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_S, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_WRAP_T, GL_CLAMP);
  glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_1D, GL_TEXTURE_MIN_FILTER, 
		  mipmaps ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR );
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  return true;
}

bool VEE_GlTexture2D::genTexture(const VEE_Image *image, 
				 bool mipmaps)
{
  const char * fname = "VEE_GlTexture2D::genTexture";

  // unallocateId();

  uint w = image->w();
  uint h = image->h();
 
  if(w != h) {
    
    if(!mipmaps)
      VEE_error("%s # %d != %d (will use mipmaps)", fname, w, h);

    mipmaps = true;
    
    // return false;
  }
 
  allocateId();

  uint s = 1;
  
  while(s < w || s < h)
    s = s << 1;
  
  m_width = m_height = s;

  glBindTexture(GL_TEXTURE_2D, m_glId);

  GLenum format;

  uint bpp = image->bytesPerPixel();

  switch(bpp) {
  case 1:
    format = GL_LUMINANCE;
    break;
  case 2:
    format = GL_LUMINANCE_ALPHA;
    break;
  case 3:
    format = GL_RGB;
    break;
  case 4:
    format = GL_RGBA;
    break;
  default:
    VEE_error("%s # Bad image format 0 %d bpp", fname, bpp);
    return false;
  }

  if(mipmaps) {
    GLint r = gluBuild2DMipmaps
      (GL_TEXTURE_2D, format, w, h, format, GL_UNSIGNED_BYTE, image->image());
    if(r) {
      VEE_error("%s # Error building mipmaps %s", 
		fname, gluErrorString(r));
    }
  }
  else 
    glTexImage2D(GL_TEXTURE_2D, 0, format, w, h, 0, format, GL_UNSIGNED_BYTE,
		 image->image());
  
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, 
		  mipmaps ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR );
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  
  int e = glGetError();
  if(e) 
    VEE_error("VEE_GlTexture2D::genTexture # GL ERROR ALREADY \"%s\"", 
	      gluErrorString(e));

  // VEE_trace("%s # 0x%p Created texture %d", fname, this, (int) m_glId);

  return true;
}

/** Allocates an OpenGL texture id. */

void VEE_GlTexture2D::allocateId()
{
  if(m_glId == (GLuint) -1)
    glGenTextures(1, &m_glId);
}

void VEE_GlTexture2D::unallocateId()
{
  // VEE_trace("VEE_GlTexture2D::unallocateId # %p %d", this, (int) m_glId);
  if(m_glId != (GLuint) -1) {
    glDeleteTextures(1, &m_glId);
    m_glId = (GLuint) -1;
  }  
}

void VEE_GlTexture2D::allocateArea(int minW, int minH, GLenum format)
{
  allocateId();
  
  glBindTexture(GL_TEXTURE_2D, m_glId);

  int s = 1;
  
  while(s < minW || s < minH)
    s = s << 1;
  
  m_width = m_height = s;

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  glTexImage2D(GL_TEXTURE_2D, 0, format, s, s,
	       0, format, GL_UNSIGNED_BYTE, 0);
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

VEE_GlTexture2DBank::VEE_GlTexture2DBank()
{}

VEE_GlTexture2DBank::~VEE_GlTexture2DBank()
{
  clear();
}

/** Get a texture by name. 

    @return This method returns null if the texture was not
    found. Otherwise it returns pointer to the OpenGL texture.
*/

VEE_GlTexture2D *VEE_GlTexture2DBank::getTexture(const std::string &name)
{
  std::map<std::string, VEE_GlTexture2D>::iterator tex = m_textures.find(name);
  
  if(tex == m_textures.end())
    return 0;

  return & (*tex).second;
}

/** Get a texture by name. If the texture does not exist yet, then one
    is created.

    @return This method returns pointer to the OpenGL texture.
*/

VEE_GlTexture2D *VEE_GlTexture2DBank::buildTexture(const std::string &name)
{
  return & m_textures[name];
}

void VEE_GlTexture2DBank::addTexture(const std::string &name, 
				     const VEE_GlTexture2D &tex)
{
  VEE_trace("VEE_GlTexture2DBank::addTexture # \"%s\"", name.c_str());

  VEE_GlTexture2D *old = getTexture(name);

  if(old) {
    VEE_trace("VEE_GlTexture2DBank::addTexture # Deleting old texture %p %d",
	      old, (int) old->id());
    old->unallocateId();
  }

  m_textures[name] = tex;
}

void VEE_GlTexture2DBank::clear()
{
  std::map<std::string, VEE_GlTexture2D>::iterator tex = m_textures.begin();
  std::map<std::string, VEE_GlTexture2D>::iterator sentinel = m_textures.end();

  while(tex != sentinel) {
    (*tex).second.unallocateId();
    tex++;
  }
}

void VEE_GlTexture2DBank::loadTextures(const std::string & path,
				       const std::string & name,
				       const std::string & suffix,
				       int number)
{
  // const char *fname = "VEE_GlTexture2DBank::loadTextures";

  char buf[16];

  VEE_Image image;

  VEE_GlTexture2D tex;

  for(int i=0; i < number; i++) {

    sprintf(buf, "%d", i);
    std::string fullname = name + buf;
    std::string filename(path + "/" + fullname + suffix);
    bool ok = loadTexture(filename, fullname);
    if(!ok)
      VEE_error("VEE_GlTexture2DBank::loadTextures # Failed to load \"%s\"",
		filename.c_str());
  }
}

bool VEE_GlTexture2DBank::loadTexture(const std::string & filename,
				      const std::string & name)
{
  const char * fname = "VEE_GlTexture2DBank::loadTexture";

  std::map<std::string, std::string>::iterator it = 
    m_fileTextures.find(filename);

  if(it != m_fileTextures.end()) {

    std::string oldName = (*it).second;

    if(oldName == name)
      return true;

    VEE_GlTexture2D *tex = getTexture(oldName);

    if(tex) {

      VEE_trace("%s # Reusing texture for file %s", fname, filename.c_str());

      addTexture(name, *tex);

      return true;
    }

    VEE_error("%s # There should be a texture for %s", 
	      fname, filename.c_str());
  }

  VEE_Image image;

  VEE_GlTexture2D tex;

  bool ok = image.load(filename.c_str());

  if(!ok) {
    VEE_error("%s # Failed to load file \"%s\"", fname, filename.c_str());
    return false;
  }

  image.mirrorXY();
  
  tex.forgetId();
  tex.genTexture(&image, true);
  
  addTexture(name, tex);
  
  VEE_trace("%s # Loaded file \"%s\" -> %d", 
	    fname, filename.c_str(), (int) tex.id());
  tex.forgetId();

  m_fileTextures[filename] = name;

  return true;
}
