
/* COPYRIGHT
 *
 * This file is part of the Simple Object Loader and Renderer - SOLAR
 *
 * Read the file "SOLAR-LICENSE" for the license.
 *
 * Authors & Copyright:
 *
 * Tommi Ilmonen, 2001-2003, except LWO-code by RJ, 2001-2003
 *
 * Tommi's email: firstname.lastname@hut.fi
 *
 */

#include <GraphicTriangleMesh.hpp>

#include <Renderer.hpp>
#include <Stream.hpp>

#include <iostream>

namespace Solar {

void GraphicTriangleMesh::Triangle::clear()
{
  m_vertices[0] = m_vertices[1] = m_vertices[2] = 0;
}

void GraphicTriangleMesh::Triangle::write(std::ostream *s)
{
  (*s) << "TRIMESH::TRIANGLE  " 
       << m_vertices[0] << "   " <<  m_vertices[1] << "   "
       << m_vertices[2];
}

void GraphicTriangleMesh::VertexData::write(std::ostream *s)
{
  (*s) << "TRIMESH::VERTEX  p = ";
  Write(s, m_location);
  (*s) << "  n = ";
  Write(s, m_normal);
  (*s) << "  u = ";
  Write(s, m_uv);
    
}

/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////

GraphicTriangleMesh::GraphicTriangleMesh()
{
  clear();
}

GraphicTriangleMesh::~GraphicTriangleMesh()
{
  clear();
}

Graphic::Type GraphicTriangleMesh::type() const
{ 
  return Graphic::TriangleMesh;
}

void GraphicTriangleMesh::getBoundingBox(BoundingBox *bbox)
{
  (*bbox) = m_boundingBox;
}

int GraphicTriangleMesh::nofPrimitives() const
{
  return m_triangles.size();
}

void GraphicTriangleMesh::render(Renderer *renderer)
{
  renderer->renderTriangleMesh(this);
}

void GraphicTriangleMesh::scale(const Fluid::Vector3 &v)
{
  for(uint i = m_vertices.size(); i--; )
    m_vertices[i].scale(v);

  m_boundingBox.scale(v);
}

void GraphicTriangleMesh::move(const Fluid::Vector3 &v)
{
  for(uint i = m_vertices.size(); i--; )
    m_vertices[i].move(v);

  m_boundingBox.move(v);
}

void GraphicTriangleMesh::transform(const Fluid::Matrix3 &m)
{
  for(uint i = m_vertices.size(); i--; )
    m_vertices[i].transform(m);

}

void GraphicTriangleMesh::write(std::ostream *s)
{
  (*s) << "TRIANGLEMESH\nVERTICES\n";
  
  uint i;

  {
    VertexData *vertex = &* m_vertices.begin();
    VertexData *sentinel = &* m_vertices.end();

    i=0;

    while(vertex != sentinel) {
      (*s) << i++ << "   ";
      vertex->write(s);
      (*s) << "\n";
      vertex++;
    }
  }

  (*s) << "TRIANGLES\n";

  {
    Triangle *tri = &* m_triangles.begin();
    Triangle *sentinel = &* m_triangles.end();

    i=0;

    while(tri != sentinel) {
      (*s) << i++ << "   ";
      tri->write(s);
      (*s) << "\n";
      tri++;
    }
  }
}

void GraphicTriangleMesh::clear()
{
  m_triangles.resize(0);
  m_vertices.resize(0);

  m_boundingBox.clear();
}

void GraphicTriangleMesh::addVertex(const VertexData &vertex)
{
  if(!m_vertices.size())
    m_boundingBox.clear(vertex.m_location);
  else 
    m_boundingBox.expand(vertex.m_location);

  m_vertices.push_back(vertex);
  
  /*
  printf("GraphicTriangleMesh::addVertex # BBOX now [%f %f %f] [%f %f %f] [%f %f %f]\n",
	 m_boundingBox.low()[0],m_boundingBox.low()[1],m_boundingBox.low()[2],
	 m_boundingBox.high()[0],
	 m_boundingBox.high()[1],
	 m_boundingBox.high()[2],
	 vertex.m_location[0], vertex.m_location[1], vertex.m_location[2]); 
  */
}

void GraphicTriangleMesh::addTriangle(const Triangle &tri)
{
  m_triangles.push_back(tri);
}

void GraphicTriangleMesh::addTriangle(uint i, uint j, uint k)
{
  Triangle t;
  t.m_vertices[0] = i;
  t.m_vertices[1] = j;
  t.m_vertices[2] = k;

  addTriangle(t);
}
  
void GraphicTriangleMesh::recalculateBoundingBox()
{
  if(!m_vertices.size())
    m_boundingBox.clear();
  else {
    
    m_boundingBox.clear(m_vertices[0].location());

    if(m_vertices.size() > 1) {
      for(VertexData *v = &* m_vertices.begin(); 
	  v != &* m_vertices.end(); v++) {
	m_boundingBox.expand(v->location());
      }
    }
  }
}

} // namespace Solar
